C*******************************************************************
C     **** CNTAAT ... COUNT NONZERO ENTRIES OF  A*Atransp ****
C*******************************************************************
C
      SUBROUTINE CNTAAT(M,MAXM,MAXN,MAXNZA,NZL,
     X TRIANG,AATPNT,MARKER,TEMP,STAVAR,
     X CLPNTS,RWNMBS,RWHEAD,RWLINK,CLNMBS,LENCOL,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 M,MAXM,MAXN,MAXNZA,NZL,TRIANG,IOLOG
      INTEGER*4 AATPNT(MAXM+1),MARKER(MAXM),TEMP(MAXM)
      INTEGER*4 CLPNTS(MAXN+1),RWHEAD(MAXM+1),RWLINK(MAXNZA)
C
C *** The following arrays can be half-length integer.
      INTEGER*2 STAVAR(MAXN)
      INTEGER*2 RWNMBS(MAXNZA),CLNMBS(MAXNZA),LENCOL(MAXN)
C
C
C *** LOCAL VARIABLES
      INTEGER*4 IROW,IR,K,KBEG,KEND,JCOL,J
      INTEGER*4 LENAAT,LENROW
      CHARACTER*100 BUFFER
C
C
C *** COMMON ARREAS
C     Markers for linking rows.
C     COMMON /ICGRAD/ MSPLIT(100000)
C     INTEGER*2       MSPLIT
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     M       Number of rows of the LP constraint matrix
C             (and the dimension of  A*Atransp).
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     MAXN    Maximum column dimension of the LP constraint matrix.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     TRIANG  Indicator of how much of  A*Atransp sparsity pattern
C             is required:
C             0  if square matrix (except its diagonal) is needed;
C             1  if only strictly upper triangle is needed.
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C            -k  free variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicate the position of the original variable.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A.
C     RWHEAD  Headers to the row linked lists of matrix A.
C     RWLINK  Row linked lists of matrix A.
C     CLNMBS  Column numbers of nonzeros in rows of matrix A.
C     LENCOL  Lengths of columns of matrix A.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     NZL     Number of nonzero entries in adjacency structure
C             A*Atransp.
C     AATPNT  Pointers to rows of  A*Atransp.
C
C
C     WORK ARRAYS:
C     MARKER  Array used to mark the rows of  A that are adjacent
C             to a given one.
C     TEMP    Array used to handle sparsity structure of rows.
C
C
C *** SUBROUTINES CALLED:
C     MYWRT
C
C
C *** PURPOSE:
C     This routine counts nonzeros of adjacency structure  A*Atransp.
C
C
C *** NOTES:
C     1. Two different rows  i and  j are said to be adjacent
C        if there exists a column in which they both have
C        a nonzero entry. The sparsity pattern array  AATPAT
C        contains then an entry  j in row  i (and, by symmetry,
C        an entry  i in row  j, if a square matrix is built).
C
C
C *** REFERENCES:
C     Duff I.S., Erisman A.M., Reid J.K. (1989). Direct methods
C        for sparse matrices, Clarendon Press, Oxford 1989,
C        chapters  2 and  10.
C     Gondzio J. (1993). Implementing Cholesky factorization for
C        interior point methods of linear programming, Optimization
C        27, pp. 121-140.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  May 19, 1992
C     Last modified: May 17, 1995
C
C
C
C *** BODY OF (CNTAAT) ***
C
C
C
C
C     Initialize for building the sparsity pattern of  A*Atransp.
C     Zero MARKER array (it will indicate rows adjacent
C     to a given one).
      DO 100 IROW=1,M
         MARKER(IROW)=0
  100 CONTINUE
C
C
C     Set the parameters controlling the progress
C     of building  AATPAT array.
C     LENAAT  is the current length of  AATPAT array.
C     LENROW  is the length of a given row of  AATPAT array.
      LENAAT=0
      LENROW=0
C
C
C
C
C
C     Main loop begins here (loop over rows of  A).
C     For every row  IROW, a list of other rows adjacent
C     to a given one is created. Row IROW itself is omitted.
      DO 500 IROW=1,M
         LENROW=0
         AATPNT(IROW)=LENAAT+1
         MARKER(IROW)=1
C
C
C     Scan row IROW. Every column that intersects it, indicates
C     rows adjacent to a given one. Every row that appears
C     for the first time may then be added to the temporary
C     list of rows adjacent to IROW.
         J=RWHEAD(IROW)
  200    IF(J.EQ.0) GO TO 350
         JCOL=CLNMBS(J)
C
C     Omit fixed columns.
         IF(STAVAR(JCOL).EQ.6) GO TO 300
         KBEG=CLPNTS(JCOL)
         KEND=KBEG+LENCOL(JCOL)-1
         DO 250 K=KBEG,KEND
            IR=RWNMBS(K)
            IF(MARKER(IR).EQ.1) GO TO 250
C
C     Omit adjacencies of linking and structural rows.
C           IF(MSPLIT(IR)+MSPLIT(IROW).GE.1) GO TO 250
C
C     Here if the row appears for the first time.
C     Add it to the adjacency list and mark the adjacent row.
            LENROW=LENROW+1
            TEMP(LENROW)=IR
            MARKER(IR)=1
  250    CONTINUE
  300    J=RWLINK(J)
         GO TO 200
C
C     Restore zero value of MARKER array.
  350    IF(LENROW.EQ.0) GO TO 450
         DO 400 K=1,LENROW
            LENAAT=LENAAT+1
            IR=TEMP(K)
            MARKER(IR)=0
  400    CONTINUE
C
C     Decide how much of the sparsity pattern of  A*Atransp
C     is required. If only upper triangle of it is to be built,
C     then all rows that have already been scanned should
C     be excluded from further search.
  450    IF(TRIANG.EQ.0) MARKER(IROW)=0
C
C
C
C
C
C     End of main loop.
  500 CONTINUE
      AATPNT(M+1)=LENAAT+1
C
C
C
C
C     Write problem statistics.
      NZL=LENAAT/2
      IF(TRIANG.EQ.1) NZL=LENAAT
      WRITE(BUFFER,501) NZL
  501 FORMAT(1X,'CNTAAT: A*Atransp will have ',I13,
     X  ' subdiagonal elts.')
      CALL MYWRT(IOLOG,BUFFER)
C
C
C
      RETURN
C
C
C
C *** LAST CARD OF (CNTAAT) ***
      END
