C**********************************************************
C     *** DTSRTA ... DUOBLE TRANSPOSE SORT OF MATRIX  A ***
C**********************************************************
C
      SUBROUTINE DTSRTA(MAXM,MAXN,MAXNZA,M,N,
     X ACOEFF,CLPNTS,RWNMBS,LENCOL,
     X RWHEAD,RWLINK,CLNMBS,
     X ACOPY,CPCOPY,STAVAR,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 MAXM,MAXN,MAXNZA,M,N,IOLOG
      DOUBLE PRECISION ACOEFF(MAXNZA),ACOPY(MAXNZA)
      INTEGER*4 CLPNTS(MAXN+1),CPCOPY(MAXN+1)
      INTEGER*4 RWHEAD(MAXM+1),RWLINK(MAXNZA)
C
C *** The following arrays can be half-length integer.
      INTEGER*2 RWNMBS(MAXNZA),CLNMBS(MAXNZA),LENCOL(MAXN)
      INTEGER*2 STAVAR(MAXN)
C
C
C *** LOCAL VARIABLES
      INTEGER*4 IROW,JCOL,K,KBEG,KEND,IPOS,NONZ
      CHARACTER*100 BUFFER
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     MAXN    Maximum column dimension of the LP constraint matrix.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     M       Number of rows of the LP constraint matrix.
C     N       Number of columns of the LP constraint matrix.
C     ACOEFF  Nonzero entries of an  LP constraint matrix.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A.
C     RWHEAD  Headers to the row linked lists of entries of matrix A.
C     RWLINK  Row linked lists of entries of matrix A.
C     CLNMBS  Column numbers of nonzeros in a given row of matrix A.
C     LENCOL  Lengths of columns of matrix A.
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C            -k  free variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicate the position of the original variable.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     Reordered  LP constraint matrix.
C     ACOEFF  Nonzero entries of an  LP constraint matrix.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A
C             (in an increasing order).
C     RWHEAD  Headers to the row linked lists of matrix A.
C     RWLINK  Row linked lists of matrix A.
C     CLNMBS  Column numbers of nonzeros in rows of matrix A.
C
C     WORK ARRAYS:
C     ACOPY   A copy of nonzero entries of an  LP constraint matrix.
C     CPCOPY  A copy of pointers to the beginning of columns of A.
C
C
C *** SUBROUTINES CALLED:
C     MYSTOP
C
C
C *** PURPOSE:
C     This routine implements a double transpose sort of matrix
C     A handled as a collection of sparse columns to obtain the 
C     increasing order of row numbers within each column.
C
C
C *** NOTES:
C     We assume that a row-wise access to matrix A is ensured via
C     RWHEAD, RWLINK and CLNMBS arrays on entry to this routine.
C     Hence we can omit the first transpose.
C
C
C *** REFERENCES:
C     Duff I.S., Erisman A.M., Reid J.K. (1989). Direct methods
C        for sparse matrices, Clarendon Press, Oxford 1989,
C        chapter 2.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  January 13, 1994
C     Last modified: June 9, 1995
C
C
C
C *** BODY OF (DTSRTA) ***
C
C
C
C     Determine the number of matrix nonzero entries.
      NONZ=CLPNTS(N+1)-1
C
C     Save copy of nonzero entries of A.
      DO 100 K=1,NONZ
         ACOPY(K)=ACOEFF(K)
  100 CONTINUE
C
C     Save copy of pointers to the beginning of columns of A.
      DO 200 JCOL=1,N+1
         CPCOPY(JCOL)=CLPNTS(JCOL)
  200 CONTINUE
C
C     Move the matrix back to a form of a collection of sparse
C     columns. Access rows in an increasing order.
      DO 500 IROW=1,M
         K=RWHEAD(IROW)
  400    IF(K.LE.0) GO TO 500
            JCOL=CLNMBS(K)
            IPOS=CPCOPY(JCOL)
            RWNMBS(IPOS)=IROW
            ACOEFF(IPOS)=ACOPY(K)
            CPCOPY(JCOL)=CPCOPY(JCOL)+1
         K=RWLINK(K)
         GO TO 400
  500 CONTINUE
C
C     Restore row linked lists for matrix A.
      DO 600 IROW=1,M
         RWHEAD(IROW)=0
  600 CONTINUE
      DO 800 JCOL=1,N
C
C     Omit all  FIXED variables.
         IF(STAVAR(JCOL).GE.6) GO TO 800
         KBEG=CLPNTS(JCOL)
         KEND=KBEG+LENCOL(JCOL)-1
         DO 700 K=KBEG,KEND
            IROW=RWNMBS(K)
            RWLINK(K)=RWHEAD(IROW)
            RWHEAD(IROW)=K
  700    CONTINUE
  800 CONTINUE
C
C *** DEBUGGING
C     DO 940 JCOL=1,N
C        IF(STAVAR(JCOL).GE.6) GO TO 940
C        KBEG=CLPNTS(JCOL)
C        KEND=KBEG+LENCOL(JCOL)-2
C        DO 920 K=KBEG,KEND
C           IF(RWNMBS(K+1).LE.RWNMBS(K)) THEN
C              WRITE(BUFFER,921) JCOL,K,RWNMBS(K),RWNMBS(K+1)
C 921          FORMAT(1X,'cl=',I6,' pos=',I6,' rw1=',I6,' rw2=',I6)
C              CALL MYWRT(IOLOG,BUFFER)
C              CALL MYSTOP(79)
C           ENDIF
C 920    CONTINUE
C 940 CONTINUE
C
      RETURN
C
C
C
C *** LAST CARD OF (DTSRTA) ***
      END
