C*******************************************************
C     *** ELCNST ... ELIMINATE REDUNDANT CONSTRAINTS ***
C*******************************************************
C
      SUBROUTINE ELCNST(IOLOG,MSGLEV,ITIGHT,
     X MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X LNHIST,MXHIST,INHIST,DPHIST,
     X RWORK,IWORK,RMAP,IMAP,IROW,RELT,
     X IMTMP1,INTMP1,INTMP2,RMTMP1,RNTMP1,
     X B,RANGES,C,LOBND,UPBND,BNDBIG,
     X ACOEFF,CLPNTS,RWNMBS,
     X RWHEAD,RWLINK,CLNMBS,LENCOL,
     X P,Q,PRLVAR,STAVAR,RWSTAT,STAROW,RWNAME,CLNAME,
     X PERM,INVP,LENROW)
C
C
C
C
C *** HIDDEN DATA STRUCTURES
      INTEGER*4 IWORK(*),IMAP(*),RMAP(*)
      DOUBLE PRECISION RWORK(*)
C
C *** HIDDEN DATA STRUCTURES DESCRIPTION
C     RWORK   Real array that contains real  LP problem data.
C     IWORK   Integer array that contains integer  LP problem data.
C     RMAP    Map of RWORK array.
C     IMAP    Map of IWORK array.
C
C
C
C
C *** PARAMETERS
      INTEGER*4 IOLOG,MSGLEV,ITIGHT,MAXM,MAXN,MAXNZA
      INTEGER*4 M,N,NSTRCT,LNHIST,MXHIST
      INTEGER*2 INHIST(MXHIST)
      DOUBLE PRECISION DPHIST(MXHIST)
      INTEGER*4 IROW(MAXN),IMTMP1(MAXM),INTMP1(MAXN)
      INTEGER*2 INTMP2(MAXN)
      DOUBLE PRECISION RELT(MAXN),RMTMP1(MAXM),RNTMP1(MAXN)
      DOUBLE PRECISION ACOEFF(MAXNZA),B(MAXM),RANGES(MAXM)
      DOUBLE PRECISION C(MAXN),LOBND(MAXN),UPBND(MAXN),BNDBIG
      INTEGER*4 CLPNTS(MAXN+1),RWHEAD(MAXM+1),RWLINK(MAXNZA)
      INTEGER*2 RWNMBS(MAXNZA),CLNMBS(MAXNZA),LENCOL(MAXN)
      DOUBLE PRECISION P(MAXM),Q(MAXM),PRLVAR(MAXN)
      CHARACTER*8 RWNAME(MAXM),CLNAME(MAXN)
      INTEGER*2 STAVAR(MAXN),STAROW(MAXM),RWSTAT(MAXM)
      INTEGER*2 INVP(MAXM),PERM(MAXM),LENROW(MAXM)
C
C
C
C *** LOCAL VARIABLES
      INTEGER*4 ROWLEN,LIMIT,NFIXED,NELIM,NTIGHT,NT0,M1,N1
      INTEGER*4 I,IKX,IPOS,IR,IRUN,J,K,KOK,KOUT
      INTEGER*4 KBEG,KEND,MNEW,KRWBEG,LNKUPD
      INTEGER*4 NNEG,NPOS,NNEGBG,NPOSBG,KNEGBG,KPOSBG
      DOUBLE PRECISION BIG,BIGNEW,X0,BNDNEW,BNDJUP,RHS0,RNRM
      DOUBLE PRECISION BLOWER,BUPPER,FSBTOL,BNDTOL,SMALLA
      CHARACTER*100 BUFFER
      CHARACTER*2   RTYPE
C
C
C
C
C *** PARAMETERS DESCRIPTION
C
C *** ON INPUT:
C     IOLOG   Output unit number where log messages are to be written.
C     MSGLEV  The level of PRE_SOLVE information desired:
C             0  only final problem dimensions are printed;
C             1  numbers of eliminated rows and columns are printed;
C             2  names of eliminated rows and columns are printed;
C             3  detailed DEBUGGING information is printed.
C     ITIGHT  A control of tightening UPPER bounds.
C             1  incomplete analysis (no tightening UPPER bounds);
C             2  maximum analysis possible.
C     MAXM    Maximum number of constraints.
C     MAXN    Maximum number of variables.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     M       Number of constraints.
C     N       Number of variables (total, i.e. including slacks,
C             surplus and artificials).
C     NSTRCT  Number of structural variables (excluding slacks,
C             surplus and artificials).
C     LNHIST  Length of the PRE_SOLVE history list.
C     MXHIST  Maximum number of entries in the PRE_SOLVE history list.
C     INHIST  Integer PRE_SOLVE history information.
C     DPHIST  Double precision PRE_SOLVE history information.
C     ACOEFF  Array of nonzero elements for each column.
C     B       Right hand side of the linear program.
C     RANGES  Array of constraint ranges.
C     C       Objective function coefficients.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWHEAD  Headers to the row linked lists of matrix A.
C     RWLINK  Row linked lists of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A.
C     CLNMBS  Column numbers of nonzeros in rows of matrix A.
C     LENCOL  Lengths of (sparse) columns of matrix A.
C     LOBND   Array of lower bounds.
C     UPBND   Array of upper bounds. Note that the upper bound
C             is changed when the variable has a lower bound. 
C     BNDBIG  Value of an unacceptably large implicit bound.
C     P       LOWER bounds on shadow prices (dual variables).
C     Q       UPPER bounds on shadow prices (dual variables).
C     PRLVAR  Primal variables of the linear program.
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C            -k  FREE variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicates the position of the original variable.
C     RWSTAT  Array of row types:
C             1  row type is = ;
C             2  row type is >= ;
C             3  row type is <= ;
C             4  objective row;
C             5  other free row.
C     STAROW  Array of row status:
C             0  row has been removed (it indicates a free row);
C             1  row has not been removed.
C     RWNAME  Array of row names (increasing order sort).
C     CLNAME  Array of column names (unordered).
C
C *** ON OUTPUT:
C
C
C
C
C *** WORK ARRAYS:
C     IROW and  RELT are the arrays for temporary handling of rows
C             and columns of the constraint matrix. They are primarily
C             intended to handle sparse vectors (in packed form)
C             but may also be used for storing dense ones.
C     IMTMP1  Integer work array of size MAXM.
C     INTMP1  Integer work array of size MAXN
C     INTMP2  Half-length integer work array of size MAXN.
C     RMTMP1  Double precision work array of size MAXM.
C     RNTMP1  Double precision work array of size MAXN.
C     PERM    Half-length integer work array of size MAXM.
C     INVP    Half-length integer work array of size MAXM.
C     LENROW  Half-length integer work array of size MAXM.
C
C
C
C
C *** PURPOSE
C     This routine computes bounds on the LP constraints and uses
C     them to eliminate redundant constraints. Next, it uses these
C     values to adjust variables' bounds. 
C
C
C
C *** SUBROUTINES CALLED
C     MYWRT,GETCOL,GETROW,DABS,EMPTYR,REORDA,REORDI,REORDV,MYSTOP
C
C
C *** NOTES
C     This routine is given direct access to the matrix A.
C     It alters hidden data structures.
C
C
C
C *** REFERENCES:
C     Gondzio J. (1994). Presolve analysis of linear programs prior 
C        to applying an interior point method, Technical Report 
C        No 1994.3, Department of Management Studies, University 
C        of Geneva, 102, Bd. Carl-Vogt, 1211 Geneva, Switzerland, 
C        February 1994, revised in December 1994.
C
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  May 17, 1993
C     Last modified: June 9, 1995
C
C
C
C
C *** BODY OF (ELCNST) ***
C
C
C
C     Initialize.
      BIG=1.0D+30
      BIGNEW=1.0D+20
      FSBTOL=5.0D-8
      BNDTOL=1.0D-5
      SMALLA=1.0D-8
      NFIXED=0
      NELIM=0
      NTIGHT=0
C
      IF(MSGLEV.LE.3) GO TO 140
      DO 130 J=1,N
         IF(STAVAR(J).LT.6) GO TO 130
         WRITE(BUFFER,131) J,LENCOL(J),STAVAR(J),
     X    LOBND(J),UPBND(J),PRLVAR(J)
  131    FORMAT(1X,'col=',I6,' ln=',I4,' st=',I6,' LO=',D10.3,
     X    ' UP=',D10.3,' X=',D10.3)
         CALL MYWRT(IOLOG,BUFFER)
  130 CONTINUE
  140 CONTINUE
C
C
C     Compute norms of the  LP constraints.
      DO 150 I=1,M
         RMTMP1(I)=1.0D-4+DABS(B(I))
  150 CONTINUE
      DO 180 J=1,NSTRCT
         KBEG=CLPNTS(J)
         KEND=KBEG+LENCOL(J)-1
         DO 160 K=KBEG,KEND
            I=RWNMBS(K)
            IF(DABS(ACOEFF(K)).GT.RMTMP1(I)) RMTMP1(I)=DABS(ACOEFF(K))
  160    CONTINUE
  180 CONTINUE
      DO 190 I=1,M
         RNRM=SMALLA*RMTMP1(I)
         IF(RNRM.LE.FSBTOL) RNRM=FSBTOL
         IF(RNRM.GE.1.0D-5) RNRM=1.0D-5
         IF(DABS(B(I)).LE.RNRM) B(I)=0.0D0
         IF(RMTMP1(I).LE.1.0E-15) GO TO 190
C        IF(RMTMP1(I).LE.1.0E-4) THEN
C           WRITE(BUFFER,191) I,RMTMP1(I)
C 191       FORMAT(1X,'   ELCNST: row=',I6,' has norm=',D10.3)
C           CALL MYWRT(IOLOG,BUFFER)
C        ENDIF
  190 CONTINUE
C
C
C
C
C
C
C
C     First main loop begins here.
C     Loop over all LP constraints.
C     Eliminate redundant constraints.
C     LNKUPD equal to 1 forces update of row linked lists.
      IRUN=1
      LNKUPD=0
  200 NT0=NTIGHT
      M1=M
      DO 1000 I=1,M
         ROWLEN=0
         KRWBEG=RWHEAD(I)
         IPOS=KRWBEG
         IF(RWSTAT(I).GE.2) THEN
            ROWLEN=1
            IPOS=RWLINK(KRWBEG)
         ENDIF
C
C     Compute LOWER and UPPER limits of the LP constraint.
C     Loop over nonzero entries of row I.
         BLOWER=0.0D0
         BUPPER=0.0D0
  300    IF(IPOS.EQ.0) GO TO 400
            ROWLEN=ROWLEN+1
            J=CLNMBS(IPOS)
            K=STAVAR(J)
            IF(K.GE.6) GO TO 350
            BNDJUP=BIG
            IF(K.EQ.1.OR.K.EQ.3) BNDJUP=UPBND(J)
            IF(ACOEFF(IPOS).LT.0.0D0) THEN
               BLOWER=BLOWER+BNDJUP*ACOEFF(IPOS)
            ELSE
               BUPPER=BUPPER+BNDJUP*ACOEFF(IPOS)
            ENDIF
  350    IPOS=RWLINK(IPOS)
         GO TO 300
C
  400    CONTINUE
         RNRM=SMALLA*(RMTMP1(I)+DABS(B(I)))
         IF(RNRM.LE.FSBTOL) RNRM=FSBTOL
         IF(RNRM.GE.1.0D-5) RNRM=1.0D-5
         IF(DABS(BLOWER-B(I)).LE.RNRM) BLOWER=B(I)
         IF(DABS(BUPPER-B(I)).LE.RNRM) BUPPER=B(I)
         LENROW(I)=ROWLEN
C        WRITE(BUFFER,401) I,RWNAME(I),ROWLEN,RWSTAT(I)
C 401    FORMAT(1X,'ELCNST: Row ',I6,' (name=',A8,
C    X    ') len=',I6,' RWSTAT=',I6)
C        CALL MYWRT(IOLOG,BUFFER)
C
C
C
C     The following notation is used below:
C     LIMIT =  0  corresponds to FORCING row (RHS = BLOWER);
C     LIMIT =  1  corresponds to FORCING row (RHS = BUPPER);
C     LIMIT = -1  corresponds to REDUNDANT row.
         IF(RWSTAT(I).EQ.1) THEN
C
C     Here for EQUALITY type constraint.
            RTYPE='EQ'
            IF(BLOWER-B(I).GT.-FSBTOL) THEN
               IF(BLOWER-B(I).GT.FSBTOL) GO TO 9010
               LIMIT=0
               GO TO 500
            ENDIF
            IF(BUPPER-B(I).LT.FSBTOL) THEN
               IF(BUPPER-B(I).LT.-FSBTOL) GO TO 9010
               LIMIT=1
               GO TO 500
            ENDIF
            GO TO 1000
         ENDIF
C
C
C
         IF(RWSTAT(I).EQ.2) THEN
C
C     Here for GREATER OR EQUAL type constraint.
            RTYPE='GE'
            IF(BUPPER-B(I).LT.FSBTOL) THEN
               IF(BUPPER-B(I).LT.-FSBTOL) GO TO 9010
               LIMIT=1
               GO TO 500
            ENDIF
            IF(BLOWER-B(I).GT.-FSBTOL) THEN
C
C     Ranged row need to satify one more condition to be eliminated.
               IF(RANGES(I).LE.BIGNEW) THEN
                  IF(BUPPER.GT.B(I)+RANGES(I)+FSBTOL) GO TO 1000
               ENDIF
               LIMIT=-1
               GO TO 500
            ENDIF
         ENDIF
C
C
C
         IF(RWSTAT(I).EQ.3) THEN
C
C     Here for LESS OR EQUAL type constraint.
            RTYPE='LE'
            IF(BLOWER-B(I).GT.-FSBTOL) THEN
               IF(BLOWER-B(I).GT.FSBTOL) GO TO 9010
               LIMIT=0
               GO TO 500
            ENDIF
            IF(BUPPER-B(I).LT.FSBTOL) THEN
C
C     Ranged row need to satify one more condition to be eliminated.
               IF(RANGES(I).LE.BIGNEW) THEN
                  IF(BLOWER.LT.B(I)-RANGES(I)-FSBTOL) GO TO 1000
               ENDIF
               LIMIT=-1
               GO TO 500
            ENDIF
         ENDIF
C
C
C
         GO TO 1000
C
C
C
C     Here to eliminate the LP constraint.
  500    NELIM=NELIM+1
         RWHEAD(I)=-RWHEAD(I)
C
C *** DEBUGGING
         IF(MSGLEV.LE.1) GO TO 505
         WRITE(BUFFER,501) I,RWNAME(I),RTYPE
  501    FORMAT(1X,'ELCNST: Row      ',I6,' (name=',A8,
     X    ' type=',A2,') is eliminated.')
         CALL MYWRT(IOLOG,BUFFER)
         IF(MSGLEV.LE.2) GO TO 505
         WRITE(BUFFER,502) I,ROWLEN,RTYPE,BLOWER,BUPPER,B(I)
  502    FORMAT(1X,'row=',I6,' ln=',I6,' type=',A2,
     X    ' BLO=',D10.3,' BUP=',D10.3,' RHS=',D10.3)
         CALL MYWRT(IOLOG,BUFFER)
C        IF(RANGES(I).GE.BIGNEW) GO TO 505
         WRITE(BUFFER,503) I,RWNAME(I),RANGES(I)
  503    FORMAT(1X,'ELCNST: Row      ',I6,' name=',A8,' range=',D12.6)
         CALL MYWRT(IOLOG,BUFFER)
  505    CONTINUE
C
         IF(LIMIT.EQ.-1) THEN
            NFIXED=NFIXED+1
            J=CLNMBS(KRWBEG)
            STAVAR(J)=14
            PRLVAR(J)=0.0D0
            GO TO 1000
         ENDIF
C
C     Here to eliminate the constraint and fix variables.
C     Loop over nonzero entries of row I.
         IPOS=KRWBEG
  600    IF(IPOS.EQ.0) GO TO 800
            J=CLNMBS(IPOS)
            IF(STAVAR(J).GE.6) GO TO 750
            IF(ACOEFF(IPOS).LT.0.0D0) THEN
               IF(LIMIT.EQ.0) X0=UPBND(J)
               IF(LIMIT.EQ.1) X0=0.0D0
            ELSE
               IF(LIMIT.EQ.0) X0=0.0D0
               IF(LIMIT.EQ.1) X0=UPBND(J)
            ENDIF
C
C     Fix and eliminate column J. Omit already FIXED variables.
C     Update RHS array.
            NFIXED=NFIXED+1
            PRLVAR(J)=X0
            KBEG=CLPNTS(J)
            KEND=KBEG+LENCOL(J)-1
            IF(STAVAR(J).NE.1.AND.STAVAR(J).NE.3) THEN
C
C     Reinitialize bounds on shadow prices.
C              WRITE(BUFFER,701) J1,CLNAME(J1),STAVAR(J1)
C 701          FORMAT(1X,'ELCNST: Variable ',I6,' (name=',A8,
C    X          ' st=',I6,')')
C              CALL MYWRT(IOLOG,BUFFER)
               DO 700 K=KBEG,KEND
                  IR=RWNMBS(K)
                  P(IR)=-BIG
                  Q(IR)=BIG
                  IF(RANGES(IR).LE.BIGNEW) GO TO 700
                  IF(RWSTAT(IR).EQ.2) P(IR)=0.0D0
                  IF(RWSTAT(IR).EQ.3) Q(IR)=0.0D0
  700          CONTINUE
            ENDIF
            IF(DABS(X0).LE.FSBTOL) X0=0.0D0
            STAVAR(J)=6
            IF(J.GT.NSTRCT) THEN 
               STAVAR(J)=14
               PRLVAR(J)=0.0D0
               GO TO 720
            ENDIF
            IF(DABS(X0).LE.FSBTOL) GO TO 720
            DO 710 K=KBEG,KEND
               IR=RWNMBS(K)
               B(IR)=B(IR)-X0*ACOEFF(K)
               IF(DABS(B(IR)).LE.SMALLA) B(IR)=0.0D0
  710       CONTINUE
  720       CONTINUE
            IF(MSGLEV.LE.1) GO TO 722
            WRITE(BUFFER,721) J,CLNAME(J),X0
  721       FORMAT(1X,'ELCNST: Variable ',I6,' (name=',A8,
     X       ') is being FIXED on X=',D14.6)
            CALL MYWRT(IOLOG,BUFFER)
  722       CONTINUE
C
  750       IPOS=RWLINK(IPOS)
            GO TO 600
C
  800    CONTINUE
C
C
C
C
C
C     End of the first main loop.
 1000 CONTINUE
      N1=NFIXED
C
C
C
C
C
C
C     Determine the permutation that puts all empty and inactive
C     rows at the end of the list.
C
      I=3
      IF(MSGLEV.LE.1) I=4
      CALL EMPTYR(MAXM,M,MNEW,I,
     X RWHEAD,STAROW,PERM,INVP,IOLOG)
C
C
C     Reorder the rows of the  LP constraint matrix according to
C     the permutation resulting from the analysis of EMPTYR.
      IF(MNEW.LT.M) THEN
C
         CALL REORDA(MAXM,MAXN,MAXNZA,M,N,
     X    CLPNTS,RWNMBS,
     X    RWHEAD,RWLINK,CLNMBS,LENCOL,
     X    PERM,INVP,IMTMP1,IROW,RELT,
     X    RWNAME,STAROW,RWSTAT,RANGES,B,IOLOG)
C
C     Reorder bounds on shadow prices P and Q.
         CALL REORDV(MAXM,M,
     X    PERM,INVP,P,RELT,IOLOG)
         CALL REORDV(MAXM,M,
     X    PERM,INVP,Q,RELT,IOLOG)
C
C     Reorder elements within each column of the  LP constraint
C     matrix in such a way that those of the active part of  A
C     are at the beginning of the lists. The column lengths will
C     later be decreased to forget inactive part of matrix  A.
C     Set the new row linked lists of nonzero elements of matrix  A.
C     Recall that  CLPNTS(j) indicates the first entry of column j.
C     Zero  RWHEAD and LENROW arrays.
         DO 1200 I=1,M
            RWHEAD(I)=0
            LENROW(I)=0
 1200    CONTINUE
C
C     Reorder nonzero elements within each column.
         DO 1500 J=1,N
            IF(STAVAR(J).GE.6) GO TO 1500
            KBEG=CLPNTS(J)-1
            KOK=0
            KOUT=0
            DO 1300 IKX=1,LENCOL(J)
               K=KBEG+IKX
               I=RWNMBS(K)
               IF(I.LE.MNEW) THEN
                  KOK=KOK+1
                  IROW(KOK)=RWNMBS(K)
                  RELT(KOK)=ACOEFF(K)
               ELSE
                  IPOS=LENCOL(J)-KOUT
                  KOUT=KOUT+1
                  IROW(IPOS)=RWNMBS(K)
                  RELT(IPOS)=ACOEFF(K)
               ENDIF
 1300       CONTINUE
            LENCOL(J)=KOK
C
C     Set the row linked lists.
C     Count nonzero elements in all rows of  A.
            DO 1400 IKX=1,LENCOL(J)
               K=KBEG+IKX
               I=IROW(IKX)
               RWNMBS(K)=I
               ACOEFF(K)=RELT(IKX)
               RWLINK(K)=RWHEAD(I)
               RWHEAD(I)=K
               LENROW(I)=LENROW(I)+1
 1400       CONTINUE
 1500    CONTINUE
C
C     Set the new number of rows of the constraint matrix.
C     Observe that row linked lists are OK.
         M=MNEW
         LNKUPD=0
C
      ENDIF
C
C
C     Remove numerical errors frm RHS.
      DO 1600 I=1,M
         IF(DABS(B(I)).LE.FSBTOL) B(I)=0.0D0
 1600 CONTINUE
C
C
C
C
C
C
C     Second main loop begins here.
C     Loop over all LP constraints.
C     Tighten bounds on variables.
      DO 3000 I=1,M
         RHS0=B(I)
         RTYPE='EQ'
         IF(RWSTAT(I).EQ.2) RTYPE='GE'
         IF(RWSTAT(I).EQ.3) RTYPE='LE'
C
C     Compute LOWER and UPPER limits of the LP constraint.
         KRWBEG=RWHEAD(I)
         IF(RWSTAT(I).GE.2) KRWBEG=RWLINK(KRWBEG)
         BLOWER=0.0D0
         BUPPER=0.0D0
         NPOS=0
         NNEG=0
         NPOSBG=0
         NNEGBG=0
C
C     Loop over nonzero entries of row I.
         IPOS=KRWBEG
 2100    IF(IPOS.EQ.0) GO TO 2200
            J=CLNMBS(IPOS)
            K=STAVAR(J)
            IF(K.GE.6) GO TO 2150
            BNDJUP=BIG
            IF(K.EQ.1.OR.K.EQ.3) BNDJUP=UPBND(J)
            IF(ACOEFF(IPOS).LT.0.0D0) THEN
               NNEG=NNEG+1
               IF(BNDJUP.GT.BIGNEW) THEN
                  NNEGBG=NNEGBG+1
                  KNEGBG=IPOS
               ELSE
                  BLOWER=BLOWER+BNDJUP*ACOEFF(IPOS)
               ENDIF
            ELSE
               NPOS=NPOS+1
               IF(BNDJUP.GT.BIGNEW) THEN 
                  NPOSBG=NPOSBG+1
                  KPOSBG=IPOS
               ELSE
                  BUPPER=BUPPER+BNDJUP*ACOEFF(IPOS)
               ENDIF
            ENDIF
 2150       IPOS=RWLINK(IPOS)
            GO TO 2100
C
 2200    CONTINUE
         RNRM=SMALLA*(RMTMP1(I)+DABS(B(I)))
         IF(RNRM.LE.FSBTOL) RNRM=FSBTOL
         IF(RNRM.GE.1.0D-5) RNRM=1.0D-5
         IF(DABS(BLOWER-B(I)).LE.RNRM) BLOWER=B(I)
         IF(DABS(BUPPER-B(I)).LE.RNRM) BUPPER=B(I)
C
C *** DEBUGGING
         IF(MSGLEV.LE.2) GO TO 2210
         WRITE(BUFFER,2201) I,RTYPE,BLOWER,BUPPER,B(I)
 2201    FORMAT(1X,'Row=',I6,' type=',A2,
     X    ' BLO=',D10.3,' BUP=',D10.3,' RHS=',D10.3)
         CALL MYWRT(IOLOG,BUFFER)
         WRITE(BUFFER,2202) NPOS,NNEG,NPOSBG,NNEGBG
 2202    FORMAT(1X,' Npos=',I5,' Nneg=',I5,
     X    '   Nposbg=',I5,' Nnegbg=',I5)
         CALL MYWRT(IOLOG,BUFFER)
         IF(RANGES(I).GE.BIGNEW) GO TO 2210
         WRITE(BUFFER,2203) I,RWNAME(I),RANGES(I)
 2203    FORMAT(1X,'ELCNST: Row      ',I6,' name=',A8,' range=',D12.6)
         CALL MYWRT(IOLOG,BUFFER)
 2210    CONTINUE
C
C
C
         IF(RWSTAT(I).EQ.1.OR.RWSTAT(I).EQ.3) THEN
C
C     Here for EQUALITY type or LESS OR EQUAL type constraint.
            IF(BLOWER-RHS0.GT.FSBTOL.AND.NNEGBG.EQ.0) GO TO 9010
            IF(NNEGBG.GE.1) GO TO 2400
C
C
C     Here if there are no negative entries with BIG Uj.
C     Loop over nonzero entries of row I.
            BLOWER=RHS0-BLOWER
            IPOS=KRWBEG
 2300       IF(IPOS.EQ.0) GO TO 2380
               J=CLNMBS(IPOS)
               K=STAVAR(J)
               IF(K.GE.6) GO TO 2360
               IF(ACOEFF(IPOS).GT.0.0D0) THEN
C
C     Implicit UPPER bound can be defined for each variable
C     refering to POSITIVE entry of row I. Ignore large bound.
                  BNDJUP=BIG
                  IF(K.EQ.1.OR.K.EQ.3) BNDJUP=UPBND(J)
                  BNDNEW=BLOWER/ACOEFF(IPOS)
                  IF(BNDNEW.GE.BNDJUP-BNDTOL) GO TO 2360
C                 IF(BNDNEW.GE.BNDBIG) GO TO 2360
                  IF(BNDNEW.GE.BNDBIG) THEN 
                     IF(K.EQ.0.OR.K.EQ.2) GO TO 2360
                  ENDIF
                  IF(ITIGHT.LE.1) GO TO 2360
                  NTIGHT=NTIGHT+1
                  UPBND(J)=BNDNEW
                  IF(MSGLEV.LE.2) GO TO 2304
                  WRITE(BUFFER,2301) J,STAVAR(J),BNDJUP,BNDNEW
 2301             FORMAT(1X,'cl=',I6,' st=',I6,' oldUPj=',D16.8,
     X             ' newUPj=',D16.8)
                  CALL MYWRT(IOLOG,BUFFER)
                  WRITE(BUFFER,2303) J,CLNAME(J),BNDNEW
 2303             FORMAT(1X,'ELCNST: Variable ',I6,' (name=',A8,
     X             ') has new UPPER bound=',D14.6)
                  CALL MYWRT(IOLOG,BUFFER)
 2304             CONTINUE
C
C     Reinitialize bounds on shadow prices.
                  KBEG=CLPNTS(J)
                  KEND=KBEG+LENCOL(J)-1
                  DO 2310 IKX=KBEG,KEND
                     IR=RWNMBS(IKX)
                     P(IR)=-BIG
                     Q(IR)=BIG
                     IF(RANGES(IR).LE.BIGNEW) GO TO 2310
                     IF(RWSTAT(IR).EQ.2) P(IR)=0.0D0
                     IF(RWSTAT(IR).EQ.3) Q(IR)=0.0D0
 2310             CONTINUE
C
                  IF(BNDNEW.LE.FSBTOL) THEN
C
C     Fix variable J on its LOWER bound.
                     NFIXED=NFIXED+1
                     X0=0.0D0
                     PRLVAR(J)=X0
                     STAVAR(J)=6
                     IF(MSGLEV.LE.1) GO TO 2312
                     WRITE(BUFFER,2311) J,CLNAME(J),X0
 2311                FORMAT(1X,'ELCNST: Variable ',I6,' (name=',A8,
     X                ') is being FIXED on X=',D14.6)
C                    CALL MYWRT(0,BUFFER)
                     CALL MYWRT(IOLOG,BUFFER)
 2312                CONTINUE
                     GO TO 2360
                  ENDIF
                  IF(K.EQ.1.OR.K.EQ.3) GO TO 2360
                  STAVAR(J)=STAVAR(J)+1
C
               ELSE
C
C     Implicit LOWER bound can be defined for each variable
C     refering to NEGATIVE entry of row I.
                  BNDJUP=UPBND(J)
                  BNDNEW=BNDJUP+BLOWER/ACOEFF(IPOS)
                  IF(BNDNEW.LE.BNDTOL) GO TO 2360
                  NTIGHT=NTIGHT+1
                  LOBND(J)=LOBND(J)+BNDNEW
                  UPBND(J)=UPBND(J)-BNDNEW
                  STAVAR(J)=3
C
C     Save the new LOWER bound in a PRE_SOLVE history list.
                  IF(LNHIST.GE.MXHIST) GO TO 9200
                  LNHIST=LNHIST+1
                  INHIST(LNHIST)=-J
                  DPHIST(LNHIST)=BNDNEW
C
C     Modify RHS (take account of the new LOWER bound on Xj).
                  KBEG=CLPNTS(J)
                  KEND=KBEG+LENCOL(J)-1
                  DO 2340 IKX=KBEG,KEND
                     IR=RWNMBS(IKX)
                     IF(IR.GT.0) THEN
                        B(IR)=B(IR)-BNDNEW*ACOEFF(IKX)
                        IF(DABS(B(IR)).LE.SMALLA) B(IR)=0.0D0
                     ENDIF
 2340             CONTINUE
C
C *** DEBUGGING
                  IF(MSGLEV.LE.2) GO TO 2342
                  WRITE(BUFFER,2341) J,STAVAR(J),BNDNEW
 2341             FORMAT(1X,'cl=',I6,' st=',I6,' oldLOj=0.0D0',
     X             ' newLOj=',D16.8)
                  CALL MYWRT(IOLOG,BUFFER)
 2342             IF(MSGLEV.LE.2) GO TO 2344
                  WRITE(BUFFER,2343) J,CLNAME(J),BNDNEW
 2343             FORMAT(1X,'ELCNST: Variable ',I6,' (name=',A8,
     X             ') has new LOWER bound=',D14.6)
                  CALL MYWRT(IOLOG,BUFFER)
 2344             CONTINUE
C
               ENDIF
 2360          IPOS=RWLINK(IPOS)
               GO TO 2300
C
 2380       CONTINUE
            BLOWER=RHS0-BLOWER
            GO TO 2500
C
C
C     Here if there exist negative entries with BIG Uj.
C     If only one variable has big UPPER bound, then its LOWER
C     bound can be improved. KNEGBG indicates its position.
 2400       IF(NNEGBG.GE.2) GO TO 2500
            J=CLNMBS(KNEGBG)
            K=STAVAR(J)
            IF(K.GE.6) GO TO 2500
            IF(K.LT.0) GO TO 2500
            BNDNEW=(RHS0-BLOWER)/ACOEFF(KNEGBG)
            IF(BNDNEW.LE.BNDTOL) GO TO 2500
            NTIGHT=NTIGHT+1
            LOBND(J)=LOBND(J)+BNDNEW
            STAVAR(J)=2
C
C     Save the new LOWER bound in a PRE_SOLVE history list.
            IF(LNHIST.GE.MXHIST) GO TO 9200
            LNHIST=LNHIST+1
            INHIST(LNHIST)=-J
            DPHIST(LNHIST)=BNDNEW
C
C     Modify RHS (take account of the new LOWER bound on Xj).
            KBEG=CLPNTS(J)
            KEND=KBEG+LENCOL(J)-1
            DO 2420 IKX=KBEG,KEND
               IR=RWNMBS(IKX)
               IF(IR.GT.0) THEN
                  B(IR)=B(IR)-BNDNEW*ACOEFF(IKX)
                  IF(DABS(B(IR)).LE.SMALLA) B(IR)=0.0D0
               ENDIF
 2420       CONTINUE
C
C *** DEBUGGING
            IF(MSGLEV.LE.2) GO TO 2422
            WRITE(BUFFER,2421) J,STAVAR(J),BNDNEW
 2421       FORMAT(1X,'5 BIG Uj, cl=',I6,' st=',I6,' oldLOj=0.0D0',
     X       ' newLOj=',D16.8)
            CALL MYWRT(IOLOG,BUFFER)
 2422       IF(MSGLEV.LE.2) GO TO 2424
            WRITE(BUFFER,2423) J,CLNAME(J),BNDNEW
 2423       FORMAT(1X,'ELCNST: Variable ',I6,' (name=',A8,
     X       ') has new LOWER bound=',D14.6)
            CALL MYWRT(IOLOG,BUFFER)
 2424       CONTINUE
C
         ENDIF
C
C
C
 2500    CONTINUE
         IF(RWSTAT(I).EQ.1.OR.RWSTAT(I).EQ.2) THEN
C
C     Here for EQUALITY type or GREATER OR EQUAL type constraint.
            IF(BUPPER-RHS0.LT.-FSBTOL.AND.NPOSBG.EQ.0) GO TO 9010
            IF(NPOSBG.GE.1) GO TO 2700
C
C
C     Here if there are no positive entries with BIG Uj.
C     Loop over nonzero entries of row I.
            BUPPER=RHS0-BUPPER
            IPOS=KRWBEG
 2600       IF(IPOS.EQ.0) GO TO 2680
               J=CLNMBS(IPOS)
               K=STAVAR(J)
               IF(K.GE.6) GO TO 2660
               IF(ACOEFF(IPOS).LT.0.0D0) THEN
C
C     Implicit UPPER bound can be defined for each variable
C     refering to NEGATIVE entry of row I. Ignore large bound.
                  BNDJUP=BIG
                  IF(K.EQ.1.OR.K.EQ.3) BNDJUP=UPBND(J)
                  BNDNEW=BUPPER/ACOEFF(IPOS)
                  IF(BNDNEW.GE.BNDJUP-BNDTOL) GO TO 2660
C                 IF(BNDNEW.GE.BNDBIG) GO TO 2660
                  IF(BNDNEW.GE.BNDBIG) THEN 
                     IF(K.EQ.0.OR.K.EQ.2) GO TO 2660
                  ENDIF
                  IF(ITIGHT.LE.1) GO TO 2660
                  NTIGHT=NTIGHT+1
                  UPBND(J)=BNDNEW
                  IF(MSGLEV.LE.2) GO TO 2604
                  WRITE(BUFFER,2601) J,STAVAR(J),BNDJUP,BNDNEW
 2601             FORMAT(1X,'cl=',I6,' st=',I6,' oldUPj=',D16.8,
     X             ' newUPj=',D16.8)
                  CALL MYWRT(IOLOG,BUFFER)
                  WRITE(BUFFER,2603) J,CLNAME(J),BNDNEW
 2603             FORMAT(1X,'ELCNST: Variable ',I6,' (name=',A8,
     X             ') has new UPPER bound=',D14.6)
                  CALL MYWRT(IOLOG,BUFFER)
 2604             CONTINUE
C
C     Reinitialize bounds on shadow prices.
                  KBEG=CLPNTS(J)
                  KEND=KBEG+LENCOL(J)-1
                  DO 2610 IKX=KBEG,KEND
                     IR=RWNMBS(IKX)
                     P(IR)=-BIG
                     Q(IR)=BIG
                     IF(RANGES(IR).LE.BIGNEW) GO TO 2610
                     IF(RWSTAT(IR).EQ.2) P(IR)=0.0D0
                     IF(RWSTAT(IR).EQ.3) Q(IR)=0.0D0
 2610             CONTINUE
C
                  IF(BNDNEW.LE.FSBTOL) THEN
C
C     Fix variable J on its LOWER bound.
                     NFIXED=NFIXED+1
                     X0=0.0D0
                     PRLVAR(J)=X0
                     STAVAR(J)=6
                     IF(MSGLEV.LE.1) GO TO 2612
                     WRITE(BUFFER,2611) J,CLNAME(J),X0
 2611                FORMAT(1X,'ELCNST: Variable ',I6,' (name=',A8,
     X                ') is being FIXED on X=',D14.6)
C                    CALL MYWRT(0,BUFFER)
                     CALL MYWRT(IOLOG,BUFFER)
 2612                CONTINUE
                     GO TO 2660
                  ENDIF
                  IF(K.EQ.1.OR.K.EQ.3) GO TO 2660
                  STAVAR(J)=STAVAR(J)+1
C
               ELSE
C
C     Implicit LOWER bound can be defined for each variable
C     refering to POSITIVE entry of row I.
                  BNDJUP=UPBND(J)
                  BNDNEW=BNDJUP+BUPPER/ACOEFF(IPOS)
                  IF(BNDNEW.LE.BNDTOL) GO TO 2660
                  NTIGHT=NTIGHT+1
                  LOBND(J)=LOBND(J)+BNDNEW
                  UPBND(J)=UPBND(J)-BNDNEW
                  STAVAR(J)=3
C
C     Save the new LOWER bound in a PRE_SOLVE history list.
                  IF(LNHIST.GE.MXHIST) GO TO 9200
                  LNHIST=LNHIST+1
                  INHIST(LNHIST)=-J
                  DPHIST(LNHIST)=BNDNEW
C
C     Modify RHS (take account of the new LOWER bound on Xj).
                  KBEG=CLPNTS(J)
                  KEND=KBEG+LENCOL(J)-1
                  DO 2640 IKX=KBEG,KEND
                     IR=RWNMBS(IKX)
                     IF(IR.GT.0) THEN 
                        B(IR)=B(IR)-BNDNEW*ACOEFF(IKX)
                        IF(DABS(B(IR)).LE.SMALLA) B(IR)=0.0D0
                     ENDIF
 2640             CONTINUE
C
C *** DEBUGGING
                  IF(MSGLEV.LE.2) GO TO 2642
                  WRITE(BUFFER,2641) J,STAVAR(J),BNDNEW
 2641             FORMAT(1X,'cl=',I6,' st=',I6,' oldLOj=0.0D0',
     X             ' newLOj=',D16.8)
                  CALL MYWRT(IOLOG,BUFFER)
 2642             IF(MSGLEV.LE.2) GO TO 2644
                  WRITE(BUFFER,2643) J,CLNAME(J),BNDNEW
 2643             FORMAT(1X,'ELCNST: Variable ',I6,' (name=',A8,
     X             ') has new LOWER bound=',D14.6)
                  CALL MYWRT(IOLOG,BUFFER)
 2644             CONTINUE
C
               ENDIF
 2660          IPOS=RWLINK(IPOS)
               GO TO 2600
C
 2680       CONTINUE
            BUPPER=RHS0-BUPPER
            GO TO 3000
C
C
C     Here if there exist positive entries with BIG Uj.
C     If only one variable has big UPPER bound, then its LOWER
C     bound can be improved. KPOSBG indicates its position.
 2700       IF(NPOSBG.GE.2) GO TO 3000
            J=CLNMBS(KPOSBG)
            K=STAVAR(J)
            IF(K.GE.6) GO TO 3000
            IF(K.LT.0) GO TO 3000
            BNDNEW=(RHS0-BUPPER)/ACOEFF(KPOSBG)
            IF(BNDNEW.LE.BNDTOL) GO TO 3000
            NTIGHT=NTIGHT+1
            LOBND(J)=LOBND(J)+BNDNEW
            STAVAR(J)=2
C
C     Save the new LOWER bound in a PRE_SOLVE history list.
            IF(LNHIST.GE.MXHIST) GO TO 9200
            LNHIST=LNHIST+1
            INHIST(LNHIST)=-J
            DPHIST(LNHIST)=BNDNEW
C
C     Modify RHS (take account of the new LOWER bound on Xj).
            KBEG=CLPNTS(J)
            KEND=KBEG+LENCOL(J)-1
            DO 2720 IKX=KBEG,KEND
               IR=RWNMBS(IKX)
               IF(IR.GT.0) THEN
                  B(IR)=B(IR)-BNDNEW*ACOEFF(IKX)
                  IF(DABS(B(IR)).LE.SMALLA) B(IR)=0.0D0
               ENDIF
 2720       CONTINUE
C
C *** DEBUGGING
            IF(MSGLEV.LE.2) GO TO 2722
            WRITE(BUFFER,2721) J,STAVAR(J),BNDNEW
 2721       FORMAT(1X,'8 BIG Uj, cl=',I6,' st=',I6,' oldLOj=0.0D0',
     X       ' newLOj=',D16.8)
            CALL MYWRT(IOLOG,BUFFER)
 2722       IF(MSGLEV.LE.2) GO TO 2724
            WRITE(BUFFER,2723) J,CLNAME(J),BNDNEW
 2723       FORMAT(1X,'ELCNST: Variable ',I6,' (name=',A8,
     X       ') has new LOWER bound=',D14.6)
            CALL MYWRT(IOLOG,BUFFER)
 2724       CONTINUE
C
         ENDIF
C
C
C
C
C
C     End of the second main loop.
 3000 CONTINUE
      IF(NFIXED.GT.N1) LNKUPD=1
C
C
C
C     Check if 1000 and 3000 loops should be repeated.
      IF(M.LT.M1) THEN
         IRUN=IRUN+1
         GO TO 200
      ENDIF
      IF(100*(NTIGHT-NT0).GE.N) THEN
         IRUN=IRUN+1
         IF(IRUN.GE.6) GO TO 3100
         GO TO 200
      ELSE
         GO TO 3100
      ENDIF
C
C
C
C
C
 3100 CONTINUE
      IF(LNKUPD.EQ.1) THEN
C
C     Restore linked lists of rows of A (new FIXED variables
C     have to be removed). Zero RWHEAD and LENROW arrays.
         DO 3200 I=1,M
            RWHEAD(I)=0
            LENROW(I)=0
 3200    CONTINUE
C
C     Set the row linked lists.
         DO 3400 J=1,N
            IF(STAVAR(J).GE.6) GO TO 3400
            KBEG=CLPNTS(J)-1
            DO 3300 IKX=1,LENCOL(J)
               K=KBEG+IKX
               I=RWNMBS(K)
               RWLINK(K)=RWHEAD(I)
               RWHEAD(I)=K
               LENROW(I)=LENROW(I)+1
 3300       CONTINUE
 3400    CONTINUE
         N1=NFIXED
         LNKUPD=0
      ENDIF
C
C
C     Check if there are inequality type rows to be eliminated.
C     Check if the eliminated rows were not violated.
      FSBTOL=1.0D-7
      DO 3500 I=1,M
         IF(DABS(B(I)).LE.FSBTOL) B(I)=0.0D0
         K=RWHEAD(I)
C        WRITE(BUFFER,3501) I,RWSTAT(I),LENROW(I),K
C3501    FORMAT(1X,'row=',I6,'  st=',I2,'  ln=',I6,'  K=',I8)
C        CALL MYWRT(IOLOG,BUFFER)
         IF(RWSTAT(I).EQ.1) THEN
C
C     Here for EQUALITY constraint.
            IF(K.GT.0) GO TO 3500
            IF(DABS(B(I)).GT.FSBTOL) GO TO 9030
            GO TO 3500
         ENDIF
         IF(LENROW(I).GE.2) GO TO 3500
C        WRITE(BUFFER,3502) I,RWSTAT(I),LENROW(I),K,CLNMBS(K)
C3502    FORMAT(1X,'rw=',I6,' st=',I2,' ln=',I6,' hd=',I8,' cl=',I6)
C        CALL MYWRT(IOLOG,BUFFER)
         IF(RWSTAT(I).EQ.2) THEN
C
C     Here for GREATER OR EQUAL type constraint.
            IF(B(I).GT.FSBTOL) GO TO 9030
            IF(K.LE.0) GO TO 3500
            NELIM=NELIM+1
            RWHEAD(I)=-RWHEAD(I)
            J=CLNMBS(K)
            STAVAR(J)=14
            NFIXED=NFIXED+1
            GO TO 3500
         ENDIF
         IF(RWSTAT(I).EQ.3) THEN
C
C     Here for LESS OR EQUAL type constraint.
            IF(B(I).LT.-FSBTOL) GO TO 9030
            IF(K.LE.0) GO TO 3500
            NELIM=NELIM+1
            RWHEAD(I)=-RWHEAD(I)
            J=CLNMBS(K)
            STAVAR(J)=14
            NFIXED=NFIXED+1
            GO TO 3500
         ENDIF
 3500 CONTINUE
C
C
C
      IF(NFIXED.GT.N1) THEN
C
C     Determine the permutation that puts all empty and inactive
C     rows at the end of the list.
C
         IRUN=3
         IF(MSGLEV.LE.1) IRUN=4
         CALL EMPTYR(MAXM,M,MNEW,IRUN,
     X    RWHEAD,STAROW,PERM,INVP,IOLOG)
C
C
C     Reorder the rows of the  LP constraint matrix according to
C     the permutation resulting from the analysis of EMPTYR.
C     Reorder LENROW array accordingly.
         IF(MNEW.LT.M) THEN
C
            CALL REORDA(MAXM,MAXN,MAXNZA,M,N,
     X       CLPNTS,RWNMBS,
     X       RWHEAD,RWLINK,CLNMBS,LENCOL,
     X       PERM,INVP,IMTMP1,IROW,RELT,
     X       RWNAME,STAROW,RWSTAT,RANGES,B,IOLOG)
C
C     Reorder bounds on shadow prices P and Q.
            CALL REORDV(MAXM,M,
     X       PERM,INVP,P,RELT,IOLOG)
            CALL REORDV(MAXM,M,
     X       PERM,INVP,Q,RELT,IOLOG)
C
C     Reorder LENROW array.
            CALL REORDI(MAXM,M,
     X       PERM,INVP,LENROW,IMTMP1(1),IOLOG)
C
C     Set the new number of rows of the constraint matrix.
            M=MNEW
C
         ENDIF
      ENDIF
C
C
C
C
C
C
C     Here if a successful run of the loop has been completed.
      IF(MSGLEV.LE.0) GO TO 5010
      WRITE(BUFFER,5001) NELIM
 5001 FORMAT(1X,'ELCNST: Constraints eliminated: ',I9)
C     CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,5002) NFIXED
 5002 FORMAT(1X,'        Variables eliminated:   ',I9)
C     CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,5003) NTIGHT
 5003 FORMAT(1X,'        Variable bounds improved:',I8)
C     CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
 5010 CONTINUE
C
C
C
      RETURN
C
C
 9010 WRITE(BUFFER,9011) RWNAME(I),RTYPE,BLOWER,BUPPER,B(I)
 9011 FORMAT(1X,'ELCNST: Row=',A8,' type=',A2,
     X ' BLO=',D10.3,' BUP=',D10.3,' RHS=',D10.3)
      CALL ERRWRT(IOLOG,BUFFER)
      WRITE(BUFFER,9012)
 9012 FORMAT(1X,'ELCNST: Problem is infeasible.')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(5)
C
 9030 WRITE(BUFFER,9031) I,RWNAME(I),B(I)
 9031 FORMAT(1X,'ELCNST: Constraint ',I6,' (name=',A8,
     X ') is violated, B=',D12.6)
      CALL ERRWRT(IOLOG,BUFFER)
      WRITE(BUFFER,9032)
 9032 FORMAT(1X,'ELCNST: Problem is infeasible.')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(5)
C
 9200 WRITE(BUFFER,9201)
 9201 FORMAT(1X,'ELCNST: Please increase space for PRE_SOLVE ',
     X 'history list.')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(6)
C
C
C
C
C *** LAST CARD OF (ELCNST) ***
      END
