C***************************************************************
C     *** ELVRBL ... ELIMINATE VARIABLES FROM THE LP PROBLEM ***
C***************************************************************
C
      SUBROUTINE ELVRBL(IOLOG,MSGLEV,ICALL,
     X MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X LNHIST,MXHIST,INHIST,DPHIST,
     X IMTMP1,IROW,RELT,
     X B,RANGES,C,UPBND,
     X ACOEFF,CLPNTS,RWNMBS,
     X RWHEAD,RWLINK,CLNMBS,LENCOL,
     X P,Q,PRLVAR,STAVAR,RWSTAT,STAROW,RWNAME,CLNAME,
     X PERM,INVP,LENROW)
C
C
C
C *** PARAMETERS
      INTEGER*4 IOLOG,MSGLEV,ICALL,MAXM,MAXN,MAXNZA,M,N,NSTRCT
      INTEGER*4 LNHIST,MXHIST
      INTEGER*2 INHIST(MXHIST)
      DOUBLE PRECISION DPHIST(MXHIST)
      DOUBLE PRECISION ACOEFF(MAXNZA),B(MAXM),RANGES(MAXM)
      DOUBLE PRECISION RELT(MAXN),C(MAXN),UPBND(MAXN)
      INTEGER*4 IMTMP1(MAXM+1),IROW(MAXN)
      INTEGER*4 CLPNTS(MAXN+1),RWHEAD(MAXM+1),RWLINK(MAXNZA)
      INTEGER*2 RWNMBS(MAXNZA),CLNMBS(MAXNZA),LENCOL(MAXN)
      DOUBLE PRECISION P(MAXM),Q(MAXM),PRLVAR(MAXN)
      CHARACTER*8 RWNAME(MAXM),CLNAME(MAXN)
      INTEGER*2 STAVAR(MAXN),STAROW(MAXM),RWSTAT(MAXM)
      INTEGER*2 INVP(MAXM),PERM(MAXM),LENROW(MAXM)
C
C
C
C *** LOCAL VARIABLES
      INTEGER*4 I,IR,IRUN,J,K,KBEG,KEND,LBIG
      INTEGER*4 MEQ,MFREE,MNEW,NFIXED,NTIGHT,NT0,NNEG,NPOS,NBIG
      DOUBLE PRECISION BIG,BIGNEW,X0,FSBTOL,OPTTOL,SMALLA
      DOUBLE PRECISION PNEW,QNEW,PJ,QJ,PPOSJ,PNEGJ,PPOSJ0,PNEGJ0
      CHARACTER*100 BUFFER
C
C
C
C
C *** PARAMETERS DESCRIPTION
C
C *** ON INPUT:
C     IOLOG   Output unit number where log messages are to be written.
C     MSGLEV  The level of PRE_SOLVE information desired:
C             0  only final problem dimensions are printed;
C             1  numbers of eliminated rows and columns are printed;
C             2  names of eliminated rows and columns are printed;
C             3  detailed DEBUGGING information is printed.
C     ICALL   Number of call of the ELVRBL routine (bounds on shadow
C             prices are initialized in a first call).
C     MAXM    Maximum number of constraints.
C     MAXN    Maximum number of variables.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     M       Number of constraints.
C     N       Number of variables (total, i.e. including slacks,
C             surplus and artificials).
C     NSTRCT  Number of structural variables (excluding slacks,
C             surplus and artificials).
C     LNHIST  Length of the PRE_SOLVE history list.
C     MXHIST  Maximum number of entries in the PRE_SOLVE history list.
C     INHIST  Integer PRE_SOLVE history information.
C     DPHIST  Double precision PRE_SOLVE history information.
C     ACOEFF  Array of nonzero elements for each column.
C     B       Right hand side of the linear program.
C     RANGES  Array of constraint ranges.
C     C       Objective function coefficients.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWHEAD  Headers to the row linked lists of matrix A.
C     RWLINK  Row linked lists of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A.
C     CLNMBS  Column numbers of nonzeros in rows of matrix A.
C     LENCOL  Lengths of (sparse) columns of matrix A.
C     UPBND   Array of upper bounds. Note that the upper bound
C             is changed when the variable has a lower bound. 
C     PRLVAR  Primal variables of the linear program.
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C            -k  FREE variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicates the position of the original variable.
C     RWSTAT  Array of row types:
C             1  row type is = ;
C             2  row type is >= ;
C             3  row type is <= ;
C             4  objective row;
C             5  other free row.
C     STAROW  Array of row status:
C             0  row has been removed (it indicates a free row);
C             1  row has not been removed.
C     RWNAME  Array of row names (increasing order sort).
C     CLNAME  Array of column names (unordered).
C
C *** ON OUTPUT:
C     P       LOWER bounds on shadow prices (dual variables).
C     Q       UPPER bounds on shadow prices (dual variables).
C
C
C
C
C *** WORK ARRAYS:
C     IROW and  RELT are the arrays for temporary handling of rows
C             and columns of the constraint matrix. They are primarily
C             intended to handle sparse vectors (in packed form)
C             but may also be used for storing dense ones.
C     IMTMP1  Integer work array of size MAXM.
C     LENROW  Half-length integer work array of size MAXM.
C
C
C
C
C *** PURPOSE
C     This routine computes bounds on shadow prices (dual variables)
C     and uses them to compute limits for the reduced costs. Those
C     are used to eliminate colulmns (to fix them on their bounds).
C
C
C
C *** SUBROUTINES CALLED
C     MYWRT,GETCOL,GETROW,DABS,EMPTYR,REORDA,REORDV,MYSTOP
C
C
C *** NOTES
C     This routine is given direct access to the matrix A.
C     It alters hidden data structures.
C
C
C
C *** REFERENCES:
C     Gondzio J. (1994). Presolve analysis of linear programs prior 
C        to applying an interior point method, Technical Report 
C        No 1994.3, Department of Management Studies, University 
C        of Geneva, 102, Bd. Carl-Vogt, 1211 Geneva, Switzerland, 
C        February 1994, revised in December 1994.
C
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  May 17, 1993
C     Last modified: June 9, 1995
C
C
C
C
C *** BODY OF (ELVRBL) ***
C
C
C
C *** DEBUGGING
      IF(MSGLEV.LE.3) GO TO 12
      WRITE(BUFFER,1) 
    1 FORMAT(1X,'ELVRBL starts !!!!')
      CALL MYWRT(IOLOG,BUFFER)
      DO 10 J=1,NSTRCT
         IF(STAVAR(J).GE.6) GO TO 10
         IF(LENCOL(J).LE.0.OR.LENCOL(J).GT.M) THEN
            WRITE(BUFFER,2) J,LENCOL(J)
    2       FORMAT(1X,'col=',I6,'  LENCOL(J)=',I6)
            CALL MYWRT(IOLOG,BUFFER)
            CALL MYSTOP(91)
         ENDIF
         KBEG=CLPNTS(J)
         KEND=KBEG+LENCOL(J)-1
         DO 5 K=KBEG,KEND
            IR=RWNMBS(K)
            IF(IR.LE.0.OR.IR.GT.M) THEN
               WRITE(BUFFER,3) J,K,IR
    3          FORMAT(1X,'col=',I6,'  pos=',I6,'  IR=',I6)
               CALL MYWRT(IOLOG,BUFFER)
               CALL MYSTOP(91)
            ENDIF
    5    CONTINUE
   10 CONTINUE
      WRITE(BUFFER,11) 
   11 FORMAT(1X,'ELVRBL analysis O.K. !!!!')
      CALL MYWRT(IOLOG,BUFFER)
   12 CONTINUE
C
C
C
C
C     Initialize.
      BIG=1.0D+30
      BIGNEW=1.0D+20
      SMALLA=1.0D-8
      FSBTOL=1.0D-7
      OPTTOL=1.0D-7
      NTIGHT=0
      NFIXED=0
C
C
C
C     Initialize bounds on dual variables (shadow prices).
      IF(ICALL.GE.2) GO TO 110
      DO 100 I=1,M
         P(I)=-BIG
         Q(I)=BIG
         IF(RWSTAT(I).EQ.1) GO TO 100
         IF(RANGES(I).LE.BIGNEW) GO TO 100
         IF(RWSTAT(I).EQ.2) P(I)=0.0D0
         IF(RWSTAT(I).EQ.3) Q(I)=0.0D0
  100 CONTINUE
  110 CONTINUE
C
C
C
C
C
C     First main loop begins here.
C     Loop over all structural columns of  A.
C     Eliminate empty columns.
C     Tighten bounds on shadow prices for single-element columns.
      DO 1000 J=1,NSTRCT
         IF(STAVAR(J).GE.6) GO TO 1000
         IF(LENCOL(J).GT.1) GO TO 1000
C
C     Check if an empty column is found.
         IF(LENCOL(J).EQ.0) THEN
            IF(MSGLEV.LE.2) GO TO 102
            WRITE(BUFFER,101) J,CLNAME(J)
  101       FORMAT(1X,'ELVRBL: Column ',I6,' (name=',A8,') is empty.')
            CALL MYWRT(IOLOG,BUFFER)
  102       CONTINUE
C
            IF(C(J).LT.0.0D0) THEN
C
C     Fix variable J on its UPPER bound.
               IF(STAVAR(J).EQ.1.OR.STAVAR(J).EQ.3) THEN
                  X0=UPBND(J)
                  GO TO 800
               ELSE
                  GO TO 9010
               ENDIF
            ELSE
C
C     Fix variable J on its LOWER bound.
               IF(STAVAR(J).GE.0) THEN
                  X0=0.0D0
                  GO TO 800
               ELSE
                  GO TO 9020
               ENDIF
            ENDIF
         ENDIF
C
C     Analyse the constraint with an entry in column J.
         KBEG=CLPNTS(J)
         IR=RWNMBS(KBEG)
         IF(RWSTAT(IR).GE.4) GO TO 1000
C
C        Note of 30.03.95. I am not sure about the need of this line:
C        IF(DABS(RANGES(IR)).LE.BIGNEW) GO TO 1000
C
C     Here for an LP constraint.
         IF(MSGLEV.LE.2) GO TO 104
         WRITE(BUFFER,103) J,CLNAME(J),STAVAR(J)
  103    FORMAT(1X,'ELVRBL: snglt cl=',I6,' (nm=',A8,' st=',I6,')')
         CALL MYWRT(IOLOG,BUFFER)
  104    CONTINUE
C
C     Compute new bounds on shadow prices P and Q.
         IF(STAVAR(J).EQ.1.OR.STAVAR(J).EQ.3) GO TO 1000
         PNEW=-BIG
         QNEW=BIG
         IF(ACOEFF(KBEG).LT.0.0D0) THEN
            PNEW=C(J)/ACOEFF(KBEG)
         ELSE
            QNEW=C(J)/ACOEFF(KBEG)
         ENDIF
C
         IF(MSGLEV.LE.2) GO TO 203
         WRITE(BUFFER,201) IR,RWNAME(IR),RWSTAT(IR)
  201    FORMAT(15X,'rw=',I6,' (nm=',A8,' st=',I6,')')
         CALL MYWRT(IOLOG,BUFFER)
         WRITE(BUFFER,202) P(IR),Q(IR),PNEW,QNEW
  202    FORMAT(1X,'Pi=',D12.5,' Qi=',D12.5,
     X    ' Pnew=',D12.5,' Qnew=',D12.5)
         CALL MYWRT(IOLOG,BUFFER)
  203    CONTINUE
C
         IF(PNEW.GT.P(IR)) P(IR)=PNEW
         IF(QNEW.LT.Q(IR)) Q(IR)=QNEW
C
C     Here if the variable cannot be eliminated.
         GO TO 1000
C
C     Eliminate column J. Update RHS.
  800    NFIXED=NFIXED+1
         PRLVAR(J)=X0
         STAVAR(J)=6
         IF(MSGLEV.LE.1) GO TO 802
         WRITE(BUFFER,801) J,CLNAME(J),X0
  801    FORMAT(1X,'ELVRBL: Variable ',I6,' (name=',A8,
     X    ') is being FIXED on X=',D14.6)
         CALL MYWRT(IOLOG,BUFFER)
  802    CONTINUE
C
C
C
C     End of the first main loop.
 1000 CONTINUE
C
C
C
C
C
C     Second main loop begins here.
C     Loop over all structural columns of  A.
C     Tighten bounds on shadow prices for all unbounded columns.
      IRUN=1
 2000 NT0=NTIGHT
      DO 3000 J=1,NSTRCT
         K=STAVAR(J)
         IF(K.GE.6) GO TO 3000
         IF(K.EQ.1.OR.K.EQ.3) GO TO 3000
C
C     Here if J is not bounded.
         KBEG=CLPNTS(J)
         KEND=KBEG+LENCOL(J)-1
C
C     Compute LOWER and UPPER limits of the dual constraint.
         PJ=0.0
         QJ=0.0
         PPOSJ=0.0
         PNEGJ=0.0
         NPOS=0
         NNEG=0
         DO 2200 K=KBEG,KEND
            IR=RWNMBS(K)
            IF(ACOEFF(K).LT.0.0D0) THEN
               NNEG=NNEG+1
               PJ=PJ+Q(IR)*ACOEFF(K)
               PNEGJ=PNEGJ+Q(IR)*ACOEFF(K)
               QJ=QJ+P(IR)*ACOEFF(K)
            ELSE
               NPOS=NPOS+1
               PJ=PJ+P(IR)*ACOEFF(K)
               PPOSJ=PPOSJ+P(IR)*ACOEFF(K)
               QJ=QJ+Q(IR)*ACOEFF(K)
            ENDIF
 2200    CONTINUE
C
C *** DEBUGGING
         IF(MSGLEV.LE.2) GO TO 2210
         WRITE(BUFFER,2201) J,PJ,QJ,C(J)
 2201    FORMAT(1X,'col=',I6,' PJ=',D10.3,' QJ=',D10.3,
     X    ' Cj=',D10.3)
         CALL MYWRT(IOLOG,BUFFER)
         WRITE(BUFFER,2202) PJ,PPOSJ,PNEGJ
 2202    FORMAT(1X,'PJ=',D10.3,'=  PPOSJ=',D10.3,' +  PNEGJ=',D10.3)
         CALL MYWRT(IOLOG,BUFFER)
         WRITE(BUFFER,2203) LENCOL(J),NPOS,NNEG
 2203    FORMAT(1X,'len=',I6,' Npos=',I5,' Nneg=',I5)
         CALL MYWRT(IOLOG,BUFFER)
 2210    CONTINUE
C
C
C     Check if the variable will be FIXED on one of its bounds.
C     If so, then do not tighten bounds on shadow prices.
         IF(C(J)-PJ.LE.-OPTTOL) GO TO 3000
         IF(C(J)-QJ.GE.OPTTOL) GO TO 3000
C
C
C
C     Check if it is possible to tighten bounds on shadow prices
C     refering to rows in which column J has negative entries.
         IF(PPOSJ.LE.-BIGNEW) GO TO 2700
         IF(NNEG.EQ.0) GO TO 2700
C
C     Count negative elements with large (infinite) Qj.
         NBIG=0
         LBIG=0
         PNEGJ0=0.0D0
         DO 2500 K=KBEG,KEND
            IF(ACOEFF(K).LT.0.0D0) THEN
               IR=RWNMBS(K)
               IF(Q(IR).GE.BIGNEW) THEN
                  NBIG=NBIG+1
                  LBIG=K
               ELSE
                  PNEGJ0=PNEGJ0+Q(IR)*ACOEFF(K)
               ENDIF
            ENDIF
 2500    CONTINUE
C        WRITE(BUFFER,2501) J,NNEG,NBIG
C2501    FORMAT(1X,'col=',I6,'  NNEG=',I6,'  Nbg=',I6)
C        CALL MYWRT(IOLOG,BUFFER)
C
C
C     Check if bounds can be tightened.
         IF(NBIG.GE.2) GO TO 2700
         IF(NBIG.EQ.0) THEN
C
C     Bounds for all negative entries can be tightened.
            X0=C(J)-PJ
            DO 2600 K=KBEG,KEND
               IF(ACOEFF(K).LT.0.0D0) THEN
                  IR=RWNMBS(K)
                  PNEW=Q(IR)+X0/ACOEFF(K)
                  IF(PNEW.GE.P(IR)+OPTTOL) THEN
                     NTIGHT=NTIGHT+1
C                    WRITE(BUFFER,2601) J,IR,P(IR),PNEW
C2601                FORMAT(1X,'col=',I6,' row=',I6,' Pi=',D10.3,
C    X                ' is improved, Pnew=',D10.3)
C                    CALL MYWRT(0,BUFFER)
C                    CALL MYWRT(IOLOG,BUFFER)
                     P(IR)=PNEW
                  ENDIF
               ENDIF
 2600       CONTINUE
C
         ELSE
C
C     Bound for only one entry can be tightened.
            X0=C(J)-PPOSJ-PNEGJ0
            IR=RWNMBS(LBIG)
            PNEW=X0/ACOEFF(LBIG)
            IF(PNEW.GE.P(IR)+OPTTOL) THEN
               NTIGHT=NTIGHT+1
C              WRITE(BUFFER,2602) J,IR,P(IR),PNEW
C2602          FORMAT(1X,'col=',I6,' row=',I6,' Pi=',D10.3,
C    X          ' is improved, Pnew=',D10.3)
C              CALL MYWRT(0,BUFFER)
C              CALL MYWRT(IOLOG,BUFFER)
               P(IR)=PNEW
            ENDIF
C
         ENDIF
C
C
C
C     Check if it is possible to tighten bounds on shadow prices
C     refering to rows in which column J has positive entries.
 2700    CONTINUE
         IF(PNEGJ.LE.-BIGNEW) GO TO 3000
         IF(NPOS.EQ.0) GO TO 3000
C
C     Count positive elements with large (infinite) Pj.
         NBIG=0
         LBIG=0
         PPOSJ0=0.0D0
         DO 2800 K=KBEG,KEND
            IF(ACOEFF(K).GT.0.0D0) THEN
               IR=RWNMBS(K)
               IF(P(IR).LE.-BIGNEW) THEN
                  NBIG=NBIG+1
                  LBIG=K
               ELSE
                  PPOSJ0=PPOSJ0+P(IR)*ACOEFF(K)
               ENDIF
            ENDIF
 2800    CONTINUE
C        WRITE(BUFFER,2801) J,NPOS,NBIG
C2801    FORMAT(1X,'col=',I6,'  NPOS=',I6,'  Nbg=',I6)
C        CALL MYWRT(IOLOG,BUFFER)
C
C
C     Check if bounds can be tightened.
         IF(NBIG.GE.2) GO TO 3000
         IF(NBIG.EQ.0) THEN
C
C     Bounds for all positive entries can be tightened.
            X0=C(J)-PJ
            DO 2900 K=KBEG,KEND
               IF(ACOEFF(K).GT.0.0D0) THEN
                  IR=RWNMBS(K)
                  QNEW=P(IR)+X0/ACOEFF(K)
                  IF(QNEW.LE.Q(IR)-OPTTOL) THEN
                     NTIGHT=NTIGHT+1
C                    WRITE(BUFFER,2901) J,IR,Q(IR),QNEW
C2901                FORMAT(1X,'col=',I6,' row=',I6,' Qi=',D10.3,
C    X                ' is improved, Qnew=',D10.3)
C                    CALL MYWRT(0,BUFFER)
C                    CALL MYWRT(IOLOG,BUFFER)
                     Q(IR)=QNEW
                  ENDIF
               ENDIF
 2900       CONTINUE
C
         ELSE
C
C     Bound for only one entry can be tightened.
            X0=C(J)-PNEGJ-PPOSJ0
            IR=RWNMBS(LBIG)
            QNEW=X0/ACOEFF(LBIG)
            IF(QNEW.LE.Q(IR)-OPTTOL) THEN
               NTIGHT=NTIGHT+1
C              WRITE(BUFFER,2902) J,IR,Q(IR),QNEW
C2902          FORMAT(1X,'col=',I6,' row=',I6,' Qi=',D10.3,
C    X          ' is improved, Qnew=',D10.3)
C              CALL MYWRT(0,BUFFER)
C              CALL MYWRT(IOLOG,BUFFER)
               Q(IR)=QNEW
            ENDIF
C
         ENDIF
C
C
C
C     End of the second main loop.
 3000 CONTINUE
C
C
C
C     Check if the loop should be repeated.
C     IF(IRUN.EQ.3) GO TO 4000
C     IF(10*(NTIGHT-NT0).GE.M.OR.IRUN*10*NTIGHT.GE.M) THEN
C        IRUN=IRUN+1
C        GO TO 2000
C     ENDIF
      IF(IRUN.EQ.10) GO TO 4000
      IF(100*(NTIGHT-NT0).GE.M) THEN
         IRUN=IRUN+1
         GO TO 2000
      ENDIF
C
C
C
C
C
C
C     Third main loop begins here.
C     Loop over all structural columns of  A.
C     Eliminate variables with strictly positive
C     (or strictly negative) reduced costs.
C     New option is added here (Oct. 93). Weakly dominated
C     columns are identified and eliminated.
 4000 DO 5000 J=1,NSTRCT
         IF(STAVAR(J).GE.6) GO TO 5000
         KBEG=CLPNTS(J)
         KEND=KBEG+LENCOL(J)-1
C
C     Define LOWER and UPPER costs for variable J.
C     Count slacks egligible to move while pushing a variable
C     to its bound, MFREE. Count entries in EQUALITY type rows.
         PJ=0.0
         QJ=0.0
         MEQ=0
         MFREE=0
         DO 4100 K=KBEG,KEND
            IR=RWNMBS(K)
            IF(RWSTAT(IR).EQ.1) MEQ=MEQ+1
            IF(ACOEFF(K).LT.0.0D0) THEN
               IF(RWSTAT(IR).EQ.2) MFREE=MFREE+1
               PJ=PJ+Q(IR)*ACOEFF(K)
               QJ=QJ+P(IR)*ACOEFF(K)
            ELSE
               IF(RWSTAT(IR).EQ.3) MFREE=MFREE+1
               PJ=PJ+P(IR)*ACOEFF(K)
               QJ=QJ+Q(IR)*ACOEFF(K)
            ENDIF
 4100    CONTINUE
C
         IF(MSGLEV.LE.2) GO TO 4102
         WRITE(BUFFER,4101) J,PJ,QJ,C(J)
 4101    FORMAT(1X,'column=',I6,' PJ=',D10.3,' QJ=',D10.3,
     X    ' Cj=',D10.3)
         CALL MYWRT(IOLOG,BUFFER)
 4102    CONTINUE
C
C
C     Check if the column is strongly dominated.
         IF(C(J)-PJ.LE.-OPTTOL) THEN
C
C     Fix variable J on its UPPER bound.
            IF(MSGLEV.LE.2) GO TO 4210
            WRITE(BUFFER,4201) J,PJ,QJ,C(J),STAVAR(J)
 4201       FORMAT(1X,'cl=',I6,' PJ=',D10.3,' QJ=',D10.3,
     X       ' Cj=',D10.3,' st=',I6,' goes to UPbnd.')
            CALL MYWRT(IOLOG,BUFFER)
            WRITE(BUFFER,4202) J,C(J)-PJ
 4202       FORMAT(1X,'Strongly dominated cl=',I6,' Cj-Pj=',D10.3)
            CALL MYWRT(IOLOG,BUFFER)
 4210       CONTINUE
            IF(STAVAR(J).EQ.1.OR.STAVAR(J).EQ.3) THEN
               X0=UPBND(J)
               GO TO 4800
            ELSE
               GO TO 9010
            ENDIF
         ENDIF
C
         IF(C(J)-QJ.GE.OPTTOL) THEN
C
C     Fix variable J on its LOWER bound.
            IF(MSGLEV.LE.2) GO TO 4310
            WRITE(BUFFER,4301) J,PJ,QJ,C(J),STAVAR(J)
 4301       FORMAT(1X,'cl=',I6,' PJ=',D10.3,' QJ=',D10.3,
     X       ' Cj=',D10.3,' st=',I6,' goes to LObnd.')
            CALL MYWRT(IOLOG,BUFFER)
            WRITE(BUFFER,4302) J,C(J)-QJ
 4302       FORMAT(1X,'Strongly dominated cl=',I6,' Cj-Qj=',D10.3)
            CALL MYWRT(IOLOG,BUFFER)
 4310       CONTINUE
            IF(STAVAR(J).GE.0) THEN
               X0=0.0D0
               GO TO 4800
            ELSE
               GO TO 9020
            ENDIF
         ENDIF
C
C
C     Check if the column is weakly dominated. Omit FREE columns,
C     columns with entries in EQUALITY rows and singleton columns.
         IF(LENCOL(J).LE.1) GO TO 5000
         IF(STAVAR(J).LT.0) GO TO 5000
         IF(MEQ.GE.1) GO TO 5000
C
         IF(MFREE.EQ.0.AND.C(J)-PJ.LE.OPTTOL) THEN
C
C     Fix variable J on its UPPER bound.
            IF(MSGLEV.LE.2) GO TO 4410
            WRITE(BUFFER,4401) J,PJ,QJ,C(J),STAVAR(J)
 4401       FORMAT(1X,'cl=',I6,' PJ=',D10.3,' QJ=',D10.3,
     X       ' Cj=',D10.3,' st=',I6,' goes to UPbnd.')
            CALL MYWRT(IOLOG,BUFFER)
            WRITE(BUFFER,4402) LENCOL(J),MEQ,MFREE,C(J)-PJ
 4402       FORMAT(1X,'Weak domination, ln=',I6,' meq=',I6,
     X       ' mfr=',I6,' Cj-Pj=',D10.3)
            CALL MYWRT(IOLOG,BUFFER)
 4410       CONTINUE
            IF(STAVAR(J).EQ.1.OR.STAVAR(J).EQ.3) THEN
               X0=UPBND(J)
               GO TO 4800
            ELSE
C              WRITE(BUFFER,4411) J,UPBND(J)
C4411          FORMAT(1X,'Weak domination, col=',I6,' UP=',D10.3)
C              CALL MYWRT(IOLOG,BUFFER)
               GO TO 5000
            ENDIF
         ENDIF
C
         IF(MFREE.EQ.LENCOL(J).AND.C(J)-QJ.GE.-OPTTOL) THEN
C
C     Fix variable J on its LOWER bound.
            IF(MSGLEV.LE.2) GO TO 4510
            WRITE(BUFFER,4501) J,PJ,QJ,C(J),STAVAR(J)
 4501       FORMAT(1X,'cl=',I6,' PJ=',D10.3,' QJ=',D10.3,
     X       ' Cj=',D10.3,' st=',I6,' goes to LObnd.')
            CALL MYWRT(IOLOG,BUFFER)
            WRITE(BUFFER,4502) LENCOL(J),MEQ,MFREE,C(J)-QJ
 4502       FORMAT(1X,'Weak domination, ln=',I6,' meq=',I6,
     X       ' mfr=',I6,' Cj-Qj=',D10.3)
            CALL MYWRT(IOLOG,BUFFER)
 4510       CONTINUE
            IF(STAVAR(J).GE.0) THEN
               X0=0.0D0
               GO TO 4800
            ELSE
C              WRITE(BUFFER,4511) J
C4511          FORMAT(1X,'Weak domination, col=',I6,' no LO bnd.')
C              CALL MYWRT(IOLOG,BUFFER)
               GO TO 5000
            ENDIF
         ENDIF
C
         GO TO 5000
C
C     Eliminate column J. Update RHS.
 4800    NFIXED=NFIXED+1
         PRLVAR(J)=X0
         STAVAR(J)=6
         IF(DABS(X0).LE.FSBTOL) THEN
            X0=0.0D0
            GO TO 4900
         ENDIF
         DO 4850 K=KBEG,KEND
            IR=RWNMBS(K)
            B(IR)=B(IR)-X0*ACOEFF(K)
            IF(DABS(B(IR)).LE.SMALLA) B(IR)=0.0D0
 4850    CONTINUE
 4900    CONTINUE
         IF(MSGLEV.LE.1) GO TO 4902
         WRITE(BUFFER,4901) J,CLNAME(J),X0
 4901    FORMAT(1X,'ELVRBL: Variable ',I6,' (name=',A8,
     X    ') is being FIXED on X=',D14.6)
         CALL MYWRT(IOLOG,BUFFER)
 4902    CONTINUE
C
C
C
C     End of the third main loop.
 5000 CONTINUE
C
C
C
C
C     Here if a successful run has been completed.
      IF(MSGLEV.LE.0) GO TO 5110
      WRITE(BUFFER,5105) NFIXED
 5105 FORMAT(1X,'ELVRBL: Variables eliminated:   ',I9)
C     CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,5106) NTIGHT
 5106 FORMAT(1X,'        Bounds on shadow prices:',I9)
C     CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
 5110 CONTINUE
C
C
C
C
C
C     Zero  RWHEAD and LENROW arrays.
      IF(NFIXED.EQ.0) GO TO 5900
      DO 5200 I=1,M
         RWHEAD(I)=0
         LENROW(I)=0
 5200 CONTINUE
C
C     Set the row linked lists.
C     Count nonzero elements in all rows of  A.
      DO 5300 J=1,N
C
C     Omit  FIXED variables.
         IF(STAVAR(J).GE.6) GO TO 5300
         KBEG=CLPNTS(J)
         KEND=KBEG+LENCOL(J)-1
         DO 5250 K=KBEG,KEND
            I=RWNMBS(K)
            RWLINK(K)=RWHEAD(I)
            CLNMBS(K)=J
            RWHEAD(I)=K
            LENROW(I)=LENROW(I)+1
 5250    CONTINUE
 5300 CONTINUE
C
C
C     Check if there are inequality type rows to be eliminated.
C     Check if the eliminated rows were not violated.
      DO 5800 I=1,M
         IF(DABS(B(I)).LE.FSBTOL) B(I)=0.0D0
         K=RWHEAD(I)
         IF(RWSTAT(I).EQ.1) THEN
C
C     Here for EQUALITY constraint.
            IF(K.NE.0) GO TO 5800
            IF(DABS(B(I)).GT.FSBTOL) GO TO 9030
            GO TO 5800
         ENDIF
         IF(LENROW(I).GE.2) GO TO 5800
         IF(RWSTAT(I).EQ.2) THEN
C
C     Here for GREATER OR EQUAL type constraint.
            IF(B(I).GT.FSBTOL) GO TO 9030
            RWHEAD(I)=-RWHEAD(I)
            J=CLNMBS(K)
            STAVAR(J)=14
            GO TO 5800
         ENDIF
         IF(RWSTAT(I).EQ.3) THEN
C
C     Here for LESS OR EQUAL type constraint.
            IF(B(I).LT.-FSBTOL) GO TO 9030
            RWHEAD(I)=-RWHEAD(I)
            J=CLNMBS(K)
            STAVAR(J)=14
            GO TO 5800
         ENDIF
 5800 CONTINUE
 5900 CONTINUE
C
C
C
C     Determine the permutation that puts all empty rows
C     at the end of the list.
C
      IRUN=3
      IF(MSGLEV.LE.1) IRUN=4
      CALL EMPTYR(MAXM,M,MNEW,IRUN,
     X RWHEAD,STAROW,PERM,INVP,IOLOG)
C
C
C     Reorder the rows of the  LP constraint matrix according to
C     the permutation resulting from the analysis of EMPTYR.
C     Reorder LENROW array accordingly.
      IF(MNEW.LT.M) THEN
C
         CALL REORDA(MAXM,MAXN,MAXNZA,M,N,
     X    CLPNTS,RWNMBS,
     X    RWHEAD,RWLINK,CLNMBS,LENCOL,
     X    PERM,INVP,IMTMP1,IROW,RELT,
     X    RWNAME,STAROW,RWSTAT,RANGES,B,IOLOG)
C
C     Reorder bounds on shadow prices P and Q.
         CALL REORDV(MAXM,M,
     X    PERM,INVP,P,RELT,IOLOG)
         CALL REORDV(MAXM,M,
     X    PERM,INVP,Q,RELT,IOLOG)
C
C     Set the new number of rows of the constraint matrix.
         M=MNEW
C
      ENDIF
C
C
      RETURN
C
 9010 WRITE(BUFFER,9011) J,CLNAME(J)
 9011 FORMAT(1X,'ELVRBL: Var. ',I6,' (name=',A8,
     X ') has no UPPER bound.')
      CALL ERRWRT(IOLOG,BUFFER)
      WRITE(BUFFER,9012)
 9012 FORMAT(1X,'ELVRBL: Primal is unbounded (or dual infeasible).')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(5)
C
 9020 WRITE(BUFFER,9021) J,CLNAME(J)
 9021 FORMAT(1X,'ELVRBL: Var. ',I6,' (name=',A8,
     X ') has no LOWER bound.')
      CALL ERRWRT(IOLOG,BUFFER)
      WRITE(BUFFER,9022)
 9022 FORMAT(1X,'ELVRBL: Primal is unbounded (or dual infeasible).')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(5)
C
 9030 WRITE(BUFFER,9031) I,RWNAME(I),B(I)
 9031 FORMAT(1X,'ELVRBL: Constraint ',I6,' (name=',A8,
     X ') is violated, B=',D12.6)
      CALL ERRWRT(IOLOG,BUFFER)
      WRITE(BUFFER,9032)
 9032 FORMAT(1X,'ELVRBL: Problem is infeasible.')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(5)
C
C
C
C
C *** LAST CARD OF (ELVRBL) ***
      END
