C**********************************************************
C     ****  EMPTYR ... REMOVE EMPTY ROWS FROM  A  ****
C**********************************************************
C
      SUBROUTINE EMPTYR(MAXM,MOLD,MNEW,IRUN,
     X RWHEAD,STAROW,PERM,INVP,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 MAXM,MOLD,MNEW,IRUN,IOLOG
      INTEGER*4 RWHEAD(MAXM)
C
C *** The following arrays can be half-length integer.
      INTEGER*2 STAROW(MAXM),PERM(MAXM),INVP(MAXM)
C
C
C *** LOCAL VARIABLES
      INTEGER*4 I,IX,K,NEMPTY,NROWOK
      CHARACTER*100 BUFFER
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     MOLD    Number of all rows of the LP constraint matrix.
C     IRUN    Number of the run of the EMPTYR routine.
C     RWHEAD  Headers to the row linked lists of matrix A.
C     STAROW  Array of row status:
C             0  row is to be removed (it indicates a free row);
C             1  row is not to be removed.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     MNEW    Number of non-empty rows of the LP constraint matrix.
C     PERM    Permutation that moves empty rows into the end of list.
C     INVP    Inverse permutation.
C
C
C *** SUBROUTINES CALLED:
C     MYWRT
C
C
C *** PURPOSE:
C     This routine reorders the rows of the  LP constraint
C     in order to put the empty ones at the end of the list.
C
C
C *** NOTES:
C
C
C *** REFERENCES:
C     Gondzio J. (1993). Implementing Cholesky factorization
C        for interior point methods of linear programming,
C        Optimization 27, pp. 121-140.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  April 14, 1991
C     Last modified: May 17, 1995
C
C
C
C *** BODY OF (EMPTYR) ***
C
C
C
C     Set up initial permutations.
C     Check if there are empty rows in  LP constraint matrix.
C     Zero their row status (STAROW array) to mark that they
C     have to be removed.
      DO 100 I=1,MOLD
C        WRITE(BUFFER,101) I,STAROW(I),RWHEAD(I)
C 101    FORMAT(1X,'EMPTYR: i=',I5,'  strw=',I6,'  rwhd=',I6)
C        CALL MYWRT(IOLOG,BUFFER)
         K=RWHEAD(I)
         IF(K.LE.0) THEN 
            STAROW(I)=0
            RWHEAD(I)=-K
C           WRITE(BUFFER,102) I
C 102       FORMAT(1X,'EMPTYR: Row to eliminate found i=',I6)
C           CALL MYWRT(IOLOG,BUFFER)
         ENDIF
  100 CONTINUE
C
C
C     Determine the permutation that puts all empty rows
C     at the end of the list.
      NEMPTY=0
      NROWOK=0
      DO 200 I=1,MOLD
         K=STAROW(I)
         IF(K.EQ.0) THEN
C           WRITE(BUFFER,201) I,MOLD-NEMPTY
C 201       FORMAT(1X,'EMPTYR: Row i=',I6,' is put to ',I6)
C           CALL MYWRT(IOLOG,BUFFER)
            INVP(I)=MOLD-NEMPTY
            NEMPTY=NEMPTY+1
         ELSE
            NROWOK=NROWOK+1
            INVP(I)=NROWOK
         ENDIF
  200 CONTINUE
C
C
      DO 300 I=1,MOLD
         IX=INVP(I)
         PERM(IX)=I
  300 CONTINUE
C
C
C
      MNEW=MOLD-NEMPTY
      IF(IRUN.EQ.1) NEMPTY=NEMPTY-1
      IF(IRUN.LE.2) THEN
         IF(NEMPTY.GT.0) THEN
            WRITE(BUFFER,501) NEMPTY
  501       FORMAT(1X,'EMPTYR:',I9,' empty rows found in  A.')
            CALL MYWRT(IOLOG,BUFFER)
         ENDIF
      ENDIF
      IF(IRUN.EQ.3) THEN
         IF(NEMPTY.GT.0) THEN
            WRITE(BUFFER,502) NEMPTY
  502       FORMAT(1X,'EMPTYR: Rows eliminated from A',3X,I8)
            CALL MYWRT(IOLOG,BUFFER)
         ENDIF
      ENDIF
C
      RETURN
C
C *** LAST CARD OF (EMPTYR) ***
      END
