C********************************************************
C     **** FSAX ... FAST (sparse)A * (dense)X ****
C********************************************************
C
      SUBROUTINE FSAX(MAXM,MAXN,MAXNZA,X,N,Y,M,
     X ACOEFF,CLPNTS,RWNMBS,LENCOL,VUSED,IOLOG)
C
C *** PARAMETERS
      INTEGER*4 MAXM,MAXN,MAXNZA,M,N,IOLOG
      DOUBLE PRECISION X(N),Y(M),ACOEFF(MAXNZA)
      INTEGER*4 CLPNTS(MAXN+1)
      INTEGER*2 RWNMBS(MAXNZA),LENCOL(MAXN)
      LOGICAL VUSED(MAXN)
C
C *** LOCAL VARIABLES
      INTEGER*4 IR,J,K,KBEG,KEND
C
C
C *** PURPOSE
C     This routine computes the product of a sparse matrix  A and
C     a dense vector  X and saves the result in a dense vector  Y.
C     It is given direct access to matrix  A.
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     MAXN    Maximum column dimension of the LP constraint matrix.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     M       Number of rows of the LP constraint matrix.
C     N       Number of columns of the LP constraint matrix.
C     X       Dense vector of dimension N.
C     ACOEFF  Nonzero elements of matrix A.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A.
C     LENCOL  Lengths of columns of matrix A.
C     VUSED   An indicator if a variable is active in the optimization
C             process:
C             .TRUE.   active variable;
C             .FALSE.  FIXED variable.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     Y       Dense vector of dimension M (Y = A * X).
C
C
C *** SUBROUTINES CALLED:
C     NONE
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Last modified: May 17, 1995
C
C
C *** BODY OF (FSAX) ***
C
      DO 100 IR=1,M
         Y(IR)=0.0D0
  100 CONTINUE
      DO 300 J=1,N
         IF(.NOT.VUSED(J)) GO TO 300
         KBEG=CLPNTS(J)
         KEND=KBEG+LENCOL(J)-1
         DO 200 K=KBEG,KEND
            IR=RWNMBS(K)
            Y(IR)=Y(IR)+X(J)*ACOEFF(K)
  200    CONTINUE
  300 CONTINUE
      RETURN
C
C *** LAST CARD OF (FSAX) ***
      END
