C**************************************************************
C     **** GETCOL ... GET THE  J-th COLUMN OF MATRIX  A ****
C**************************************************************
C
      SUBROUTINE GETCOL(J,RWORK,IWORK,RMAP,IMAP,
     X IROW,RELT,COLLEN,MAXN,IOLOG)
C
C
C
C *** PARAMETERS
      INTEGER*4 J,MAXN,COLLEN,IOLOG
      INTEGER*4 IROW(MAXN)
      DOUBLE PRECISION RELT(MAXN)
      INTEGER*4 IWORK(*),IMAP(*),RMAP(*)
      DOUBLE PRECISION RWORK(*)
C
C
C
C *** LOCAL VARIABLES
      INTEGER*4 MI1,MI2,MI6,MR1
C
C
C
C *** PARAMETERS DESCRIPTION
C     ON INPUT:
C     J       Index of the column to be created.
C     RWORK   Real work array that contain almost all real
C             LP problem data.
C     IWORK   Integer work array that contain almost all integer
C             LP problem data.
C     RMAP    Map of RWORK.
C     IMAP    Map of IWORK.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     IROW    Row indices of nonzero entries of the column to be created.
C     RELT    Nonzero entries of the column to be created.
C     COLLEN  Number of nonzero entries of the column to be created.
C
C
C
C *** SUBROUTINES CALLED
C     XGTCOL
C
C
C *** NOTES
C
C     Map of IWORK array:
C     IMAP(1)   Points to CLPNTS array.
C     IMAP(2)   Points to RWNMBS array.
C     IMAP(3)   Points to RWHEAD array.
C     IMAP(4)   Points to RWLINK array.
C     IMAP(5)   Points to CLNMBS array.
C     IMAP(6)   Points to LENCOL array.
C     IMAP(7)   Points to the first empty cell of IWORK array.
C
C     Map of RWORK array:
C     RMAP(1)   Points to ACOEFF array.
C     RMAP(2)   Points to COBJ array.
C     RMAP(3)   Points to RHS array.
C     RMAP(4)   Points to the first empty cell of RWORK array.
C
C
C
C *** REFERENCES:
C     Gondzio J., Tachat D. (1994). The design and application of 
C        IPMLO - a FORTRAN library for linear optimization with 
C        interior point methods, RAIRO Recherche Operationnelle 28, 
C        No 1, pp. 37-56.
C
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  March 21, 1992
C     Last modified: May 17, 1995
C
C
C
C
C *** BODY OF (GETCOL) ***
C
C
C     Set pointers to the arrays in the hidden data structures.
C     IMAP(1)   Points to CLPNTS array.
C     IMAP(2)   Points to RWNMBS array.
C     IMAP(6)   Points to LENCOL array.
C     RMAP(1)   Points to ACOEFF array.
      MI1=IMAP(1)
      MI2=IMAP(2)
      MI6=IMAP(6)
      MR1=RMAP(1)
C
C     Call the lower level routine.
C     SUBROUTINE XGTCOL(J,ACOEFF,
C    X CLPNTS,RWNMBS,LENCOL,
C    X IROW,RELT,COLLEN,IOLOG)
C
      CALL XGTCOL(J,RWORK(MR1),
     X IWORK(MI1),IWORK(MI2),IWORK(MI6),
     X IROW,RELT,COLLEN,IOLOG)
C
      RETURN
C
C *** LAST CARD OF (GETCOL) ***
      END
