C**************************************************************
C     ****  GETROW ... GET THE  I-th ROW OF MATRIX  A  ****
C**************************************************************
C
      SUBROUTINE GETROW(I,RWORK,IWORK,RMAP,IMAP,
     X JCOL,RELT,ROWLEN,MAXN,IOLOG)
C
C
C
C *** PARAMETERS
      INTEGER*4 I,MAXN,ROWLEN,IOLOG
      INTEGER*4 JCOL(MAXN)
      DOUBLE PRECISION RELT(MAXN)
      INTEGER*4 IWORK(*),IMAP(*),RMAP(*)
      DOUBLE PRECISION RWORK(*)
C
C
C
C *** LOCAL VARIABLES
      INTEGER*4 MI3,MI4,MI5,MR1
C
C
C
C *** PARAMETERS DESCRIPTION
C     ON INPUT:
C     I       Index of the row to be created.
C     RWORK   Real work array that contain almost all real
C             LP problem data.
C     IWORK   Integer work array that contain almost all integer
C             LP problem data.
C     RMAP    Map of RWORK.
C     IMAP    Map of IWORK.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     JCOL    Column indices of nonzero entries of the row to be created.
C     RELT    Nonzero entries of the row to be created.
C     ROWLEN  Number of nonzero entries of the row to be created.
C
C
C
C *** SUBROUTINES CALLED
C     XGTROW
C
C
C *** NOTES
C
C     Map of IWORK array:
C     IMAP(1)   Points to CLPNTS array.
C     IMAP(2)   Points to RWNMBS array.
C     IMAP(3)   Points to RWHEAD array.
C     IMAP(4)   Points to RWLINK array.
C     IMAP(5)   Points to CLNMBS array.
C     IMAP(6)   Points to LENCOL array.
C     IMAP(7)   Points to the first empty cell of IWORK array.
C
C     Map of RWORK array:
C     RMAP(1)   Points to ACOEFF array.
C     RMAP(2)   Points to COBJ array.
C     RMAP(3)   Points to RHS array.
C     RMAP(4)   Points to the first empty cell of RWORK array.
C
C
C
C *** REFERENCES:
C     Gondzio J., Tachat D. (1994). The design and application of 
C        IPMLO - a FORTRAN library for linear optimization with 
C        interior point methods, RAIRO Recherche Operationnelle 28, 
C        No 1, pp. 37-56.
C
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  March 21, 1992
C     Last modified: May 17, 1995
C
C
C
C
C *** BODY OF (GETROW) ***
C
C
C     Set pointers to the arrays in the hidden data structures.
C     IMAP(3)   Points to RWHEAD array.
C     IMAP(4)   Points to RWLINK array.
C     IMAP(5)   Points to CLNMBS array.
C     RMAP(1)   Points to ACOEFF array.
      MI3=IMAP(3)
      MI4=IMAP(4)
      MI5=IMAP(5)
      MR1=RMAP(1)
C
C     Call the lower level routine.
C     SUBROUTINE XGTROW(I,ACOEFF,
C    X RWHEAD,RWLINK,CLNMBS,
C    X JCOL,RELT,ROWLEN,IOLOG)
C
      CALL XGTROW(I,RWORK(MR1),
     X IWORK(MI3),IWORK(MI4),IWORK(MI5),
     X JCOL,RELT,ROWLEN,IOLOG)
C
      RETURN
C
C *** LAST CARD OF (GETROW) ***
      END
