C*************************************************************
C     ***  This is the main program of the  HOPDM library  ***
C     ***  Version 2.13 of  February 8, 1996               ***
C*************************************************************
C
C     PROGRAM HMAIN2
C
C
C *** PURPOSE
C     This is the main program of the  HOPDM library.
C
C     HOPDM is an implementation of a Higher Order 
C     Primal-Dual logarithmic barrier Method.
C
C
C
C
C *** GLOBAL PARAMETERS
      INTEGER*4 IOLOG,LIWORK,LRWORK,LIMAP,LRMAP
      INTEGER*4 LMAX,MDIM,NDIM,NZDIM,MAXM,MAXN,MAXNZA,MAXNZL,MXHIST
      DOUBLE PRECISION OPTTOL
      PARAMETER (IOLOG=11,LIMAP=10,LRMAP=10,LMAX=2)
C
C
C
C     NOTE: 
C        If you would like to solve problems which sizes exceed
C     32767 rows/columns, then you should change all the INTEGER*2
C     declarations into INTEGER*4 ones. 
C     You can use a script 'Two2four' to do it automatically. 
C     Do not forget to modify INTSZ parameter accordingly.
C        If you would like to solve problems with more than 
C     100000 rows, then you should change declaration of MSPLIT 
C     array in ICGRAD common block.
C
C
C     for computers with  32 MB of virtual memory:
C     --------------------------------------------
      PARAMETER (MDIM=20000,NDIM=32767,NZDIM=200000,MAXNZL=600000)
      PARAMETER (MXHIST=MDIM + 2*NDIM)
      PARAMETER (LIWORK=MDIM + 2*NDIM + 3*NZDIM + 9) 
      PARAMETER (LRWORK=MDIM +   NDIM +   NZDIM + 1) 
C
C     for computers with  64 MB of virtual memory:
C     --------------------------------------------
C     PARAMETER (MDIM=60000,NDIM=100000,NZDIM=400000,MAXNZL=2320000)
C     PARAMETER (MXHIST=MDIM + 2*NDIM)
C     PARAMETER (LIWORK=MDIM + 2*NDIM + 3*NZDIM + 9) 
C     PARAMETER (LRWORK=MDIM +   NDIM +   NZDIM + 1) 
C
C     for computers with  128 MB of virtual memory:
C     ---------------------------------------------
C     PARAMETER (MDIM=100000,NDIM=300000,NZDIM=1000000,MAXNZL=5000000)
C     PARAMETER (MXHIST=MDIM + 2*NDIM)
C     PARAMETER (LIWORK=MDIM + 2*NDIM + 3*NZDIM + 9) 
C     PARAMETER (LRWORK=MDIM +   NDIM +   NZDIM + 1) 
C
C
C
C *** GLOBAL PARAMETERS DESCRIPTION
C     LIWORK  Dimension of the integer work array IWORK.
C     LRWORK  Dimension of the real work array RWORK.
C     LIMAP   Dimension of the map of the integer work array, IMAP.
C     LRMAP   Dimension of the map of the real work array, RMAP.
C     LMAX    Maximum order of polynomial used in the method.
C     MDIM    Maximum number of constraints (see also MAXM).
C     NDIM    Maximum number of variables (see also MAXN).
C     NZDIM   Maximum number of non-zeros in the LP constraint matrix
C             (see also MAXNZA).
C     MAXNZL  Maximum number of non-zeros in the Cholesky factor.
C     MXHIST  Maximum number of entries in the PRE_SOLVE history list.
C     IOLOG   Output unit number where log messages are to be written.
C     OPTTOL  Optimality tolerance (default 1.D-8 set in RDSPEC).
C
C
C
C *** VARIABLES AND ARRAYS ASSOCIATED WITH THE MPS FILE
      CHARACTER*13 FILMPS,FILSPC,FILLOG,FILSOL
      CHARACTER*9 NAMEC,NAMEB,NAMBND,NAMRAN,NAMMPS
      CHARACTER*8 RWNAME(MDIM),RWCOPY(MDIM),CLNAME(NDIM)
      LOGICAL VUSED(NDIM),VBNDED(NDIM)
      INTEGER*2 STAVAR(NDIM),STAROW(MDIM),RWSTAT(MDIM)
      DOUBLE PRECISION RANGES(MDIM),UPBND(NDIM),LOBND(NDIM)
      DOUBLE PRECISION BIG,DLOBND,DUPBND
C
C *** MPS VARIABLES DESCRIPTION
C     FILSPC  Specifications file name.
C     FILMPS  MPS input file name.
C     FILLOG  Log file name.
C     FILSOL  Solution file name.
C     NAMEC   The name of the desired objective function.
C     NAMEB   The name of the right hand side section chosen.
C     NAMBND  The name of the bound section chosen.
C     NAMRAN  The name of the range section chosen.
C     NAMMPS  The problem name.
C     RWNAME  Array of row names.
C     RWCOPY  A copy of row names (keeps trace of the original order).
C     RWSTAT  Array of row types:
C             1  row type is = ;
C             2  row type is >= ;
C             3  row type is <= ;
C             4  objective row;
C             5  other free row.
C     STAROW  Array of row status:
C             0  row has been removed (it indicates a free row);
C             1  row has not been removed.
C     CLNAME  Array of column names.
C     VUSED   An indicator if a variable is active in the optimization
C             process:
C             .TRUE.   active variable;
C             .FALSE.  FIXED variable.
C     VBNDED  An indicator if a variable has an UPPER bound:
C             .TRUE.   UPPER bounded variable;
C             .FALSE.  UNBOUNDED variable;
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C             7  (or larger) PRESUMED OPTIMAL variable i.e.: x = x0;
C            -k  FREE variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicates the position of the original variable.
C     RANGES  Array of constraint ranges.
C     UPBND   Array of upper bounds. Note that the upper bound
C             is changed when the variable has a lower bound. 
C     LOBND   Array of lower bounds.
C
C
C
C *** HIDDEN DATA STRUCTURES
      INTEGER*4 IWORK(LIWORK),IMAP(LIMAP),RMAP(LRMAP)
      DOUBLE PRECISION RWORK(LRWORK)
C
C *** HIDDEN DATA STRUCTURES DESCRIPTION
C     RWORK   Real work array containing almost all real
C             LP problem data.
C     IWORK   Integer work array containing almost all integer
C             LP problem data.
C     RMAP    Map of RWORK.
C     IMAP    Map of IWORK.
C
C     Map of IWORK array:
C     IMAP(1)   Points to CLPNTS array.
C     IMAP(2)   Points to RWNMBS array.
C     IMAP(3)   Points to RWHEAD array.
C     IMAP(4)   Points to RWLINK array.
C     IMAP(5)   Points to CLNMBS array.
C     IMAP(6)   Points to LENCOL array.
C     IMAP(7)   Points to the first empty cell of IWORK array.
C
C     Map of RWORK array:
C     RMAP(1)   Points to ACOEFF array.
C     RMAP(2)   Points to C array.
C     RMAP(3)   Points to RHS array.
C     RMAP(4)   Points to the first empty cell of RWORK array.
C
C
C
C *** WORK ARRAYS
      INTEGER*4 INTMP1(NDIM),IROW(NDIM)
      INTEGER*2 INTMP2(NDIM),INTMP3(NDIM)
      INTEGER*4 IMTMP1(MDIM+1),IMTMP2(MDIM+1)
      DOUBLE PRECISION X(NDIM),S(NDIM)
      DOUBLE PRECISION Y(MDIM),Z(NDIM),W(NDIM),YPROX(MDIM)
      DOUBLE PRECISION DELTAX(NDIM,LMAX),DELTAS(NDIM,LMAX)
      DOUBLE PRECISION DELTAY(MDIM,LMAX),RSCALE(MDIM),CSCALE(NDIM)
      DOUBLE PRECISION DELTAZ(NDIM,LMAX),DELTAW(NDIM,LMAX),RELT(NDIM)
      DOUBLE PRECISION THETA(NDIM),XIB(MDIM),XIC(NDIM),XIU(NDIM)
      DOUBLE PRECISION P(MDIM),Q(MDIM),OSCALE
      DOUBLE PRECISION RMTMP1(MDIM),RMTMP2(MDIM)
      DOUBLE PRECISION RMTMP3(MDIM),RMTMP4(MDIM)
      DOUBLE PRECISION RNTMP1(NDIM),RNTMP2(NDIM),RNTMP3(NDIM)
      DOUBLE PRECISION RNTMP4(NDIM),RNTMP5(NDIM),RNTMP6(NDIM)
      DOUBLE PRECISION COLNRM(NDIM)
C
C *** VARIABLES FOR GENQMD and GENMMD ROUTINES
      INTEGER*4 DHEAD(MDIM+100),PERM0(MDIM),INVP0(MDIM)
      INTEGER*4 NBRHD(MDIM),QSIZE(MDIM),QLINK(MDIM)
C
C *** The following arrays can be half-length integer.
      INTEGER*2 PERM(MDIM),INVP(MDIM)
      INTEGER*2 HEADER(MDIM+1),LINKFD(MDIM+1),LINKBK(MDIM+1)
C
C *** WORK ARRAYS DESCRIPTION
C     INTMP1  Integer work array of size MAXN.
C     INTMP2  Half-length integer work array of size MAXN.
C     INTMP3  Half-length integer work array of size MAXN.
C     IMTMP1  Integer work array of size MAXM.
C     IMTMP2  Integer work array of size MAXM.
C     IROW and  RELT are the arrays for temporary handling of rows
C             and columns of the constraint matrix. They are primarily
C             intended to handle sparse vectors (in packed form)
C             but may also be used for storing dense ones.
C     RMTMP1  Double precision work array of size MAXM.
C     RMTMP2  Double precision work array of size MAXM.
C     RMTMP3  Double precision work array of size MAXM.
C     RMTMP4  Double precision work array of size MAXM.
C     RNTMP1  Double precision work array of size MAXN.
C     RNTMP2  Double precision work array of size MAXN.
C     RNTMP3  Double precision work array of size MAXN.
C     RNTMP4  Double precision work array of size MAXN.
C     RNTMP5  Double precision work array of size MAXN.
C     RNTMP6  Double precision work array of size MAXN.
C
C     X       Primal variables of the linear program.
C     S       Primal slack variables of the linear program.
C     Y       Dual variables of the linear program.
C     Z       Dual slack variables of the linear program (var. Z).
C     W       Dual slack variables of the linear program (var. W).
C     DELTAX(*,L)  L-th component of deltaX.
C     DELTAS(*,L)  L-th component of deltaS.
C     DELTAY(*,L)  L-th component of deltaY.
C     DELTAZ(*,L)  L-th component of deltaZ.
C     DELTAW(*,L)  L-th component of deltaW.
C     YPROX   Dual proximal point. 
C             After the optimization: dual solution in original order.
C
C     XIB     Violation of primal constraints, i.e.  b - A * x
C     XIC     Violation of dual   constraints, i.e.  c - At*y - z + w
C     XIU     Violation of variable bounds, i.e.     UPBND - x - s
C     RSCALE  Row scaling factors.
C     CSCALE  Column scaling factors.
C     OSCALE  Objective row scaling factor.
C     THETA   Diagonal weight matrix.
C     P       LOWER bounds on shadow prices (dual variables).
C     Q       UPPER bounds on shadow prices (dual variables).
C
C     PERM    Permutation resulting from the minimum degree ordering.
C     INVP    Inverse permutation.
C     HEADER  Header of the doubly linked lists.
C     LINKFD  Forward linked lists.
C     LINKBK  Backward linked lists.
C
C     TCODE   Termination code:
C             0  OPTIMAL solution found;
C             1  Primal INFEASIBLE (or dual UNBOUNDED);
C             2  Primal UNBOUNDED (or dual INFEASIBLE);
C             3  SUBOPTIMAL solution found (accuracy problems);
C             4  Excess iterations/time limit.
C     ITER    Iteration counter.
C
C
C
C *** PRESOLVE PARAMETERS
      INTEGER*4 LNHIST,LEVPRS,MSGLEV
      INTEGER*2 INHIST(MXHIST)
      DOUBLE PRECISION DPHIST(MXHIST)
C
C     LEVPRS  The level of presolve analysis desired:
C             7  corresponds to maximum available presolve analysis
C                (with no limits on implied variables' bounds), 
C             6  corresponds to DEFAULT presolve analysis,
C             5  turns off splitting dense columns, 
C             4  additionally turns off the heuristic to make A sparser,
C             3  additionally turns off column aggregation technique,
C             2  additionally turns off the elimin. of dominated cols, 
C             1  additionally turns off the elimin. of redundant rows, 
C             0  turns off any presolve analysis.
C     LNHIST  Length of the PRE_SOLVE history list.
C     MXHIST  Maximum number of entries in the PRE_SOLVE history list.
C     INHIST  Integer PRE_SOLVE history information.
C     DPHIST  Double precision PRE_SOLVE history information.
C
C
C
C *** LOCAL VARIABLES
      INTEGER*4 IROBJ,M,M0,MFINAL,N,NSTRCT,NZA,MAXCOL,NZL
      INTEGER*4 I,ICALL,INTSZ,TCODE,ITER,LORD,IWRITE
      CHARACTER*100 BUFFER
      DOUBLE PRECISION MULT
      INTEGER*4 IDIGTS
      DOUBLE PRECISION ERRORB,ERRORU,ERRORC,ACSMLL,ACLRGE
C
C     IROBJ   Index of the objective row (original row order).
C     M       Current number of constraints.
C     M0      Original number of constraints.
C     N       Number of variables (total, i.e. including slacks, surplus
C             and artificials).
C     NZA     Current number of nonzeros of the LP constraint matrix.
C     NSTRCT  Number of structural variables (excluding slacks, surplus
C             and artificials).
C     INTSZ   A size of INTEGER used to store indices.
C             2 means that half-length integers are used,
C             4 means that full-length integers are used.
C
C
C
C *** DATA STRUCTURES FOR CHOLESKY FACTOR
      DOUBLE PRECISION LCOEFF(MAXNZL)
      DOUBLE PRECISION LDIAG(MDIM),LDSQRT(MDIM)
      INTEGER*4 LCLPTS(MDIM+1),LLINKS(MAXNZL)
      INTEGER*2 LRWNBS(MAXNZL)
      EQUIVALENCE (LCOEFF(1),LLINKS(1))
C
C     LCOEFF  Off-diagonal nonzero coefficients of Cholesky matrix.
C     LCLPTS  Pointers to columns of the Cholesky factor.
C     LRWNBS  Row numbers of nonzeros in columns of matrix  L.
C     LLINKS  Linked lists for Cholesky factor.
C     LDIAG   Diagonal elements of Cholesky factor.
C     LDSQRT  Square roots of the diagonal elements of Cholesky factor.
C
C
C
C *** COMMON ARREAS
C     Cholesky factorization parameters.
      COMMON /CHFCT/   CSMALL,PIVTOL,TAU,DENSE,PIVMAX,PIVMIN,IDNSRW
      DOUBLE PRECISION CSMALL,PIVTOL,TAU,DENSE,PIVMAX,PIVMIN
      INTEGER*4        IDNSRW
C
C     Additional Cholesky fact. parameters (interface to HYBRID).
      COMMON /CHHYB/   RO,FLOPS,IREG,NZCHL,RTCD
      DOUBLE PRECISION RO,FLOPS
      INTEGER*4        IREG,NZCHL,RTCD
C
C     An indicator if the elimination routine has been used.
      COMMON /ELMNTE/  IELIM
      INTEGER*4        IELIM
C
C
C
C     For DOS, the integer array IDATIM is used by subroutine TIMEPF
C     to store the current date, time and elapsed time.
C     For UNIX, the real scalar ELTIME is used by subroutine DATTIM
C     to store the elapsed time.
C
C     Only for DOS
      COMMON/IDTM/ IDATIM
      INTEGER*4    IDATIM(9)
C
C     Only for UNIX
      COMMON /TIME/ ELTIME
      REAL ELTIME(3)
C
C     Markers for linking rows.
      COMMON /ICGRAD/ MSPLIT(100000)
      INTEGER*2       MSPLIT
C
C
C
C *** SUBROUTINES CALLED
C     MYWRT,RDSPEC,CHCKDM,SETMAP,RDMPS1,RDMPS2,PCPDM (or ACPDM),
C     PRESOL,PREPRO,INITSC,SCALEA,UNSCLA,POSTSL,WRTSOL,MYSTOP
C
C
C *** NOTES
C
C
C
C *** REFERENCES:
C     Altman A., Gondzio J. (1993). An efficient implementation of 
C        a higher order primal-dual interior point method for large 
C        sparse linear programs, Archives of Control Sciences 2, 
C        No 1-2, pp. 23-40. 
C     Altman A., Gondzio J. (1993). HOPDM - A higher order primal-
C        dual method for large scale linear programmming, European
C        Journal of Operational Research 66 (1993) pp 158-160.
C     Gondzio J. (1992). Splitting dense columns of the constraint
C        matrix in interior point methods for large scale linear
C        programming, Optimization 24, pp. 285-297.
C     Gondzio J. (1993). Implementing Cholesky factorization for 
C        interior point methods of linear programming, Optimization 
C        27, pp. 121-140.
C     Gondzio J. (1994). Presolve analysis of linear programs prior 
C        to applying an interior point method, Technical Report 
C        No 1994.3, Department of Management Studies, University 
C        of Geneva, 102, Bd. Carl-Vogt, 1211 Geneva, Switzerland, 
C        February 1994, revised in December 1994, ORSA Journal on
C        Computing (to appear). 
C     Gondzio J. (1994). Multiple centrality corrections in a primal-
C        dual method for linear programming, Technical Report 
C        No 1994.20, Department of Management Studies, University 
C        of Geneva, 102, Bd. Carl-Vogt, 1211 Geneva, Switzerland, 
C        November 1994, revised in May 1995, Computational 
C        Optimization and Applications (to appear). 
C     Gondzio J. (1995). HOPDM (version 2.12) - A fast LP solver 
C        based on a primal-dual interior point method, European
C        Journal of Operational Research 85, pp. 221-225.
C     Gondzio J., Makowski M. (1995). Solving a class of LP problems
C        with a primal-dual logarithmic barrier method, European
C        Journal of Operational Research 80, pp. 184-192.
C     Gondzio J., Tachat D. (1994). The design and application of 
C        IPMLO - a FORTRAN library for linear optimization with 
C        interior point methods, RAIRO Recherche Operationnelle 28, 
C        No 1, pp. 37-56.
C
C
C
C *** HISTORY:
C     The first version of this routine (called HOMAIN) was
C     written by:    Anna Altman & Jacek Gondzio,
C                    Systems Research Institute,
C                    Polish Academy of Sciences,
C                    Newelska 6, 01-447 Warsaw, Poland.
C     Date written:  May 18, 1992
C
C     This is a second release of it (called HMAIN2),
C     written by:    Jacek Gondzio,
C                    Systems Research Institute,
C                    Polish Academy of Sciences,
C                    Newelska 6, 01-447 Warsaw, Poland.
C     Date written:  November 18, 1993
C     Last modified: February 8, 1996
C
C
C
C
C
C
C *** BODY OF (HMAIN2) ***
C
C
C
C     Print the author's names at the console.
      WRITE(BUFFER,51)
   51 FORMAT(1X)
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(0,BUFFER)
      WRITE(BUFFER,52)
   52 FORMAT(2X,' HOPDM - Higher Order Primal-Dual Method, ver 2.13')
      CALL MYWRT(0,BUFFER)
      WRITE(BUFFER,53)
   53 FORMAT(2X,'--------------------------------------------------')
      CALL MYWRT(0,BUFFER)
      WRITE(BUFFER,51)
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(0,BUFFER)
      WRITE(BUFFER,54)
   54 FORMAT(2X,'Written by:    Jacek Gondzio,')
      CALL MYWRT(0,BUFFER)
      WRITE(BUFFER,55)
   55 FORMAT(17X,'Systems Research Institute,')
      CALL MYWRT(0,BUFFER)
      WRITE(BUFFER,56)
   56 FORMAT(17X,'Polish Academy of Sciences,')
      CALL MYWRT(0,BUFFER)
      WRITE(BUFFER,57)
   57 FORMAT(17X,'Newelska 6, 01-447 Warsaw, Poland')
      CALL MYWRT(0,BUFFER)
      WRITE(BUFFER,51)
      CALL MYWRT(0,BUFFER)
      WRITE(BUFFER,58)
   58 FORMAT(2X,'Last modified: February 8, 1996')
      CALL MYWRT(0,BUFFER)
      WRITE(BUFFER,51)
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(0,BUFFER)
C
C
C
C     Set the maximum order of Taylor polynomial.
      LORD=2
C
C     Set an indicator if the elimination routine has been used.
      IELIM=0
C
C     Set an indicator of how small pivots are to be handled.
C     IREG=-1
C
C     Set the level of printing pre_solve report desired.
      MSGLEV=0
C
C
C
C     Set default problem dimensions.
      MAXM=MDIM-1
      MAXN=NDIM-1
      MAXNZA=NZDIM-1
C
C     Read the problem specifications.
      FILSPC='spc'
C
      CALL RDSPEC(FILMPS,FILSPC,FILLOG,FILSOL,
     X MAXM,MAXN,MAXNZA,NAMEC,NAMEB,NAMBND,NAMRAN,
     X MULT,BIG,DLOBND,DUPBND,OPTTOL,LEVPRS)
C
C     Open the output files.
      OPEN(IOLOG,FILE=FILLOG,ACCESS='SEQUENTIAL')
      OPEN(99,FILE='fort.99',ACCESS='SEQUENTIAL')
C
C     Check if dimensions do not exceed allocated space.
      CALL CHCKDM(MDIM,NDIM,NZDIM,MAXM,MAXN,MAXNZA,IOLOG)
C
C
C
C     Set up the maps of the hidden problem data.
C     INTSZ   A size of INTEGER used to store indices.
C             2 means that half-length integers are used,
C             4 means that full-length integers are used.
C
      INTSZ=2
      CALL SETMAP(MAXM,MAXN,MAXNZA,
     X IMAP,RMAP,LIWORK,LRWORK,INTSZ,IOLOG)
C
C     Print the current date, time and elapsed time.
      WRITE(BUFFER,101)
  101 FORMAT(1X)
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,102)
  102 FORMAT(1X,'HOMAIN: Reading the MPS file.')
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
C
      CALL MYTIME(0,IOLOG)
      CALL MYTIME(0,0)
C
C
C
C     Read the MPS input file.
C
      CALL RDMPS1(MAXM,MAXN,MAXNZA,
     X M,N,NZA,IROBJ,IOLOG,
     X BIG,DLOBND,DUPBND,  
     X NAMEC,NAMEB,NAMRAN,NAMBND,NAMMPS,FILMPS,
     X RWNAME,CLNAME,STAVAR,RWSTAT,
     X HEADER,LINKFD,INTMP2,INTMP3,
     X IWORK(IMAP(2)),IWORK(IMAP(1)),IROW,
     X RWORK(RMAP(1)),RWORK(RMAP(3)),RANGES,
     X UPBND,LOBND,RELT)
C
      CALL RDMPS2(RWORK(RMAP(2)),RWORK(RMAP(3)),RANGES,
     X IWORK(IMAP(1)),IWORK(IMAP(2)),RWORK(RMAP(1)),
     X IWORK(IMAP(3)),IWORK(IMAP(4)),IWORK(IMAP(5)),IWORK(IMAP(6)),
     X MAXM,MAXN,MAXNZA,M,M0,N,NZA,NSTRCT,MULT,
     X INTMP1(1),IMTMP1,IMTMP2,
     X STAVAR,UPBND,LOBND,P,Q,BIG,IROBJ,
     X RWNAME,RWCOPY,RWSTAT,STAROW,CLNAME,
     X RWORK,IWORK,RMAP,IMAP,IROW,RELT,
     X X,IOLOG)
C
C     Print the current date, time and elapsed time.
      WRITE(BUFFER,103)
  103 FORMAT(1X,'HOMAIN: Reading completed.')
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
C
      CALL MYTIME(1,IOLOG)
      CALL MYTIME(1,0)
C
C
C
C     Print the current date, time and elapsed time.
      WRITE(BUFFER,201)
  201 FORMAT(1X)
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,202)
  202 FORMAT(1X,'HOMAIN: Presolve analysis.')
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
C
      CALL MYTIME(0,IOLOG)
      CALL MYTIME(0,0)
C
C
C
C     Go perform the PRE_SOLVE analysis.
C     LEVPRS is the level of presolve analysis desired.
C     MAXCOL is the threshold length for columns to be split.
C     LNHIST is the lenght of the PRE_solve history list.
C     MSGLEV is the desired level of output messages.
C
      MAXCOL=100
      LNHIST=0
      CALL PRESOL(MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X LNHIST,MXHIST,INHIST,DPHIST,
     X IWORK,RWORK,IMAP,RMAP,LIWORK,LRWORK,IROW,RELT,
     X X,RWORK(RMAP(3)),RWORK(RMAP(2)),CLNAME,UPBND,LOBND,
     X INTMP1,INTMP2,INTMP3,IMTMP1,IMTMP2,
     X RMTMP1,P,Q,RNTMP1,RNTMP2,RNTMP3,
     X PERM,INVP,HEADER,LINKFD,LINKBK,
     X STAVAR,RWNAME,STAROW,RWSTAT,RANGES,
     X MAXCOL,MSGLEV,LEVPRS,IOLOG)
C
C     Print the current date, time and elapsed time.
      WRITE(BUFFER,203)
  203 FORMAT(1X,'HOMAIN: Analysis completed.')
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
C
      CALL MYTIME(1,IOLOG)
      CALL MYTIME(1,99)
      CALL MYTIME(1,0)
C
C
C
C     Check if the PRE_SOLVE analysis has solved the problem.
      IF(M.EQ.0) THEN
         WRITE(BUFFER,211)
  211    FORMAT(1X)
         CALL MYWRT(0,BUFFER)
         CALL MYWRT(IOLOG,BUFFER)
         WRITE(BUFFER,212)
  212    FORMAT(1X,'HOMAIN: Problem solved by a PRE_SOLVE analysis!')
         CALL MYWRT(0,BUFFER)
         CALL MYWRT(IOLOG,BUFFER)
         WRITE(BUFFER,213)
  213    FORMAT(1X,'HOMAIN: Optimal solution found after    0',
     X    ' iterations.')
         CALL MYWRT(IOLOG,BUFFER)
         CALL MYWRT(99,BUFFER)
         WRITE(BUFFER,214)
  214    FORMAT(1X)
         CALL MYWRT(0,BUFFER)
         GO TO 600
      ENDIF
C
C
C
C     Print the current date, time and elapsed time.
      WRITE(BUFFER,301)
  301 FORMAT(1X)
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,302)
  302 FORMAT(1X,'HOMAIN: Preprocessing for Cholesky factorization.')
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
C
      CALL MYTIME(1,IOLOG)
      CALL MYTIME(1,0)
C
C
C
C     Go preprocess the  LP constraint matrix (reorder it according
C     to the permutation resulting from the mininmum degree heuristic
C     and perform the symbolic Cholesky factorization to set up data
C     structures for triangular factor  L).
C
      ICALL=0
      CALL PREPRO(MAXM,MAXN,MAXNZA,MAXNZL,M,N,
     X IWORK,RWORK,IMAP,RMAP,LIWORK,LRWORK,IROW,RELT,
     X INTMP1,IMTMP1,IMTMP2,RMTMP1,
     X DHEAD,PERM0,INVP0,NBRHD,QSIZE,QLINK,
     X PERM,INVP,HEADER,LINKFD,LINKBK,
     X LCLPTS,LRWNBS,LLINKS,
     X STAVAR,VUSED,VBNDED,P,Q,RWNAME,STAROW,RWSTAT,RANGES,
     X NZL,ICALL,IOLOG)
C
C     Print the current date, time and elapsed time.
      WRITE(BUFFER,303)
  303 FORMAT(1X,'HOMAIN: Preprocessing completed.')
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
C
      CALL MYTIME(1,IOLOG)
      CALL MYTIME(1,99)
      CALL MYTIME(1,0)
C
C
C
C
C     Scale the LP constraint matrix.
C
      WRITE(BUFFER,401)
  401 FORMAT(1X)
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,402)
  402 FORMAT(1X,'HOMAIN: Scaling the linear program.')
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
C
C     Initialize scaling factors.
C     Even if you do not scale matrix A, you have to initialize 
C     scaling factors and the infinity norms of the columns of A.
      CALL INITSC(M,N,COLNRM,CSCALE,RSCALE,OSCALE)
C
      CALL SCALEA(IOLOG,
     X MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X RMTMP1,RMTMP2,COLNRM,RNTMP2,
     X RWORK(RMAP(3)),RANGES,RWORK(RMAP(2)),UPBND,CSCALE,RSCALE,OSCALE,
     X RWORK(RMAP(1)),IWORK(IMAP(1)),IWORK(IMAP(2)),P,Q,
     X IWORK(IMAP(6)),STAVAR)
C
      WRITE(BUFFER,403)
  403 FORMAT(1X,'HOMAIN: Scaling done.')
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
C
C *** DEBUGGING
C     WRITE(IOLOG,461) 
C 461 FORMAT(1X/1X,'HMAIN2: LP rows before solution:')
C     DO 463 I=1,M
C        WRITE(BUFFER,462) I,RWNAME(I),RWSTAT(I),
C    X    P(I),Y(I),Q(I)
C 462    FORMAT(1X,'rw=',I4,', ',A8,' st=',I2,
C    X    ' Pi=',D14.8,' Yi=',D14.8,' Qi=',D14.8)
C        CALL MYWRT(IOLOG,BUFFER)
C 463 CONTINUE
C
C
C
C     Print the current date, time and elapsed time.
      WRITE(BUFFER,501)
  501 FORMAT(1X)
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,502)
  502 FORMAT(1X,'HOMAIN: Solution of the  LP problem starts.')
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
C
      CALL MYTIME(1,IOLOG)
      CALL MYTIME(1,0)
C
C
C
C     Go solve the  LP problem.
C
      CALL PCPDM(OPTTOL,TCODE,ITER,
     X LORD,MAXM,MAXN,MAXNZA,MAXNZL,M,MFINAL,N,NSTRCT,NZA,
     X IWORK,RWORK,IMAP,RMAP,LIWORK,LRWORK,IROW,RELT,
     X INTMP1,INTMP2,INTMP3,IMTMP1,IMTMP2,
     X RMTMP1,RMTMP2,RMTMP3,RNTMP1,RNTMP2,RNTMP3,
     X PERM,INVP,HEADER,LINKFD,LINKBK,
     X VUSED,VBNDED,RWORK(RMAP(2)),UPBND,RWORK(RMAP(3)),RANGES,
     X THETA,XIB,XIC,XIU,RMTMP4,RNTMP4,RNTMP5,RNTMP6,
     X COLNRM,X,S,Y,Z,W,
     X DELTAX,DELTAS,DELTAY,DELTAZ,DELTAW,YPROX,
     X LCOEFF,LDIAG,LDSQRT,LCLPTS,LRWNBS,LLINKS,
     X RSCALE,CSCALE,STAVAR,P,Q,STAROW,RWSTAT,RWNAME,IOLOG)
C
C     IDIGTS=3
C     CALL ACPDM(OPTTOL,TCODE,ITER,
C    X LORD,MAXM,MAXN,MAXNZA,MAXNZL,M,MFINAL,N,NSTRCT,NZA,
C    X IDIGTS,ERRORB,ERRORU,ERRORC,ACSMLL,ACLRGE,
C    X IWORK,RWORK,IMAP,RMAP,LIWORK,LRWORK,IROW,RELT,
C    X INTMP1,INTMP2,INTMP3,IMTMP1,IMTMP2,
C    X RMTMP1,RMTMP2,RMTMP3,RNTMP1,RNTMP2,RNTMP3,
C    X PERM,INVP,HEADER,LINKFD,LINKBK,
C    X VUSED,VBNDED,RWORK(RMAP(2)),UPBND,RWORK(RMAP(3)),RANGES,
C    X THETA,XIB,XIC,XIU,RMTMP4,RNTMP4,RNTMP5,RNTMP6,
C    X COLNRM,X,S,Y,Z,W,
C    X DELTAX,DELTAS,DELTAY,DELTAZ,DELTAW,YPROX,
C    X LCOEFF,LDIAG,LDSQRT,LCLPTS,LRWNBS,LLINKS,
C    X RSCALE,CSCALE,STAVAR,P,Q,STAROW,RWSTAT,RWNAME,IOLOG)
C
C     Print the current date, time and elapsed time.
      WRITE(BUFFER,503)
  503 FORMAT(1X)
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,504) TCODE
  504 FORMAT(1X,'HOMAIN: Exit HOPDM, termination code =',I3)
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
      CALL MYWRT(99,BUFFER)
C     WRITE(BUFFER,505) ITER
C 505 FORMAT(1X,'HOMAIN: Exit HOPDM after',I4,' iterations.')
C     CALL MYWRT(0,BUFFER)
C     CALL MYWRT(IOLOG,BUFFER)
C     CALL MYWRT(99,BUFFER)
C     DO 520 J=1,N
C        IF(INTMP1(J).EQ.1) THEN
C           WRITE(BUFFER,521) J
C 521       FORMAT(1X,' variable J=',I6,'  goes to lower bnd.')
C           CALL MYWRT(IOLOG,BUFFER)
C        ENDIF
C 520 CONTINUE
C
      CALL MYTIME(1,IOLOG)
      CALL MYTIME(1,99)
      CALL MYTIME(1,0)
C
C
C
C     Unscale the LP constraint matrix.
C
      CALL UNSCLA(IOLOG,MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X X,Y,S,Z,W,RWORK(RMAP(3)),RANGES,RWORK(RMAP(2)),UPBND,STAVAR,
     X P,Q,CSCALE,RSCALE,OSCALE,RWORK(RMAP(1)),IWORK(IMAP(1)),
     X IWORK(IMAP(3)),IWORK(IMAP(4)),IWORK(IMAP(5)),IWORK(IMAP(6)))
C
  600 CONTINUE
C
C
C
C *** DEBUGGING
C     WRITE(IOLOG,601) MFINAL,M
C 601 FORMAT(1X/1X,'after solution,  Mfinal=',I6,'  M=',I6)
C     DO 605 I=1,MFINAL
C        IF(Y(I).LE.P(I)-1.D-6.OR.Y(I).GE.Q(I)+1.D-6) THEN
C           WRITE(BUFFER,602) I,RWNAME(I),RWSTAT(I),
C    X       P(I),Y(I),Q(I)
C 602       FORMAT(1X,'rw=',I4,', ',A8,' st=',I2,
C    X       ' Pi=',D14.8,' Yi=',D14.8,' Qi=',D14.8)
C           CALL MYWRT(IOLOG,BUFFER)
C           IF(Y(I).LE.P(I)-1.D-6.OR.Y(I).GE.Q(I)+1.D-6) THEN
C              WRITE(BUFFER,603) I
C 603          FORMAT(1X,'Yi in row=',I6,' is out of bounds.')
C              CALL MYWRT(IOLOG,BUFFER)
C           ENDIF
C        ENDIF
C 605 CONTINUE
C
C
C
C     Go postprocess the  LP constraint matrix.
C
      CALL POSTSL(IOLOG,MSGLEV,
     X MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X LNHIST,MXHIST,INHIST,DPHIST,
     X RWORK,IWORK,RMAP,IMAP,IROW,RELT,
     X IMTMP1,INTMP1,INTMP2,RNTMP1,
     X RWORK(RMAP(3)),RANGES,RWORK(RMAP(2)),LOBND,UPBND,
     X RWORK(RMAP(1)),IWORK(IMAP(1)),IWORK(IMAP(2)),
     X IWORK(IMAP(3)),IWORK(IMAP(4)),IWORK(IMAP(5)),IWORK(IMAP(6)),
     X X,STAVAR,RWSTAT,STAROW,RWNAME,CLNAME)
C
C
C
C     Write the solution.
C     Standard MPS output file can be created only if the option
C     of rows and columns elimination is disabled. If this option
C     is used, complete output is not available.
      IWRITE=1
C
      CALL WRTSOL(FILSOL,IROBJ,M,M0,MFINAL,N,NSTRCT,MAXM,MAXN,
     X STAVAR,RWSTAT,STAROW,RWNAME,RWCOPY,CLNAME,NAMMPS,MULT,
     X LOBND,UPBND,RWORK(RMAP(3)),RWORK(RMAP(2)),X,Y,YPROX,
     X RWORK,IWORK,RMAP,IMAP,IROW,RELT,
     X RMTMP1,RMTMP2,IWRITE,IOLOG)
C
C
C
C
C     Close the files.
      CLOSE(IOLOG)
      CLOSE(99)
C
      CALL MYSTOP(TCODE)
C
C
C *** LAST CARD OF (HMAIN2) ***
      END
