C*****************************************************************
C     *** IRSOLV ... SOLVE EQUATION WITH  A*THETA*Atransp ***
C         USE ITERATIVE REFINEMENT TO IMPROVE THE ACCURACY
C*****************************************************************
C
      SUBROUTINE IRSOLV(LCOEFF,LCLPTS,LRWNBS,LDIAG,
     X MAXM,MAXN,MAXNZA,MAXNZL,M,N,ITREF,IALARM,
     X IWORK,RWORK,IMAP,RMAP,LIWORK,LRWORK,IROW,RELT,
     X THETA,STAVAR,VUSED,
     X RMWRK1,RMWRK2,RMWRK3,RNWRK1,RNWRK2,RNWRK3,
     X FNEW,DDD,DELX,DELY,RESX,RESY,IOLOG)
C
C *** PARAMETERS
      INTEGER*4 MAXM,MAXN,MAXNZA,MAXNZL,M,N,LIWORK,LRWORK
      INTEGER*4 ITREF,IALARM,IOLOG
      DOUBLE PRECISION LCOEFF(*),LDIAG(MAXM)
      INTEGER*4 LCLPTS(MAXM+1)
      INTEGER*2 LRWNBS(MAXNZL)
      INTEGER*4 IROW(MAXN)
      DOUBLE PRECISION RELT(MAXN),THETA(MAXN)
      INTEGER*2 STAVAR(MAXN)
      LOGICAL VUSED(MAXN)
      DOUBLE PRECISION RMWRK1(MAXM),RMWRK2(MAXM),RMWRK3(MAXM)
      DOUBLE PRECISION RNWRK1(MAXN),RNWRK2(MAXN),RNWRK3(MAXN)
      DOUBLE PRECISION FNEW(MAXN),DDD(MAXM),DELX(MAXN),DELY(MAXM)
      DOUBLE PRECISION RESX,RESY
C
C *** HIDDEN DATA STRUCTURES
      INTEGER*4 IWORK(LIWORK),IMAP(*),RMAP(*)
      DOUBLE PRECISION RWORK(LRWORK)
C
C *** LOCAL VARIABLES
      INTEGER*4 I,ITER,J
      DOUBLE PRECISION RESX0,RESY0,RO2,RO3
      CHARACTER*100 BUFFER
C
C     Additional Cholesky fact. parameters (interface to HYBRID).
      COMMON /CHHYB/   RO,FLOPS,IREG,NZCHL,RTCD
      DOUBLE PRECISION RO,FLOPS
      INTEGER*4        IREG,NZCHL,RTCD
C
C     Markers for linking rows.
      COMMON /ICGRAD/ MSPLIT(100000)
      INTEGER*2       MSPLIT
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     MAXN    Maximum column dimension of the LP constraint matrix.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     MAXNZL  Maximum number of nonzeros of the Cholesky factor.
C     M       Number of rows of the LP constraint matrix
C             (and the dimension of  A*Atransp).
C     N       Number of columns of the LP constraint matrix.
C     ITREF   Number of steps of the iterative refinement to be
C             done to improve the accuracy of solution with
C             the Cholesky factorization of A*THETA*Atransp.
C     IALARM  Indicator if there are errors in the Cholesky factors:
C             0 if Cholesky factor is sufficiently stable;
C             1 if the iterative refinement process does not improve 
C               the accuracy;
C             2 if there are very large errors in Cholesky factor.
C     LCOEFF  Off-diagonal nonzero coefficients of Cholesky matrix.
C     LCLPTS  Pointers to columns of the Cholesky factor.
C     LRWNBS  Row numbers of nonzeros in columns of matrix  L.
C     LDIAG   Diagonal elements of Cholesky factor.
C     THETA   Diagonal weight matrix.
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C            -k  free variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicate the position of the original variable.
C     VUSED   An indicator if a variable is active in the optimization
C             process:
C             .TRUE.   active variable;
C             .FALSE.  FIXED variable.
C     FNEW    Right-hand-side of the equation (part refering to X).
C     DDD     Right-hand-side of the equation (part refering to Y).
C     IOLOG   Output unit number where log messages are to be written.
C
C *** COMMON ARREAS
C     IREG    Regularization:
C             0  add RO to all diagonal elements and increase small
C                pivots to TAUADD (used by HYBRID);
C             1  increase small pivots to TAUADD (used by HYBRID);
C            -1  increase very small pivots to TAUADD (used by HOPDM).
C     RO      Regularization parameter.
C
C
C     ON OUTPUT:
C     DELX    Solution of the equation (deltaX).
C     DELY    Solution of the equation (deltaY).
C     RESX    Residuum of the solution (part refering to deltaX).
C     RESY    Residuum of the solution (part refering to deltaY).
C
C     WORK ARRAYS:
C     IROW  and  RELT are the arrays for temporary handling
C             of rows/columns of the constraint matrix. They
C             are primarily intended to handle sparse vectors
C             (in packed form) but may also be used for storing
C             dense ones.
C     RMWRK1  Double precision work array of size MAXM.
C     RMWRK2  Double precision work array of size MAXM.
C     RMWRK3  Double precision work array of size MAXM.
C     RNWRK1  Double precision work array of size MAXN.
C     RNWRK2  Double precision work array of size MAXN.
C     RNWRK3  Double precision work array of size MAXN.
C
C     Map of IWORK array:
C     IMAP(1)   Points to CLPNTS array.
C     IMAP(2)   Points to RWNMBS array.
C     IMAP(3)   Points to RWHEAD array.
C     IMAP(4)   Points to RWLINK array.
C     IMAP(5)   Points to CLNMBS array.
C     IMAP(6)   Points to LENCOL array.
C     IMAP(7)   Points to the first empty cell of IWORK array.
C
C     Map of RWORK array:
C     RMAP(1)   Points to ACOEFF array.
C     RMAP(2)   Points to C array.
C     RMAP(3)   Points to RHS array.
C     RMAP(4)   Points to the first empty cell of RWORK array.
C
C
C *** SUBROUTINES CALLED:
C     DABS, SAX (or FSAX), SATY (or FSATY), SOLAAT
C
C *** PURPOSE:
C     This routine solves the equation with  A*THETA*Atransp.
C     It uses the Cholesky decomposition  L*D*Ltransp of the
C     above matrix. It performs the required number of steps
C     of the iterative refinement on the augmented system
C     formulation.
C
C *** NOTES:
C
C *** REFERENCES:
C     Altman A., Gondzio J. (1993). An efficient implementation of 
C        a higher order primal-dual interior point method for large 
C        sparse linear programs, Archives of Control Sciences 2, 
C        No 1-2, pp. 23-40. 
C     Altman A., Gondzio J. (1993). HOPDM - A higher order primal-
C        dual method for large scale linear programmming, European
C        Journal of Operational Research 66 (1993) pp 158-160.
C     Gondzio J. (1993). Implementing Cholesky factorization
C        for interior point methods of linear programming,
C        Optimization 27, pp. 121-140.
C     Gondzio J. (1994). Multiple centrality corrections in a primal-
C        dual method for linear programming, Technical Report 
C        No 1994.20, Department of Management Studies, University 
C        of Geneva, 102, Bd. Carl-Vogt, 1211 Geneva, Switzerland, 
C        November 1994, revised in May 1995, Computational 
C        Optimization and Applications (to appear). 
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Last modified: September 25, 1995
C
C
C *** BODY OF (IRSOLV) ***
C
C
      IALARM=0
      RESX=0.0D0
      RESY=0.0D0
C
C
C
C     Compute the right hand side of the normal eqations.
      DO 100 J=1,N
         IF(VUSED(J)) RNWRK1(J)=FNEW(J)*THETA(J)
  100 CONTINUE
      CALL FSAX(MAXM,MAXN,MAXNZA,RNWRK1,N,RMWRK1,M,
     X RWORK(RMAP(1)),IWORK(IMAP(1)),IWORK(IMAP(2)),IWORK(IMAP(6)),
     X VUSED,IOLOG)
      DO 200 I=1,M
         RMWRK1(I)=RMWRK1(I)+DDD(I)
  200 CONTINUE
C
C     Solve normal equations for deltaY.
      CALL SOLAAT(LCOEFF,LCLPTS,LRWNBS,LDIAG,
     X MAXNZL,MAXM,M,DELY,RMWRK1,IOLOG)
C
C     Compute deltaX.
      CALL FSATY(MAXM,MAXN,MAXNZA,DELY,M,RNWRK1,N,
     X RWORK(RMAP(1)),IWORK(IMAP(1)),IWORK(IMAP(2)),IWORK(IMAP(6)),
     X VUSED,IOLOG)
      DO 300 J=1,N
         IF(VUSED(J)) DELX(J)=(RNWRK1(J)-FNEW(J))*THETA(J)
  300 CONTINUE
C
C     Save current solution in RMWRK3 and RNWRK3 arrays.
      DO 400 I=1,M
         RMWRK3(I)=DELY(I)
  400 CONTINUE
      DO 500 J=1,N
         IF(VUSED(J)) RNWRK3(J)=DELX(J)
  500 CONTINUE
C
C
C
C
C
C     Main loop begins here.
      RO2=1.0D+1
      RO3=1.0D+3
      ITER=0
 1000 ITER=ITER+1
C
C
C
C     Compute the residual of the current solution in 
C     the augmented system of  KKT equations.
C     Use  RMWRK2  to store a residual for part Y (null space error).
      CALL FSAX(MAXM,MAXN,MAXNZA,DELX,N,RMWRK2,M,
     X RWORK(RMAP(1)),IWORK(IMAP(1)),IWORK(IMAP(2)),IWORK(IMAP(6)),
     X VUSED,IOLOG)
C
      RESX=0.0D0
      RESY=0.0D0
      DO 1200 I=1,M
         RMWRK2(I)=DDD(I)-RMWRK2(I)
         IF(MSPLIT(I).EQ.0) THEN
            RMWRK2(I)=RMWRK2(I)-RO*DELY(I)
         ELSE
            IF(MSPLIT(I).EQ.1) THEN
               RMWRK2(I)=RMWRK2(I)-RO2*DELY(I)
            ELSE
               RMWRK2(I)=RMWRK2(I)-RO3*DELY(I)
            ENDIF
         ENDIF
         IF(DABS(RMWRK2(I)).GT.RESY) RESY=DABS(RMWRK2(I))
 1200 CONTINUE
C
C *** DEBUGGING
      WRITE(BUFFER,1201) ITER,RESY
 1201 FORMAT(1X,'IRSOLV: Iter=',I2,' null space error=',1PD9.2)
C     CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
C
C     Save the residuals for the initial solution.
      IF(ITER.EQ.1) THEN
         RESX0=RESX
         RESY0=RESY
      ENDIF
      IF(RESX+RESY.LE.1.0D-8) GO TO 2000
      IF(RESX+RESY.GE.1.0D+6) THEN
         IALARM=2
         GO TO 2000
      ENDIF
      IF(RESX+RESY.LE.1.0D-7.AND.ITER.GE.2) GO TO 2000
      IF(ITER.GT.ITREF) GO TO 2000
C
C     Restore the initial solution if the residuals have increased
C     after the iterative refinemet step. Set IALARM parameter.
      IF(RESX+RESY.GT.RESX0+RESY0+1.0D-12) THEN
         DO 1300 I=1,M
            DELY(I)=RMWRK3(I)
 1300    CONTINUE
         DO 1400 J=1,N
            IF(VUSED(J)) DELX(J)=RNWRK3(J)
 1400    CONTINUE
         RESX=RESX0
         RESY=RESY0
C        WRITE(BUFFER,1401)
C1401    FORMAT(1X,'IRSOLV: Error growth in Cholesky factorization.')
C        CALL MYWRT(IOLOG,BUFFER)
         IALARM=1
         GO TO 2000
      ENDIF
      IF(RESX+RESY.GT.1.0D-1*(RESX0+RESY0).AND.ITER.GE.2) THEN
         IALARM=1
      ENDIF
C
C     Give up if residuals are too large.
      IF(RESX+RESY.GE.1.0D+1) GO TO 2000
C
C
C
C     Repeat solution of the augmented system for residuals.
C     Use  RNWRK1  to store a correction for part X.
C     Use  RMWRK1  to store a correction for part Y.
C
C     Solve normal equations for correction of deltaY.
      CALL SOLAAT(LCOEFF,LCLPTS,LRWNBS,LDIAG,
     X MAXNZL,MAXM,M,RMWRK1,RMWRK2,IOLOG)
C
C     Compute correction of deltaX.
      CALL FSATY(MAXM,MAXN,MAXNZA,RMWRK1,M,RNWRK1,N,
     X RWORK(RMAP(1)),IWORK(IMAP(1)),IWORK(IMAP(2)),IWORK(IMAP(6)),
     X VUSED,IOLOG)
C
C
C
C     Add corrections to the current solution.
      DO 1800 I=1,M
         DELY(I)=DELY(I)+RMWRK1(I)
 1800 CONTINUE
      DO 1900 J=1,N
         IF(VUSED(J)) DELX(J)=DELX(J)+RNWRK1(J)*THETA(J)
 1900 CONTINUE
C
C
C
C
C
C     End of the main loop.
      GO TO 1000
C
C
 2000 CONTINUE
      RETURN
C
C
C *** LAST CARD OF (IRSOLV) ***
      END
