C************************************************************
C     ****    NUMFCT ... NUMERICAL FACTORIZATION    ****
C************************************************************
C
      SUBROUTINE NUMFCT(LCOEFF,LCLPTS,LRWNBS,LDIAG,LDSQRT,
     X MAXNZL,MAXM,M,MKSQRT,RTCODE,
     X HEADER,LINKFD,LINKBK,RSTART,DPWORK,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 MAXNZL,MAXM,M,MKSQRT,RTCODE,IOLOG
      DOUBLE PRECISION LCOEFF(*),LDIAG(MAXM),LDSQRT(MAXM),DPWORK(MAXM)
      INTEGER*4 LCLPTS(MAXM+1),RSTART(MAXM)
C
C *** The following arrays can be half-length integer.
      INTEGER*2 LRWNBS(MAXNZL)
      INTEGER*2 HEADER(MAXM),LINKFD(MAXM),LINKBK(MAXM)
C
C
C *** LOCAL VARIABLES
      INTEGER*4 IROW,IRWACT,IX,JCOL,K,KBEG,KEND,LENROW,NEXT,NXTACT
      INTEGER*4 IRWMAX,IRWMIN,I,IRMV
      DOUBLE PRECISION DP,ELTMAX,PVCAND,CGROW,RGROW,SMALLP,TAUADD
      DOUBLE PRECISION RO2,RO3
      CHARACTER*100 BUFFER
C
C
C *** COMMON ARREAS
C     Cholesky factorization parameters.
      COMMON /CHFCT/   CSMALL,PIVTOL,TAU,DENSE,PIVMAX,PIVMIN,IDNSRW
      DOUBLE PRECISION CSMALL,PIVTOL,TAU,DENSE,PIVMAX,PIVMIN
      INTEGER*4        IDNSRW
C
C     Additional Cholesky fact. parameters (interface to HYBRID).
      COMMON /CHHYB/   RO,FLOPS,IREG,NZCHL,RTCD
      DOUBLE PRECISION RO,FLOPS
      INTEGER*4        IREG,NZCHL,RTCD
C
C     Markers for linking rows.
      COMMON /ICGRAD/ MSPLIT(100000)
      INTEGER*2       MSPLIT
C
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     MAXNZL  Maximum number of nonzeros of the Cholesky factor.
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     M       Number of rows of the LP constraint matrix
C             (and the dimension of  A*Atransp).
C     MKSQRT  Parameter indicating if square roots of LDIAG are to be
C             computed:
C             0  no square roots necessary;
C             1  compute square roots of diagonal matrix.
C     LCOEFF  Off-diagonal nonzero coefficients of  A*THETA*Atransp
C             matrix (fill-in positions are zeroed).
C     LCLPTS  Pointers to columns of the Cholesky factor.
C     LRWNBS  Row numbers of nonzeros in columns of matrix  L.
C     LDIAG   Diagonal elements of  A*THETA*Atransp.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     RTCODE  Indicator if there are errors in the Cholesky factors:
C             0 if Cholesky factor is sufficiently stable;
C             1 if Cholesky factorization failed.
C     LCOEFF  Off-diagonal nonzero coefficients of Cholesky matrix.
C     LCLPTS  Pointers to columns of the Cholesky factor.
C     LRWNBS  Row numbers of nonzeros in columns of matrix  L.
C     LDIAG   Diagonal elements of Cholesky factor.
C     LDSQRT  Square roots of the diagonal elements of Cholesky factor.
C
C     WORK ARRAYS:
C     HEADER  Header of the doubly linked lists of rows that have their
C             next active off-diagonal entries in the same columns.
C     LINKFD  Forward linked lists.
C     LINKBK  Backward linked lists.
C     RSTART  Pointer to the next active off-diagonal entry
C             of a given row.
C     DPWORK  Temporary work array.
C
C
C
C *** Some parameters used by the Cholesky decomposition.
C     CSMALL  During the Cholesky decomposition all numbers smaller
C             than  CSMALL (in the absolute value) are presumed
C             to be numerical errors only and are set to zero.
C             CSMALL is initialized to the computer relative precision.
C     PIVTOL  The tolerance for pivots in Cholesky factor  L.
C             Pivots smaller than  PIVTOL are rejected and the matrix
C             is presumed to be singular. The factorization is not
C             terminated, however. Pivot element is replaced with
C             a small positive value.
C     TAU     To avoid unpredicted exit from the Cholesky decomposition
C             a small multiple of the identity matrix is added to the 
C             A*THETA*Atransp matrix before its factorization.
C             It has a value equal to TAU times the largest diagonal
C             element of the matrix to be decomposed. TAU is
C             initialized to the value of computer relative precision.
C     DENSE   Threshold value for a column to be treated as dense.
C     IDNSRW  Index of row of the Cholesky factor for which a switch
C             is made to dense code.
C     IREG    Regularization:
C             0  add RO to all diagonal elements and increase small
C                pivots to TAUADD (used by HYBRID);
C             1  increase small pivots to TAUADD (used by HYBRID);
C            -1  increase very small pivots to TAUADD (used by HOPDM).
C     RO      Regularization parameter.
C
C
C
C *** SUBROUTINES CALLED:
C     MYWRT,DABS,DSQRT,DCOPY,DAXPY
C
C
C *** PURPOSE:
C     This routine implements the numerical factorization
C     for a symmetric positive definite matrix.
C     It decomposes positive definite symmetric martix
C     to the form:  L * D * Ltransp.
C
C
C *** NOTES:
C     1. This routine follows Duff et al. (1989) description
C        of the Cholesky factorization. It thus reflects
C        the multifrontal approach to it.
C     2. The lower right corner of the Cholesky factor is stored
C        as a dense matrix (double addressing is thus avoided).
C        IDNSRW (from CHFACT common block) is a number of the
C        first row of a dense window.
C
C
C *** REFERENCES:
C     Duff I.S., Erisman A.M., Reid J.K. (1989). Direct methods
C        for sparse matrices, Clarendon Press, Oxford 1989,
C        chapters  3 and  10.
C     Gondzio J. (1993). Implementing Cholesky factorization
C        for interior point methods of linear programming,
C        Optimization 27, pp. 121-140.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  April 15, 1991
C     Last modified: February 6, 1996
C
C
C
C *** BODY OF (NUMFCT) ***
C
C
C
      RTCODE=0
C
C
C
C
C     Find the largest and the smallest diagonal entries
C     of  A*THETA*Atransp.
      RO2=1.0D+1
      RO3=1.0D+3
      IX=1
      IRWMAX=1
      IRWMIN=1
      DP=DABS(LDIAG(1))
      PIVMAX=DP
      PIVMIN=DP
      DO 40 IROW=2,M
         DP=LDIAG(IROW)
         IF(DP.GT.PIVMAX) THEN
            PIVMAX=DP
            IRWMAX=IROW
         ENDIF
         IF(DP.LT.PIVMIN) THEN
            PIVMIN=DP
            IRWMIN=IROW
         ENDIF
   40 CONTINUE
C
C     WRITE(BUFFER,41) IRWMAX,LDIAG(IRWMAX)
C  41 FORMAT(1X,'NUMFCT: Max. diag. elt in row ',I6,',  Dii=',1PD12.4)
C     CALL MYWRT(IOLOG,BUFFER)
C     WRITE(BUFFER,42) IRWMIN,LDIAG(IRWMIN)
C  42 FORMAT(1X,'NUMFCT: Min. diag. elt in row ',I6,',  Dii=',1PD12.4)
C     CALL MYWRT(IOLOG,BUFFER)
C
C
C
C     Set the tolerance for acceptable pivot element.
C     Pivot elements smaller than PIVTOL will be replaced with TAUADD.
C
      IF(IREG.EQ.-1) THEN
C *** Here for HOPDM: correct only very small pivots of A*Atransp.
         TAUADD=TAU*LDIAG(IRWMAX)
         PIVTOL=1.0D-10*TAUADD
      ENDIF
C
      IF(IREG.GE.0) THEN
C *** Here for HYBRID: correct too small pivots of A*Atransp.
         TAUADD=TAU*LDIAG(IRWMAX)
         PIVTOL=TAUADD
         IF(IREG.EQ.0) THEN
C     Add regularizing term to A*Atransp (proximal algorithm).
C           DO 60 IROW=1,M
C              LDIAG(IROW)=LDIAG(IROW)+RO
C  60       CONTINUE
         ENDIF
C
C     Modification: June 10, 1994 for proximal point algorithm.
         PIVTOL=1.0D-10*TAUADD
C
      ENDIF
      SMALLP=0.999999D0*PIVTOL
C     WRITE(BUFFER,61) PIVTOL,TAUADD,RO
C  61 FORMAT(1X,'NUMFCT: PIVTOL=',1PD12.4,'  TAUADD=',1PD12.4,
C    X ' RO=',1PD12.4)
C     CALL MYWRT(IOLOG,BUFFER)
C
C
C
C     Zero  HEADER and  RSTART arrays.
      DO 100 IROW=1,M
         HEADER(IROW)=0
         RSTART(IROW)=0
         MSPLIT(IROW)=0
  100 CONTINUE
C
C
C     Set the doubly linked lists of rows that have the next
C     subdiagonal entry in the same columns. Recall that  LCLPTS(i)
C     indicates the first off-diagonal entry of row i.
C     Save also the locations of these elements in  RSTART array.
      DO 200 IROW=1,IDNSRW-1
         KBEG=LCLPTS(IROW)
         KEND=LCLPTS(IROW+1)-1
         IF(KBEG.GT.KEND) GO TO 200
         JCOL=LRWNBS(KBEG)
         NEXT=HEADER(JCOL)
         LINKFD(IROW)=NEXT
         HEADER(JCOL)=IROW
         IF(NEXT.GT.0) LINKBK(NEXT)=IROW
         LINKBK(IROW)=-JCOL
         RSTART(IROW)=KBEG
  200 CONTINUE
C
C
C
C
C
C     The first main loop begins here.
C     Loop over SPARSE rows of Cholesky factor.
C     The contents of a given row  IROW are first unpacked into
C     a full array  DPWORK. All rows that have an off-diagonal
C     entry in column  IROW are then scanned. Their contributions
C     are added to  DPWORK array. Having completed the above
C     calculations the elements of  DPWORK array are packed back
C     to the data structures for Cholesky factor. All the rows
C     involved in a pivotal step (excluding the pivotal one)
C     are added to the linked lists determined by their next
C     off-diagonal entry.
C
      DO 1000 IROW=1,IDNSRW-1
C
C
C     Unpack row  IROW of Cholesky matrix to a work array.
         KBEG=LCLPTS(IROW)
         KEND=LCLPTS(IROW+1)-1
         DO 300 K=KBEG,KEND
            JCOL=LRWNBS(K)
            DPWORK(JCOL)=LCOEFF(K)
  300    CONTINUE
C
C
C     Scan all rows of Cholesky matrix involved in this pivotal step
C     (i.e. those which have the next active element in column  IROW).
C     IRWACT is a number of such an active row.
C     NXTACT is a number of the next active row.
C     ELTMAX is a maximum element in an actvive row.
         IRWACT=HEADER(IROW)
         ELTMAX=0.0D0
  350    IF(IRWACT.EQ.0) GO TO 500
         NXTACT=LINKFD(IRWACT)
         KBEG=RSTART(IRWACT)
         KEND=LCLPTS(IRWACT+1)-1
         IF(DABS(LCOEFF(KBEG)).GE.ELTMAX) THEN
            ELTMAX=DABS(LCOEFF(KBEG))
         ENDIF
         DP=LDIAG(IRWACT)*LCOEFF(KBEG)
C        LDIAG(IROW)=LDIAG(IROW)-DP*LCOEFF(KBEG)
         DO 400 K=KBEG+1,KEND
            JCOL=LRWNBS(K)
            DPWORK(JCOL)=DPWORK(JCOL)-DP*LCOEFF(K)
  400    CONTINUE
C
C
C     Here if the contribution of row  IRWACT has been added.
C     Find the next active entry of this row (i.e.  JCOL)
C     and add the row to the appropriate linked list.
         IF(KBEG+1.GT.KEND) GO TO 450
         JCOL=LRWNBS(KBEG+1)
         NEXT=HEADER(JCOL)
         HEADER(JCOL)=IRWACT
         LINKFD(IRWACT)=NEXT
         LINKBK(IRWACT)=-JCOL
         IF(NEXT.GT.0) LINKBK(NEXT)=IRWACT
C
C     Save the position of the next active entry.
         RSTART(IRWACT)=KBEG+1
C
C
C     Continue the scanning of all rows involved in this pivotal step.
  450    IRWACT=NXTACT
         GO TO 350
C
C
C
C     Here if the pivot row is fully determined.
C     Pack it back to the data structures for Cholesky factor.
  500    KBEG=LCLPTS(IROW)
         KEND=LCLPTS(IROW+1)-1
         IF(LDIAG(IROW).GE.1.0D+2*TAUADD) GO TO 800
C
C
C     Handling small pivots.
         IF(LDIAG(IROW).LE.PIVTOL) LDIAG(IROW)=PIVTOL
C
C     Analyse growth factor in a row.
         IRMV=0
         RGROW=ELTMAX/(LDIAG(IROW)+RO)
         IF(RGROW.GE.1.0D+12) THEN
            LDIAG(IROW)=LDIAG(IROW)+RO2-RO
            MSPLIT(IROW)=1
            WRITE(BUFFER,701) IROW,LDIAG(IROW),RGROW
  701       FORMAT(1X,'May95,s rw=',I6,' Dii=',1PD12.4,' rgr=',1PD8.1)
            CALL MYWRT(IOLOG,BUFFER)
         ENDIF
         RGROW=ELTMAX/LDIAG(IROW)
C
C     Analyse growth factor in a column. Choose the minimum value
C     of the pivot that keeps AAt positive definite.
         CGROW=0.0D0
         PVCAND=0.0D0
         DO 710 K=KBEG,KEND
            JCOL=LRWNBS(K)
            IF(DABS(DPWORK(JCOL)).GE.CGROW) CGROW=DABS(DPWORK(JCOL))
            DP=DPWORK(JCOL)*DPWORK(JCOL)/LDIAG(JCOL)
            IF(DP.GE.PVCAND) PVCAND=DP
  710    CONTINUE
         CGROW=CGROW/LDIAG(IROW)
         IF(CGROW.GE.1.0D+7) THEN
            WRITE(BUFFER,711) IROW,MSPLIT(IROW),CGROW
  711       FORMAT(1X,'sparse: rw=',I6,' msplit=',I2,' CGROW=',1PD8.1)
            CALL MYWRT(IOLOG,BUFFER)
            IF(MSPLIT(IROW).EQ.0) THEN
               LDIAG(IROW)=LDIAG(IROW)+RO2-RO
               MSPLIT(IROW)=1
               WRITE(BUFFER,712) IROW,LDIAG(IROW)
  712          FORMAT(1X,'May95,3s: rw=',I6,' Dii=',1PD12.4)
               CALL MYWRT(IOLOG,BUFFER)
            ELSE
               WRITE(BUFFER,713) CGROW
  713          FORMAT(1X,'sparse: rw should be removed, cgr=',1PD8.1)
               CALL MYWRT(IOLOG,BUFFER)
            ENDIF
         ENDIF
         IF(PVCAND.GE.0.9999D0*LDIAG(IROW)) THEN
            WRITE(BUFFER,721) IROW,LDIAG(IROW),PVCAND
  721       FORMAT(1X,'irow=',I6,' Dii=',1PD16.8,
     X       ' pivot candidate p=',1PD16.8)
            CALL MYWRT(IOLOG,BUFFER)
            IF(PVCAND.GE.1.0002D0*LDIAG(IROW)) THEN
               IF(PVCAND.GE.0.99D0*RO2) THEN 
                  WRITE(BUFFER,722) PVCAND
  722             FORMAT(1X,'sparse: rw shd be removed, pcnd=',1PD8.1)
                  CALL MYWRT(IOLOG,BUFFER)
                  IF(PVCAND.GE.0.99D0*RO3) IRMV=1
                  LDIAG(IROW)=LDIAG(IROW)+RO3-RO
                  MSPLIT(IROW)=2
               ELSE
                  LDIAG(IROW)=LDIAG(IROW)+RO2-RO
                  MSPLIT(IROW)=1
               ENDIF
               IF(PVCAND.GE.1.0D+16) THEN 
                  RTCODE=1
                  GO TO 6000
               ENDIF
            ENDIF
            WRITE(BUFFER,723) IROW,MSPLIT(IROW),LDIAG(IROW)
  723       FORMAT(1X,'May95,2s: rw=',I6,' msplit=',I2,' Dii=',1PD9.1)
            CALL MYWRT(IOLOG,BUFFER)
         ENDIF
  800    LDIAG(IROW)=LDIAG(IROW)+RO
         IF(IRMV.EQ.1) THEN
            WRITE(BUFFER,801) IROW,LDIAG(IROW)
  801       FORMAT(1X,'sparse: rw=',I6,' Dii=',1PD12.4,' is removed.')
            CALL MYWRT(IOLOG,BUFFER)
            DO 820 K=KBEG,KEND
               JCOL=LRWNBS(K)
               DPWORK(JCOL)=0.0D0
  820       CONTINUE
         ENDIF
         DP=1.0/LDIAG(IROW)
         DO 900 K=KBEG,KEND
            JCOL=LRWNBS(K)
            LCOEFF(K)=DP*DPWORK(JCOL)
            LDIAG(JCOL)=LDIAG(JCOL)-DP*DPWORK(JCOL)*DPWORK(JCOL)
            IF(LDIAG(JCOL).LE.SMALLP) LDIAG(JCOL)=SMALLP
  900    CONTINUE
C
C
C
C     End of main loop.
 1000 CONTINUE
C
C
C
C
C
C     The second main loop begins here.
C     Loop over DENSE rows of Cholesky factor.
C     The contents of a given row  IROW are first unpacked into
C     a full array  DPWORK. All rows that have an off-diagonal
C     entry in column  IROW are then scanned. Their contributions
C     are added to  DPWORK array. Having completed the above
C     calculations the elements of  DPWORK array are packed back
C     to the data structures for Cholesky factor. All the sparse rows
C     involved in a pivotal step are added to the linked lists
C     determined by their next off-diagonal entry.
C
      DO 2000 IROW=IDNSRW,M
C
C
C     Unpack row  IROW of Cholesky matrix to a work array.
         KBEG=LCLPTS(IROW)
         LENROW=LCLPTS(IROW+1)-KBEG
C        CALL DCOPY(LCOEFF(KBEG),DPWORK(IROW+1),LENROW)
         call dcopy(LENROW,LCOEFF(KBEG),ix,DPWORK(IROW+1),ix)
C
C
C     Scan all rows of Cholesky matrix involved in this pivotal step
C     (i.e. those which have the next active element in column  IROW).
C
C
C     First, take account of sparse rows.
C     IRWACT is a number of such an active row.
C     NXTACT is a number of the next active row.
C     ELTMAX is a maximum element in an actvive row.
         IRWACT=HEADER(IROW)
         ELTMAX=0.0D0
 1350    IF(IRWACT.EQ.0) GO TO 1500
         NXTACT=LINKFD(IRWACT)
         KBEG=RSTART(IRWACT)
         KEND=LCLPTS(IRWACT+1)-1
         IF(DABS(LCOEFF(KBEG)).GE.ELTMAX) THEN
            ELTMAX=DABS(LCOEFF(KBEG))
         ENDIF
         DP=LDIAG(IRWACT)*LCOEFF(KBEG)
C        LDIAG(IROW)=LDIAG(IROW)-DP*LCOEFF(KBEG)
         DO 1400 K=KBEG+1,KEND
            JCOL=LRWNBS(K)
            DPWORK(JCOL)=DPWORK(JCOL)-DP*LCOEFF(K)
 1400    CONTINUE
C
C
C     Here if the contribution of row  IRWACT has been added.
C     Find the next active entry of this row (i.e.  JCOL)
C     and add the row to the appropriate linked list.
         IF(KBEG+1.GT.KEND) GO TO 1450
         JCOL=LRWNBS(KBEG+1)
         NEXT=HEADER(JCOL)
         HEADER(JCOL)=IRWACT
         LINKFD(IRWACT)=NEXT
         LINKBK(IRWACT)=-JCOL
         IF(NEXT.GT.0) LINKBK(NEXT)=IRWACT
C
C     Save the position of the next active entry.
         RSTART(IRWACT)=KBEG+1
C
C
C     Continue the scanning of all rows involved in this pivotal step.
 1450    IRWACT=NXTACT
         GO TO 1350
C
C
C     Take account of dense rows.
 1500    DO 1600 IRWACT=IDNSRW,IROW-1
         KBEG=LCLPTS(IRWACT+1)-LENROW-1
         IF(DABS(LCOEFF(KBEG)).GE.ELTMAX) THEN
            ELTMAX=DABS(LCOEFF(KBEG))
         ENDIF
         DP=LDIAG(IRWACT)*LCOEFF(KBEG)
C        LDIAG(IROW)=LDIAG(IROW)-DP*LCOEFF(KBEG)
C        CALL DAXPY(LCOEFF(KBEG+1),DPWORK(IROW+1),LENROW,-DP)
         call daxpy(LENROW,-DP,LCOEFF(KBEG+1),ix,DPWORK(IROW+1),ix)
 1600 CONTINUE
C
C
C
C     Here if the pivot row is fully determined.
C     Pack it back to the data structures for Cholesky factor.
         KBEG=LCLPTS(IROW)
         LENROW=LCLPTS(IROW+1)-KBEG
         KBEG=KBEG-1
         IF(LDIAG(IROW).GE.1.0D+2*TAUADD) GO TO 1800
C
C
C     Handling small pivots.
         IF(LDIAG(IROW).LE.PIVTOL) LDIAG(IROW)=PIVTOL
C
C     Analyse growth factor in a row.
         IRMV=0
         RGROW=ELTMAX/(LDIAG(IROW)+RO)
         IF(RGROW.GE.1.0D+12) THEN
            LDIAG(IROW)=LDIAG(IROW)+RO2-RO
            MSPLIT(IROW)=1
            WRITE(BUFFER,1701) IROW,LDIAG(IROW),RGROW
 1701       FORMAT(1X,'May95,d rw=',I6,' Dii=',1PD12.4,' rgr=',1PD8.1)
            CALL MYWRT(IOLOG,BUFFER)
         ENDIF
         RGROW=ELTMAX/LDIAG(IROW)
C
C     Analyse growth factor in a column.
         CGROW=0.0D0
         PVCAND=0.0D0
         DO 1710 I=1,LENROW
            JCOL=IROW+I
            IF(DABS(DPWORK(JCOL)).GE.CGROW) CGROW=DABS(DPWORK(JCOL))
            DP=DPWORK(JCOL)*DPWORK(JCOL)/LDIAG(JCOL)
            IF(DP.GE.PVCAND) PVCAND=DP
 1710    CONTINUE
         CGROW=CGROW/LDIAG(IROW)
         IF(CGROW.GE.1.0D+7) THEN
            WRITE(BUFFER,1711) IROW,MSPLIT(IROW),CGROW
 1711       FORMAT(1X,'dense:  rw=',I6,' msplit=',I2,' CGROW=',1PD8.1)
            CALL MYWRT(IOLOG,BUFFER)
            IF(MSPLIT(IROW).EQ.0) THEN
               LDIAG(IROW)=LDIAG(IROW)+RO2-RO
               MSPLIT(IROW)=1
               WRITE(BUFFER,1712) IROW,LDIAG(IROW)
 1712          FORMAT(1X,'May95,3d: rw=',I6,' Dii=',1PD12.4)
               CALL MYWRT(IOLOG,BUFFER)
            ELSE
               WRITE(BUFFER,1713) CGROW
 1713          FORMAT(1X,'dense:  rw should be removed, cgr=',1PD8.1)
               CALL MYWRT(IOLOG,BUFFER)
            ENDIF
         ENDIF
         IF(PVCAND.GE.0.9999D0*LDIAG(IROW)) THEN
            WRITE(BUFFER,1721) IROW,LDIAG(IROW),PVCAND
 1721       FORMAT(1X,'irow=',I6,' Dii=',1PD16.8,
     X       ' pivot candidate p=',1PD16.8)
            CALL MYWRT(IOLOG,BUFFER)
            IF(PVCAND.GE.1.0002D0*LDIAG(IROW)) THEN
               IF(PVCAND.GE.0.99D0*RO2) THEN 
                  WRITE(BUFFER,1722) PVCAND
 1722             FORMAT(1X,'dense:  rw shd be removed, pcnd=',1PD8.1)
                  CALL MYWRT(IOLOG,BUFFER)
                  IF(PVCAND.GE.0.99D0*RO3) IRMV=1
                  LDIAG(IROW)=LDIAG(IROW)+RO3-RO
                  MSPLIT(IROW)=2
               ELSE
                  LDIAG(IROW)=LDIAG(IROW)+RO2-RO
                  MSPLIT(IROW)=1
               ENDIF
               IF(PVCAND.GE.1.0D+16) THEN 
                  RTCODE=1
                  GO TO 6000
               ENDIF
            ENDIF
            WRITE(BUFFER,1723) IROW,MSPLIT(IROW),LDIAG(IROW)
 1723       FORMAT(1X,'May95,2d: rw=',I6,' msplit=',I2,' Dii=',1PD9.1)
            CALL MYWRT(IOLOG,BUFFER)
         ENDIF
 1800    LDIAG(IROW)=LDIAG(IROW)+RO
         IF(IRMV.EQ.1) THEN
            WRITE(BUFFER,1801) IROW,LDIAG(IROW)
 1801       FORMAT(1X,'dense:  rw=',I6,' Dii=',1PD12.4,' is removed.')
            CALL MYWRT(IOLOG,BUFFER)
            DO 1820 K=KBEG,KEND
               JCOL=LRWNBS(K)
               DPWORK(JCOL)=0.0D0
 1820       CONTINUE
         ENDIF
         DP=1.0/LDIAG(IROW)
         DO 1900 I=1,LENROW
            JCOL=IROW+I
            K=KBEG+I
            LCOEFF(K)=DP*DPWORK(JCOL)
            LDIAG(JCOL)=LDIAG(JCOL)-DP*DPWORK(JCOL)*DPWORK(JCOL)
            IF(LDIAG(JCOL).LE.SMALLP) LDIAG(JCOL)=SMALLP
 1900    CONTINUE
C
C
C
C     End of main loop.
 2000 CONTINUE
      IF(LDIAG(M).LE.1.0D+2*TAUADD) THEN
         RGROW=ELTMAX/LDIAG(M)
         WRITE(BUFFER,2001) LDIAG(M),RGROW
 2001    FORMAT(1X,'last pivot  Dii=',1PD12.4,' rgr=',1PD12.4)
         CALL MYWRT(IOLOG,BUFFER)
         IF(RGROW.GE.1.0D+8) THEN
            IF(MSPLIT(M).EQ.0) THEN
               LDIAG(M)=LDIAG(M)+RO2-RO
               MSPLIT(M)=1
            ENDIF
         ENDIF
      ENDIF
C
C
C
C
C
C     Find the largest and the smallest diagonal entries
C     of the Cholesky factor.
      IRWMAX=1
      IRWMIN=1
      DP=DABS(LDIAG(1))
      PIVMAX=DP
      PIVMIN=DP
      DO 5200 IROW=2,M
         DP=DABS(LDIAG(IROW))
         IF(DP.GT.PIVMAX) THEN
            PIVMAX=DP
            IRWMAX=IROW
         ENDIF
         IF(DP.LT.PIVMIN) THEN
            PIVMIN=DP
            IRWMIN=IROW
         ENDIF
 5200 CONTINUE
C
      WRITE(BUFFER,5201) IRWMAX,LDIAG(IRWMAX)
 5201 FORMAT(1X,'NUMFCT: Max. pivot in row ',I6,',  Dii=',1PD12.4)
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,5202) IRWMIN,LDIAG(IRWMIN)
 5202 FORMAT(1X,'NUMFCT: Min. pivot in row ',I6,',  Dii=',1PD12.4)
      CALL MYWRT(IOLOG,BUFFER)
C
C
C
C     Find the largest off-diagonal element of the Cholesky factor.
C     DP=0.0
C     DO 5300 K=1,LCLPTS(M+1)-1
C        IF(DABS(LCOEFF(K)).GT.DP) DP=DABS(LCOEFF(K))
C5300 CONTINUE
C     WRITE(BUFFER,5301) DP
C5301 FORMAT(1X,'NUMFCT: Max. elt in L=',1PD12.4)
C     CALL MYWRT(IOLOG,BUFFER)
C
C
C
C     Compute the square roots of diagonal elements if necessary.
      IF(MKSQRT.EQ.0) GO TO 5500
      DO 5400 IROW=1,M
         LDSQRT(IROW)=DSQRT(LDIAG(IROW))
 5400 CONTINUE
 5500 CONTINUE
C
C
C
C
C
C
 6000 CONTINUE
      RETURN
C
C
C
C     Here to write error message.
C9000 WRITE(BUFFER,9001) IROW,LDIAG(IROW)
C9001 FORMAT(1X,'NUMFCT ERROR: Diagonal entry of row ',I6,
C    X ' is too small ',D12.4)
C     CALL ERRWRT(IOLOG,BUFFER)
C     CALL MYSTOP(99)
C
C
C *** LAST CARD OF (NUMFCT) ***
      END
