C*************************************************************
C     *** PCDIR ... COMPUTE THE STEP DIRECTION             ***
C     ***           PREDICTOR-CORRECTOR PRIMAL-DUAL METHOD ***
C*************************************************************
C
      SUBROUTINE PCDIR(IDIR,BARR,OLDBAR,ALPHAP,ALPHAD,SMALLX,
     X MAXM,MAXN,MAXNZA,MAXNZL,M,N,ITREF,IALARM,
     X IWORK,RWORK,IMAP,RMAP,LIWORK,LRWORK,IROW,RELT,
     X LCOEFF,LCLPTS,LRWNBS,LDIAG,LDSQRT,
     X STAVAR,VUSED,VBNDED,THETA,XIB,XIC,XIU,DDD,GGG,HHH,FNEW,
     X X,Y,S,Z,W,DELTAX,DELTAS,DELTAY,DELTAZ,DELTAW,YPROX,
     X RMTMP1,RMTMP2,RMTMP3,RNTMP1,RNTMP2,RNTMP3,
     X RESX,RESY,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 IDIR,MAXM,MAXN,MAXNZA,MAXNZL,M,N
      INTEGER*4 LIWORK,LRWORK,ITREF,IALARM,IOLOG
      INTEGER*2 STAVAR(MAXN)
      LOGICAL VUSED(MAXN),VBNDED(MAXN)
      DOUBLE PRECISION THETA(MAXN),XIB(MAXM),XIC(MAXN),XIU(MAXN)
      DOUBLE PRECISION BARR,OLDBAR,ALPHAP,ALPHAD,SMALLX
      DOUBLE PRECISION DDD(MAXM),GGG(MAXN),HHH(MAXN),FNEW(MAXN)
      INTEGER*4 IROW(MAXN)
      DOUBLE PRECISION RELT(MAXN)
      DOUBLE PRECISION RMTMP1(MAXM),RMTMP2(MAXM),RMTMP3(MAXM)
      DOUBLE PRECISION RNTMP1(MAXN),RNTMP2(MAXN),RNTMP3(MAXN)
      DOUBLE PRECISION X(MAXN),S(MAXN),Y(MAXM),Z(MAXN),W(MAXN)
      DOUBLE PRECISION DELTAX(MAXN,2),DELTAS(MAXN,2)
      DOUBLE PRECISION DELTAY(MAXM,2),YPROX(MAXM)
      DOUBLE PRECISION DELTAZ(MAXN,2),DELTAW(MAXN,2)
      DOUBLE PRECISION RESX,RESY
C
C *** DATA STRUCTURES FOR CHOLESKY FACTOR
      DOUBLE PRECISION LCOEFF(MAXNZL)
      DOUBLE PRECISION LDIAG(MAXM),LDSQRT(MAXM)
      INTEGER*4 LCLPTS(MAXM+1)
      INTEGER*2 LRWNBS(MAXNZL)
C
C *** HIDDEN DATA STRUCTURES
      INTEGER*4 IWORK(LIWORK),IMAP(*),RMAP(*)
      DOUBLE PRECISION RWORK(LRWORK)
C
C
C *** LOCAL VARIABLES
      DOUBLE PRECISION DP,DX,DS,DZ,DW,XZSW,ALP,ALD,RO2,RO3
      INTEGER*4        I,J,K,KSMALL,KLARGE
      CHARACTER*100    BUFFER
C
C     Additional Cholesky fact. parameters (interface to HYBRID).
      COMMON /CHHYB/   RO,FLOPS,IREG,NZCHL,RTCD
      DOUBLE PRECISION RO,FLOPS
      INTEGER*4        IREG,NZCHL,RTCD
C
C     Markers for linking rows.
      COMMON /ICGRAD/ MSPLIT(100000)
      INTEGER*2       MSPLIT
C
C
C *** PARAMETERS DESCRIPTION
C
C     IDIR    Index of the Newton's step component desired:
C             1  primal-dual affine scaling (small centering added);
C             2  Mehrotra's corrector term (predictor-corrector);
C             3  Gondzio's  corrector term (high order corrector).
C             4  pure primal-dual direction.
C     BARR    Barrier parameter.
C     SMALLX  The threshold number for primal variables X and S.
C
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     MAXN    Maximum column dimension of the LP constraint matrix.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     MAXNZL  Maximum number of nonzeros of the Cholesky factor.
C     M       Number of rows of the LP constraint matrix
C             (and the dimension of  A*Atransp).
C     N       Number of columns of the LP constraint matrix.
C     ITREF   Number of steps of the iterative refinement process
C             to be done to improve the accuracy of solutions
C             with the Cholesky factorization of A*THETA*Atransp.
C     IALARM  Indicator if there are errors in the Cholesky factors:
C             0 if Cholesky factor is sufficiently stable;
C             1 if the iterative refinement process does not improve 
C               the accuracy;
C             2 if there are very large errors in Cholesky factor.
C
C     LCOEFF  Off-diagonal nonzero coefficients of Cholesky matrix.
C     LCLPTS  Pointers to columns of the Cholesky factor.
C     LRWNBS  Row numbers of nonzeros in columns of matrix  L.
C     LDIAG   Diagonal elements of Cholesky factor.
C     LDSQRT  Square roots of the diagonal elements of Cholesky factor.
C
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C             7  (or larger) PRESUMED OPTIMAL variable i.e.: x = x0;
C            -k  free variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status.
C     VUSED   An indicator if a variable is active in the optimization
C             process:
C             .TRUE.   active variable;
C             .FALSE.  FIXED variable.
C     VBNDED  An indicator if a variable has an UPPER bound:
C             .TRUE.   UPPER bounded variable;
C             .FALSE.  UNBOUNDED variable;
C     THETA   Diagonal weight matrix.
C     XIB     Violation of primal constraints, i.e.  b - A * x
C     XIC     Violation of dual   constraints, i.e.  c - At*y - z + w
C     XIU     Violation of variable bounds, i.e.     UPBND - x - s
C     DDD     Work array. It stores:
C             XIB                             (affine dir);
C             zero                            (corr., p-c algorithm).
C             zero                            (corr., pure p-d step).
C     GGG     Work array. It stores:
C             -X*Z*e                          (affine dir.);
C             BARR*e - deltaX*deltaZ*e        (corr., p-c algorithm).
C             BARR*e                          (corr., pure p-d step).
C     HHH     Work array. It stores:
C             -S*W*e                          (affine dir.);
C             BARR*e - deltaS*deltaW*e        (corr., p-c algorithm).
C             BARR*e                          (corr., pure p-d step).
C     FNEW    Work array. It stores:
C             XIC-X**(-1)*GGG+S**(-1)*HHH-S**(-1)*W*XIU (affine dir);
C             -X**(-1)*GGG+S**(-1)*HHH        (any corrector step).
C
C     X       Primal variables of the linear program.
C     S       Primal slack variables of the linear program.
C     Y       Dual variables of the linear program.
C     Z       Dual slack variables of the linear program.
C     W       Dual slack variables of the linear program.
C     DELTAX(*,L)  L-th component of deltaX.
C     DELTAS(*,L)  L-th component of deltaS.
C     DELTAY(*,L)  L-th component of deltaY.
C     DELTAZ(*,L)  L-th component of deltaZ.
C     DELTAW(*,L)  L-th component of deltaW.
C     YPROX   Dual proximal point.
C
C     RESX    Residuum of the solution (part refering to deltaX).
C     RESY    Residuum of the solution (part refering to deltaY).
C     IOLOG   Output unit number where log messages are to be written.
C
C
C *** COMMON ARREAS
C     IREG    Regularization:
C             0  add RO to all diagonal elements and increase small
C                pivots to TAUADD (used by HYBRID);
C             1  increase small pivots to TAUADD (used by HYBRID);
C            -1  increase very small pivots to TAUADD (used by HOPDM).
C     RO      Regularization parameter.
C
C
C *** HIDDEN DATA STRUCTURES DESCRIPTION
C     RWORK   Real work array containing almost all real
C             LP problem data.
C     IWORK   Integer work array containing almost all integer
C             LP problem data.
C     RMAP    Map of RWORK.
C     IMAP    Map of IWORK.
C
C     Map of IWORK array:
C     IMAP(1)   Points to CLPNTS array.
C     IMAP(2)   Points to RWNMBS array.
C     IMAP(3)   Points to RWHEAD array.
C     IMAP(4)   Points to RWLINK array.
C     IMAP(5)   Points to CLNMBS array.
C     IMAP(6)   Points to LENCOL array.
C     IMAP(7)   Points to the first empty cell of IWORK array.
C
C     Map of RWORK array:
C     RMAP(1)   Points to ACOEFF array.
C     RMAP(2)   Points to COBJ array.
C     RMAP(3)   Points to RHS array.
C     RMAP(4)   Points to the first empty cell of RWORK array.
C
C
C *** WORK ARRAYS.
C     IROW  and  RELT are the arrays for temporary handling
C             of rows/columns of the constraint matrix. They
C             are primarily intended to handle sparse vectors
C             (in packed form) but may also be used for storing
C             dense ones.
C     RMTMP1  Double precision work array of size MAXM.
C     RMTMP2  Double precision work array of size MAXM.
C     RMTMP3  Double precision work array of size MAXM.
C     RNTMP1  Double precision work array of size MAXN.
C     RNTMP2  Double precision work array of size MAXN.
C     RNTMP3  Double precision work array of size MAXN.
C
C
C *** SUBROUTINES CALLED:
C     IRSOLV
C
C
C *** PURPOSE:
C     This subroutine computes the components of the step direction.
C
C
C *** REFERENCES:
C     Altman A., Gondzio J. (1993). HOPDM - A higher order primal-
C        dual method for large scale linear programmming, European
C        Journal of Operational Research 66 (1993) pp 158-160.
C     Gondzio J. (1994). Multiple centrality corrections in a primal-
C        dual method for linear programming, Technical Report 
C        No 1994.20, Department of Management Studies, University 
C        of Geneva, 102, Bd. Carl-Vogt, 1211 Geneva, Switzerland, 
C        November 1994, revised in May 1995, Computational 
C        Optimization and Applications (to appear). 
C     Gondzio J. (1995). HOPDM (version 2.12) - A fast LP solver 
C        based on a primal-dual interior point method, European
C        Journal of Operational Research 85, pp. 221-225.
C     Lustig I., Marsten R., Shanno D.F. (1992). On implementing
C        Mehrotra's predictor-corrector interior point method for
C        linear programming, SIAM Journal on Optimization 2, 
C        No 3, pp. 435-449.
C     Mehrotra S. (1992): On the Implementation of a Primal-Dual 
C        Interior Point Method, SIAM Journal on Optimization 2, 
C        No 4, pp. 575-601.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio,
C                    Systems Research Institute,
C                    Polish Academy of Sciences,
C                    Newelska 6, 01-447 Warsaw, Poland.
C     Last modified: September 25, 1995
C
C
C
C *** BODY OF (PCDIR) ***
C
C
C
C
C
      RO2=1.0D+1
      RO3=1.0D+3
C     WRITE(BUFFER,1) RO,RO2,RO3
C   1 FORMAT(1X,'  RO=',1PD10.2,'  RO2=',1PD10.2,'  RO3=',1PD10.2)
C     CALL MYWRT(IOLOG,BUFFER)
C
C
C
C
C
      IF(IDIR.GE.2) GO TO 2000
C
C
C
C     Here if a primal-dual affine scaling direction
C     is to be computed. Note small centering added.
C     **********************************************
C     GGG     BARR-X*Z*e 
C     HHH     BARR-S*W*e
C     FNEW    XIC-X**(-1)*GGG+S**(-1)*HHH-S**(-1)*W*XIU
C
C
C     Compute  GGG, HHH and FNEW.
      DO 1200 J=1,N
         IF(VUSED(J)) THEN
            GGG(J)=BARR-X(J)*Z(J)
            FNEW(J)=XIC(J)-GGG(J)/X(J)
            IF(VBNDED(J)) THEN
               HHH(J)=BARR-S(J)*W(J)
               FNEW(J)=FNEW(J)+HHH(J)/S(J)-W(J)*XIU(J)/S(J)
            ENDIF
         ENDIF
 1200 CONTINUE
      DO 1400 I=1,M
         IF(MSPLIT(I).EQ.0) THEN
            DDD(I)=XIB(I)+RO*(Y(I)-YPROX(I))
         ELSE
            IF(MSPLIT(I).EQ.1) THEN
               DDD(I)=XIB(I)+RO2*(Y(I)-YPROX(I))
            ELSE
               DDD(I)=XIB(I)+RO3*(Y(I)-YPROX(I))
            ENDIF
         ENDIF
 1400 CONTINUE
C
C
C     Solve the augmented system for  deltaX and  deltaY.
C     Use normal equations in solve for  deltaY and an iterative
C     refinement on the augmented system to improve the accuracy.
C
      CALL IRSOLV(LCOEFF,LCLPTS,LRWNBS,LDIAG,
     X MAXM,MAXN,MAXNZA,MAXNZL,M,N,ITREF,IALARM,
     X IWORK,RWORK,IMAP,RMAP,LIWORK,LRWORK,IROW,RELT,
     X THETA,STAVAR,VUSED,
     X RMTMP1,RMTMP2,RMTMP3,RNTMP1,RNTMP2,RNTMP3,
     X FNEW,DDD,DELTAX(1,1),DELTAY(1,1),RESX,RESY,IOLOG)
C
C
C     Compute  deltaZ = X**(-1)*GGG - X**(-1)*Z*deltaX
C     Compute  deltaS = XIU - deltaX
C     Compute  deltaW = S**(-1)*HHH - S**(-1)*W*deltaS
      DO 1600 J=1,N
         IF(VUSED(J)) THEN
            DELTAZ(J,1)=(GGG(J)-Z(J)*DELTAX(J,1))/X(J)
            IF(VBNDED(J)) THEN
               DELTAS(J,1)=XIU(J)-DELTAX(J,1)
               DELTAW(J,1)=(HHH(J)-W(J)*DELTAS(J,1))/S(J)
            ENDIF
         ENDIF
 1600 CONTINUE
C
      GO TO 9000
C
C
C
C
C
 2000 IF(IDIR.GE.3) GO TO 3000
C
C
C
C     Here if Mehrotra's corrector step is to be computed.
C     ****************************************************
C     GGG     (BARR-OLDBAR)*e - deltaX*deltaZ*e
C     HHH     (BARR-OLDBAR)*e - deltaS*deltaW*e
C     FNEW    -X**(-1)*GGG+S**(-1)*HHH
C
C     (Higher order) predictor-corrector.
      DO 2200 J=1,N
         IF(VUSED(J)) THEN
            GGG(J)=-DELTAX(J,1)*DELTAZ(J,1)
            GGG(J)=GGG(J)+BARR-OLDBAR
            FNEW(J)=-GGG(J)/X(J)
            IF(VBNDED(J)) THEN
               HHH(J)=-DELTAS(J,1)*DELTAW(J,1)
               HHH(J)=HHH(J)+BARR-OLDBAR
               FNEW(J)=FNEW(J)+HHH(J)/S(J)
            ENDIF
         ENDIF
 2200 CONTINUE
      DO 2400 I=1,M
         DDD(I)=0.0D0
 2400 CONTINUE
C
C
C     Solve the augmented system for  deltaX and  deltaY.
C     Use normal equations in solve for  deltaY and an iterative
C     refinement on the augmented system to improve the accuracy.
C
      CALL IRSOLV(LCOEFF,LCLPTS,LRWNBS,LDIAG,
     X MAXM,MAXN,MAXNZA,MAXNZL,M,N,ITREF,IALARM,
     X IWORK,RWORK,IMAP,RMAP,LIWORK,LRWORK,IROW,RELT,
     X THETA,STAVAR,VUSED,
     X RMTMP1,RMTMP2,RMTMP3,RNTMP1,RNTMP2,RNTMP3,
     X FNEW,DDD,DELTAX(1,2),DELTAY(1,2),RESX,RESY,IOLOG)
C
C
C     Compute  deltaZ = X**(-1)*(GGG-Z*deltaX)
C     Compute  deltaS = - deltaX
C     Compute  deltaW = S**(-1)*(HHH-W*deltaS)
      DO 2600 J=1,N
         IF(VUSED(J)) THEN
            DELTAZ(J,2)=(GGG(J)-Z(J)*DELTAX(J,2))/X(J)
            IF(VBNDED(J)) THEN
               DELTAS(J,2)=-DELTAX(J,2)
               DELTAW(J,2)=(HHH(J)-W(J)*DELTAS(J,2))/S(J)
            ENDIF
         ENDIF
 2600 CONTINUE
C
      GO TO 9000
C
 3000 CONTINUE
C
C
C
C     Here if Gondzio's corrector step is to be computed.
C     ***************************************************
C     GGG     TARGET*e - Xprim*Zprim*e
C     HHH     TARGET*e - Sprim*Wprim*e
C     FNEW    -X**(-1)*GGG+S**(-1)*HHH
C
C
C     Compute the minimum complementarity gap that can be achieved
C     when moving in a current predictor direction:
C     XZSW=(x+ALPHAP*dx)t*(z+ALPHAD*dz)+(s+ALPHAP*ds)t*(w+ALPHAD*dw).
C
      XZSW=0.0D0
      K=0
      DO 3100 J=1,N
         IF(VUSED(J)) THEN
            DX=DELTAX(J,1)
            DZ=DELTAZ(J,1)
            XZSW=XZSW+(X(J)+ALPHAP*DX)*(Z(J)+ALPHAD*DZ)
            K=K+1
            IF(VBNDED(J)) THEN
               DS=DELTAS(J,1)
               DW=DELTAW(J,1)
               XZSW=XZSW+(S(J)+ALPHAP*DS)*(W(J)+ALPHAD*DW)
               K=K+1
            ENDIF
         ENDIF
 3100 CONTINUE
      XZSW=XZSW/DBLE(K)
C     WRITE(BUFFER,3101) BARR,XZSW
C3101 FORMAT(1X,' barrier=',1PD10.3,'   predicted cmpl.=',1PD10.3)
C     CALL MYWRT(IOLOG,BUFFER)
C     IF(XZSW.GE.2.0D0*BARR) XZSW=2.0D0*BARR
C
C
C     (Higher order) predictor-corrector.
      ALP=ALPHAP*1.08D0+0.08D0
      IF(ALP.GE.1.0D0) ALP=1.0D0
      ALD=ALPHAD*1.08D0+0.08D0
      IF(ALD.GE.1.0D0) ALD=1.0D0
      KSMALL=0
      KLARGE=0
      DO 3200 J=1,N
         IF(VUSED(J)) THEN
            DX=DELTAX(J,1)
            DZ=DELTAZ(J,1)
            DP=(X(J)+ALP*DX)*(Z(J)+ALD*DZ)
            GGG(J)=0.0D0
            IF(DP.LE.1.0D-1*BARR) THEN
               KSMALL=KSMALL+1
C              GGG(J)=2.0D0*BARR-DP
               GGG(J)=BARR-DP
            ENDIF
            IF(DP.GE.10.0D0*BARR) THEN
               KLARGE=KLARGE+1
               GGG(J)=-5.0D0*BARR
C              IF(DP.GE.50.0D0*BARR) GGG(J)=-10.0D0*BARR
            ENDIF
            FNEW(J)=-GGG(J)/X(J)
            IF(VBNDED(J)) THEN
               DS=DELTAS(J,1)
               DW=DELTAW(J,1)
               DP=(S(J)+ALP*DS)*(W(J)+ALD*DW)
               HHH(J)=0.0D0
               IF(DP.LE.1.0D-1*BARR) THEN 
                  KSMALL=KSMALL+1
C                 HHH(J)=2.0D0*BARR-DP
                  HHH(J)=BARR-DP
               ENDIF
               IF(DP.GE.10.0D0*BARR) THEN
                  KLARGE=KLARGE+1
                  HHH(J)=-5.0D0*BARR
C                 IF(DP.GE.50.0D0*BARR) HHH(J)=-10.0D0*BARR
               ENDIF
               FNEW(J)=FNEW(J)+HHH(J)/S(J)
            ENDIF
         ENDIF
 3200 CONTINUE
C     WRITE(BUFFER,3201) KSMALL,KLARGE
C3201 FORMAT(1X,'PCDIR: complement. pairs, KSMALL=',I6,' KLARGE=',I6)
C     CALL MYWRT(IOLOG,BUFFER)
      DO 3400 I=1,M
         DDD(I)=0.0D0
 3400 CONTINUE
C
C
C     Solve the augmented system for  deltaX and  deltaY.
C     Use normal equations in solve for  deltaY and an iterative
C     refinement on the augmented system to improve the accuracy.
C
      CALL IRSOLV(LCOEFF,LCLPTS,LRWNBS,LDIAG,
     X MAXM,MAXN,MAXNZA,MAXNZL,M,N,ITREF,IALARM,
     X IWORK,RWORK,IMAP,RMAP,LIWORK,LRWORK,IROW,RELT,
     X THETA,STAVAR,VUSED,
     X RMTMP1,RMTMP2,RMTMP3,RNTMP1,RNTMP2,RNTMP3,
     X FNEW,DDD,DELTAX(1,2),DELTAY(1,2),RESX,RESY,IOLOG)
C
C
C     Compute  deltaZ = X**(-1)*(GGG-Z*deltaX)
C     Compute  deltaS = - deltaX
C     Compute  deltaW = S**(-1)*(HHH-W*deltaS)
      DO 3600 J=1,N
         IF(VUSED(J)) THEN
            DELTAZ(J,2)=(GGG(J)-Z(J)*DELTAX(J,2))/X(J)
            IF(VBNDED(J)) THEN
               DELTAS(J,2)=-DELTAX(J,2)
               DELTAW(J,2)=(HHH(J)-W(J)*DELTAS(J,2))/S(J)
            ENDIF
         ENDIF
 3600 CONTINUE
C
      GO TO 9000
C
C
C
C
C
C
C
C
C
C
 9000 CONTINUE
      RETURN
C
C
C *** LAST CARD OF (PCDIR) ***
      END
