C*********************************************************************
C     * PCINIT ... INITIALIZE FOR THE PREDICTOR-CORRECTOR P-D METHOD *
C*********************************************************************
C
      SUBROUTINE PCINIT(IXY,LORD,MAXM,MAXN,MAXNZA,MAXNZL,M,N,
     X IWORK,RWORK,IMAP,RMAP,LIWORK,LRWORK,IROW,RELT,
     X LCOEFF,LCLPTS,LRWNBS,LDIAG,LDSQRT,
     X VUSED,VBNDED,HEADER,LINKFD,LINKBK,
     X COLNRM,C,STAVAR,P,Q,B,UPBND,THETA,X,Y,S,Z,W,
     X DELTAX,DELTAS,DELTAY,DELTAZ,DELTAW,
     X INTMP1,RMTMP1,RMTMP2,RNTMP1,RNTMP2,IOLOG)
C
C
C
C *** PARAMETERS
      INTEGER*4 IXY,LORD,MAXM,MAXN,MAXNZA,MAXNZL,M,N
      INTEGER*4 LIWORK,LRWORK,IOLOG
      INTEGER*4 INTMP1(MAXN),IROW(MAXN)
      DOUBLE PRECISION RMTMP1(MAXM),RMTMP2(MAXM)
      DOUBLE PRECISION RNTMP1(MAXN),RNTMP2(MAXN),RELT(MAXN)
      INTEGER*2 HEADER(MAXM+1),LINKFD(MAXM+1),LINKBK(MAXM+1)
      LOGICAL VUSED(MAXN),VBNDED(MAXN)
      INTEGER*2 STAVAR(MAXN)
      DOUBLE PRECISION COLNRM(MAXN),P(MAXM),Q(MAXM)
      DOUBLE PRECISION C(MAXN),B(MAXM),X(MAXN),S(MAXN),Y(MAXM)
      DOUBLE PRECISION Z(MAXN),W(MAXN),UPBND(MAXN),THETA(MAXN)
      DOUBLE PRECISION DELTAX(MAXN,LORD),DELTAS(MAXN,LORD)
      DOUBLE PRECISION DELTAY(MAXM,LORD)
      DOUBLE PRECISION DELTAZ(MAXN,LORD),DELTAW(MAXN,LORD)
C
C *** DATA STRUCTURES FOR CHOLESKY FACTOR
      DOUBLE PRECISION LCOEFF(MAXNZL)
      DOUBLE PRECISION LDIAG(MAXM),LDSQRT(MAXM)
      INTEGER*4 LCLPTS(MAXM+1)
      INTEGER*2 LRWNBS(MAXNZL)
C
C *** HIDDEN DATA STRUCTURES
      INTEGER*4 IWORK(LIWORK),IMAP(*),RMAP(*)
      DOUBLE PRECISION RWORK(LRWORK)
C
C
C
C *** LOCAL VARIABLES
      INTEGER*4 I,J,K,LX,LS,MKSQRT,RTCODE
      DOUBLE PRECISION XMIN,SMIN,ZMIN,WMIN
      DOUBLE PRECISION DP,DD,XJP,SJP,ZJD,WJD,SX,SS,SZ,SW,XZSW
      CHARACTER*100 BUFFER
C
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C
C     IXY     Indicates the method of setting initial values:
C             1   Altman and Gondzio (92) combined with Mehrotra (91).
C             2   y(i)=0  x(j)=s(j)=N/||Aj|| and  z(j)=w(j)=||Aj||.
C             4   FAP: y(i)=0 and x(j):=min{x0,UPBND(j)/2}, where x0=1.
C     LORD    The highest degree of computed derivatives of  x,s,y,z,w
C             (order of Mehrotra's method).
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     MAXN    Maximum column dimension of the LP constraint matrix.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     MAXNZL  Maximum number of nonzeros of the Cholesky factor.
C     M       Number of rows of the LP constraint matrix
C             (and the dimension of  A*Atransp).
C     N       Number of columns of the LP constraint matrix.
C
C     LCOEFF  Off-diagonal nonzero coefficients of Cholesky matrix.
C     LCLPTS  Pointers to columns of the Cholesky factor.
C     LRWNBS  Row numbers of nonzeros in columns of matrix  L.
C     LDIAG   Diagonal elements of Cholesky factor.
C     LDSQRT  Square roots of the diagonal elements of Cholesky factor.
C
C     HEADER  Header of the doubly linked lists.
C     LINKFD  Forward linked lists.
C     LINKBK  Backward linked lists.
C
C     INTMP1  Integer work array of size MAXN.
C     IROW  and  RELT are the arrays for temporary handling
C             of rows/columns of the constraint matrix. They
C             are primarily intended to handle sparse vectors
C             (in packed form) but may also be used for storing
C             dense ones.
C     RMTMP1  Double precision work array of size MAXM.
C     RMTMP2  Double precision work array of size MAXM.
C     RNTMP1  Double precision work array of size MAXN.
C     RNTMP2  Double precision work array of size MAXN.
C
C     COLNRM  Infinity morms of columns of  A.
C     C       Objective function coefficients.
C     P       LOWER bounds on shadow prices (dual variables).
C     Q       UPPER bounds on shadow prices (dual variables).
C     VUSED   An indicator if a variable is active in the optimization
C             process:
C             .TRUE.   active variable;
C             .FALSE.  FIXED variable.
C     VBNDED  An indicator if a variable has an UPPER bound:
C             .TRUE.   UPPER bounded variable;
C             .FALSE.  UNBOUNDED variable;
C     STAVAR  Array of variable status.
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C             7  (or larger) PRESUMED OPTIMAL variable i.e.: x = x0;
C            -k  free variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status.
C     B       Right-hand-side of the linear program.
C     UPBND   Upper bounds for primal variables X.
C     THETA   Diagonal weight matrix.
C
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C
C     X       Primal variables of the linear program.
C     S       Primal slack variables of the linear program.
C     Y       Dual variables of the linear program.
C     Z       Dual slack variables of the linear program.
C     W       Dual slack variables of the linear program.
C     DELTAX(*,L)  L-th derivative of x(alpha) (for alpha=1) with 
C                  respect to alpha divided by l!. 
C     DELTAS(*,L)  L-th derivative of s(alpha) (for alpha=1) with 
C                  respect to alpha divided by l!. 
C     DELTAY(*,L)  L-th derivative of y(alpha) (for alpha=1) with 
C                  respect to alpha divided by l!. 
C     DELTAZ(*,L)  L-th derivative of z(alpha) (for alpha=1) with 
C                  respect to alpha divided by l!. 
C     DELTAW(*,L)  L-th derivative of w(alpha) (for alpha=1) with 
C                  respect to alpha divided by l!. 
C
C
C
C
C *** HIDDEN DATA STRUCTURES DESCRIPTION
C     RWORK   Real work array containing almost all real
C             LP problem data.
C     IWORK   Integer work array containing almost all integer
C             LP problem data.
C     RMAP    Map of RWORK.
C     IMAP    Map of IWORK.
C
C     Map of IWORK array:
C     IMAP(1)   Points to CLPNTS array.
C     IMAP(2)   Points to RWNMBS array.
C     IMAP(3)   Points to RWHEAD array.
C     IMAP(4)   Points to RWLINK array.
C     IMAP(5)   Points to CLNMBS array.
C     IMAP(6)   Points to LENCOL array.
C     IMAP(7)   Points to the first empty cell of IWORK array.
C
C     Map of RWORK array:
C     RMAP(1)   Points to ACOEFF array.
C     RMAP(2)   Points to COBJ array.
C     RMAP(3)   Points to RHS array.
C     RMAP(4)   Points to the first empty cell of RWORK array.
C
C
C
C *** LOCAL VARIABLES DESCRIPTION
C     DP        Stepsize in primal space.
C     DD        Stepsize in dual space.
C     XZSW      Conmplementarity gap (xt*z+st*w).
C     XMIN      Minimal value of x(j),j=1,...,n.
C     SMIN      Minimal value of s(j),j=1,...,n.
C     ZMIN      Minimal value of z(j),j=1,...,n.
C     WMIN      Minimal value of w(j),j=1,...,n.
C
C
C
C *** SUBROUTINES CALLED:
C     FACTOR,SOLAAT,GETCOL,SAX,SATY,SAXPY,DABS,DMAX1
C
C
C *** PURPOSE:
C     This routine initializes primal variable X, dual variable Y
C     and all the slack variables S, Z, W for the higher order
C     primal-dual logarithmic barrier interior point method of
C     Mehrotra (1991).
C
C     We start from the solution of the following optimization
C     problems:
C     min{|(x,s)|:A*x=b, x+s=upbnd} and
C     min{|(z,w)|:At*y+z-w=c}.
C     Solution of the first problem:
C     v = (A*At)**(-1)*(A*upbnd-2*b),
C     x = (upbnd-At*v)/2,
C     s = upbnd - x.
C     Solution of the second problem:
C     y = (A*At)**(-1)*(A*c),
C     z = (At*y-c)/2,
C     w=-z.
C
C     Next we define stepsizes in primal and dual spaces:
C     dp:=max{-1.5*min[x(j)],-1.5*min[s(j)],0} and
C     dd:=max{-1.5*min[z(j)],-1.5*min[w(j)],0}.
C     and modify dp and dd:
C     dp:=dp+0.5*xzsw/(sz+sw),
C     dd:=dd+0.5*xzsw/(sx+ss), where
C     xzsw=(x+dp*e)t*(z+dd*e)+(s+dp*e)t*(w+dd*e),
C     sx=sum_{j=1}^{n}{x(j)},
C     ss=sum_{j=1}^{n}{s(j)},
C     sz=sum_{j=1}^{n}{z(j)},
C     sw=sum_{j=1}^{n}{w(j)},  and
C     x:=x+dp*e, s:=s+dp*e, z:= z+dd*e, w:=w+dd*e (y is unaltered).
C
C
C
C
C
C *** REFERENCES:
C     Altman A., Gondzio J. (1992). An efficient implementation
C        of a higher order primal-dual interior point method
C        for large sparse linear programs, Archives of Control
C        Sciences (to appear).
C     Altman A., Gondzio J. (1993). HOPDM - A higher order primal-
C        dual method for large scale linear programmming, European
C        Journal of Operational Research 66 (1993) pp 158-160.
C     Lustig I., Marsten R., Shanno D.F. (1992). On implementing
C        Mehrotra's predictor-corrector interior point method for
C        linear programming, SIAM Journal on Optimization 2, 
C        No 3, pp. 435-449.
C     Mehrotra S. (1992): On the Implementation of a Primal-Dual 
C        Interior Point Method, SIAM Journal on Optimization 2, 
C        No 4, pp. 575-601.
C     Mehrotra S. (1991): Higher Order Methods and their Performance,
C        Technical Report 90-16R1, Department of Industrial Engineering
C        and Management Sciences, Northwestern University, Evanston,
C        Illinois 60208-3119, U.S.A.
C
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio,
C                    Systems Research Institute,
C                    Polish Academy of Sciences,
C                    Newelska 6, 01-447 Warsaw, Poland.
C     Date written:  March 12, 1992
C     Last modified: September 25, 1995
C
C
C
C
C
C *** BODY OF (PCINIT) ***
C
C
C
C     WRITE(BUFFER,101) M,N,MAXM,MAXN
C 101 FORMAT(1X,'PCINIT: M=',I6,' N=',I6,' MAXM=',I6,' MAXN=',I6)
C     CALL MYWRT(IOLOG,BUFFER)
C     DO 200 J=1,N
C        WRITE(BUFFER,201) J,STAVAR(J),C(J)
C 201    FORMAT(1X,'PCINIT: cl=',I6,' st=',I6,' Cj=',D10.3)
C        CALL MYWRT(IOLOG,BUFFER)
C 200 CONTINUE
C
C
C
C
C     IXY=2
      IF(IXY.EQ.1) THEN
C
C     Here for a combination of Altman & Gondzio '92 and Mehrotra '91
C     starting point.
C
C     Let  Au and  An denote columns of A that refer to
C     UPPER bounded and UNBOUNDED variables, respectively.
C     Note that modified UPPER bounds (RNTMP2 array) will be used.
C
C     Before initializing primal and dual variables, set THETA
C     array appropriate for an auxiliary QP problem.
C     Define RNTMP2 array i.e. modified variables' UPPER bounds.
C     Compute Au*THETA*upbnd and store in RMTMP2.
C     Compute THETA*c and store in RNTMP1.
C
      DO 1100 I=1,M
         RMTMP2(I)=B(I)
 1100 CONTINUE
      DO 1200 J=1,N
         THETA(J)=0.0D0
         IF(VUSED(J)) THEN
            THETA(J)=1.0D0
            RNTMP1(J)=C(J)
            IF(VBNDED(J)) THEN
               THETA(J)=0.5
               RNTMP1(J)=0.5*C(J)
               RNTMP2(J)=UPBND(J)
               IF(UPBND(J).LE.1.0D-2) RNTMP2(J)=1.0D-2
               IF(UPBND(J).GE.1.0D+3) RNTMP2(J)=1.0D+3
C              WRITE(BUFFER,1201) J,C(J),RNTMP1(J),UPBND(J),RNTMP2(J)
C1201          FORMAT(1X,'J=',I6,' Cj=',D10.3,' RNTMP1=',D10.3,
C    X          ' UP=',D10.3,' UPnew=',D10.3)
C              CALL MYWRT(IOLOG,BUFFER)
               DP=-0.5*RNTMP2(J)
               CALL GETCOL(J,RWORK,IWORK,RMAP,IMAP,
     X          IROW,RELT,K,MAXN,IOLOG)
               CALL SAXPY(IROW,RELT,K,RMTMP2,DP)
            ENDIF
         ENDIF
 1200 CONTINUE
C
C
C     Factorize  A*THETA*Atransp matrix.
C
      MKSQRT=0
      CALL FACTOR(MAXM,MAXN,MAXNZA,MAXNZL,M,
     X IWORK,RWORK,IMAP,RMAP,LIWORK,LRWORK,
     X IROW,RELT,HEADER,LINKFD,LINKBK,THETA,STAVAR,
     X LCOEFF,LDIAG,LDSQRT,LCLPTS,LRWNBS,MKSQRT,RTCODE,IOLOG)
C
      IF(RTCODE.EQ.1) THEN
         WRITE(BUFFER,1211) 
 1211    FORMAT(1X,'PCINIT: Fatal accuracy problem',
     X    ' (large errors in LLt decomposition).')
         CALL MYWRT(0,BUFFER)
         CALL MYWRT(IOLOG,BUFFER)
         CALL MYWRT(99,BUFFER)
         CALL MYSTOP(13)
      ENDIF
C
C
C     Recall that  RMTMP2:=b-Au*THETA*upbnd
C     Compute y
         CALL SOLAAT(LCOEFF,LCLPTS,LRWNBS,LDIAG,
     X    MAXNZL,MAXM,M,Y,RMTMP2,IOLOG)
C
C     Compute  z:=At*y
         CALL SATY(RWORK,IWORK,RMAP,IMAP,Y,M,Z,N,
     X    IROW,RELT,MAXN,IOLOG)
C
C     Compute x:= THETA*(upbnd+At*y)
         DO 1300 J=1,N
            IF(VUSED(J)) THEN
               X(J)=THETA(J)*Z(J)
               IF(VBNDED(J)) THEN
                  X(J)=THETA(J)*(RNTMP2(J)+Z(J))
                  S(J)=UPBND(J)-X(J)
               ENDIF
            ENDIF
 1300    CONTINUE
C
C
C     Set up initial values of the dual variables Y, Z and W.
C
C     Compute RMTMP2=A*THETA*c
         CALL SAX(RWORK,IWORK,RMAP,IMAP,STAVAR,RNTMP1,N,RMTMP2,M,
     X    IROW,RELT,MAXN,IOLOG)
C
C     Compute  y
         CALL SOLAAT(LCOEFF,LCLPTS,LRWNBS,LDIAG,
     X    MAXNZL,MAXM,M,Y,RMTMP2,IOLOG)
C
C     Correct infeasible dual variables.
         LX=0
         LS=0
         DO 1400 I=1,M
            IF(Y(I).LT.P(I)) THEN
C              WRITE(BUFFER,1401) I,P(I),Y(I),Q(I)
C1401          FORMAT(1X,'rw=',I6,' Pi=',D12.4,' Yi=',D12.4,
C    X          ' Qi=',D12.4)
C              CALL MYWRT(IOLOG,BUFFER)
               LX=LX+1
C              Y(I)=(P(I)+Y(I))/2.0
               Y(I)=P(I)
            ENDIF
            IF(Y(I).GT.Q(I)) THEN
C              WRITE(BUFFER,1402) I,P(I),Y(I),Q(I)
C1402          FORMAT(1X,'rw=',I6,' Pi=',D12.4,' Yi=',D12.4,
C    X          ' Qi=',D12.4)
C              CALL MYWRT(IOLOG,BUFFER)
               LS=LS+1
C              Y(I)=(Q(I)+Y(I))/2.0
               Y(I)=Q(I)
            ENDIF
 1400    CONTINUE
C
C *** DEBUGGING
C        WRITE(BUFFER,1403) LX,LS
C1403    FORMAT(1X,'PCINIT: Dual var. corrected:  Pi=',I6,' Qi=',I6)
C        CALL MYWRT(IOLOG,BUFFER)
C
C     Compute  z:=At*y
         CALL SATY(RWORK,IWORK,RMAP,IMAP,Y,M,Z,N,
     X    IROW,RELT,MAXN,IOLOG)
C
C     Compute  z:=THETA*(c-At*y)
C     Compute  w:= -z
         DO 1500 J=1,N
            IF(VUSED(J)) THEN
               Z(J)=C(J)-Z(J)
               IF(VBNDED(J)) THEN
                  Z(J)=THETA(J)*Z(J)
                  W(J)=-Z(J)
               ENDIF
C              IF(J.GT.50) GO TO 1500
C              WRITE(BUFFER,1501) J,C(J),Z(J),W(J)
C1501          FORMAT(1X,'1501F:  J=',I6,' C=',D10.3,
C    X          ' Z=',D10.3,' W=',D10.3)
C              CALL MYWRT(IOLOG,BUFFER)
            ENDIF
 1500    CONTINUE
C
C     Compute  XMIN,SMIN,ZMIN and WMIN.
         XMIN=1.0D10
         ZMIN=1.0D10
         SMIN=1.0D10
         WMIN=1.0D10
         DO 1600 J=1,N
            IF(VUSED(J)) THEN
               IF(X(J).LT.XMIN) XMIN=X(J)
               IF(Z(J).LT.ZMIN) ZMIN=Z(J)
               IF(VBNDED(J)) THEN
                  IF(S(J).LT.SMIN) SMIN=S(J)
                  IF(W(J).LT.WMIN) WMIN=W(J)
               ENDIF
            ENDIF
 1600    CONTINUE	
         DP=DMAX1(-1.5*XMIN,-1.5*SMIN,0.001D0)
         DD=DMAX1(-1.5*ZMIN,-1.5*WMIN,0.001D0)
C        WRITE(BUFFER,1601) DP,DD
C1601    FORMAT(1X,'after 1600 loop:  DP=',D10.3,'  DD=',D10.3)
C        CALL MYWRT(IOLOG,BUFFER)
C
C     Compute  xzsw=(x+dp*e)t*(z+dd*e)+(s+dp*e)t*(w+dd*e)
C     sx=sum_{j=1}^{n}{x(J)},
C     ss=sum_{j=1}^{n}{s(J)},
C     sz=sum_{j=1}^{n}{z(J)},
C     sw=sum_{j=1}^{n}{w(J)}.
         SX=0.
         SS=0.
         SZ=0.
         SW=0.
         XZSW=0.
         DO 1700 J=1,N
            IF(VUSED(J)) THEN
               XJP=X(J)+DP
               ZJD=Z(J)+DD
               SX=SX+XJP
               SZ=SZ+ZJD
               XZSW=XZSW+XJP*ZJD
               IF(VBNDED(J)) THEN
                  SJP=S(J)+DP
                  WJD=W(J)+DD
                  SS=SS+SJP
                  SW=SW+WJD
                  XZSW=XZSW+SJP*WJD
               ENDIF
C              IF(J.GT.50) GO TO 1700
C              WRITE(BUFFER,1701) J,STAVAR(J),X(J),S(J),Z(J),W(J)
C1701          FORMAT(1X,'1701F, J=',I6,' st=',I6,' X=',D10.3,' S=',
C    X          D10.3,' Z=',D10.3' W=',D10.3)
C              CALL MYWRT(IOLOG,BUFFER)
            ENDIF
 1700    CONTINUE
C        WRITE(BUFFER,1702) SX,SS,SZ,SW
C1702    FORMAT(1X,'FINAL:  SX=',D10.3,'  SS=',D10.3,
C    X    '  SZ=',D10.3,'  SW=',D10.3)
C        CALL MYWRT(IOLOG,BUFFER)
         DP=DP+0.5*XZSW/(SZ+SW)
         DD=DD+0.5*XZSW/(SX+SS)
C        WRITE(BUFFER,1703) DP,DD
C1703    FORMAT(1X,'after 1700 loop:  DP=',D10.3,'  DD=',D10.3)
C        CALL MYWRT(IOLOG,BUFFER)
         DO 1800 J=1,N
            IF(VUSED(J)) THEN
               X(J)=X(J)+DP
               Z(J)=Z(J)+DD
               IF(VBNDED(J)) THEN
                  S(J)=S(J)+DP
                  W(J)=W(J)+DD
               ENDIF
C              WRITE(BUFFER,1801) J,STAVAR(J),X(J),S(J),Z(J),W(J)
C1801          FORMAT(1X,'1801F, J=',I6,' st=',I6,' X=',D10.3,' S=',
C    X          D10.3,' Z=',D10.3' W=',D10.3)
C              CALL MYWRT(IOLOG,BUFFER)
            ENDIF
 1800    CONTINUE
         GO TO 8000
C
      ENDIF
C
C
C
C
      IF(IXY.EQ.2) THEN
C
C     Here for a new primitive starting point.
C
C     Initialize primal variables:  Xj, Sj = N/||Aj||.
C     Initialize dual variables:    Zj, Wj = ||Aj||.
C     Bound them all away from zero.
C
         DP=DBLE(N)
         XMIN=1.0D0
         ZMIN=1.0D0
         SMIN=1.0D0
         WMIN=1.0D0
         DO 2100 J=1,N
            IF(VUSED(J)) THEN
               X(J)=DP/COLNRM(J)
               IF(X(J).LE.XMIN) X(J)=XMIN
               Z(J)=COLNRM(J)
               IF(Z(J).LE.ZMIN) Z(J)=ZMIN
               IF(VBNDED(J)) THEN
                  S(J)=UPBND(J)-X(J)
                  IF(S(J).LE.X(J)) S(J)=X(J)
                  W(J)=COLNRM(J)
                  IF(W(J).LE.WMIN) W(J)=WMIN
               ENDIF
            ENDIF
 2100    CONTINUE
C
C
C     Initialize dual variables:  Yi = 0.0 and correct
C     if infeasible.
         LX=0
         LS=0
         DO 2200 I=1,M
            Y(I)=0.0D0
            IF(Y(I).LT.P(I)) THEN
C              WRITE(BUFFER,2201) I,P(I),Y(I),Q(I)
C2201          FORMAT(1X,'rw=',I6,' Pi=',D12.4,' Yi=',D12.4,
C    X          ' Qi=',D12.4)
C              CALL MYWRT(IOLOG,BUFFER)
               LX=LX+1
C              Y(I)=(P(I)+Y(I))/2.0
               Y(I)=P(I)
            ENDIF
            IF(Y(I).GT.Q(I)) THEN
C              WRITE(BUFFER,2202) I,P(I),Y(I),Q(I)
C2202          FORMAT(1X,'rw=',I6,' Pi=',D12.4,' Yi=',D12.4,
C    X          ' Qi=',D12.4)
C              CALL MYWRT(IOLOG,BUFFER)
               LS=LS+1
C              Y(I)=(Q(I)+Y(I))/2.0
               Y(I)=Q(I)
            ENDIF
 2200    CONTINUE
C
C *** DEBUGGING
         WRITE(BUFFER,2203) LX,LS
 2203    FORMAT(1X,'PCINIT: Dual var. corrected:  Pi=',I6,' Qi=',I6)
         CALL MYWRT(IOLOG,BUFFER)
         GO TO 8000
C
      ENDIF
C
C
C
C
      IF(IXY.EQ.4) THEN
C
C     Here for FAP starting point.
C
C     Initialize primal variables:  Xj = 0.5,  Sj = Uj-Xj
C     and bound them away from zero.
C
         DP=0.5D0
         DO 4100 J=1,N
            IF(VUSED(J)) THEN
               X(J)=DP
               IF(STAVAR(J).EQ.0) X(J)=2.0D1
               IF(VBNDED(J)) THEN
                  IF(UPBND(J).GE.1.0D+2) X(J)=2.0D1
                  S(J)=UPBND(J)-X(J)
                  IF(S(J).LE.DP) S(J)=DP
               ENDIF
            ENDIF
 4100    CONTINUE
C
C     Initialize dual variables:  Yi = 0.0
         DO 4200 I=1,M
            Y(I)=0.0D0
 4200    CONTINUE
C
C     Initialize dual slack variables:  Zj - Wj = Cj 
C     and bound them both away from zero.
         DP=0.2D0
         DO 4300 J=1,N
            IF(VUSED(J)) THEN
               IF(C(J).GE.DP) THEN
                  Z(J)=C(J)
               ELSE
                  Z(J)=DP
               ENDIF
               IF(VBNDED(J)) THEN
                  IF(C(J).GE.0.0D0) THEN
                     W(J)=DP
                     Z(J)=C(J)+W(J)
                  ELSE
                     Z(J)=DP
                     W(J)=Z(J)-C(J)
                  ENDIF
               ENDIF
            ENDIF
 4300    CONTINUE
         GO TO 8000
C
      ENDIF
C
C
C
C
C     Zero all unused components of DELTAs.
 8000 DO 8500 I=1,LORD
         DO 8400 J=1,N
            DELTAX(J,I)=0.0
            DELTAZ(J,I)=0.0
            DELTAS(J,I)=0.0
            DELTAW(J,I)=0.0
C           IF(I.NE.1) GO TO 8400
C           WRITE(BUFFER,8001) J,STAVAR(J),X(J),S(J),Z(J),W(J)
C8001       FORMAT(1X,'J=',I6,' st=',I6,' X=',D10.3,' S=',D10.3,
C    X       ' Z=',D10.3' W=',D10.3)
C           CALL MYWRT(IOLOG,BUFFER)
 8400    CONTINUE
 8500 CONTINUE
C
C
C
      RETURN
C
C
C *** LAST CARD OF (PCINIT) ***
      END
