C******************************************************
C     ***  POSTSL ... POST_OPTIMIZATION PROCESSING  ***
C******************************************************
C
      SUBROUTINE POSTSL(IOLOG,MSGLEV,
     X MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X LNHIST,MXHIST,INHIST,DPHIST,
     X RWORK,IWORK,RMAP,IMAP,IROW,RELT,
     X IMTMP1,INTMP1,INTMP2,RNTMP1,
     X B,RANGES,C,LOBND,UPBND,
     X ACOEFF,CLPNTS,RWNMBS,
     X RWHEAD,RWLINK,CLNMBS,LENCOL,
     X PRLVAR,STAVAR,RWSTAT,STAROW,RWNAME,CLNAME)
C
C
C
C
C *** HIDDEN DATA STRUCTURES
      INTEGER*4 IWORK(*),IMAP(*),RMAP(*)
      DOUBLE PRECISION RWORK(*)
C
C *** HIDDEN DATA STRUCTURES DESCRIPTION
C     RWORK   Real array that contains real  LP problem data.
C     IWORK   Integer array that contains integer  LP problem data.
C     RMAP    Map of RWORK array.
C     IMAP    Map of IWORK array.
C
C
C
C
C *** PARAMETERS
      INTEGER*4 IOLOG,MSGLEV,MAXM,MAXN,MAXNZA
      INTEGER*4 M,N,NSTRCT,LNHIST,MXHIST
      INTEGER*2 INHIST(MXHIST)
      DOUBLE PRECISION DPHIST(MXHIST)
      INTEGER*4 IROW(MAXN),IMTMP1(MAXM),INTMP1(MAXN)
      INTEGER*2 INTMP2(MAXN)
      DOUBLE PRECISION RELT(MAXN),RNTMP1(MAXN)
      DOUBLE PRECISION ACOEFF(MAXNZA),B(MAXM),RANGES(MAXM)
      DOUBLE PRECISION C(MAXN),LOBND(MAXN),UPBND(MAXN)
      INTEGER*4 CLPNTS(MAXN+1),RWHEAD(MAXM+1),RWLINK(MAXNZA)
      INTEGER*2 RWNMBS(MAXNZA),CLNMBS(MAXNZA),LENCOL(MAXN)
      DOUBLE PRECISION PRLVAR(MAXN)
      CHARACTER*8 RWNAME(MAXM),CLNAME(MAXN)
      INTEGER*2 STAVAR(MAXN),STAROW(MAXM),RWSTAT(MAXM)
C
C
C
C *** LOCAL VARIABLES
      INTEGER*4 I,IENTRY,IPOS,J,JCOL,KBEG
      DOUBLE PRECISION DP,VALUE
      CHARACTER*100 BUFFER
C
C
C
C
C *** PARAMETERS DESCRIPTION
C
C *** ON INPUT:
C     IOLOG   Output unit number where log messages are to be written.
C     MSGLEV  The level of PRE_SOLVE information desired:
C             0  only final problem dimensions are printed;
C             1  numbers of eliminated rows and columns are printed;
C             2  names of eliminated rows and columns are printed;
C             3  detailed DEBUGGING information is printed.
C     MAXM    Maximum number of constraints.
C     MAXN    Maximum number of variables.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     M       Number of constraints.
C     N       Number of variables (total, i.e. including slacks,
C             surplus and artificials).
C     NSTRCT  Number of structural variables (excluding slacks,
C             surplus and artificials).
C     LNHIST  Length of the PRE_SOLVE history list.
C     MXHIST  Maximum number of entries in the PRE_SOLVE history list.
C     INHIST  Integer PRE_SOLVE history information.
C     DPHIST  Double precision PRE_SOLVE history information.
C     ACOEFF  Array of nonzero elements for each column.
C     B       Right hand side of the linear program.
C     RANGES  Array of constraint ranges.
C     C       Objective function coefficients.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWHEAD  Headers to the row linked lists of matrix A.
C     RWLINK  Row linked lists of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A.
C     CLNMBS  Column numbers of nonzeros in rows of matrix A.
C     LENCOL  Lengths of (sparse) columns of matrix A.
C     LOBND   Array of lower bounds.
C     UPBND   Array of upper bounds. Note that the upper bound
C             is changed when the variable has a lower bound. 
C     PRLVAR  Primal variables of the linear program.
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C            -k  FREE variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicates the position of the original variable.
C     RWSTAT  Array of row types:
C             1  row type is = ;
C             2  row type is >= ;
C             3  row type is <= ;
C             4  objective row;
C             5  other free row.
C     STAROW  Array of row status:
C             0  row has been removed (it indicates a free row);
C             1  row has not been removed.
C     RWNAME  Array of row names (increasing order sort).
C     CLNAME  Array of column names (unordered).
C
C *** ON OUTPUT:
C
C
C
C
C *** WORK ARRAYS:
C     IROW and  RELT are the arrays for temporary handling of rows
C             and columns of the constraint matrix. They are primarily
C             intended to handle sparse vectors (in packed form)
C             but may also be used for storing dense ones.
C     IMTMP1  Integer work array of size MAXM.
C     INTMP1  Integer work array of size MAXN
C     INTMP2  Half-length integer work array of size MAXN.
C     RNTMP1  Double precision work array of size MAXN.
C
C
C
C
C *** PURPOSE
C     This routine does POST_PROCESSING on the optimal solution. 
C     It performs an 'undo' operation on a stack-type PRE_SOLVE
C     history list. 
C     Recall that PRE_SOLVE history contains two types of entries.
C
C     Positive entries of INHIST array indicate eliminated FREE
C     (or implied FREE) variables. Appropriate DPHIST entries
C     store pivot elements.
C
C     Negative entries of INHIST array indicate variables for
C     which LOWER bound has been pushed. DPHIST entries hadle
C     bound corrections in such a case.
C
C
C
C *** SUBROUTINES CALLED
C     MYWRT,DABS
C
C
C *** NOTES
C     This routine is given direct access to the matrix A
C     but it does not alter hidden data structures.
C
C
C
C *** REFERENCES:
C     Gondzio J. (1994). Presolve analysis of linear programs prior 
C        to applying an interior point method, Technical Report 
C        No 1994.3, Department of Management Studies, University 
C        of Geneva, 102, Bd. Carl-Vogt, 1211 Geneva, Switzerland, 
C        February 1994, revised in December 1994.
C
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  November 18, 1993
C     Last modified: May 15, 1995
C
C
C
C
C *** BODY OF (POSTSL) ***
C
C
C
C
C
C
      IF(MSGLEV.LE.3) GO TO 140
      DO 130 J=1,N
         IF(STAVAR(J).LT.6) GO TO 130
         WRITE(BUFFER,131) J,LENCOL(J),STAVAR(J),
     X    LOBND(J),UPBND(J),PRLVAR(J)
  131    FORMAT(1X,'col=',I6,' ln=',I4,' st=',I6,' LO=',D10.3,
     X    ' UP=',D10.3,' X=',D10.3)
         CALL MYWRT(IOLOG,BUFFER)
  130 CONTINUE
  140 CONTINUE
C
C
C
C     Prepare data structures used to update optimal values
C     of the primal variables.
C     RNTMP1 array handles (updated) primal solution.
      DO 200 J=1,N
         RNTMP1(J)=PRLVAR(J)
  200 CONTINUE
C
C
C
C
C
C
C     Main loop begins here.
C     Loop over all PRE_SOLVE history list.
      DO 1000 IENTRY=LNHIST,1,-1
         J=INHIST(IENTRY)
         IF(J.LE.0) THEN
C
C
C
C     Here if the variable has its LOWER BOUND pushed.
C     Add the bound's contibution to the variable.
            J=-J
            RNTMP1(J)=RNTMP1(J)+DPHIST(IENTRY)
            IF(STAVAR(J).EQ.15) LOBND(J)=LOBND(J)-DPHIST(IENTRY)
C
         ELSE
C
C
C
C     Here if the variable has been made FREE (or implied FREE).
C     Look for pivot element.
            KBEG=CLPNTS(J)
            I=RWNMBS(KBEG)
            IF(MSGLEV.LE.2) GO TO 310
            WRITE(BUFFER,301) I,RWNAME(I),RWSTAT(I)
  301       FORMAT(1X,'POSTSL: Row ',I6,' (name=',A8,
     X       ') RWSTAT=',I6)
            CALL MYWRT(IOLOG,BUFFER)
            WRITE(BUFFER,302) J,I,ACOEFF(KBEG),DPHIST(IENTRY)
  302       FORMAT(1X,'cl=',I6,' rw=',I6,' elt=',D10.3,
     X       ' dphist=',D10.3)
            CALL MYWRT(IOLOG,BUFFER)
  310       CONTINUE
C
C     Restore objective and compute the value of the variable.
            VALUE=B(I)
            IPOS=RWHEAD(I)
            DP=C(J)/DPHIST(IENTRY)
  500       IF(IPOS.EQ.0) GO TO 700
               JCOL=CLNMBS(IPOS)
               IF(JCOL.EQ.J) GO TO 600
C              WRITE(BUFFER,501) IPOS,JCOL,C(JCOL),-DP*ACOEFF(IPOS)
C 501          FORMAT(1X,'ipos=',I6,' j=',I6,' c=',D10.3,
C    X          ' corr.=',D10.3)
C              CALL MYWRT(IOLOG,BUFFER)
               C(JCOL)=C(JCOL)+DP*ACOEFF(IPOS)
               VALUE=VALUE-RNTMP1(JCOL)*ACOEFF(IPOS)
  600       IPOS=RWLINK(IPOS)
            GO TO 500
  700       CONTINUE
            RNTMP1(J)=VALUE/DPHIST(IENTRY)
C
         ENDIF
C
C
C
C     End of main loop.
 1000 CONTINUE
C
C
C
C
C
C
C     Recover optimal values of all FREE variables.
      DO 1200 J=1,NSTRCT
         IF(STAVAR(J).EQ.15) PRLVAR(J)=RNTMP1(J)
 1200 CONTINUE
C
C
C
C
      RETURN
C
C
C
C *** LAST CARD OF (POSTSL) ***
      END
