C**************************************************************
C     **** PRESOL ... PRESOLVE ANALYSIS OF THE  LP PROBMEM ****
C**************************************************************
C
      SUBROUTINE PRESOL(MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X LNHIST,MXHIST,INHIST,DPHIST,
     X IWORK,RWORK,IMAP,RMAP,LIWORK,LRWORK,IROW,RELT,
     X X,B,C,CLNAME,UPBND,LOBND,
     X INTMP1,INTMP2,INTMP3,IMTMP1,IMTMP2,
     X RMTMP1,P,Q,RNTMP1,RNTMP2,RNTMP3,
     X PERM,INVP,HEADER,LINKFD,LINKBK,
     X STAVAR,RWNAME,STAROW,RWSTAT,RANGES,
     X MAXCOL,MSGLEV,LEVPRS,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 MAXM,MAXN,MAXNZA,M,N,NSTRCT,LIWORK,LRWORK
      INTEGER*4 LNHIST,MXHIST,MAXCOL,MSGLEV,LEVPRS,IOLOG
      INTEGER*4 INTMP1(MAXN),IROW(MAXN),IMTMP1(MAXM+1),IMTMP2(MAXM+1)
      INTEGER*2 INTMP2(MAXN),INTMP3(MAXN)
      DOUBLE PRECISION RMTMP1(MAXM),P(MAXM),Q(MAXM),RELT(MAXN)
      DOUBLE PRECISION RNTMP1(MAXN),RNTMP2(MAXN),RNTMP3(MAXN)
      INTEGER*2 PERM(MAXM),INVP(MAXM)
      INTEGER*2 HEADER(MAXM+1),LINKFD(MAXM+1),LINKBK(MAXM+1)
      CHARACTER*8 RWNAME(MAXM),CLNAME(MAXN)
      DOUBLE PRECISION X(MAXN),B(MAXM),RANGES(MAXM)
      DOUBLE PRECISION C(MAXN),UPBND(MAXN),LOBND(MAXN)
      INTEGER*2 STAVAR(MAXN),STAROW(MAXM),RWSTAT(MAXM)
      INTEGER*2 INHIST(MXHIST)
      DOUBLE PRECISION DPHIST(MXHIST)
C
C *** HIDDEN DATA STRUCTURES
      INTEGER*4 IWORK(LIWORK),IMAP(*),RMAP(*)
      DOUBLE PRECISION RWORK(LRWORK)
C
C
C
C *** LOCAL VARIABLES
      DOUBLE PRECISION BNDBIG
      INTEGER*4     I,IRUN,ITIGHT,NPASS
      INTEGER*4     M0,N0,NZ0,M1,N1,NZ1,MC,NC,NZC
      INTEGER*4     MJ0,NJ0,NZJ0,MJ1,NJ1,NZJ1,MKSP
      CHARACTER*100 BUFFER
C
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     MAXN    Maximum column dimension of the LP constraint matrix.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     M       Number of rows of the LP constraint matrix
C             (and the dimension of  A*Atransp).
C     N       Number of columns of the LP constraint matrix.
C     NSTRCT  Number of structural variables (excluding slacks, surplus
C             and artificials).
C     LNHIST  Length of the PRE_SOLVE history list;
C     MXHIST  Maximum number of entries in the PRE_SOLVE history list.
C     INHIST  Integer PRE_SOLVE history information.
C     DPHIST  Double precision PRE_SOLVE history information.
C     IROW and  RELT are the arrays for temporary handling of rows
C             and columns of the constraint matrix. They are primarily
C             intended to handle sparse vectors (in packed form)
C             but may also be used for storing dense ones.
C     INTMP1  Integer work array of size MAXN.
C     INTMP2  Half-length integer work array of size MAXN.
C     INTMP3  Half-length integer work array of size MAXN.
C     IMTMP1  Integer work array of size MAXM.
C     IMTMP2  Integer work array of size MAXM.
C     RMTMP1  Double precision work array of size MAXM.
C     RNTMP1  Double precision work array of size MAXN.
C     RNTMP2  Double precision work array of size MAXN.
C     RNTMP3  Double precision work array of size MAXN.
C     X       Primal variables of the linear program.
C     B       Right hand side of the linear program.
C     C       Objective function coefficients.
C     UPBND   Array of upper bounds. Note that the upper bound
C             is changed when the variable has a lower bound. 
C     LOBND   Array of lower bounds.
C     PERM    Permutation of rows of A.
C     INVP    Inverse permutation.
C     HEADER  Header of the doubly linked lists.
C     LINKFD  Forward linked lists.
C     LINKBK  Backward linked lists.
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C            -k  FREE variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicates the position of the original variable.
C     RWNAME  Array of row names (increasing order sort).
C     CLNAME  Array of column names (unordered).
C     STAROW  Array of row status:
C             0  row has been removed (it indicates a free row);
C             1  row has not been removed.
C     RWSTAT  Array of row types (sort as before):
C             1  row type is = ;
C             2  row type is >= ;
C             3  row type is <= ;
C             4  row type is objective or free.
C     RANGES  Array of constraint ranges.
C     MAXCOL  Threshold length for columns to be split.
C     MSGLEV  The level of PRE_SOLVE information desired:
C             0  only final problem dimensions are printed;
C             1  numbers of eliminated rows and columns are printed;
C             2  names of eliminated rows and columns are printed;
C             3  detailed DEBUGGING information is printed.
C     LEVPRS  The level of PRE_SOLVE analysis desired:
C             7  corresponds to maximum available presolve analysis
C                (with no limits on implied variables' bounds), 
C             6  corresponds to DEFAULT presolve analysis,
C             5  turns off splitting dense columns, 
C             4  additionally turns off the heuristic to make A sparser,
C             3  additionally turns off column aggregation technique,
C             2  additionally turns off the elimin. of dominated cols, 
C             1  additionally turns off the elimin. of redundant rows, 
C             0  turns off any presolve analysis.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     P       LOWER bounds on shadow prices (dual variables).
C     Q       UPPER bounds on shadow prices (dual variables).
C
C
C
C
C *** HIDDEN DATA STRUCTURES DESCRIPTION
C     RWORK   Real work array containing almost all real
C             LP problem data.
C     IWORK   Integer work array containing almost all integer
C             LP problem data.
C     RMAP    Map of RWORK.
C     IMAP    Map of IWORK.
C
C     Map of IWORK array:
C     IMAP(1)   Points to CLPNTS array.
C     IMAP(2)   Points to RWNMBS array.
C     IMAP(3)   Points to RWHEAD array.
C     IMAP(4)   Points to RWLINK array.
C     IMAP(5)   Points to CLNMBS array.
C     IMAP(6)   Points to LENCOL array.
C     IMAP(7)   Points to the first empty cell of IWORK array.
C
C     Map of RWORK array:
C     RMAP(1)   Points to ACOEFF array.
C     RMAP(2)   Points to COBJ array.
C     RMAP(3)   Points to RHS array.
C     RMAP(4)   Points to the first empty cell of RWORK array.
C
C
C
C
C *** SUBROUTINES CALLED:
C     MYWRT,GETDIM,FDIDEN,ELVRBL,ELCNST,RCLSNG,RRWSNG,FDAGGR,
C     MKSPAR,DETSPL,SPLIT
C
C
C *** PURPOSE:
C     This routine preprocesses the linear programming problem.
C     The following actions are performed: 
C     CLEAN:
C     - determine (and later tighten) bounds on shadow prices,
C     - eliminate dominated (and weakly dominated) variables,
C     - eliminate singleton rows,
C     - eliminate singleton columns (implied FREE variables),
C     - find identical columns and aggregate them,
C     - find hidden split FREE variables,
C     - eliminate redundant (dominated or forcing) constraints,
C     - tighten bounds on variables,
C     MAKE SPARSER:
C     - pivot out some nonzero entries of A,
C     SPLIT:
C     - split dense columns into shorter pieces.
C
C
C
C *** NOTES:
C
C
C *** REFERENCES:
C     Gondzio J. (1992). Splitting dense columns of the constraint
C        matrix in interior point methods for large scale linear
C        programming, Optimization 24, pp. 285-297.
C     Gondzio J. (1993). Implementing Cholesky factorization
C        for interior point methods of linear programming,
C        Optimization 27, pp. 121-140.
C     Gondzio J. (1994). Presolve analysis of linear programs prior
C        to applying an interior point method, Technical Report
C        No 1994.3, Department of Management Studies, University
C        of Geneva, 102, Bd. Carl-Vogt, 1211 Geneva, Switzerland,
C        February 1994, revised in December 1994.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  March 15, 1993
C     Last modified: May 17, 1995
C
C
C
C *** BODY OF (PRESOL) ***
C
C     WRITE(BUFFER,1) LEVPRS
C   1 FORMAT(1X,'PRESOL: Presolve level=',I2)
C     CALL MYWRT(IOLOG,BUFFER)
C
C
C
C     Set large values of M0, N0 and NZ0 to force at least
C     one pass of presolve analysis.
      M0=10000000
      N0=10000000
      NZ0=100000000
C
C     Force tightening upper bounds (1 disables tightening).
      ITIGHT=2
      IF(LEVPRS.LE.0) GO TO 4000
C
C
C     Determine problem dimensions.
      CALL GETDIM(IOLOG,
     X MAXM,MAXN,M,N,NSTRCT,
     X M1,N1,NZ1,B,C,IWORK(IMAP(6)),STAVAR)
C
C     Save initial problem dimensions.
      MJ0=M1
      NJ0=N1
      NZJ0=NZ1
C
C
C
C
C     Main loop begins here.
      IRUN=1
  100 DO 1000 NPASS=1,10
      BNDBIG=-1.0D0
      IF(NPASS.GE.2) BNDBIG=1.0D-6
C
C
C
C     Print statistics on the current formulation of the problem.
      IF(MSGLEV.LT.0) GO TO 110
      WRITE(BUFFER,101) NPASS-1,M1,N1,NZ1
  101 FORMAT(1X,'PRESOL: PASS=',I2,'   M=',I7,'   N=',I7,'   NZ=',I12)
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(99,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
  110 CONTINUE
C
C
C
C     Check if the last pass reduced the problem size.
      IF(M1.EQ.M0.AND.N1.EQ.N0.AND.NZ1.EQ.NZ0) GO TO 1100
      IF(M1.EQ.0) GO TO 3000
C
C
C
C     Save problem dimensions and continue the analysis.
      M0=M1
      N0=N1
      NZ0=NZ1
C
C
C
C     Eliminate dominated (and weakly dominated) variables.
C
      IF(NPASS.GT.1) GO TO 200
      IF(LEVPRS.LE.2) GO TO 200
C
      I=NPASS+IRUN-1
      CALL ELVRBL(IOLOG,MSGLEV,I,
     X MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X LNHIST,MXHIST,INHIST,DPHIST,
     X IMTMP1,IROW,RELT,
     X RWORK(RMAP(3)),RANGES,RWORK(RMAP(2)),UPBND,
     X RWORK(RMAP(1)),IWORK(IMAP(1)),IWORK(IMAP(2)),
     X IWORK(IMAP(3)),IWORK(IMAP(4)),IWORK(IMAP(5)),IWORK(IMAP(6)),
     X P,Q,X,STAVAR,RWSTAT,STAROW,RWNAME,CLNAME,
     X PERM,INVP,LINKFD)
C
C
C
C     Eliminate singleton FREE variables.
  200 CONTINUE
C
      CALL RCLSNG(IOLOG,MSGLEV,
     X MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X LNHIST,MXHIST,INHIST,DPHIST,
     X RWORK,IWORK,RMAP,IMAP,IROW,RELT,
     X IMTMP1,INTMP1,INTMP2,RNTMP1,
     X RWORK(RMAP(3)),RANGES,RWORK(RMAP(2)),LOBND,UPBND,
     X RWORK(RMAP(1)),IWORK(IMAP(1)),IWORK(IMAP(2)),
     X IWORK(IMAP(3)),IWORK(IMAP(4)),IWORK(IMAP(5)),IWORK(IMAP(6)),
     X P,Q,X,STAVAR,RWSTAT,STAROW,RWNAME,CLNAME,
     X PERM,INVP,HEADER)
C
C
C
C     Determine current problem dimensions.
      CALL GETDIM(IOLOG,
     X MAXM,MAXN,M,N,NSTRCT,
     X MC,NC,NZC,B,C,IWORK(IMAP(6)),STAVAR)
C
C
C
C     Eliminate singleton constraints.
C
      CALL RRWSNG(IOLOG,MSGLEV,
     X MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X LNHIST,MXHIST,INHIST,DPHIST,
     X RWORK,IWORK,RMAP,IMAP,IROW,RELT,IMTMP1,IMTMP2,
     X RWORK(RMAP(3)),RANGES,LOBND,UPBND,
     X RWORK(RMAP(1)),IWORK(IMAP(1)),IWORK(IMAP(2)),
     X IWORK(IMAP(3)),IWORK(IMAP(4)),IWORK(IMAP(5)),IWORK(IMAP(6)),
     X P,Q,X,STAVAR,RWSTAT,STAROW,RWNAME,CLNAME,
     X PERM,INVP,INTMP1,LINKFD,LINKBK)
C
C
C
C     If RRWSNG routine reduced the problem size, then try singleton
C     variables elimination again (uniquely in the first pass).
      IF(NPASS.GT.1) GO TO 300
      IF(M.EQ.MC) GO TO 300
C
      CALL RCLSNG(IOLOG,MSGLEV,
     X MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X LNHIST,MXHIST,INHIST,DPHIST,
     X RWORK,IWORK,RMAP,IMAP,IROW,RELT,
     X IMTMP1,INTMP1,INTMP2,RNTMP1,
     X RWORK(RMAP(3)),RANGES,RWORK(RMAP(2)),LOBND,UPBND,
     X RWORK(RMAP(1)),IWORK(IMAP(1)),IWORK(IMAP(2)),
     X IWORK(IMAP(3)),IWORK(IMAP(4)),IWORK(IMAP(5)),IWORK(IMAP(6)),
     X P,Q,X,STAVAR,RWSTAT,STAROW,RWNAME,CLNAME,
     X PERM,INVP,HEADER)
C
C
C
C     Eliminate dominated (and weakly dominated) variables.
  300 IF(LEVPRS.LE.2) GO TO 350
C
      I=2
      CALL ELVRBL(IOLOG,MSGLEV,I,
     X MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X LNHIST,MXHIST,INHIST,DPHIST,
     X IMTMP1,IROW,RELT,
     X RWORK(RMAP(3)),RANGES,RWORK(RMAP(2)),UPBND,
     X RWORK(RMAP(1)),IWORK(IMAP(1)),IWORK(IMAP(2)),
     X IWORK(IMAP(3)),IWORK(IMAP(4)),IWORK(IMAP(5)),IWORK(IMAP(6)),
     X P,Q,X,STAVAR,RWSTAT,STAROW,RWNAME,CLNAME,
     X PERM,INVP,LINKFD)
C
C
C
C     Find variables of identical structure.
C     Do it only in the first PASS (expensive search).
  350 IF(NPASS.GE.2) GO TO 400
      IF(NC.GE.5*MC) GO TO 400
      IF(LEVPRS.LE.3) GO TO 400
C
      CALL FDIDEN(IOLOG,MSGLEV,
     X MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X LNHIST,MXHIST,INHIST,DPHIST,
     X RWORK(RMAP(3)),RWORK(RMAP(2)),LOBND,UPBND,
     X RWORK(RMAP(1)),IWORK(IMAP(1)),IWORK(IMAP(2)),
     X IWORK(IMAP(3)),IWORK(IMAP(4)),IWORK(IMAP(5)),IWORK(IMAP(6)),
     X P,Q,X,STAVAR,RWSTAT,CLNAME,RANGES,
     X PERM,INVP,RMTMP1)
C
C
C
C     Determine current problem dimensions.
  400 CONTINUE
      CALL GETDIM(IOLOG,
     X MAXM,MAXN,M,N,NSTRCT,
     X M1,N1,NZ1,B,C,IWORK(IMAP(6)),STAVAR)
C
C
C
C     Eliminate redundant constraints.
      IF(M1.EQ.M0.AND.N1.EQ.N0.AND.NZ1.EQ.NZ0) THEN
C
C     Enable tightening UPPER bounds if no other possibility 
C     for reduction exists.
         IF(NPASS.GE.2) BNDBIG=1.0D+2
      ENDIF
      IF(LEVPRS.LE.1) GO TO 450
      IF(LEVPRS.GE.7.AND.NPASS.GE.2) BNDBIG=1.0D+8
C
      CALL ELCNST(IOLOG,MSGLEV,ITIGHT,
     X MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X LNHIST,MXHIST,INHIST,DPHIST,
     X RWORK,IWORK,RMAP,IMAP,IROW,RELT,
     X IMTMP1,INTMP1,INTMP2,RMTMP1,RNTMP1,
     X RWORK(RMAP(3)),RANGES,RWORK(RMAP(2)),LOBND,UPBND,BNDBIG,
     X RWORK(RMAP(1)),IWORK(IMAP(1)),IWORK(IMAP(2)),
     X IWORK(IMAP(3)),IWORK(IMAP(4)),IWORK(IMAP(5)),IWORK(IMAP(6)),
     X P,Q,X,STAVAR,RWSTAT,STAROW,RWNAME,CLNAME,
     X PERM,INVP,HEADER)
C
C
C
C     Find aggregate variables.
C     Do it only in the first two PASSes (expensive search).
  450 IF(NPASS.GE.3) GO TO 500
      IF(N1.GE.100*M1.AND.IRUN.EQ.2) GO TO 500
      IF(LEVPRS.LE.3) GO TO 500
C
      CALL FDAGGR(IOLOG,MSGLEV,ITIGHT,
     X MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X LNHIST,MXHIST,INHIST,DPHIST,
     X RWORK,IWORK,RMAP,IMAP,IROW,RELT,
     X IMTMP1,RMTMP1,INTMP1,RNTMP1,RNTMP2,RNTMP3,
     X RWORK(RMAP(3)),RANGES,RWORK(RMAP(2)),LOBND,UPBND,BNDBIG,
     X RWORK(RMAP(1)),IWORK(IMAP(1)),IWORK(IMAP(2)),
     X IWORK(IMAP(3)),IWORK(IMAP(4)),IWORK(IMAP(5)),IWORK(IMAP(6)),
     X P,Q,X,STAVAR,RWSTAT,STAROW,RWNAME,CLNAME,
     X INTMP2,PERM,INVP)
C
C
C
C     Determine current problem dimensions.
  500 CONTINUE
      CALL GETDIM(IOLOG,
     X MAXM,MAXN,M,N,NSTRCT,
     X M1,N1,NZ1,B,C,IWORK(IMAP(6)),STAVAR)
C
C
C
C
C     End of main loop.
 1000 CONTINUE
 1100 CONTINUE
C
C
C
C     Save problem dimensions and continue the analysis.
      M0=M1
      N0=N1
      NZ0=NZ1
C
C
C
C     Make the LP constraint matrix sparser.
C     Analyse EQUALITY type constraints.
      IF(LEVPRS.LE.4) GO TO 2000
      IF(IRUN.EQ.2) GO TO 2000
C
      CALL MKSPAR(IOLOG,MSGLEV,
     X MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X M1,N1,NZ1,IROW,RELT,
     X IMTMP1,INTMP1,INTMP2,
     X RWORK(RMAP(3)),RANGES,
     X RWORK(RMAP(1)),IWORK(IMAP(1)),IWORK(IMAP(2)),
     X IWORK(IMAP(3)),IWORK(IMAP(4)),IWORK(IMAP(5)),IWORK(IMAP(6)),
     X P,Q,STAVAR,RWSTAT,STAROW,RWNAME,
     X PERM,INVP,HEADER,LINKFD,LINKBK)
C
C     Save the number of nonzeros removed by MKSPAR routine.
      IF(IRUN.EQ.1) MKSP=NZ0-NZ1
C
C
C
C     Check if the MKSPAR routine reduced the problem size.
      IF(M1.EQ.M0.AND.N1.EQ.N0.AND.NZ1.EQ.NZ0) THEN
C
C     Further reduction is not possible.
         GO TO 2000
      ELSE
C
C     Repeat the analysis.
         IRUN=IRUN+1
         GO TO 100
      ENDIF
C
C
C
C     Split long columns of the constraint matrix if there are any.
 2000 CONTINUE
C
C
C
C     Determine optimal length for split columns.
      CALL DETSPL(IOLOG,
     X MAXM,MAXN,M,N,NSTRCT,
     X MAXCOL,IWORK(IMAP(6)),STAVAR)
C
C
C
C
      IF(LEVPRS.LE.5) GO TO 3000
C
      CALL SPLIT(MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X LNHIST,MXHIST,INHIST,DPHIST,
     X RWORK(RMAP(1)),IWORK(IMAP(1)),IWORK(IMAP(2)),
     X IWORK(IMAP(3)),IWORK(IMAP(4)),IWORK(IMAP(5)),IWORK(IMAP(6)),
     X HEADER,LINKFD,LINKBK,INTMP1,IROW,
     X MAXCOL,IMTMP1,RMTMP1,IMTMP2,RELT,
     X P,Q,CLNAME,STAVAR,X,RWORK(RMAP(2)),UPBND,LOBND,
     X RWNAME,STAROW,RWSTAT,RANGES,RWORK(RMAP(3)),IOLOG)
C
C
C
C
C     Determine final problem dimensions.
 3000 CONTINUE
      CALL GETDIM(IOLOG,
     X MAXM,MAXN,M,N,NSTRCT,
     X M1,N1,NZ1,B,C,IWORK(IMAP(6)),STAVAR)
C
C     Save final problem dimensions.
      MJ1=M1
      NJ1=N1
      NZJ1=NZ1
C
C     WRITE(BUFFER,3001) MJ0,NJ0,NZJ0,MJ1,NJ1,NZJ1,MKSP
C3001 FORMAT(1X,'qqqa',I5,' &',I6,' &',I7,
C    X ' &',I5,' &',I6,' &',I7,' &',I5)
C     CALL MYWRT(99,BUFFER)
C     CALL MYWRT(IOLOG,BUFFER)
C
C
C     Check if there are any FREE variables in the LP problem.
      NJ0=0
      DO 3100 I=1,N
         IF(STAVAR(I).LT.0) NJ0=NJ0+1
 3100 CONTINUE
C     WRITE(BUFFER,3101) NJ0/2
C3101 FORMAT(1X,'qqqc there are',I5,' FREE variables.')
C     CALL MYWRT(99,BUFFER)
C     CALL MYWRT(IOLOG,BUFFER)
C
C
C
C
 4000 CONTINUE
      RETURN
C
C
C
C *** LAST CARD OF (PRESOL) ***
      END
