C****************************************************
C     ****  RDMPS2 ... READ THE  MPS FILE  ****
C****************************************************
C
      SUBROUTINE RDMPS2(C,B,RANGES,
     X CLPNTS,RWNMBS,ACOEFF,
     X RWHEAD,RWLINK,CLNMBS,LENCOL,
     X MAXM,MAXN,MAXNZA,M,M0,N,NZA,NSTRCT,MULT,
     X IMTMP1,IMTMP2,IMTMP3,
     X STAVAR,UPBND,LOBND,P,Q,BIG,IROBJ,
     X RWNAME,RWCOPY,RWSTAT,STAROW,CLNAME,
     X RWORK,IWORK,RMAP,IMAP,IROW,RELT,
     X PRLVAR,IOLOG)
C
C
C
C *** VARIABLES AND ARRAYS ASSOCIATED WITH THE MPS FILE
      INTEGER*4 MAXM,MAXN,MAXNZA,M,M0,N,NZA,NSTRCT,IROBJ
      CHARACTER*8 RWNAME(MAXM),RWCOPY(MAXM),CLNAME(MAXN)
      INTEGER*2 STAVAR(MAXN),STAROW(MAXM),RWSTAT(MAXM)
      DOUBLE PRECISION RANGES(MAXM),UPBND(MAXN),LOBND(MAXN)
      DOUBLE PRECISION P(MAXM),Q(MAXM),BIG,MULT
C
C *** MPS VARIABLES DESCRIPTION
C     MAXM    Maximum number of constraints.
C     MAXN    Maximum number of variables.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     M       Current number of constraints.
C     M0      Original number of constraints.
C     N       Number of variables (total, i.e. including slacks, surplus
C             and artificials).
C     NZA     Current number of nonzeros of the LP constraint matrix.
C     NSTRCT  Number of structural variables (excluding slacks, surplus
C             and artificials).
C     MULT    Direction of optimization:
C             +1 means minimization;
C             -1 means maximization.
C     RWNAME  Array of row names.
C     RWCOPY  A copy of row names (keeps trace of the original order).
C     RWSTAT  Array of row types:
C             1  row type is = ;
C             2  row type is >= ;
C             3  row type is <= ;
C             4  objective row;
C             5  other free row.
C     STAROW  Array of row status:
C             0  row has been removed (it indicates a free row);
C             1  row has not been removed.
C     CLNAME  Array of column names.
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C            -k  FREE variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicates the position of the original variable.
C     RANGES  Array of constraint ranges.
C     UPBND   Array of upper bounds. Note that the upper bound
C             is changed when the variable has a lower bound. 
C     LOBND   Array of lower bounds.
C     P       LOWER bounds on shadow prices (dual variables).
C     Q       UPPER bounds on shadow prices (dual variables).
C
C
C
C *** INPUT/OUTPUT FILES
      INTEGER*4 IOLOG
C
C *** INPUT/OUTPUT FILES DESCRIPTION
C     IOLOG   Output unit number where log messages are to be written.
C
C
C
C *** HIDDEN DATA STRUCTURES
      INTEGER*4 IWORK(*),IMAP(*),RMAP(*)
      DOUBLE PRECISION RWORK(*)
C
C *** HIDDEN DATA STRUCTURES DESCRIPTION
C     RWORK   Real work array containing almost all real
C             LP problem data.
C     IWORK   Integer work array containing almost all integer
C             LP problem data.
C     RMAP    Map of RWORK.
C     IMAP    Map of IWORK.
C
C     Map of IWORK array:
C     IMAP(1)   Points to CLPNTS array.
C     IMAP(2)   Points to RWNMBS array.
C     IMAP(3)   Points to RWHEAD array.
C     IMAP(4)   Points to RWLINK array.
C     IMAP(5)   Points to CLNMBS array.
C     IMAP(6)   Points to LENCOL array.
C     IMAP(7)   Points to the first empty cell of IWORK array.
C
C     Map of RWORK array:
C     RMAP(1)   Points to ACOEFF array.
C     RMAP(2)   Points to COBJ array.
C     RMAP(3)   Points to RHS array.
C     RMAP(4)   Points to the first empty cell of RWORK array.
C
C
C
C *** LP DATA
      DOUBLE PRECISION ACOEFF(MAXNZA),C(MAXN),B(MAXM)
      INTEGER*4 CLPNTS(MAXN+1),RWHEAD(MAXM),RWLINK(MAXNZA)
      INTEGER*2 RWNMBS(MAXNZA),CLNMBS(MAXNZA),LENCOL(MAXN)
C
C *** LP DATA DESCRIPTION
C     ACOEFF  Array of nonzero elements for each column.
C     C       Objective function coefficients.
C     B       Right hand side of the linear program.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWHEAD  Headers to the row linked lists of matrix A.
C     RWLINK  Row linked lists of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A.
C     CLNMBS  Column numbers of nonzeros in rows of matrix A.
C     LENCOL  Lengths of (sparse) columns of matrix A.
C
C
C
C *** WORK ARRAYS
      INTEGER*4 IMTMP1(MAXM),IMTMP2(MAXM),IMTMP3(MAXM)
      INTEGER*4 IROW(MAXN)
      DOUBLE PRECISION RELT(MAXN),PRLVAR(MAXN)
C
C *** WORK ARRAYS DESCRIPTION
C     IMTMP1  Integer work array of size MAXM.
C     IMTMP2  Integer work array of size MAXM.
C     IMTMP3  Integer work array of size MAXM
C     IROW and  RELT are the arrays for temporary handling of rows
C             and columns of the constraint matrix. They are primarily
C             intended to handle sparse vectors (in packed form)
C             but may also be used for storing dense ones.
C     PRLVAR  Primal variables of the LP problem.
C
C
C
C *** LOCAL VARIABLES
      INTEGER*4 MEQ,MGE,MLE,MFREE,MNEW,I,J,COLLEN,INDEX
      INTEGER*4 ITRACE,IKX,IPOS,K,KBEG,KEND,KNEW,KOK,KOUT
      DOUBLE PRECISION BIGNEW,VAL1
      CHARACTER*100 BUFFER
C
C
C
C *** LOCAL VARIABLES DESCRIPTION
C     MEQ     Number of equality constraints.
C     MGE     Number of constraints of type  greater or equal.
C     MLE     Number of constraints of type  less or equal.
C     MFREE   Number of free constraints.
C     MNEW    Number of constraints after removing the free ones.
C     ITRACE  Trace parameter:
C             0  means no message at all;
C             1  means writing  MPS file statistics;
C             2  means detailed tracing the MPS file input.
C
C
C
C *** PURPOSE
C     This routine sets up HOPDM data structures based 
C     on the information read earlier from the MPS input file.
C
C
C
C *** SUBROUTINES CALLED
C     SDOT,GETCOL,GETROW,MYSTOP
C
C
C
C *** NOTES
C
C
C
C *** REFERENCES:
C     Altman A., Gondzio J. (1993). An efficient implementation of 
C        a higher order primal-dual interior point method for large 
C        sparse linear programs, Archives of Control Sciences 2, 
C        No 1-2, pp. 23-40. 
C     Altman A., Gondzio J. (1993). HOPDM - A higher order primal-
C        dual method for large scale linear programmming, European
C        Journal of Operational Research 66 (1993) pp. 158-160.
C     Gondzio J., Tachat D. (1994). The design and application of 
C        IPMLO - a FORTRAN library for linear optimization with 
C        interior point methods, RAIRO Recherche Operationnelle 28, 
C        No 1, pp. 37-56.
C     Gondzio J. (1995). HOPDM (version 2.12) - A fast LP solver 
C        based on a primal-dual interior point method, European
C        Journal of Operational Research 85, pp. 221-225.
C     Murtagh B. (1981). Advanced Linear Programming, McGrew-Hill,
C        New York, 1981.
C     Murtagh B., Saunders M. (1983). MINOS 5.0 User's guide,
C        Technical Report SOL 83-20, Department of Operations Research,
C        Stanford University, Stanford, 1983.
C
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  November 15, 1992
C     Last modified: July 9, 1995
C
C
C
C
C
C
C *** BODY OF (RDMPS2) ***
C
C
C
C     Save the original sequence of rows.
      M0=M
      DO 20 I=1,M
         RWCOPY(I)=RWNAME(I)
C        WRITE(BUFFER,21) I,RWSTAT(I),RWNAME(I),RWCOPY(I)
C  21    FORMAT(1X,'rw=',I6,',  st=',I2,',  name=',A8,'  copy=',A8)
C        CALL MYWRT(IOLOG,BUFFER)
   20 CONTINUE
C
C
C
C     Initialize.
      ITRACE=1
      MEQ=0
      MGE=0
      MLE=0
      MFREE=0
      BIGNEW=0.999*BIG
C
C     Initialize bounds on shadow prices.
      DO 50 I=1,M
         P(I)=-BIG
         Q(I)=BIG
   50 CONTINUE
C
C
C     Count constraints of different types.
      DO 100 I=1,M
         STAROW(I)=0
C
         IF(RWSTAT(I).EQ.1) THEN
            MEQ=MEQ+1
            GO TO 100
         ENDIF
C
         IF(RWSTAT(I).EQ.2) THEN
            MGE=MGE+1
            GO TO 100
         ENDIF
C
         IF(RWSTAT(I).EQ.3) THEN
            MLE=MLE+1
            GO TO 100
         ENDIF
C
         IF(RWSTAT(I).GE.4) THEN
            MFREE=MFREE+1
            GO TO 100
         ENDIF
  100 CONTINUE
C
C
C     Write the  MPS statistics.
      IF(ITRACE.EQ.0) GO TO 120
C
      WRITE(BUFFER,101) 
  101 FORMAT(1X,'RDMPS:  MPS file statistics:')
      CALL MYWRT(IOLOG,BUFFER)
      CALL MYWRT(99,BUFFER)
      WRITE(BUFFER,102) M-1
  102 FORMAT(8X,I9,' constraints in the  LP problem:')
      CALL MYWRT(IOLOG,BUFFER)
      CALL MYWRT(99,BUFFER)
      WRITE(BUFFER,103) MEQ
  103 FORMAT(8X,I9,' of equality type,')
      CALL MYWRT(IOLOG,BUFFER)
      CALL MYWRT(99,BUFFER)
      WRITE(BUFFER,104) MGE
  104 FORMAT(8X,I9,' of type greater or equal to,')
      CALL MYWRT(IOLOG,BUFFER)
      CALL MYWRT(99,BUFFER)
      WRITE(BUFFER,105) MLE
  105 FORMAT(8X,I9,' of type less or equal to,')
      CALL MYWRT(IOLOG,BUFFER)
      CALL MYWRT(99,BUFFER)
      WRITE(BUFFER,106) MFREE-1
  106 FORMAT(8X,I9,' free rows (excluding objective);')
      CALL MYWRT(IOLOG,BUFFER)
      CALL MYWRT(99,BUFFER)
      WRITE(BUFFER,107) N
  107 FORMAT(8X,I9,' variables in the  LP problem;')
      CALL MYWRT(IOLOG,BUFFER)
      CALL MYWRT(99,BUFFER)
      NZA=CLPNTS(N+1)-1
      WRITE(BUFFER,108) NZA
  108 FORMAT(8X,I9,' nonzero elts in the  LP problem.')
      CALL MYWRT(IOLOG,BUFFER)
      CALL MYWRT(99,BUFFER)
C
  120 CONTINUE
C
C
C     Initialize objective function.
      DO 160 J=1,MAXN
         C(J)=0.0D0
  160 CONTINUE
C
C
C     Retrieve objective function from the constraint matrix.
C     Count nonzero elements in all LP constraints.
      DO 200 J=1,N
         KBEG=CLPNTS(J)
         KEND=CLPNTS(J+1)-1
         DO 180 INDEX=KBEG,KEND
            I=RWNMBS(INDEX)
            STAROW(I)=STAROW(I)+1
            IF(I.EQ.IROBJ) C(J)=MULT*ACOEFF(INDEX)
  180    CONTINUE
C        WRITE(IOLOG,181) J,C(J)
C 181    FORMAT(1X,'J=',I5,'  Cj=',D14.6)
  200 CONTINUE
C
C
C     Set the row linked lists of nonzero elements of matrix  A.
C     Set up  LENCOL array (do not count entries of free rows).
C     Recall that  CLPNTS(j) indicates the first entry of column j.
C     Zero  RWHEAD array.
      DO 420 I=1,M
         RWHEAD(I)=0
  420 CONTINUE
C
C     Set the row linked lists.
      DO 460 J=1,N
         LENCOL(J)=0
         KBEG=CLPNTS(J)
         KEND=CLPNTS(J+1)-1
         DO 440 K=KBEG,KEND
            I=RWNMBS(K)
            IF(RWSTAT(I).LE.3) LENCOL(J)=LENCOL(J)+1
            RWLINK(K)=RWHEAD(I)
            CLNMBS(K)=J
            RWHEAD(I)=K
  440    CONTINUE
C        WRITE(IOLOG,441) J,LENCOL(J),CLPNTS(J),CLPNTS(J+1)
C 441    FORMAT(1X,'J=',I5,'  ln=',I6,'  pnts=',I6,2X,I6)
  460 CONTINUE
C
C
C
C     Modify  STAROW array. Zero value indicates empty (or free) row.
      DO 480 I=1,M
         IF(STAROW(I).GT.0) STAROW(I)=1
         IF(RWSTAT(I).GE.4) STAROW(I)=0
  480 CONTINUE
C
C
C
C     Reorder rows of the  LP constraint matrix to eliminate
C     the empty ones.
C
C     Determine the permutation that puts all empty (or free)
C     rows at the end of the list.
C
      I=1
      CALL EMPTYR(MAXM,M,MNEW,I,
     X RWHEAD,STAROW,IROW(1),IMTMP1(1),IOLOG)
C
C
C
C     Reorder the rows of the  LP constraint matrix according to
C     the permutation resulting from the elimination of empty rows.
C
      IF(MNEW.EQ.M) GO TO 610
      CALL REORDA(MAXM,MAXN,MAXNZA,M,N,
     X CLPNTS,RWNMBS,
     X RWHEAD,RWLINK,CLNMBS,LENCOL,
     X IROW(1),IMTMP1(1),IMTMP2,IMTMP3,RELT,
     X RWNAME,STAROW,RWSTAT,RANGES,B,IOLOG)
C
C
C
C     Reorder elements within each column of the  LP constraint
C     matrix in such a way that those of the active part of  A
C     are at the beginning of the lists. The column lengths will
C     later be decreased to forget inactive part of matrix  A.
C     Set the new row linked lists of nonzero elements of matrix  A.
C     Recall that  CLPNTS(j) indicates the first entry of column j.
C     Zero  RWHEAD array.
         DO 520 I=1,M
            RWHEAD(I)=0
  520    CONTINUE
         DO 600 J=1,N
            KOK=0
            KOUT=0
            KBEG=CLPNTS(J)-1
            COLLEN=CLPNTS(J+1)-CLPNTS(J)
            DO 540 IKX=1,COLLEN
               K=KBEG+IKX
               I=RWNMBS(K)
               IF(I.LE.MNEW) THEN
                  KOK=KOK+1
                  IROW(KOK)=RWNMBS(K)
                  RELT(KOK)=ACOEFF(K)
               ELSE
                  IPOS=COLLEN-KOUT
                  KOUT=KOUT+1
                  IROW(IPOS)=RWNMBS(K)
                  RELT(IPOS)=ACOEFF(K)
               ENDIF
  540       CONTINUE
C
C     Set the row linked lists.
            DO 560 IKX=1,COLLEN
               K=KBEG+IKX
               I=IROW(IKX)
               RWNMBS(K)=I
               ACOEFF(K)=RELT(IKX)
               RWLINK(K)=RWHEAD(I)
               RWHEAD(I)=K
  560       CONTINUE
  600    CONTINUE
C
C
C     Set the new number of rows of the constraint matrix.
  610 M=MNEW
C
C
C
C
C     Define variable status and transform the problem accordingly.
C
C     STAVAR(j)=0  for STANDARD (nonnegative) variable.
C
C     STAVAR(j)=-k for FREE variable. FREE variable is split into
C     two STANDARD variables: xj-xk. Consequently, variable xk
C     is added to the problem and j-th column is replicated
C     with the negative sign.
C
C     STAVAR(j)=1  for UPPER bounded variable.
C
C     STAVAR(j)=2  for LOWER bounded variable. Such variable is
C     pushed to a zero LOWER bound (it becomes a STANDARD variable)
C     and the  RHS vector is transformed accordingly.
C
C     STAVAR(j)=3  for both LOWER and UPPER bounded variable. Such
C     variable is first pushed to a zero LOWER bound, which causes
C     changing its UPPER bound (and at the same time requires
C     some modification of the  RHS vector) and later treated as
C     an UPPER bounded variable.
C
C     STAVAR(j)=6  for FIXED variable. Such variable is eliminated
C     from the  LP problem formulation (this needs appropriate
C     modification of the RHS vector).
C
C
C
C
C
C
C     Analyse variable bounds and define  STAVAR array.
      NSTRCT=N
      NZA=CLPNTS(N+1)-1
      DO 700 J=1,NSTRCT
         STAVAR(J)=0
         PRLVAR(J)=0.0D0
C
C
C     Check if the variable is FREE.
         IF(LOBND(J).LE.-BIGNEW.AND.UPBND(J).GE.BIGNEW) THEN
            N=N+1
            IF(N.GT.MAXN) GO TO 9030
            KBEG=CLPNTS(J)
            KEND=KBEG+LENCOL(J)-1
            COLLEN=LENCOL(J)
            IF(NZA+COLLEN.GT.MAXNZA) GO TO 9020
            DO 620 K=KBEG,KEND
               KNEW=NZA+K-KBEG+1
               RWNMBS(KNEW)=RWNMBS(K)
               ACOEFF(KNEW)=-ACOEFF(K)
               I=RWNMBS(K)
               RWLINK(KNEW)=RWHEAD(I)
               CLNMBS(KNEW)=N
               RWHEAD(I)=KNEW
  620       CONTINUE
            C(N)=-C(J)
            STAVAR(N)=-J
            LOBND(N)=0.0D0
            UPBND(N)=BIG
            PRLVAR(N)=0.0D0
            STAVAR(J)=-N
            LOBND(J)=0.0D0
            UPBND(J)=BIG
            CLPNTS(N)=NZA+1
            CLNAME(N)='        '
            LENCOL(N)=COLLEN
            NZA=NZA+COLLEN
            GO TO 700
         ENDIF
C
C
C     Check if the variable has type MINUS INFINITY.
         IF(LOBND(J).LE.-BIGNEW.AND.UPBND(J).LE.BIGNEW) THEN
C
C     Observe that we only change the sign of MI type variable.
C     The next section will take contribution of nonzero lower
C     bound, if any.
            KBEG=CLPNTS(J)
            KEND=KBEG+LENCOL(J)-1
            DO 640 K=KBEG,KEND
               ACOEFF(K)=-ACOEFF(K)
  640       CONTINUE
            C(J)=-C(J)
            LOBND(J)=-UPBND(J)
            UPBND(J)=BIG
         ENDIF
C
C
C
C     Check if the variable has nonzero LOWER bound.
         IF(LOBND(J).NE.0.0D0) THEN
            STAVAR(J)=2
            VAL1=LOBND(J)
            KBEG=CLPNTS(J)
            KEND=KBEG+LENCOL(J)-1
            DO 660 K=KBEG,KEND
               I=RWNMBS(K)
               B(I)=B(I)-VAL1*ACOEFF(K)
  660       CONTINUE
C
C     Check if the variable has also finite UPPER bound.
            IF(UPBND(J).LE.BIGNEW) THEN
C
C     Here if the variable has both LOWER and UPPER bounds.
C     Observe that it has already been pushed to a zero LOWER bound.
C     Consequently, its new UPPER bound is equal to UPBND(J)-LOBND(J).
               STAVAR(J)=3
               IF(UPBND(J).LT.LOBND(J)) GO TO 9200
               UPBND(J)=UPBND(J)-LOBND(J)
C
C     Check if the UPPER bound is equal to the LOWER bound.
C     If so, then FIX the variable on its LOWER bound.
               IF(UPBND(J).EQ.0.0D0) THEN
                  PRLVAR(J)=0.0
                  STAVAR(J)=6
                  GO TO 700
               ENDIF
            ENDIF
            GO TO 700
         ENDIF
C
C
C     Check if the variable has finite UPPER bound.
         IF(UPBND(J).LE.BIGNEW) THEN
            STAVAR(J)=1
            IF(UPBND(J).LT.0.0D0) GO TO 9200
C
C     Check if the UPPER bound is equal to the LOWER bound.
C     If so, then FIX the variable on its LOWER bound.
            IF(UPBND(J).EQ.0.0D0) THEN
               PRLVAR(J)=0.0
               STAVAR(J)=6
               GO TO 700
            ENDIF
            GO TO 700
         ENDIF
C
  700 CONTINUE
      CLPNTS(N+1)=NZA+1
      NSTRCT=N
C
C
C
C
C     Check if there were any  FIXED variables. If so, then they
C     have been removed from the  LP problem formulation and,
C     consequently, should be removed from the row linked lists.
      DO 710 J=1,N
         IF(STAVAR(J).EQ.6) THEN
            GO TO 720
         ENDIF
  710 CONTINUE
C
C     Here if there are no  FIXED variables.
      GO TO 780
C
C     Here if a  FIXED variable has been found.
C     Set the new row linked lists of nonzero elements of matrix  A.
C     Recall that  CLPNTS(j) indicates the first entry of column j.
C     Zero  RWHEAD array.
  720 DO 730 I=1,M
         RWHEAD(I)=0
  730 CONTINUE
C     WRITE(BUFFER,731)
C 731 FORMAT(1X,'RDMPS2: FIXED variables found.')
C     CALL MYWRT(IOLOG,BUFFER)
C
C     Set the row linked lists.
      DO 750 J=1,N
C
C     Omit  FIXED variables.
         IF(STAVAR(J).EQ.6) GO TO 750
         KBEG=CLPNTS(J)
         KEND=KBEG+LENCOL(J)-1
         DO 740 K=KBEG,KEND
            I=RWNMBS(K)
            RWLINK(K)=RWHEAD(I)
            CLNMBS(K)=J
            RWHEAD(I)=K
  740    CONTINUE
  750 CONTINUE
C
C
C
C     Check if the removal of the FIXED variables created empty rows.
      DO 770 I=1,M
         IF(RWHEAD(I).NE.0) GO TO 770
C
C
C     Determine the permutation that puts all empty rows
C     at the end of the list.
C
         CALL EMPTYR(MAXM,M,MNEW,2,
     X    RWHEAD,STAROW,IROW(1),IMTMP1(1),IOLOG)
C
C
C     Reorder the rows of the  LP constraint matrix according to
C     the permutation resulting from the elimination of empty rows.
C
         CALL REORDA(MAXM,MAXN,MAXNZA,M,N,
     X    CLPNTS,RWNMBS,
     X    RWHEAD,RWLINK,CLNMBS,LENCOL,
     X    IROW(1),IMTMP1(1),IMTMP2,IMTMP3,RELT,
     X    RWNAME,STAROW,RWSTAT,RANGES,B,IOLOG)
C
C     Set the new number of rows of the constraint matrix.
         M=MNEW
C
         GO TO 780
  770 CONTINUE
C
C
C
C
C
C
C     Add slack and surplus variables.
C     Initialize all structural variables to zero values.
  780 CONTINUE
C
C
C
C
C     Recall that  RWSTAT is the array of row types:
C     1  row type is = ;
C     2  row type is >= ;
C     3  row type is <= ;
C     4  objective row;
C     5  other free row.
C
C     Loop over all constraints.
      DO 800 I=1,M
C
         IF(RWSTAT(I).EQ.1) THEN
C
C     Here for EQUALITY constraint.
C     Do not add any variable.
            GO TO 800
         ENDIF
C
         N=N+1
         IF(N.GE.MAXN) GO TO 9030
         IF(NZA+1.GE.MAXNZA) GO TO 9020
         KNEW=NZA+1
C
         IF(RWSTAT(I).EQ.2) THEN
C
C     Here for GREATER OR EQUAL type constraint.
C     Add the artificial variable.
            ACOEFF(KNEW)=-1.0
            C(N)=0.0
            LENCOL(N)=1
            GO TO 790
         ENDIF
C
         IF(RWSTAT(I).EQ.3) THEN
C
C     Here for LESS OR EQUAL type constraint.
C     Add the artificial variable.
            ACOEFF(KNEW)=+1.0
            C(N)=0.0
            LENCOL(N)=1
            GO TO 790
         ENDIF
C
         GO TO 9120
C
  790    RWNMBS(KNEW)=I
         RWLINK(KNEW)=RWHEAD(I)
         CLNMBS(KNEW)=N
         RWHEAD(I)=KNEW
         STAVAR(N)=0
         CLNAME(N)='logical '
         LOBND(N)=0.0D0
         UPBND(N)=BIG
         IF(RANGES(I).LE.BIGNEW) THEN
C           WRITE(BUFFER,791) I,RANGES(I)
C 791       FORMAT(1X,'RDMPS2: row=',I6,'  range=',D12.6)
C           CALL MYWRT(0,BUFFER)
C           CALL MYWRT(IOLOG,BUFFER)
            STAVAR(N)=1
            UPBND(N)=RANGES(I)
         ENDIF
         PRLVAR(N)=0.0D0
         CLPNTS(N)=NZA+1
         NZA=NZA+1
C
C
C
  800 CONTINUE
      CLPNTS(N+1)=NZA+1
C
C
C
C     Initialize primal variables and count large upper bounds.
      K=0
      DO 820 J=1,N
         PRLVAR(J)=0.0
         IF(STAVAR(J).EQ.1.OR.STAVAR(J).EQ.3) THEN
            IF(UPBND(J).GE.9999.0) K=K+1
         ENDIF
  820 CONTINUE
      IF(K.GE.1) THEN
         WRITE(BUFFER,821) K
  821    FORMAT(1X,'RDMPS:  ',I8,' variables have large UPPER bound.')
         CALL MYWRT(0,BUFFER)
         CALL MYWRT(IOLOG,BUFFER)
      ENDIF
C
C
      IF(ITRACE.EQ.0) GO TO 890
C
C     Write the  LP problem statistics.
      K=0
      DO 840 J=1,N
         IF(STAVAR(J).GE.6) GO TO 840
         K=K+LENCOL(J)
  840 CONTINUE
      WRITE(BUFFER,891)
  891 FORMAT(1X,'RDMPS:  Reformulated  LP problem statistics:')
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,892) M
  892 FORMAT(8X,I9,' constraints in the  LP problem;')
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,893) N
  893 FORMAT(8X,I9,' variables in the  LP problem,')
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,894) NSTRCT
  894 FORMAT(8X,I9,' of which are structurals;')
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,895) K
  895 FORMAT(8X,I9,' nonzero elts in matrix A.')
      CALL MYWRT(IOLOG,BUFFER)
C
  890 CONTINUE
C
C
C
C     IF(ITRACE.GE.2) THEN
C
C        WRITE(IOLOG,*)
C        WRITE(IOLOG,*) 'RDMPS:  MPS file has been read.'
C
C     LP problem constraints.
C        WRITE(IOLOG,*)
C        WRITE(IOLOG,*) 'RWNAME'
C        WRITE(IOLOG,'(5(2X,A8))') (RWNAME(I),I=1,M)
C        WRITE(IOLOG,*) 'RWSTAT'
C        WRITE(IOLOG,'(5(2X,I8))') (RWSTAT(I),I=1,M)
C        WRITE(IOLOG,*) 'B'
C        WRITE(IOLOG,'(5(2X,D12.5))') (B(I),I=1,M)
C        WRITE(IOLOG,*) 'RANGES'
C        WRITE(IOLOG,'(5(2X,D12.5))') (RANGES(I),I=1,M)
C
C     LP problem variables.
C        WRITE(IOLOG,*)
C        WRITE(IOLOG,*) 'CLNAME'
C        WRITE(IOLOG,'(5(2X,A8))') (CLNAME(J),J=1,N)
C        WRITE(IOLOG,*) 'STAVAR'
C        WRITE(IOLOG,'(5(2X,I8))') (STAVAR(J),J=1,N)
C        WRITE(IOLOG,*) 'C'
C        WRITE(IOLOG,'(5(2X,D12.5))') (C(J),J=1,N)
C        WRITE(IOLOG,*) 'UPBND'
C        WRITE(IOLOG,'(5(2X,D12.5))') (UPBND(J),J=1,N)
C        WRITE(IOLOG,333) 'LOBND'
C 333    FORMAT(1X,A8)
C        WRITE(IOLOG,'(5(2X,D12.5))') (LOBND(J),J=1,N)
C
C     LP constraint matrix.
C        WRITE(IOLOG,*)
C        WRITE(IOLOG,*) 'RWHEAD'
C        WRITE(IOLOG,'(5(2X,I8))') (RWHEAD(I),I=1,M)
C        WRITE(IOLOG,*) 'CLPNTS'
C        WRITE(IOLOG,'(5(2X,I8))') (CLPNTS(J),J=1,N+1)
C        WRITE(IOLOG,*) 'LENCOL'
C        WRITE(IOLOG,'(5(2X,I8))') (LENCOL(J),J=1,N)
C        WRITE(IOLOG,*)
C        DO 1010 K=1,NZA
C           WRITE(IOLOG,1011) K,ACOEFF(K),RWNMBS(K),
C    X       CLNMBS(K),RWLINK(K)
C1011       FORMAT(1X,'sub=',I8,'   elt=',D12.5,
C    X       '   rw=',I8,'   cl=',I8,'   rwlnk=',I8)
C1010    CONTINUE
C
C     LP constraint matrix by ROWS.
C        WRITE(IOLOG,*)
C        WRITE(IOLOG,*) 'ROWS of the LP constraint matrix'
C        DO 1020 I=1,M
C           CALL GETROW(I,RWORK,IWORK,RMAP,IMAP,
C    X       IROW,RELT,K,MAXN,IOLOG)
C           WRITE(IOLOG,'(A4,I5,A2,I4,A4,10(1X,I4))') 'ROW ',I,
C    X       ' (',K,') : ',(IROW(J),J=1,K)
C1020    CONTINUE
C
C     LP constraint matrix by COLUMNS.
C        WRITE(IOLOG,*)
C        WRITE(IOLOG,*) 'COLUMNS of the LP constraint matrix'
C        DO 1030 J=1,N
C           CALL GETCOL(J,RWORK,IWORK,RMAP,IMAP,
C    X       IROW,RELT,K,MAXN,IOLOG)
C           WRITE(IOLOG,'(A4,I5,A2,I4,A4,10(1X,I4))') 'COL ',J,
C    X       ' (',K,') : ',(IROW(I),I=1,K)
C1030    CONTINUE
C        WRITE(IOLOG,*)
C
C     ENDIF
C
C
C
C
      RETURN
C
C
C
C
C
C     Here when error occurs.
 9020 WRITE(BUFFER,9021)
 9021 FORMAT(1X,'RDMPS2 ERROR: Number of nonzeros',
     X ' of matrix A exceeds MAXNZA.')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(8)
C
 9030 WRITE(BUFFER,9031)
 9031 FORMAT(1X,'RDMPS2 ERROR: Number of variables',
     X ' in the MPS file exceeds MAXN.')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(8)
C
 9120 WRITE(BUFFER,9121) I,RWSTAT(I)
 9121 FORMAT(1X,'RDMPS2 ERROR: Constraint',I8,' has RWSTAT=',I8)
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(87)
C
 9200 WRITE(BUFFER,9201) LOBND(J),UPBND(J),CLNAME(J)
 9201 FORMAT(1X,'RDMPS2: LO bound ',D12.6,' exceeds ',
     X 'UP one ',D12.6,' (var=',A8,').')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(87)
C
C
C
C
C
C *** LAST CARD OF (RDMPS2) ***
      END
