C**********************************************************
C     ****  RDSPEC ... READ THE SPECIFICATIONS FILE  ****
C**********************************************************
C
      SUBROUTINE RDSPEC(FILMPS,FILSPC,FILLOG,FILSOL,
     X MAXM,MAXN,MAXNZA,NAMEC,NAMEB,NAMBND,NAMRAN,
     X MULT,BIG,DLOBND,DUPBND,OPTTOL,LEVPRS)
C
C
C
C *** PARAMETERS
      INTEGER*4 MAXM,MAXN,MAXNZA,LEVPRS
      CHARACTER*13 FILMPS,FILSPC,FILLOG,FILSOL
      CHARACTER*9 NAMEC,NAMEB,NAMBND,NAMRAN
      DOUBLE PRECISION MULT,BIG,DLOBND,DUPBND,OPTTOL
C
C
C *** LOCAL VARIABLES
      INTEGER*4 I,IKEY,IOSPC,LINE
      CHARACTER*3 KEY(25,2),KEYWRD
      CHARACTER*12 TEXT
      CHARACTER*100 BUFFER
C
C
C *** PARAMETERS DESCRIPTION
C     FILSPC  Specifications file name.
C     FILMPS  MPS input file name.
C     FILLOG  Log file name.
C     FILSOL  Solution file name.
C     NAMEC   The name of the desired objective function.
C     NAMEB   The name of the right hand side section chosen.
C     NAMBND  The name of the bound section chosen.
C     NAMRAN  The name of the range section chosen.
C     MULT    Direction of optimization:
C             +1 means minimization;
C             -1 means maximization.
C     BIG     "Big" number.
C     DLOBND  Default LOWER bound.
C     DUPBND  Default UPPER bound.
C     OPTTOL  Optimality tolerance (default 1.0D-8).
C     LEVPRS  The level of presolve analysis desired:
C             7  corresponds to maximum available presolve analysis
C                (with no limits on implied variables' bounds), 
C             6  corresponds to DEFAULT presolve analysis,
C             5  turns off splitting dense columns, 
C             4  additionally turns off the heuristic to make A sparser,
C             3  additionally turns off column aggregation technique,
C             2  additionally turns off the elimin. of dominated cols, 
C             1  additionally turns off the elimin. of redundant rows, 
C             0  turns off any presolve analysis.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Last modified: June 9, 1995
C
C
C
C
C *** BODY OF (RDSPEC) ***
C
C
C
C
C     Define keywords.
      KEY(1,1)='ROW'
      KEY(1,2)='row'
      KEY(2,1)='COL'
      KEY(2,2)='col'
      KEY(3,1)='ELE'
      KEY(3,2)='ele'
      KEY(4,1)='NON'
      KEY(4,2)='non'
C
      KEY(5,1)='MAX'
      KEY(5,2)='max'
      KEY(6,1)='MIN'
      KEY(6,2)='min'
      KEY(7,1)='LOB'
      KEY(7,2)='lob'
      KEY(8,1)='UPB'
      KEY(8,2)='upb'
C
      KEY(9,1)='OBJ'
      KEY(9,2)='obj'
      KEY(10,1)='RHS'
      KEY(10,2)='rhs'
      KEY(11,1)='RAN'
      KEY(11,2)='ran'
      KEY(12,1)='BOU'
      KEY(12,2)='bou'
C
      KEY(13,1)='MPS'
      KEY(13,2)='mps'
      KEY(14,1)='SOL'
      KEY(14,2)='sol'
      KEY(15,1)='ERR'
      KEY(15,2)='err'
C
      KEY(16,1)='   '
      KEY(16,2)='   '
      KEY(17,1)='   '
      KEY(17,2)='   '
      KEY(18,1)='OPT'
      KEY(18,2)='opt'
C
      KEY(19,1)='BEG'
      KEY(19,2)='beg'
      KEY(20,1)='END'
      KEY(20,2)='end'
C
      KEY(21,1)='PRE'
      KEY(21,2)='pre'
      KEY(22,1)='LOG'
      KEY(22,2)='log'
      KEY(23,1)='   '
      KEY(23,2)='   '
      KEY(24,1)='   '
      KEY(24,2)='   '
      KEY(25,1)='   '
      KEY(25,2)='   '
C
C
C
C
C     Set default file names.
      FILMPS='mps'
      FILLOG='log'
      FILSOL='sol'
C
C     Set default MPS file parameters.
      NAMEC(1:8)='        '
      NAMEB(1:8)='        '
      NAMRAN(1:8)='        '
      NAMBND(1:8)='        '
      MULT=1.0D0
C
C     Set default bounds.
      BIG=1.0D+30
      DLOBND=0.0D0
      DUPBND=BIG
C
C     Set optimality tolerance.
      OPTTOL=1.0D-8
C
C     Set presolve level.
      LEVPRS=6
C
C     Open the specifications file.
      IOSPC=15
      OPEN(IOSPC,FILE=FILSPC,STATUS='OLD',ERR=9000)
C
C
C
C
C     Read the first line of the specs file.
      LINE=0
  100 LINE=LINE+1
      READ(IOSPC,101,ERR=9010) KEYWRD,TEXT
  101 FORMAT(A3,9X,A12)
      IF(KEYWRD.EQ.KEY(19,1).OR.KEYWRD.EQ.KEY(19,2)) THEN
         GO TO 200
      ELSE
         WRITE(BUFFER,102) KEYWRD,LINE
  102    FORMAT(1X,'RDSPEC: Unexpected keyword ',A3,' at line',I4)
         CALL MYWRT(0,BUFFER)
         GO TO 100
      ENDIF
C
C
C
C
C     Main loop begins here.
C     ----------------------
  200 LINE=LINE+1
      READ(IOSPC,201,ERR=9010) KEYWRD,TEXT
  201 FORMAT(A3,9X,A12)
C
C     Check if it is the end of the specs file.
      IF(KEYWRD.EQ.KEY(20,1).OR.KEYWRD.EQ.KEY(20,2)) GO TO 3000
C
C     Determine the type of the specification read.
      IKEY=0
      DO 300 I=1,25
         IF(KEYWRD.EQ.KEY(I,1).OR.KEYWRD.EQ.KEY(I,2)) THEN
            IKEY=I
            GO TO 400
         ENDIF
  300 CONTINUE
C
  400 IF(IKEY.EQ.0) THEN
         WRITE(BUFFER,401) KEYWRD,LINE
  401    FORMAT(1X,'RDSPEC: Keyword ',A3,' (line',I4,') is ignored.')
         CALL MYWRT(0,BUFFER)
         GO TO 200
      ENDIF
C
C     Here if the keyword is identified.
C
      IF(IKEY.GE.5) GO TO 1050
      IF(IKEY.EQ.1) THEN
         READ(TEXT,1010,ERR=9020) MAXM
 1010    FORMAT(I12)
         GO TO 2000
      ENDIF
C
      IF(IKEY.EQ.2) THEN
         READ(TEXT,1020,ERR=9020) MAXN
 1020    FORMAT(I12)
         GO TO 2000
      ENDIF
C
      IF(IKEY.EQ.3.OR.IKEY.EQ.4) THEN
         READ(TEXT,1030,ERR=9020) MAXNZA
 1030    FORMAT(I12)
         GO TO 2000
      ENDIF
C
 1050 IF(IKEY.GE.10) GO TO 1095
      IF(IKEY.EQ.5) THEN
         MULT=-1.0
         GO TO 2000
      ENDIF
C
      IF(IKEY.EQ.6) THEN
         MULT=+1.0
         GO TO 2000
      ENDIF
C
      IF(IKEY.EQ.7) THEN
         READ(TEXT,1070,ERR=9020) DLOBND
 1070    FORMAT(D12.0)
         GO TO 2000
      ENDIF
C
      IF(IKEY.EQ.8) THEN
         READ(TEXT,1080,ERR=9020) DUPBND
 1080    FORMAT(D12.0)
         GO TO 2000
      ENDIF
C
      IF(IKEY.EQ.9) THEN
         READ(TEXT,1090,ERR=9020) NAMEC
 1090    FORMAT(A8)
         GO TO 2000
      ENDIF
C
 1095 IF(IKEY.EQ.10) THEN
         READ(TEXT,1100,ERR=9020) NAMEB
 1100    FORMAT(A8)
         GO TO 2000
      ENDIF
C
      IF(IKEY.EQ.11) THEN
         READ(TEXT,1110,ERR=9020) NAMRAN
 1110    FORMAT(A8)
         GO TO 2000
      ENDIF
C
      IF(IKEY.EQ.12) THEN
         READ(TEXT,1120,ERR=9020) NAMBND
 1120    FORMAT(A8)
         GO TO 2000
      ENDIF
C
      IF(IKEY.EQ.13) THEN
         READ(TEXT,1130,ERR=9020) FILMPS
 1130    FORMAT(A12)
         GO TO 2000
      ENDIF
C
      IF(IKEY.EQ.14) THEN
         READ(TEXT,1140,ERR=9020) FILSOL
 1140    FORMAT(A12)
         GO TO 2000
      ENDIF
C
      IF(IKEY.EQ.15) THEN
         READ(TEXT,1150,ERR=9020) FILLOG
 1150    FORMAT(A12)
         GO TO 2000
      ENDIF
C
      IF(IKEY.EQ.18) THEN
         READ(TEXT,1180,ERR=9020) OPTTOL
 1180    FORMAT(D10.0)
         GO TO 2000
      ENDIF
C
      IF(IKEY.EQ.21) THEN
         READ(TEXT,1210,ERR=9020) LEVPRS
 1210    FORMAT(I8)
         GO TO 2000
      ENDIF
C
      IF(IKEY.EQ.22) THEN
         READ(TEXT,1220,ERR=9020) FILLOG
 1220    FORMAT(A12)
         GO TO 2000
      ENDIF
C
C
C
C
C
C     End of main loop.
C     -----------------
 2000 CONTINUE
      GO TO 200
C
C
C
C
C     Check if there are no error settings in the specifications.
 3000 CONTINUE
C
C
C     Close the specifications file.
      CLOSE(IOSPC)
      RETURN
C
C
C
C
C
 9000 WRITE(BUFFER,9001) FILSPC
 9001 FORMAT(1X,'RDSPEC: Cannot open file = ',A13,
     X ', default settings are used!')
C     CALL ERRWRT(99,BUFFER)
      CALL MYWRT(99,BUFFER)
      RETURN
C
C     Here when error occurs.
 9010 WRITE(BUFFER,9011) LINE
 9011 FORMAT(1X,'RDSPEC: Error while reading line',I4,
     X ' of the SPC file.')
      CALL ERRWRT(99,BUFFER)
      CALL MYSTOP(82)
C
 9020 WRITE(BUFFER,9021) LINE
 9021 FORMAT(1X,'RDSPEC: Wrong specification at line',I4,
     X ' of the specs file.')
      CALL ERRWRT(99,BUFFER)
      CALL MYSTOP(82)
C
C
C
C
C *** LAST CARD OF (RDSPEC) ***
      END
      SUBROUTINE CHCKDM(MDIM,NDIM,NZDIM,MAXM,MAXN,MAXNZA,IOLOG)
C
C *** PARAMETERS
      INTEGER*4 MDIM,NDIM,NZDIM,MAXM,MAXN,MAXNZA,IOLOG
      CHARACTER*100 BUFFER
C
C *** PURPOSE:
C     This routine checks if dimensions do not exceed allocated
C     space.
C
C
C *** BODY OF (CHCKDM) ***
C
      IF(MAXM.GT.MDIM) THEN
         WRITE(BUFFER,3001) 
 3001    FORMAT(1X,'CHCKDM ERROR: MAXM exceeds MDIM.')
         CALL ERRWRT(IOLOG,BUFFER)
         CALL MYSTOP(81)
      ENDIF
C
      IF(MAXN.GT.NDIM) THEN
         WRITE(BUFFER,3002) 
 3002    FORMAT(1X,'CHCKDM ERROR: MAXN exceeds NDIM.')
         CALL ERRWRT(IOLOG,BUFFER)
         CALL MYSTOP(81)
      ENDIF
C
      IF(MAXNZA.GT.NZDIM) THEN
         WRITE(BUFFER,3003) 
 3003    FORMAT(1X,'CHCKDM ERROR: MAXNZA exceeds NZDIM.')
         CALL ERRWRT(IOLOG,BUFFER)
         CALL MYSTOP(81)
      ENDIF
C
C *** LAST CARD OF (CHCKDM) ***
      END
