C*************************************************************
C     **** REORDV ... REORDER THE ELEMENTS OF A VECTOR ***
C*************************************************************
C
      SUBROUTINE REORDV(MAXM,M,
     X PERM,INVP,VECTOR,DPWORK,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 MAXM,M,IOLOG
      INTEGER*2 PERM(MAXM),INVP(MAXM)
      DOUBLE PRECISION VECTOR(MAXM),DPWORK(MAXM)
C
C
C *** LOCAL VARIABLES
      INTEGER*4 I,IROW
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     M       Number of rows of the LP constraint matrix.
C     PERM    Permutation resulting from the minimum degree ordering.
C     INVP    Inverse permutation.
C     VECTOR  Array to be permuted.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     VECTOR  Array reordered according to a given permutation.
C
C     WORK ARRAYS:
C     DPWORK  Double precision work array.
C
C
C *** SUBROUTINES CALLED:
C     NONE
C
C
C *** PURPOSE:
C     This routine reorders the vector according to a given
C     permutation.
C
C *** NOTES:
C
C *** REFERENCES:
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  October 26, 1992
C     Last modified: May 17, 1995
C
C
C
C *** BODY OF (REORDV) ***
C
C
C     Reorder  VECTOR array.
      DO 100 I=1,M
         DPWORK(I)=VECTOR(I)
  100 CONTINUE
      DO 200 I=1,M
         IROW=INVP(I)
         VECTOR(IROW)=DPWORK(I)
  200 CONTINUE
C
C
      RETURN
C
C *** LAST CARD OF (REORDV) ***
      END
