C****************************************************
C     **** SATY ... (sparse)Atransp * (dense)Y ****
C****************************************************
C
      SUBROUTINE SATY(RWORK,IWORK,RMAP,IMAP,Y,M,X,N,
     X IROW,RELT,MAXN,IOLOG)
C
C *** PARAMETERS
      INTEGER*4 MAXN,M,N,IOLOG
      INTEGER*4 IROW(MAXN)
      DOUBLE PRECISION X(N),Y(M),RELT(MAXN)
C
C *** HIDDEN DATA STRUCTURES
      INTEGER*4 IWORK(*),IMAP(*),RMAP(*)
      DOUBLE PRECISION RWORK(*)
C
C *** LOCAL VARIABLES
      INTEGER*4 I,J,KNZ
C
C
C
C *** PURPOSE
C     This routine computes the product of a sparse matrix  Atransp
C     and a dense vector Y and saves the result in a dense vector X.
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     MAXN    Maximum column dimension of the LP constraint matrix.
C     M       Number of rows of the LP constraint matrix.
C     N       Number of columns of the LP constraint matrix.
C     IROW  and  RELT are the arrays for temporary handling
C             of rows/columns of the constraint matrix. They
C             are primarily intended to handle sparse vectors
C             (in packed form) but may also be used for storing
C             dense ones.
C     Y       Dense vector of dimension M.
C
C     ON OUTPUT:
C     X       Dense vector of dimension N (X = Atransp * Y).
C
C
C *** HIDDEN DATA STRUCTURES DESCRIPTION
C     RWORK   Real work array containing almost all real
C             LP problem data.
C     IWORK   Integer work array containing almost all integer
C             LP problem data.
C     RMAP    Map of RWORK.
C     IMAP    Map of IWORK.
C
C
C
C *** SUBROUTINES CALLED:
C     GETROW,SAXPY
C
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C
C     and            Dominique Tachat, LAMSADE,
C                    University of Paris Dauphine,
C                    Place du Marechal de Lattre de Tassigny,
C                    75775 Paris Cedex 16, France.
C
C     Date written:  May 4, 1992
C     Last modified: May 17, 1995
C
C
C
C
C *** BODY OF (SATY) ***
C
      DO 100 J=1,N
         X(J)=0.0
  100 CONTINUE
      DO 200 I=1,M
         CALL GETROW(I,RWORK,IWORK,RMAP,IMAP,
     X    IROW,RELT,KNZ,MAXN,IOLOG)
         IF(KNZ.EQ.0) GO TO 200
         CALL SAXPY(IROW,RELT,KNZ,X,Y(I))
  200 CONTINUE
      RETURN
C
C *** LAST CARD OF (SATY) ***
      END
