C**********************************************
C     **** SAX ... (sparse)A * (dense)X ****
C**********************************************
C
      SUBROUTINE SAX(RWORK,IWORK,RMAP,IMAP,STAVAR,X,N,Y,M,
     X IROW,RELT,MAXN,IOLOG)
C
C *** PARAMETERS
      INTEGER*4 MAXN,M,N,IOLOG
      INTEGER*2 STAVAR(MAXN)
      INTEGER*4 IROW(MAXN)
      DOUBLE PRECISION X(N),Y(M),RELT(MAXN)
C
C *** HIDDEN DATA STRUCTURES
      INTEGER*4 IWORK(*),IMAP(*),RMAP(*)
      DOUBLE PRECISION RWORK(*)
C
C *** LOCAL VARIABLES
      INTEGER*4 I,J,KNZ
C
C
C
C *** PURPOSE
C     This routine computes the product of a sparse matrix  A and
C     a dense vector  X and saves the result in a dense vector  Y.
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     MAXN    Maximum column dimension of the LP constraint matrix.
C     M       Number of rows of the LP constraint matrix.
C     N       Number of columns of the LP constraint matrix.
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C             7  PRESUMED OPTIMAL variable i.e.: x = x0;
C            -k  free variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicate the position of the original variable.
C     IROW  and  RELT are the arrays for temporary handling
C             of rows/columns of the constraint matrix. They
C             are primarily intended to handle sparse vectors
C             (in packed form) but may also be used for storing
C             dense ones.
C     X       Dense vector of dimension N.
C
C     ON OUTPUT:
C     Y       Dense vector of dimension M (Y = A * X).
C
C
C *** HIDDEN DATA STRUCTURES DESCRIPTION
C     RWORK   Real work array containing almost all real
C             LP problem data.
C     IWORK   Integer work array containing almost all integer
C             LP problem data.
C     RMAP    Map of RWORK.
C     IMAP    Map of IWORK.
C
C
C
C *** SUBROUTINES CALLED:
C     GETCOL,SAXPY
C
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C
C     and            Dominique Tachat, LAMSADE,
C                    University of Paris Dauphine,
C                    Place du Marechal de Lattre de Tassigny,
C                    75775 Paris Cedex 16, France.
C
C     Date written:  May 5, 1992
C     Last modified: May 17, 1995
C
C
C
C
C *** BODY OF (SAX) ***
C
      DO 100 I=1,M
         Y(I)=0.
  100 CONTINUE
      DO 200 J=1,N
         IF(STAVAR(J).GE.6) GO TO 200
         CALL GETCOL(J,RWORK,IWORK,RMAP,IMAP,
     X    IROW,RELT,KNZ,MAXN,IOLOG)
         IF(KNZ.EQ.0) GO TO 200
         CALL SAXPY(IROW,RELT,KNZ,Y,X(J))
  200 CONTINUE
      RETURN
C
C *** LAST CARD OF (SAX) ***
      END
