C******************************************************************
C     **** SAXPY ... (dense)Y = ALPHA * (sparse)X + (dense)Y ****
C******************************************************************
C
      SUBROUTINE SAXPY(IROW,RELT,KNZ,Y,ALPHA)
C
C *** PARAMETERS
      INTEGER*4 KNZ,IROW(KNZ)
      DOUBLE PRECISION Y(*),RELT(KNZ),ALPHA
C
C *** LOCAL VARIABLES
      INTEGER*4 I,IKX
C
C *** PURPOSE
C     This routine computes the following sum:
C     (dense)Y = ALPHA * (sparse)X + (dense)Y
C     Sparse vector  X is packed in  IROW and  RELT arrays.
C
C
C *** PARAMETERS DESCRIPTION
C     ON INPUT:
C     Y       Dense vector.
C     IROW    Row numbers of nonzeros in a sparse vector  X.
C     RELT    Nonzero entries a sparse vector  X.
C     KNZ     Number of nonzero entries in vector  X.
C     ALPHA   Scalar used to multiply sparse vector  X.
C     ON OUTPUT:
C     Y       Dense vector (result of the addition).
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C
C     and            Dominique Tachat, LAMSADE,
C                    University of Paris Dauphine,
C                    Place du Marechal de Lattre de Tassigny,
C                    75775 Paris Cedex 16, France.
C
C     Last modified: March 21, 1992
C
C
C
C
C *** BODY OF (SAXPY) ***
C
      DO 100 I=1,KNZ
         IKX=IROW(I)
         Y(IKX)=Y(IKX)+ALPHA*RELT(I)
  100 CONTINUE
      RETURN
C
C *** LAST CARD OF (SAXPY) ***
      END
