C***************************************************
C     *** SCALEA ... SCALE LP CONSTRAINT MATRIX  ***
C***************************************************
C
      SUBROUTINE SCALEA(IOLOG,
     X MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X RMTMP1,RMTMP2,RNTMP1,RNTMP2,
     X B,RANGES,C,UPBND,CSCALE,RSCALE,OSCALE,
     X ACOEFF,CLPNTS,RWNMBS,
     X LENCOL,STAVAR)
C
C
C
C *** PARAMETERS
      INTEGER*4 IOLOG,MAXM,MAXN,MAXNZA,M,N,NSTRCT
      DOUBLE PRECISION RMTMP1(MAXM),RMTMP2(MAXM)
      DOUBLE PRECISION RNTMP1(MAXN),RNTMP2(MAXN)
      DOUBLE PRECISION C(MAXN),UPBND(MAXN)
      DOUBLE PRECISION CSCALE(MAXN),RSCALE(MAXM),OSCALE
      DOUBLE PRECISION ACOEFF(MAXNZA),B(MAXM),RANGES(MAXM)
      INTEGER*4 CLPNTS(MAXN+1)
      INTEGER*2 RWNMBS(MAXNZA),LENCOL(MAXN),STAVAR(MAXN)
C
C
C
C *** LOCAL VARIABLES
      INTEGER*4 I,IPASS,J,K,KBEG,KEND
      DOUBLE PRECISION CSCL,RSCL,DP,ELTMIN,ELTMAX,OBJMAX,OBJMIN
      CHARACTER*100 BUFFER
C
C
C
C
C *** PARAMETERS DESCRIPTION
C
C *** ON INPUT:
C     IOLOG   Output unit number where log messages are to be written.
C     MAXM    Maximum number of constraints.
C     MAXN    Maximum number of variables.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     M       Number of constraints.
C     N       Number of variables (total, i.e. including slacks,
C             surplus and artificials).
C     NSTRCT  Number of structural variables (excluding slacks, surplus
C             and artificials).
C     ACOEFF  Array of nonzero elements for each column.
C     B       Right hand side of the linear program.
C     RANGES  Array of constraint ranges.
C     C       Objective function coefficients.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A.
C     LENCOL  Lengths of (sparse) columns of matrix A.
C     UPBND   Array of upper bounds. Note that the upper bound
C             is changed when the variable has a lower bound. 
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C            -k  FREE variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicates the position of the original variable.
C *** ON OUTPUT:
C     CSCALE  Column scaling factors.
C     RSCALE  Row scaling factors.
C     OSCALE  Objective row scaling factor.
C
C
C
C *** WORK ARRAYS:
C     RMTMP1  Double precision work array of size MAXM.
C     RMTMP2  Double precision work array of size MAXM.
C     RNTMP1  Double precision work array of size MAXN.
C     RNTMP2  Double precision work array of size MAXN.
C
C
C
C
C *** PURPOSE
C     This routine scales the LP constraint matrix.
C
C *** SUBROUTINES CALLED
C     MYWRT,DABS,DSQRT
C
C *** NOTES
C
C *** REFERENCES:
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Last modified: May 17, 1995
C
C
C
C
C *** BODY OF (SCALEA) ***
C
C
C
C
C     Find the largest and the smallest elements in columns of A.
      DO 300 J=1,N
         RNTMP1(J)=0.0D0
         RNTMP2(J)=1.0D+10
         IF(DABS(C(J)).GE.1.0D-8) THEN
            RNTMP1(J)=DABS(C(J))
            RNTMP2(J)=DABS(C(J))
         ENDIF
         IF(STAVAR(J).GE.6.OR.LENCOL(J).EQ.0) THEN
            RNTMP1(J)=1.0D0
            RNTMP2(J)=1.0D0
            GO TO 300
         ENDIF
         KBEG=CLPNTS(J)
         KEND=KBEG+LENCOL(J)-1
         DO 200 K=KBEG,KEND
            DP=DABS(ACOEFF(K))
            IF(DP.GT.RNTMP1(J)) RNTMP1(J)=DP
            IF(DP.LT.RNTMP2(J)) RNTMP2(J)=DP
  200    CONTINUE
C        WRITE(IOLOG,201) J,RNTMP1(J),RNTMP2(J)
C 201    FORMAT(1X,'col=',I5,' RNTMP1=',D10.3,' RNTMP2=',D10.3)
  300 CONTINUE
C
C
C
C
C
C     Main loop begins here.
      DO 1000 IPASS=1,2
C
C
C     Find the largest and the smallest element of A.
      ELTMAX=0.0D0
      ELTMIN=1.0D+10
      DO 320 J=1,N
         IF(STAVAR(J).GE.6) GO TO 320
         IF(RNTMP1(J).GT.ELTMAX) ELTMAX=RNTMP1(J)
         IF(RNTMP2(J).LT.ELTMIN) ELTMIN=RNTMP2(J)
  320 CONTINUE
      DP=ELTMAX/ELTMIN
      WRITE(BUFFER,321) IPASS-1,ELTMAX,ELTMIN,DP
  321 FORMAT(1X,'SCALEA: PASS=',I2,'   Amax=',1PD8.1,
     X '   Amin=',1PD8.1,'   Amax/Amin=',1PD8.1)
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
C
C
C     Scale columns of the LP constraint matrix.
C     WRITE(BUFFER,401)
C 401 FORMAT(1X,'SCALEA: Scaling columns of  A.')
C     CALL MYWRT(0,BUFFER)
C     CALL MYWRT(IOLOG,BUFFER)
C
C     Divide column  j of A by CSCALE(j).
C     Find the largest and the smallest elements in rows of A.
C     Find the largest and the smallest elements in the objective row.
      DO 400 I=1,M
         RMTMP1(I)=0.0D0
         RMTMP2(I)=1.0D+10
  400 CONTINUE
      OBJMAX=0.0D0
      OBJMIN=1.0D+10
      DO 460 J=1,N
         IF(STAVAR(J).GE.6) GO TO 460
         DP=DSQRT(RNTMP1(J)*RNTMP2(J))
C        WRITE(IOLOG,402) J,DP
C 402    FORMAT(1X,'SCALEA: column= ',I5,'   CSCALE=',D10.3)
         KBEG=CLPNTS(J)
         KEND=KBEG+LENCOL(J)-1
         DO 440 K=KBEG,KEND
            ACOEFF(K)=ACOEFF(K)/DP
            I=RWNMBS(K)
            RSCL=DABS(ACOEFF(K))
            IF(RSCL.GT.RMTMP1(I)) RMTMP1(I)=RSCL
            IF(RSCL.LT.RMTMP2(I)) RMTMP2(I)=RSCL
  440    CONTINUE
         C(J)=C(J)/DP
         IF(DABS(C(J)).GE.1.0D-8) THEN
            IF(DABS(C(J)).GT.OBJMAX) OBJMAX=DABS(C(J))
            IF(DABS(C(J)).LT.OBJMIN) OBJMIN=DABS(C(J))
         ENDIF
         UPBND(J)=UPBND(J)*DP
         CSCALE(J)=CSCALE(J)*DP
  460 CONTINUE
C
C
C     Scale rows of the LP constraint matrix.
C     WRITE(BUFFER,601)
C 601 FORMAT(1X,'SCALEA: Scaling rows of  A.')
C     CALL MYWRT(0,BUFFER)
C     CALL MYWRT(IOLOG,BUFFER)
      DO 600 I=1,M
         RMTMP1(I)=DSQRT(RMTMP1(I)*RMTMP2(I))
         IF(DABS(RANGES(I)).LE.1.0D+18) RMTMP1(I)=1.0D0
C        WRITE(IOLOG,602) I,RMTMP1(I)
C 602    FORMAT(1X,'SCALEA: row= ',I5,'   RSCALE=',D10.3)
  600 CONTINUE
C
C     Divide row  i of A by RSCALE(i) (omit slack coefficients).
C     Find the largest and the smallest elements in columns of A.
C     Divide objective row by DP=DSQRT(OBJMAX*OBJMIN).
      IF(OBJMAX.GE.1.0D-10) THEN
         DP=DSQRT(OBJMAX*OBJMIN)/1.0D+1
      ELSE
         DP=1.0D0
      ENDIF
      IF(DP.GE.1.0D+2) DP=1.0D+2
      IF(PASS.GE.2) DP=1.0D0
C     WRITE(BUFFER,603) OBJMAX,OBJMIN,DP
C 603 FORMAT(1X,'SCALEA: Omx=',1PD8.1,' Omn=',1PD8.1,' Oscl=',1PD8.1)
C     CALL MYWRT(IOLOG,BUFFER)
      OSCALE=OSCALE*DP
      DO 660 J=1,NSTRCT
         RNTMP1(J)=0.0D0
         RNTMP2(J)=1.0D+10
C        IF(DABS(C(J)).GE.1.0D-8) THEN
C           RNTMP1(J)=DABS(C(J))
C           RNTMP2(J)=DABS(C(J))
C        ENDIF
         IF(STAVAR(J).GE.6.OR.LENCOL(J).EQ.0) THEN
            RNTMP1(J)=1.0D0
            RNTMP2(J)=1.0D0
            GO TO 650
         ENDIF
         KBEG=CLPNTS(J)
         KEND=KBEG+LENCOL(J)-1
         DO 640 K=KBEG,KEND
            I=RWNMBS(K)
            ACOEFF(K)=ACOEFF(K)/RMTMP1(I)
            CSCL=DABS(ACOEFF(K))
            IF(CSCL.GT.RNTMP1(J)) RNTMP1(J)=CSCL
            IF(CSCL.LT.RNTMP2(J)) RNTMP2(J)=CSCL
  640    CONTINUE
  650    C(J)=C(J)/DP
  660 CONTINUE
      DO 680 I=1,M
         B(I)=B(I)/RMTMP1(I)
         RANGES(I)=RANGES(I)/RMTMP1(I)
         RSCALE(I)=RSCALE(I)*RMTMP1(I)
  680 CONTINUE
C
C     Do not scale slack coefficients.
      DO 700 J=NSTRCT+1,N
         RNTMP1(J)=1.0D0
         RNTMP2(J)=1.0D0
         C(J)=C(J)/DP
  700 CONTINUE
C
C
C
C
C     End of main loop.
 1000 CONTINUE
C
C
C
C
C     Find the largest and the smallest element of A.
      ELTMAX=0.0D0
      ELTMIN=1.0D+10
      DO 1100 J=1,N
         IF(STAVAR(J).GE.6) GO TO 1100
         IF(RNTMP1(J).GT.ELTMAX) ELTMAX=RNTMP1(J)
         IF(RNTMP2(J).LT.ELTMIN) ELTMIN=RNTMP2(J)
 1100 CONTINUE
      DP=ELTMAX/ELTMIN
      WRITE(BUFFER,1101) IPASS-1,ELTMAX,ELTMIN,DP
 1101 FORMAT(1X,'SCALEA: PASS=',I2,'   Amax=',1PD8.1,
     X '   Amin=',1PD8.1,'   Amax/Amin=',1PD8.1)
      CALL MYWRT(0,BUFFER)
      CALL MYWRT(IOLOG,BUFFER)
C
C
      RETURN
C
C *** LAST CARD OF (SCALEA) ***
      END
