C**************************************************************
C     ***  SMPLX ... A (ONE ROW) LOOK-AHEAD SIMPLEX METHOD  ***
C**************************************************************
C
      SUBROUTINE SMPLX(IOLOG,MSGLEV,NMAX,N,NSTRCT,ROWST,
     X COEFF,X,C,UPPER,RDCOST,RHS,P,Q,DUAL)
C
C *** PARAMETERS
      INTEGER*4 IOLOG,MSGLEV,NMAX,N,NSTRCT,ROWST
      DOUBLE PRECISION COEFF(NMAX),X(NMAX),C(NMAX),UPPER(NMAX)
      DOUBLE PRECISION RDCOST(NMAX),RHS,P,Q,DUAL
C
C *** LOCAL VARIABLES
      INTEGER*4 IBASIC,ITER,ISTEP,J,JBEST
      DOUBLE PRECISION BIGNEW,FSBTOL,OPTTOL,SMALLA
      DOUBLE PRECISION DP,DBEST,STEP,STEPB,SBEST,OBJ
      CHARACTER*100 BUFFER
C
C
C *** PARAMETERS DESCRIPTION
C *** ON INPUT:
C     IOLOG   Output unit number where log messages are to be written.
C     MSGLEV  The level of PRE_SOLVE information desired:
C             0  only final problem dimensions are printed;
C             1  numbers of eliminated rows and columns are printed;
C             2  names of eliminated rows and columns are printed;
C             3  detailed DEBUGGING information is printed.
C     NMAX    Maximum number of columns.
C     N       Number of variables (including logicals).
C     NSTRCT  Number of structural variables.
C     ROWST   Status of the row:
C             1  'EQ' row;
C             2  'GE' row;
C             3  'LE' row.
C     COEFF   LP constraint coefficients.
C     X       Primal variables.
C     C       Objective function coefficients.
C     UPPER   Variables' upper bounds.
C     RDCOST  Reduced costs.
C     RHS     Right hand side.
C     P       LOWER bound on shadow price.
C     Q       UPPER bound on shadow price.
C     DUAL    Dual variable.
C *** ON OUTPUT:
C     X       Primal variables.
C     RDCOST  Reduced costs.
C     DUAL    Dual variable.
C
C
C *** WORK ARRAYS:
C
C
C *** PURPOSE
C     This routine solves a one-row linear program.
C     It uses the simplex method.
C
C *** SUBROUTINES CALLED
C     MYWRT,MYSTOP,DABS
C
C *** NOTES
C
C
C *** REFERENCES:
C     Gondzio J. (1994). Presolve analysis of linear programs prior 
C        to applying an interior point method, Technical Report 
C        No 1994.3, Department of Management Studies, University 
C        of Geneva, 102, Bd. Carl-Vogt, 1211 Geneva, Switzerland, 
C        February 1994, revised in December 1994.
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  December 3, 1993
C     Last modified: June 9, 1995
C
C
C
C *** BODY OF (SMPLX) ***
C
C
C
C     Initialize.
      BIGNEW=1.0D+20
      FSBTOL=1.0D-8
      OPTTOL=1.0D-8
      SMALLA=1.0D-8
C
C     Initialize all primal variables on zero.
C     Compute the 'big number', a penalty for artificial variable.
      DBEST=0.0D0
      DO 100 J=1,NSTRCT
         X(J)=0.0D0
         IF(DABS(C(J)).GT.DBEST) DBEST=DABS(C(J))
C        WRITE(BUFFER,101) J,COEFF(J),C(J),UPPER(J)
C 101    FORMAT(1X,'cl=',I3,' coeff=',D14.6,' c=',D14.6,' Uj=',D14.6)
C        CALL MYWRT(IOLOG,BUFFER)
  100 CONTINUE
      DBEST=1.0D+3*(DBEST+1.0D0)
C
C
C
C     Add logical variables and construct initial feasible solution.
      N=NSTRCT
      IF(ROWST.EQ.1) THEN
         IF(DABS(RHS).LE.SMALLA) THEN
            IBASIC=1
            GO TO 200
         ENDIF
         N=NSTRCT+1
         C(N)=DBEST
         UPPER(N)=2*BIGNEW
         IF(RHS.GE.0.0D0) THEN
            COEFF(N)=1.0D0
            X(N)=RHS
         ELSE
            COEFF(N)=-1.0D0
            X(N)=-RHS
         ENDIF
         IBASIC=N
         GO TO 200
      ENDIF
C
      IF(ROWST.EQ.2) THEN
         N=NSTRCT+1
         C(N)=0.0D0
         UPPER(N)=2*BIGNEW
         COEFF(N)=-1.0D0
         IF(RHS.GE.0.0D0) THEN
            X(N)=0.0D0
            N=N+1
            C(N)=DBEST
            UPPER(N)=2*BIGNEW
            COEFF(N)=1.0D0
            X(N)=RHS
         ELSE
            X(N)=-RHS
         ENDIF
         IBASIC=N
         GO TO 200
      ENDIF
C
      IF(ROWST.EQ.3) THEN
         N=NSTRCT+1
         C(N)=0.0D0
         UPPER(N)=2*BIGNEW
         COEFF(N)=1.0D0
         IF(RHS.GE.0.0D0) THEN
            X(N)=RHS
         ELSE
            X(N)=0.0D0
            N=N+1
            C(N)=DBEST
            UPPER(N)=2*BIGNEW
            COEFF(N)=-1.0D0
            X(N)=-RHS
         ENDIF
         IBASIC=N
         GO TO 200
      ENDIF
C
C
C
C
C
C     Here when feasible solution found.
C     N is the number of LP variables (including logicals).
C     Do the feasibility check.
  200 DP=RHS
      DO 300 J=1,N
         DP=DP-COEFF(J)*X(J)
  300 CONTINUE
      IF(DABS(DP).LE.FSBTOL) GO TO 400
      WRITE(BUFFER,301) DP
  301 FORMAT(1X,'SMPLX:  Initial solution is infeasible, DP=',D10.3)
      CALL MYWRT(IOLOG,BUFFER)
      CALL MYSTOP(5)
C
C
C
C
C
C
C     Main loop begins here.
C     Simplex iterations.
C     IBASIC indicates basic variable.
  400 ITER=0
 1000 ITER=ITER+1
         IF(ITER.GE.2*N) THEN
            WRITE(BUFFER,1001)
 1001       FORMAT(1X,'SMPLX:  Excess iterations limit.')
            CALL ERRWRT(IOLOG,BUFFER)
            CALL MYSTOP(96)
         ENDIF
C
C
C     Compute dual variable.
         DUAL=C(IBASIC)/COEFF(IBASIC)
C
C
C     Price all nonbasic variables.
C     Compute the current objective.
         OBJ=0.0D0
         IF(MSGLEV.LE.2) GO TO 1003
         WRITE(BUFFER,1002) ITER,IBASIC
 1002    FORMAT(1X,'SMPLX:  iter=',I6,' ibasic=',I6)
         CALL MYWRT(IOLOG,BUFFER)
 1003    CONTINUE
         DO 1100 J=1,N
            RDCOST(J)=C(J)-DUAL*COEFF(J)
            OBJ=OBJ+C(J)*X(J)
            IF(MSGLEV.LE.2) GO TO 1012
            WRITE(BUFFER,1011) J,RDCOST(J),X(J)
 1011       FORMAT(1X,'LB: cl=',I6,' rc=',D16.8,
     X       ' X=',D12.5)
            CALL MYWRT(IOLOG,BUFFER)
 1012    CONTINUE
 1100    CONTINUE
         IF(MSGLEV.LE.1) GO TO 1022
         WRITE(BUFFER,1121) ITER,OBJ
 1121    FORMAT(1X,'SMPLX:  iter=',I6,' obj=',D10.3)
         CALL MYWRT(IOLOG,BUFFER)
 1022    CONTINUE
C
C
C     Choose a variable to enter the basis.
C     Look for a zero variable with the most negative reduced
C     cost or a variable blocked on its UPPER bound with the
C     most positive reduced cost.
C     Additionally, compute the stepsize and the predicted
C     improvement of the objective function (do a look-ahead).
C     Pick up the variable that gives maximum decrease of the
C     objective.
         DBEST=0.0D0
         SBEST=0.0D0
         JBEST=0
         DO 1500 J=1,N
            IF(DABS(RDCOST(J)).LE.OPTTOL) GO TO 1500
C
C
            IF(RDCOST(J).LE.0.0D0.AND.X(J).LE.FSBTOL) THEN
C
C
C     Here when the reduced cost is negitive
C     and a variable is on its zero LOWER bound.
C     Compute the stepsize.
               STEP=1.0D+16
               DP=-COEFF(J)/COEFF(IBASIC)
               IF(DP.GE.0) THEN
                  STEPB=(UPPER(IBASIC)-X(IBASIC))/DP
               ELSE
                  STEPB=-X(IBASIC)/DP
               ENDIF
               IF(STEPB.LT.STEP) STEP=STEPB
               IF(UPPER(J)-X(J).LT.STEP) STEP=UPPER(J)-X(J)
               IF(STEP.GE.1.0D+10) THEN
                  WRITE(BUFFER,1201)
 1201             FORMAT(1X,'SMPLX:  Problem is unbounded.')
                  CALL MYWRT(IOLOG,BUFFER)
                  CALL MYSTOP(97)
               ENDIF
C
C     Compute the objective improvement (look ahead).
               DP=RDCOST(J)*STEP
               IF(MSGLEV.LE.2) GO TO 1203
               WRITE(BUFFER,1202) J,RDCOST(J),STEP,DP
 1202          FORMAT(1X,'LB: cl=',I6,' rc=',D16.8,
     X          ' step=',D12.5,' impr=',D16.8)
               CALL MYWRT(IOLOG,BUFFER)
 1203          CONTINUE
               IF(DP.LT.DBEST) THEN
                  DBEST=DP
                  SBEST=STEP
                  JBEST=J
                  ISTEP=+1
               ENDIF
            ENDIF
C
C
            IF(RDCOST(J).GE.0.0D0.AND.UPPER(J)-X(J).LE.FSBTOL) THEN
C
C
C     Here when the reduced cost is positive
C     and a variable is on its UPPER bound.
C     Compute the stepsize.
               STEP=1.0D+16
               DP=COEFF(J)/COEFF(IBASIC)
               IF(DP.GE.0) THEN
                  STEPB=(UPPER(IBASIC)-X(IBASIC))/DP
               ELSE
                  STEPB=-X(IBASIC)/DP
               ENDIF
               IF(STEPB.LT.STEP) STEP=STEPB
               IF(X(J).LT.STEP) STEP=X(J)
               IF(STEP.GE.1.0D+10) THEN
                  WRITE(BUFFER,1301)
 1301             FORMAT(1X,'SMPLX:  Problem is unbounded.')
                  CALL MYWRT(IOLOG,BUFFER)
                  CALL MYSTOP(97)
               ENDIF
C
C     Compute the objective improvement (look ahead).
               DP=-RDCOST(J)*STEP
               IF(MSGLEV.LE.2) GO TO 1303
               WRITE(BUFFER,1302) J,RDCOST(J),STEP,DP
 1302          FORMAT(1X,'UB: cl=',I6,' rc=',D16.8,
     X          ' step=',D12.5,' impr=',D16.8)
               CALL MYWRT(IOLOG,BUFFER)
 1303          CONTINUE
               IF(DP.LT.DBEST) THEN
                  DBEST=DP
                  SBEST=STEP
                  JBEST=J
                  ISTEP=-1
               ENDIF
            ENDIF
C

 1500    CONTINUE
         IF(MSGLEV.LE.2) GO TO 1502
         WRITE(BUFFER,1501) ISTEP,JBEST,RDCOST(JBEST),DBEST
 1501    FORMAT(1X,'Istp=',I2,' Jbst=',I6,' rc=',D16.8,' impr=',D16.8)
         CALL MYWRT(IOLOG,BUFFER)
 1502    CONTINUE
C
C
C     Here to make step.
         IF(DBEST.GE.-OPTTOL) GO TO 2100
         IF(ISTEP.EQ.-1) SBEST=-SBEST
         X(IBASIC)=X(IBASIC)-SBEST*COEFF(JBEST)/COEFF(IBASIC)
         X(JBEST)=X(JBEST)+SBEST
         IF(DABS(X(JBEST)-UPPER(JBEST)).GE.SMALLA.AND.
     X    DABS(X(JBEST)).GE.SMALLA) THEN
C
C     Basis change.
            IBASIC=JBEST
         ENDIF
C
C
C
C
C
C
C     End of main loop.
 2000 GO TO 1000
 2100 CONTINUE
C
      IF(MSGLEV.LE.1) GO TO 2102
      WRITE(BUFFER,2101) ITER-1
 2101 FORMAT(1X,'SMPLX:  Optimum found after',I4,' iteration(s).')
      CALL MYWRT(IOLOG,BUFFER)
 2102 CONTINUE
C
C
      RETURN
C
C
C *** LAST CARD OF (SMPLX) ***
      END
