C*****************************************************************
C     *** SOLAAT ... SOLVE EQUATION WITH  A*THETA*Atransp ***
C*****************************************************************
C
      SUBROUTINE SOLAAT(LCOEFF,LCLPTS,LRWNBS,LDIAG,
     X MAXNZL,MAXM,M,COLUMN,RHS,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 MAXNZL,MAXM,M,IOLOG
      DOUBLE PRECISION LCOEFF(*),LDIAG(MAXM)
      DOUBLE PRECISION COLUMN(MAXM),RHS(MAXM)
      INTEGER*4 LCLPTS(MAXM+1)
C
C *** The following array can be half-length integer.
      INTEGER*2 LRWNBS(MAXNZL)
C
C
C *** LOCAL VARIABLES
      INTEGER*4 IROW
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     MAXNZL  Maximum number of nonzeros of the Cholesky factor.
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     M       Number of rows of the LP constraint matrix
C             (and the dimension of  A*Atransp).
C     LCOEFF  Off-diagonal nonzero coefficients of Cholesky matrix.
C     LCLPTS  Pointers to columns of the Cholesky factor.
C     LRWNBS  Row numbers of nonzeros in columns of matrix  L.
C     LDIAG   Diagonal elements of Cholesky factor.
C     RHS     Right-hand-side of the equation.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     COLUMN  Solution of the equation  (A*THETA*Atransp) * X = RHS.
C
C     WORK ARRAYS:
C     NONE
C
C
C *** SUBROUTINES CALLED:
C     SOLVL,SOLVLT
C
C
C *** PURPOSE:
C     This routine solves the equation with  A*THETA*Atransp.
C     It uses the Cholesky decomposition  L*D*Ltransp
C     of the above matrix.
C
C
C *** NOTES:
C     1. RHS array is destroyed by this routine.
C     2. The solution is devided into three phases:
C        (i)       L * X1 = RHS           (SOLVL  routine)
C        (ii)      D * X2 = X1
C        (iii)     Ltransp * COLUMN = X2  (SOLVLT routine)
C
C
C *** REFERENCES:
C     Duff I.S., Erisman A.M., Reid J.K. (1989). Direct methods
C        for sparse matrices, Clarendon Press, Oxford 1989,
C        chapter  6.
C     Gondzio J. (1993). Implementing Cholesky factorization
C        for interior point methods of linear programming,
C        Optimization 27, pp. 121-140.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  April 15, 1991
C     Last modified: May 17, 1995
C
C
C
C *** BODY OF (SOLAAT) ***
C
C     Solve the equation  L * X1 = RHS
C     (save  X1 in  COLUMN array).
      CALL SOLVL(LCOEFF,LCLPTS,LRWNBS,
     X MAXNZL,MAXM,M,COLUMN,RHS,IOLOG)
C
C     Solve the equation  D * X2 = X1
C     (save  X2 in  RHS array).
      DO 300 IROW=1,M
         RHS(IROW)=COLUMN(IROW)/LDIAG(IROW)
  300 CONTINUE
C
C     Solve the equation  Ltransp * COLUMN = X2
C     (COLUMN array contains the solution).
      CALL SOLVLT(LCOEFF,LCLPTS,LRWNBS,
     X MAXNZL,MAXM,M,COLUMN,RHS,IOLOG)
C
      RETURN
C
C *** LAST CARD OF (SOLAAT) ***
      END
