C************************************************************
C     ****  SPLIT ... SPLITTING DENSE COLUMNS OF  A   ****
C************************************************************
C
      SUBROUTINE SPLIT(MAXM,MAXN,MAXNZA,M,N,NSTRCT,
     X LNHIST,MXHIST,INHIST,DPHIST,
     X ACOEFF,CLPNTS,RWNMBS,RWHEAD,RWLINK,CLNMBS,LENCOL,
     X SPLCOL,ROWLEN,CSPLIT,LENCLA,LENCLB,
     X MAXCOL,SVIROW,SVRELT,IROW,RELT,
     X P,Q,CLNAME,STAVAR,PRLVAR,COBJ,UPBND,LOBND,
     X RWNAME,STAROW,RWSTAT,RANGES,RHS,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 MAXM,MAXN,MAXNZA,M,N,NSTRCT
      INTEGER*4 LNHIST,MXHIST,MAXCOL,IOLOG
      INTEGER*2 INHIST(MXHIST)
      DOUBLE PRECISION DPHIST(MXHIST)
      INTEGER*4 CLPNTS(MAXN+1),RWHEAD(MAXM+1),RWLINK(MAXNZA)
      INTEGER*4 SVIROW(MAXM),IROW(MAXM),LENCLA(MAXN),LENCLB(MAXN)
      CHARACTER*8 RWNAME(MAXM),CLNAME(MAXN)
      DOUBLE PRECISION ACOEFF(MAXNZA),SVRELT(MAXM),RELT(MAXM)
      DOUBLE PRECISION P(MAXM),Q(MAXM),RANGES(MAXM),RHS(MAXM)
      DOUBLE PRECISION PRLVAR(MAXN),COBJ(MAXN),UPBND(MAXN),LOBND(MAXN)
C
C *** The following arrays can be half-length integer.
      INTEGER*2 RWNMBS(MAXNZA),CLNMBS(MAXNZA),LENCOL(MAXN)
      INTEGER*2 STAVAR(MAXN),STAROW(MAXM),RWSTAT(MAXM)
      INTEGER*2 CSPLIT(MAXM),SPLCOL(MAXM),ROWLEN(MAXM)
C
C
C *** LOCAL VARIABLES
      INTEGER*4 I,IDUMMY,IR,IRW,INEWRW,IX,ICOLS,IND,IPARTS
      INTEGER*4 K,KBEG,KEND,KNEW,KBEGNW,NELTS,NLEFT
      INTEGER*4 IRWADD,KRWADD,KADDED,ISPLIT,JSPLIT,JCOLMN
      INTEGER*4 LNCOL,COLLEN,LNMIN,LNMAX,MAXLEN
      INTEGER*4 J,JLAST,JCOL,JPOS,JCOLNW,ISIGMA,BSTSIG,IPENAL
      INTEGER*4 BESTCL,BESTLN,NRSPLT,SVCLLN,MOVE,MVCOL
      INTEGER*4 MNEW,NNEW,NSTNEW,NZNEW
      DOUBLE PRECISION BIG
      CHARACTER*8 TEXT
      CHARACTER*100 BUFFER
C
C
C *** COMMON ARREAS
C     Markers for linking rows.
      SAVE   /ICGRAD/
      COMMON /ICGRAD/ MSPLIT(100000)
      INTEGER*2       MSPLIT
C
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     MAXN    Maximum column dimension of the LP constraint matrix.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     M       Number of rows of the LP constraint matrix
C             (and the dimension of  A*Atransp).
C     N       Number of columns of the LP constraint matrix.
C     NSTRCT  Number of structural variables (excluding slacks, surplus
C             and artificials).
C     LNHIST  Length of the PRE_SOLVE history list;
C     MXHIST  Maximum number of entries in the PRE_SOLVE history list.
C     INHIST  Integer PRE_SOLVE history information.
C     DPHIST  Double precision PRE_SOLVE history information.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A.
C     RWHEAD  Headers to the row linked lists of matrix A.
C     RWLINK  Row linked lists of matrix A.
C     CLNMBS  Column numbers of nonzeros in rows of matrix A.
C     LENCOL  Lengths of columns of matrix A.
C     MAXCOL  Threshold length for columns to be split.
C     P       LOWER bounds on shadow prices (dual variables).
C     Q       UPPER bounds on shadow prices (dual variables).
C     CLNAME  Array of column names (unordered).
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C            -k  FREE variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicates the position of the original variable.
C     UPBND   Array of upper bounds. Note that the upper bound
C             is changed when the variable has a lower bound. 
C     LOBND   Array of lower bounds.
C     COBJ    Array of cost coefficients (objective function).
C     RWNAME  Array of row names (increasing order sort).
C     STAROW  Array of row status:
C             0  row has been removed (it indicates a free row);
C             1  row has not been removed.
C     RWSTAT  Array of row types (sort as before):
C             1  row type is = ;
C             2  row type is >= ;
C             3  row type is <= ;
C             4  row type is objective or free.
C     RANGES  Array of constraint ranges.
C     RHS     LP right-hand-side.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     New  LP constraint matrix with all long columns split.
C     NSTRCT  Number of structural variables (excluding slacks, surplus
C             and artificials).
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A.
C     RWHEAD  Headers to the row linked lists of matrix A.
C     RWLINK  Row linked lists of matrix A.
C     CLNMBS  Column numbers of nonzeros in rows of matrix A.
C
C     WORK ARRAYS:
C     CSPLIT  Work array for handling the list of columns to split.
C     LENCLA  Work array for temporary handling the length of columns.
C     LENCLB  Work array for temporary handling the length of columns.
C     SVIROW  Work array used to save the contents of  IROW array.
C     SVRELT  Work array used to save the contents of  RELT array.
C     SPLCOL  Work array for splitting mechanism.
C     ROWLEN  Handles the number of nonzero etries of a given row
C             that appear in long columns.
C     IROW    Integer work array.
C     RELT    Double precision work array.
C
C
C *** SUBROUTINES CALLED:
C     MYWRT,MYSTOP
C
C
C *** PURPOSE:
C     This routine splits all long columns of the  LP constraint
C     matrix into the shorter ones. Long columns are those which
C     have at least  MAXCOL nonzero entries.
C
C
C *** NOTES:
C     1. This routine depends on data structures for handling  A.
C     2. It should not be called if the LP problem contains FREE
C        variables or if singleton FREE columns are eliminated.
C
C
C *** REFERENCES:
C     Gondzio J. (1992). Splitting dense columns of the constraint
C        matrix in interior point methods for large scale linear
C        programming, Optimization 24, pp. 285-297.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  October 12, 1990
C     Last modified: June 9, 1995
C
C
C
C *** BODY OF (SPLIT) ***
C
C
C
      MNEW=M
      NNEW=N
      NZNEW=CLPNTS(N+1)-1
      KRWADD=0
      BIG=1.0D+30
C
C
C     Initialize markers for linking rows.
      DO 100 IR=1,MAXM
         MSPLIT(IR)=0
  100 CONTINUE
C
C
C
C
C
C     Scan the whole matrix  A and count the columns to be split.
C     Split only columns with the number of nonzero entries that
C     considerably exceeds MAXCOL.
C     IRWADD is the number of rows that will be bordered to  A.
C     NRSPLT is the number of columns to be split.
      IRWADD=0
      NRSPLT=0
      MAXLEN=MAXCOL+MAXCOL/5
      DO 150 J=1,N
C
C     Omit fixed columns.
C     Check if column  J is to be split.
         IF(STAVAR(J).GE.6) GO TO 150
         IF(LENCOL(J).LE.MAXLEN) GO TO 150
C
C     Save the column as the one to be split.
         NRSPLT=NRSPLT+1
         CSPLIT(NRSPLT)=J
C        WRITE(BUFFER,151) J,CLNAME(J),LENCOL(J)
C 151    FORMAT(1x,'column ',i5,2x,a8,' (len=',i4,') will be split.')
C        CALL MYWRT(IOLOG,BUFFER)
         INEWRW=(LENCOL(J)-1)/MAXCOL
         IRWADD=IRWADD+INEWRW
  150 CONTINUE
C
C
C
C     Check if there are any columns to be split.
      MNEW=M+IRWADD
      NNEW=N+IRWADD
      NSTNEW=NSTRCT+IRWADD
      NZNEW=CLPNTS(N+1)-1+2*IRWADD
      IF(NRSPLT.EQ.0) GO TO 2000
C
C
C
C
C     Scan the whole matrix  A and check if there are any FREE
C     variables in the LP problem. If so, then disable splitting.
      DO 160 J=1,N
         IF(STAVAR(J).LT.0) THEN
            NRSPLT=0
            WRITE(BUFFER,161)
  161       FORMAT(1X,'SPLIT:  Splitting disabled,',
     X       ' FREE variables present in the LP problem.')
            CALL MYWRT(IOLOG,BUFFER)
            MNEW=M
            NNEW=N
            NSTNEW=NSTRCT
            NZNEW=CLPNTS(N+1)-1
            GO TO 2000
         ENDIF
  160 CONTINUE
C
C
C
C
C     Scan PRE_SOLVE history list and check if there were any
C     singleton FREE variables eliminated. Disable splitting, if so.
      DO 180 IX=1,LNHIST
         IF(INHIST(IX).GT.0) THEN
            NRSPLT=0
            WRITE(BUFFER,181)
  181       FORMAT(1X,'SPLIT:  Splitting disabled,',
     X       ' singleton FREE variables were eliminated.')
            CALL MYWRT(IOLOG,BUFFER)
            MNEW=M
            NNEW=N
            NSTNEW=NSTRCT
            NZNEW=CLPNTS(N+1)-1
            GO TO 2000
         ENDIF
  180 CONTINUE
C
C
C
C
C
C
C
C
C
C
C     Check if there is enough space to perform splitting.
      IF(MNEW.GT.MAXM) GO TO 9000
      IF(NNEW.GT.MAXN) GO TO 9000
      IF(NZNEW.GT.MAXNZA) GO TO 9000
C
C
C
      WRITE(BUFFER,201)
  201 FORMAT(1X,'SPLIT:  Splitting starts.')
      CALL MYWRT(0,BUFFER)
C
C
C
C     Initialize.
      DO 200 IR=1,MAXM
         RWHEAD(IR)=0
         SPLCOL(IR)=0
         ROWLEN(IR)=0
  200 CONTINUE
C
C
C
C     Scan the whole matrix  A and expand data structures
C     in such a way that empty cells are left for elements
C     added in linking rows that are bordered to  A.
C     Row linked lists are created for short columns.
C     ROWLEN(IRW) handles the number of nonzero entries
C     of row  IRW that appear in long columns.
C     MVCOL  determines how far a given column must be moved.
C     MOVE   determines how far its elements must be moved.
C     KRWADD is the numer of rows (and columns) added.
      MVCOL=IRWADD
      MOVE=2*IRWADD
      KRWADD=0
      JLAST=N
      CLPNTS(N+MVCOL+1)=CLPNTS(N+1)+MOVE
      DO 300 ISPLIT=NRSPLT,1,-1
         JSPLIT=CSPLIT(ISPLIT)
C
C     Move all columns with indices JSPLIT+1,JSPLIT+2,...,JLAST
C     to their new positions. Save their new positions in LENCLA.
         DO 250 JCOL=JLAST,JSPLIT+1,-1
            KBEG=CLPNTS(JCOL)
            KEND=KBEG+LENCOL(JCOL)-1
            KBEGNW=KBEG+MOVE
            JCOLNW=JCOL+MVCOL
            LENCLA(JCOL)=JCOLNW
            DO 220 K=KEND,KBEG,-1
               KNEW=K+MOVE
               ACOEFF(KNEW)=ACOEFF(K)
               RWNMBS(KNEW)=RWNMBS(K)
  220       CONTINUE
C
C     Update the row linked lists of short columns.
C     Omit the fixed columns.
            IF(STAVAR(JCOL).GE.6) GO TO 240
            DO 230 K=KEND,KBEG,-1
               KNEW=K+MOVE
               CLNMBS(KNEW)=JCOLNW
               IRW=RWNMBS(KNEW)
               RWLINK(KNEW)=RWHEAD(IRW)
               RWHEAD(IRW)=KNEW
  230       CONTINUE
  240       CLPNTS(JCOLNW)=KBEGNW
            CLNAME(JCOLNW)=CLNAME(JCOL)
            LENCOL(JCOLNW)=LENCOL(JCOL)
            COBJ(JCOLNW)=COBJ(JCOL)
            STAVAR(JCOLNW)=STAVAR(JCOL)
            PRLVAR(JCOLNW)=PRLVAR(JCOL)
            UPBND(JCOLNW)=UPBND(JCOL)
            LOBND(JCOLNW)=LOBND(JCOL)
  250    CONTINUE
C
C     Add new rows and new columns to a constraint matrix.
         COLLEN=LENCOL(JSPLIT)
         INEWRW=(COLLEN-1)/MAXCOL
C
C     Define new row and column names.
         DO 260 I=1,INEWRW
            KADDED=KRWADD+I
            WRITE(TEXT,251) KADDED
  251       FORMAT(I8)
            TEXT(1:5)='SPLRW'
            IRW=M+KADDED
            RWNAME(IRW)=TEXT
            MSPLIT(IRW)=1
            P(IRW)=-BIG
            Q(IRW)=BIG
            RHS(IRW)=0.0
            RANGES(IRW)=0.0
            RWSTAT(IRW)=1
            STAROW(IRW)=1
            TEXT(4:5)='CL'
            JCOL=JCOLNW-I
            CLNAME(JCOL)=TEXT
            LENCOL(JCOL)=0
            COBJ(JCOL)=0.0
            STAVAR(JCOL)=STAVAR(JSPLIT)
            PRLVAR(JCOL)=PRLVAR(JSPLIT)
            UPBND(JCOL)=UPBND(JSPLIT)
            LOBND(JCOL)=LOBND(JSPLIT)
  260    CONTINUE
C
C     Move the column to be split to its new position.
C     Leave empty cells for nonzero elements that will
C     later be added by the splitting mechanism.
C     Save its new position in LENCLA.
         MVCOL=MVCOL-INEWRW
         MOVE=MOVE-2*INEWRW
         KBEG=CLPNTS(JSPLIT)
         KEND=KBEG+LENCOL(JSPLIT)-1
         KBEGNW=KBEG+MOVE
         JCOLNW=JSPLIT+MVCOL
         LENCLA(JSPLIT)=JCOLNW
         DO 280 K=KEND,KBEG,-1
            KNEW=K+MOVE
            ACOEFF(KNEW)=ACOEFF(K)
            RWNMBS(KNEW)=RWNMBS(K)
C
C     Update the  ROWLEN array. ROWLEN(IRW) indicates the number 
C     of nonzero etries of row  IRW that appear in long columns.
            IRW=RWNMBS(KNEW)
            ROWLEN(IRW)=ROWLEN(IRW)+1
  280    CONTINUE
         CLPNTS(JCOLNW)=KBEGNW
         CLNAME(JCOLNW)=CLNAME(JSPLIT)
         LENCOL(JCOLNW)=LENCOL(JSPLIT)
         COBJ(JCOLNW)=COBJ(JSPLIT)
         STAVAR(JCOLNW)=STAVAR(JSPLIT)
         PRLVAR(JCOLNW)=PRLVAR(JSPLIT)
         UPBND(JCOLNW)=UPBND(JSPLIT)
         LOBND(JCOLNW)=LOBND(JSPLIT)
C
C     Update the number of added rows.
C     Save the new index of column to be split.
         KRWADD=KRWADD+INEWRW
         CSPLIT(ISPLIT)=JCOLNW
         JLAST=JSPLIT-1
C
  300 CONTINUE
C
C
C
C     Update the row linked lists of short columns
C     (add all columns with indices  1,2,...,JLAST).
C     Save their new positions in LENCLA.
C     Omit the fixed columns.
      DO 350 JCOL=JLAST,1,-1
         LENCLA(JCOL)=JCOL
         IF(STAVAR(JCOL).GE.6) GO TO 350
         KBEG=CLPNTS(JCOL)
         KEND=KBEG+LENCOL(JCOL)-1
         DO 330 K=KEND,KBEG,-1
            CLNMBS(K)=JCOL
            IRW=RWNMBS(K)
            RWLINK(K)=RWHEAD(IRW)
            RWHEAD(IRW)=K
  330    CONTINUE
  350 CONTINUE
C
C
C
C     Update the PRE_SOLVE history list (column numbers have changed).
      DO 380 IX=1,LNHIST
         J=-INHIST(IX)
         INHIST(IX)=-LENCLA(J)
C        WRITE(BUFFER,381) IX,J,LENCLA(J)
C 381    FORMAT(1X,'ientry=',I6,' col=',I6,' becomes col=',I6)
C        CALL MYWRT(IOLOG,BUFFER)
  380 CONTINUE
C
C
C
C
C
C
C
C     Here to perform splitting.
C     Initialize IRWADD again.
      IRWADD=0
C
C
C     Select the shortest column from those to be split.
  400 BESTCL=0
      BESTLN=32000
      IND=0
      DO 420 I=1,NRSPLT
         JCOL=CSPLIT(I)
         IF(JCOL.LE.0) GO TO 420
         COLLEN=LENCOL(JCOL)
         IF(COLLEN.GE.BESTLN) GO TO 420
         BESTCL=JCOL
         BESTLN=COLLEN
         IND=I
  420 CONTINUE
      IF(BESTCL.EQ.0) GO TO 2000
C
C
C     Get the column to be split from the data structures.
      JCOLMN=BESTCL
      CSPLIT(IND)=0
C
C
C
C     Logic for splitting.
C     First, save the long column.
C     Set up LENCLB array.
      DO 440 I=1,NNEW
         LENCLB(I)=0
  440 CONTINUE
      COLLEN=LENCOL(JCOLMN)
      KBEG=CLPNTS(JCOLMN)-1
      DO 480 I=1,COLLEN
         K=KBEG+I
         IRW=RWNMBS(K)
         SVIROW(I)=IRW
         SVRELT(I)=ACOEFF(K)
         SPLCOL(IRW)=I
         JPOS=RWHEAD(IRW)
  460    IF(JPOS.EQ.0) GO TO 480
         JCOL=CLNMBS(JPOS)
         LENCLB(JCOL)=LENCLB(JCOL)+1
         JPOS=RWLINK(JPOS)
         GO TO 460
  480 CONTINUE
      SVCLLN=COLLEN
      INEWRW=(COLLEN-1)/MAXCOL
      MAXLEN=COLLEN/(INEWRW+1)
C
C *** DEBUGGING
C     WRITE(BUFFER,461) JCOLMN,LENCOL(JCOLMN),MAXLEN
C 461 FORMAT(1X,'col=',I6,' ln=',I6,' is split, MAXLEN=',I6)
C     CALL MYWRT(IOLOG,BUFFER)
C
      MAXLEN=MAXLEN+MAXLEN/3
      IF(MAXLEN.GE.MAXCOL+MAXCOL/3) MAXLEN=MAXCOL+MAXCOL/3
C
C
C *** DEBUGGING
C     WRITE(BUFFER,481) CLNAME(JCOLMN),JCOLMN,COLLEN
C 481 FORMAT(1X,'var= ',A8,' (col=',I4,', len=',I4,') is split.')
C     CALL MYWRT(IOLOG,BUFFER)
C     DO 483 I=1,COLLEN
C        WRITE(BUFFER,482) I,SVIROW(I),SVRELT(I)
C 482    FORMAT(1X,'I=',I5,'  SVIROW=',I5,'  SVRELT=',E10.4)
C        CALL MYWRT(IOLOG,BUFFER)
C 483 CONTINUE
C
C
C     Add new (split) columns to the data structures.
      ICOLS=0
C
C
C
C
C
C     Main loop begins here.
      DO 1000 IDUMMY=1,INEWRW+1
         BESTCL=0
         BESTLN=0
         NELTS=0
         LNMIN=(INEWRW-IDUMMY)*MAXCOL
         LNMAX=LNMIN+MAXCOL
         IPARTS=INEWRW+2-IDUMMY
C
C
C     Check how many nonzero elements in the long columns left.
C     If their number is less than MAXCOL, then the sparsity
C     pattern analysis is unnecessary.
         LNCOL=0
         DO 520 IND=1,SVCLLN
            IRW=SVIROW(IND)
            IF(SPLCOL(IRW).LE.0) GO TO 520
            LNCOL=LNCOL+1
            RELT(LNCOL)=SVRELT(IND)
            IROW(LNCOL)=SVIROW(IND)
  520    CONTINUE
         COLLEN=LNCOL
         NELTS=LNCOL
         IF(LNCOL.LE.MAXCOL) GO TO 580
C
C     Restore zero value of LNCOL.
         LNCOL=0
C
C
C     Look for the column with maximum entries in rows
C     in which the column to be split has also nonzero elements.
C     BESTCL  is the index of the best column found by now;
C     BESTLN  is the number of common elements.
         BESTCL=0
         BESTLN=0
         DO 540 JCOL=1,NNEW
            IF(LENCLB(JCOL).LE.BESTLN) GO TO 540
            BESTCL=JCOL
            BESTLN=LENCLB(JCOL)
  540    CONTINUE
C
C     Check if there are still elements left in the long column.
         IF(NELTS.EQ.0) GO TO 1100
         IF(BESTCL.GT.0) GO TO 560
         LNCOL=0
         GO TO 700
C
C     Extract the partitioned column from the long one
C     and set up LENCLA array.
  560    JCOL=BESTCL
         COLLEN=LENCOL(JCOL)
         KBEG=CLPNTS(JCOL)-1
         DO 570 I=1,COLLEN
            K=KBEG+I
            IROW(I)=RWNMBS(K)
            RELT(I)=ACOEFF(K)
  570    CONTINUE
  580    LNCOL=0
         DO 600 I=1,NNEW
            LENCLA(I)=0
  600    CONTINUE
         DO 640 IX=1,COLLEN
            IR=IROW(IX)
            JCOL=SPLCOL(IR)
            IF(JCOL.GT.0) THEN
               LNCOL=LNCOL+1
               RELT(LNCOL)=SVRELT(JCOL)
               IROW(LNCOL)=IR
               SPLCOL(IR)=-JCOL
               JPOS=RWHEAD(IR)
  620          IF(JPOS.EQ.0) GO TO 640
               JCOL=CLNMBS(JPOS)
               LENCLA(JCOL)=LENCLA(JCOL)+1
               LENCLB(JCOL)=LENCLB(JCOL)-1
               JPOS=RWLINK(JPOS)
               GO TO 620
            ENDIF
  640    CONTINUE
C
C
C     Analyse whether it is profitable to add more elements
C     to the column. An element is then looked for that can
C     be moved from the set of unsplit elements to the set
C     of already split elements with the mininmum penalty.
C     To break ties, a number of nonzero entries that are
C     in the given row and in the columns to be split is
C     counted and its largest possible value is selected.
C     BSTSIG  is the penalty of the best row found by now;
C     BESTLN  is the length of the best row found by now;
C     IRW     is the index of the best row found by now.
  700    BSTSIG=1000000
         BESTLN=0
         IRW=0
         DO 800 IX=1,SVCLLN
            IR=SVIROW(IX)
            JCOL=SPLCOL(IR)
            IF(JCOL.LT.0) GO TO 800
            ISIGMA=0
            JPOS=RWHEAD(IR)
  720       IF(JPOS.EQ.0) GO TO 740
            JCOL=CLNMBS(JPOS)
            ISIGMA=ISIGMA+LENCLB(JCOL)-LENCLA(JCOL)-1
            JPOS=RWLINK(JPOS)
            GO TO 720
  740       IF(ISIGMA-BSTSIG) 780,760,800
  760       IF(ROWLEN(IR).LE.BESTLN) GO TO 800
  780       IRW=IR
            BSTSIG=ISIGMA
            BESTLN=ROWLEN(IR)
  800    CONTINUE
         IF(IRW.EQ.0) GO TO 860
C
C     If the number of elements left is too large, then (no metter
C     how large the penalty is) the element has to be added.
         NLEFT=NELTS-LNCOL
         IF(NLEFT.GT.LNMAX) GO TO 820
C
C     If the number of elements that would rest after
C     the addition of the one selected now is too small,
C     then the element can not be added.
         IF(NLEFT.LE.LNMIN+1) GO TO 860
C
C     Do not let the number of elements in a column 
C     to considerably exceed MAXCOL.
         IF(LNCOL.GE.MAXLEN) GO TO 860
C
C
C     Now check if the addition of the element selected
C     improves total penalty indicator. If the change of
C     penalty is positive (which means it is better 
C     not to add the element), then end up the column.
         IPENAL=(IPARTS*LNCOL-NELTS)/(IPARTS-1)+BSTSIG
         IF(IPENAL.GT.0) GO TO 860
C
C
C *** DEBUGGING
C        WRITE(BUFFER,821) LNCOL+1,BSTSIG,IPENAL
C 821    FORMAT(1X,'Col. augmentation (elt=',I5,'), penalty=',I6,
C    X    ',  tot. penalty=',I5)
C        CALL MYWRT(IOLOG,BUFFER)
C
C
  820    LNCOL=LNCOL+1
         JCOL=SPLCOL(IRW)
         RELT(LNCOL)=SVRELT(JCOL)
         IROW(LNCOL)=SVIROW(JCOL)
         SPLCOL(IRW)=-JCOL
C
C     Update LENCLA and LENCLB arrays.
         JPOS=RWHEAD(IRW)
  840    IF(JPOS.EQ.0) GO TO 700
         JCOL=CLNMBS(JPOS)
         LENCLA(JCOL)=LENCLA(JCOL)+1
         LENCLB(JCOL)=LENCLB(JCOL)-1
         JPOS=RWLINK(JPOS)
         GO TO 840
C
C     Augment the column and add it to the data structure.
  860    COLLEN=LNCOL
         ICOLS=ICOLS+1
         IF(ICOLS.EQ.1) THEN
            COLLEN=COLLEN+1
            IROW(COLLEN)=M+IRWADD+ICOLS
            RELT(COLLEN)=+1.0
         ELSE
            IF(ICOLS.LE.INEWRW) THEN
               COLLEN=COLLEN+2
               IROW(COLLEN-1)=M+IRWADD+ICOLS-1
               RELT(COLLEN-1)=-1.0
               IROW(COLLEN)=M+IRWADD+ICOLS
               RELT(COLLEN)=+1.0
            ELSE
               COLLEN=COLLEN+1
               IROW(COLLEN)=M+IRWADD+ICOLS-1
               RELT(COLLEN)=-1.0
            ENDIF
         ENDIF
C
C
C     Add the new column to the data structures.
         KBEG=CLPNTS(JCOLMN)
         DO 880 I=1,COLLEN
            K=KBEG+I-1
            ACOEFF(K)=RELT(I)
            IRW=IROW(I)
            RWNMBS(K)=IRW
            CLNMBS(K)=JCOLMN
            RWLINK(K)=RWHEAD(IRW)
            RWHEAD(IRW)=K
            ROWLEN(IRW)=ROWLEN(IRW)-1
  880    CONTINUE
         IF(ICOLS.LE.INEWRW) CLPNTS(JCOLMN+1)=KBEG+COLLEN
         LENCOL(JCOLMN)=COLLEN
         JCOLMN=JCOLMN+1
C
C
C
C
C
         ISIGMA=(NELTS-LNCOL-1)/MAXCOL
C
C
C *** DEBUGGING
C     The following two lines have to be uncommented to help 
C     the 'xlf' compiler of IBM POWER PC computer to produce
C     correct code for a -O option.
         WRITE(BUFFER,881) JCOLMN,COLLEN
  881    FORMAT(1X,'SPLIT: new variable=',I6,' of lenght=',I6)
C        CALL MYWRT(IOLOG,BUFFER)
C        DO 883 I=1,COLLEN
C           WRITE(BUFFER,882) I,IROW(I),RELT(I)
C 882       FORMAT(1X,'I=',I6,'  IROW=',I6,'  RELT=',E10.4)
C           CALL MYWRT(IOLOG,BUFFER)
C 883    CONTINUE
C
C
C
C
C     End of main loop.
 1000 CONTINUE
C
C
C
C
C     Restore zero value of SPLCOL array.
 1100 DO 1120 I=1,SVCLLN
         IRW=SVIROW(I)
         SPLCOL(IRW)=0
 1120 CONTINUE
C
C
C     Update the number of added rows.
      IRWADD=IRWADD+INEWRW
      GO TO 400
C
C
C
C
C     Restore the number of structural variables.
 2000 IF(NRSPLT.EQ.0) THEN 
C
         WRITE(BUFFER,2001)
 2001    FORMAT(1X,'SPLIT:  There are no columns to split.')
         CALL MYWRT(0,BUFFER)
C
      ELSE
C
         WRITE(BUFFER,2002)
 2002    FORMAT(1X,'SPLIT:  Splitting completed.')
         CALL MYWRT(0,BUFFER)
C
C     Set the row linked lists (slack should be
C     the first element of the list).
         LNCOL=0
         DO 2200 I=1,MNEW
            RWHEAD(I)=0
 2200    CONTINUE
         DO 2300 J=1,NNEW
C
C     Omit  FIXED variables.
            IF(STAVAR(J).GE.6) GO TO 2300
            LNCOL=LNCOL+LENCOL(J)
            KBEG=CLPNTS(J)
            KEND=KBEG+LENCOL(J)-1
            DO 2250 K=KBEG,KEND
               I=RWNMBS(K)
               RWLINK(K)=RWHEAD(I)
               CLNMBS(K)=J
               RWHEAD(I)=K
 2250       CONTINUE
 2300    CONTINUE
      ENDIF
C
C
C
C *** DEBUGGING
C     IF(NRSPLT.EQ.0) GO TO 3000
C     DO 2420 I=1,MAXM
C       ROWLEN(I)=0
C2420 CONTINUE
C     DO 2442 IRW=1,MNEW
C        WRITE(BUFFER,2441) IRW,RWHEAD(IRW)
C2441    FORMAT(1X,'row=',I6,'  rwhead=',I6)
C        CALL MYWRT(IOLOG,BUFFER)
C2442 CONTINUE
C     DO 2460 JCOL=1,NNEW
C        IF(STAVAR(JCOL).GE.6) GO TO 2460
C        IF(LENCOL(JCOL).LE.MAXCOL) GO TO 2460
C        LNCOL=LENCOL(JCOL)
C        KBEG=CLPNTS(JCOL)
C        KEND=KBEG+LENCOL(JCOL)-1
C        WRITE(BUFFER,2461) JCOL,LNCOL
C2461    FORMAT(1X,' SPLIT: column=',I5,'  has length=',I5)
C        CALL MYWRT(IOLOG,BUFFER)
C        DO 2463 K=KBEG,KEND
C           WRITE(BUFFER,2462) K,ACOEFF(K),RWNMBS(K),
C    X      RWLINK(K),CLNMBS(K)
C2462       FORMAT(1X,' K=',I6,'   elt=',D8.2,'   row=',I6,
C    X      '   link=',I6,'   col=',I6)
C           CALL MYWRT(IOLOG,BUFFER)
C2463    CONTINUE
C        IF(LNCOL.GT.0) ROWLEN(LNCOL)=ROWLEN(LNCOL)+1
C2460 CONTINUE
C     MAXLEN=MNEW
C     DO 2480 I=MNEW+1,1,-1
C        IF(ROWLEN(I).EQ.0) GO TO 2480
C        GO TO 2500
C2480 CONTINUE
C2500 MAXLEN=I
C     WRITE(IOLOG,2501) MAXLEN
C2501 FORMAT(1X,'SPLIT:  The longest column has ',I6,' elts.'/
C    X       1X,'        Profile of the new matrix  A:')
C     WRITE(IOLOG,2502) (ROWLEN(I),I=1,MAXLEN)
C2502 FORMAT(9X,10I6)
C     ISIGMA=0
C     DO 2521 I=1,MAXLEN
C       ISIGMA=ISIGMA+ROWLEN(I)*I
C2521 CONTINUE
C     WRITE(IOLOG,2522) ISIGMA
C2522 FORMAT(/1X,'SPLIT:  Nonzeros of matrix  A:       ',I9)
C
C
C     Take account of added linking rows.
 3000 M=MNEW
      N=NNEW
      NSTRCT=NSTNEW
C
C
C
C     Write the  MPS statistics.
      IF(NRSPLT.GT.0) THEN
         WRITE(BUFFER,3001) M
 3001    FORMAT(1X,'SPLIT:  Constraints in the mps file: ',I9)
         CALL MYWRT(IOLOG,BUFFER)
         WRITE(BUFFER,3002) KRWADD
 3002    FORMAT(1X,'        Splitting constraints added: ',I9)
         CALL MYWRT(IOLOG,BUFFER)
         WRITE(BUFFER,3003) N
 3003    FORMAT(1X,'        Variables in the mps file:   ',I9)
         CALL MYWRT(IOLOG,BUFFER)
         WRITE(BUFFER,3004) LNCOL
 3004    FORMAT(1X,'        Nonzeros of matrix  A:       ',I9)
         CALL MYWRT(IOLOG,BUFFER)
      ENDIF
C
C
      RETURN
C
C
C
C     Here if an error occurs.
 9000 WRITE(BUFFER,9001)
 9001 FORMAT(1X,'SPLIT ERROR: Please increase space for new  A')
      CALL ERRWRT(IOLOG,BUFFER)
      WRITE(BUFFER,9002) MNEW,MAXM
 9002 FORMAT(14X,'there will be',I10,
     X ' constraints (current max. is',I10,')')
      CALL ERRWRT(IOLOG,BUFFER)
      WRITE(BUFFER,9003) NNEW,MAXN
 9003 FORMAT(14X,'there will be',I10,
     X ' variables   (current max. is',I10,')')
      CALL ERRWRT(IOLOG,BUFFER)
      WRITE(BUFFER,9004) NZNEW,MAXNZA
 9004 FORMAT(14X,'there will be',I10,
     X ' nonzeros    (current max. is',I10,')')
      CALL ERRWRT(IOLOG,BUFFER)
      CALL MYSTOP(8)
C
C
C
C *** LAST CARD OF (SPLIT) ***
      END
