C**********************************************************
C     ****    SYMFCT ... SYMBOLIC FACTORIZATION    ****
C**********************************************************
C
      SUBROUTINE SYMFCT(AATPAT,AATPNT,
     X LCLPTS,LRWNBS,MAXNZL,MAXM,MAXN,MAXNZA,M,
     X HEADER,LINKFD,LINKBK,MARKER,TEMP,STAVAR,
     X CLPNTS,RWNMBS,
     X RWHEAD,RWLINK,CLNMBS,LENCOL,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 MAXNZL,MAXM,MAXN,MAXNZA,M,IOLOG
      INTEGER*4 AATPAT(MAXNZL),AATPNT(MAXM+1),LCLPTS(MAXM+1)
      INTEGER*4 MARKER(MAXM),TEMP(MAXM)
      INTEGER*4 CLPNTS(MAXN+1),RWHEAD(MAXM+1),RWLINK(MAXNZA)
C
C *** The following arrays can be half-length integer.
      INTEGER*2 LRWNBS(MAXNZL),STAVAR(MAXN)
      INTEGER*2 RWNMBS(MAXNZA),CLNMBS(MAXNZA),LENCOL(MAXN)
      INTEGER*2 HEADER(MAXM),LINKFD(MAXM),LINKBK(MAXM)
C
C
C *** LOCAL VARIABLES
      INTEGER*4 I,IROW,IR,IRWACT,JCOL,K,KBEG,KEND,KX
      INTEGER*4 LENOFL,LENOK,LENROW,TRIANG,NEXT,PREVS
      DOUBLE PRECISION A0,A1,A2,DFLOPS
      CHARACTER*100 BUFFER
C
C
C *** COMMON ARREAS
C     Cholesky factorization parameters.
      COMMON /CHFCT/   CSMALL,PIVTOL,TAU,DENSE,PIVMAX,PIVMIN,IDNSRW
      DOUBLE PRECISION CSMALL,PIVTOL,TAU,DENSE,PIVMAX,PIVMIN
      INTEGER*4        IDNSRW
C
C     Additional Cholesky fact. parameters (interface to HYBRID).
      COMMON /CHHYB/   RO,FLOPS,IREG,NZCHL,RTCD
      DOUBLE PRECISION RO,FLOPS
      INTEGER*4        IREG,NZCHL,RTCD
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     MAXNZL  Maximum number of nonzeros of the Cholesky factor.
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     MAXN    Maximum column dimension of the LP constraint matrix.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     M       Number of rows of the LP constraint matrix
C             (and the dimension of  A*Atransp).
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C            -k  free variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicate the position of the original variable.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A.
C     RWHEAD  Headers to the row linked lists of matrix A.
C     RWLINK  Row linked lists of matrix A.
C     CLNMBS  Column numbers of nonzeros in rows of matrix A.
C     LENCOL  Lengths of columns of matrix A.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     LCLPTS  Pointers to columns of the Cholesky factor.
C     LRWNBS  Row numbers of nonzeros in columns of matrix  L.
C
C     Additionally, through the CHHYB common block SYMFCT returns:
C     NZCHL   Overestimate of the number of nonzeros
C             of Cholesky factor (important if RTCD = 0).
C     RTCD    Return code from the symbolic factorization:
C             0 SYMFCT failure;
C             1 SYMFCT success.
C     FLOPS   Flops required to compute the decomposition.
C
C     WORK ARRAYS:
C     AATPAT  Triangular part of the sparsity pattern of  A*Atransp
C             handled as a collection of sparse row vectors
C             (diagonal elements are excluded from the list).
C     AATPNT  Pointers to rows of  A*Atransp.
C     HEADER  Header of the doubly linked lists of rows that have
C             their first off-diagonal entries in the same columns.
C     LINKFD  Forward linked lists.
C     LINKBK  Backward linked lists.
C     MARKER  Array used to mark columns when merging is done.
C     TEMP    Temporary array used for handling pivotal clique.
C
C
C *** SUBROUTINES CALLED:
C     MYWRT,MYSTOP,DEFAAT,DTSORT
C
C
C *** PURPOSE:
C     This routine implements the symbolic factorization
C     for a symmetric positive definite matrix.
C
C
C *** NOTES:
C     1. This routine follows Duff et al. (1989) description
C        of the minimum degree ordering. It is thus strongly
C        influenced by the multifrontal approach to the Cholesky
C        decomposition.
C     2. The lower right corner of the Cholesky factor is stored
C        as a dense matrix (double addressing is thus avoided).
C        IDNSRW (from CHFACT common block) is a number of the
C        first row of a dense window.
C
C
C *** REFERENCES:
C     Duff I.S., Erisman A.M., Reid J.K. (1989). Direct methods
C        for sparse matrices, Clarendon Press, Oxford 1989,
C        chapter 10.
C     Gondzio J. (1993). Implementing Cholesky factorization
C        for interior point methods of linear programming,
C        Optimization 27, pp. 121-140.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  May 19, 1991
C     Last modified: June 9, 1995
C
C
C
C *** BODY OF (SYMFCT) ***
C
C     Set the return code (for successful run).
      RTCD=1
C
C
C     Initialize for the symbolic factorization.
C     Set up the sarsity pattern of  A*Atransp array
C     (only triangular part of  A*Atransp is necessary).
      TRIANG=1
      CALL DEFAAT(LRWNBS,AATPNT,AATPAT,
     X MAXNZL,MAXM,MAXN,MAXNZA,M,TRIANG,
     X MARKER,TEMP,STAVAR,
     X CLPNTS,RWNMBS,RWHEAD,RWLINK,CLNMBS,LENCOL,IOLOG)
C
C
C     Copy the sparsity pattern of  A*Atransp to  AATPAT array.
      DO 20 K=1,AATPNT(M+1)-1
         AATPAT(K)=LRWNBS(K)
   20 CONTINUE
C
C
C     Zero work arrays.
      DO 60 IROW=1,M
         HEADER(IROW)=0
         MARKER(IROW)=0
   60 CONTINUE
C
C     Set the doubly linked lists of rows that have
C     the first subdiagonal entry in the same columns.
C     Recall that  AATPNT(i) indicates the first 
C     off-diagonal entry of row i.
      DO 80 IROW=1,M
         KBEG=AATPNT(IROW)
         KEND=AATPNT(IROW+1)-1
         IF(KBEG.GT.KEND) GO TO 80
         JCOL=AATPAT(KBEG)
         NEXT=HEADER(JCOL)
         LINKFD(IROW)=NEXT
         HEADER(JCOL)=IROW
         IF(NEXT.GT.0) LINKBK(NEXT)=IROW
         LINKBK(IROW)=-JCOL
   80 CONTINUE
C
C
C     Set the parameters controlling the progress of building
C     the sparsity pattern of the Cholesky factor.
C     LENOFL  is a current length of the Cholesky factor.
C     FLOPS   is a cost of the numerical phase of the factorization.
C     LENROW  is the length of a given row of Cholesky factor.
      LENOFL=0
      FLOPS=0.0D0
      LENROW=0
C
C
C
C
C
C     Main loop begins here (loop over rows of Cholesky factor).
C     For every row  IROW, its sparsity pattern is merged
C     with those of all rows that have the first off-diagonal
C     entry in the pivot column. All the already merged rows
C     are removed from the linked lists. The pivot row is added
C     to the sparsity pattern of the Cholesky matrix and its
C     index is added to the linked list determined by its first
C     off-diagonal entry.
      DO 500 IROW=1,M
         LCLPTS(IROW)=LENOFL+1
C
C
C     Create the pivot row sparsity pattern.
C     Start from the sparsity pattern of row IROW of  A*Atransp.
         MARKER(IROW)=1
         KBEG=AATPNT(IROW)
         KEND=AATPNT(IROW+1)-1
         LENROW=0
         DO 100 K=KBEG,KEND
            LENROW=LENROW+1
            IR=AATPAT(K)
            TEMP(LENROW)=IR
            MARKER(IR)=1
  100    CONTINUE
C
C
C     Merge all the rows of Cholesky matrix that have
C     the first off-diagonal entry in column  IROW with
C     the pivot row sparsity pattern.
C     IRWACT is a number of row that is being merged with a pivot one.
         IRWACT=HEADER(IROW)
  150    IF(IRWACT.EQ.0) GO TO 250
         KBEG=LCLPTS(IRWACT)
         KEND=LCLPTS(IRWACT+1)-1
         DO 200 K=KBEG,KEND
            IR=LRWNBS(K)
            IF(MARKER(IR).EQ.1) GO TO 200
            LENROW=LENROW+1
            TEMP(LENROW)=IR
            MARKER(IR)=1
  200    CONTINUE
         IRWACT=LINKFD(IRWACT)
         GO TO 150
C
C
C     Here if pivot row sparsity pattern is determined.
C     Update  FLOPS.
  250 FLOPS=FLOPS+DBLE(LENROW)*DBLE(LENROW)
C
C
C     Find the number of column of its first off-diagonal
C     entry (and save it in  JCOL).
         IF(LENROW.LE.1) GO TO 350
         JCOL=M+1
         DO 300 IR=1,LENROW
            IF(TEMP(IR).GE.JCOL) GO TO 300
            IRWACT=IR
            JCOL=TEMP(IR)
  300    CONTINUE
C
C
C     Place the first off-diagonal entry at the beginning of the list.
         TEMP(IRWACT)=TEMP(1)
         TEMP(1)=JCOL
C
C
C     Copy the pivot row sparsity pattern to LRWNBS array.
  350    IF(LENROW.EQ.0) GO TO 500
         IF(LENOFL+LENROW.GT.MAXNZL) GO TO 9000
         DO 400 K=1,LENROW
            LENOFL=LENOFL+1
            IR=TEMP(K)
            LRWNBS(LENOFL)=IR
            MARKER(IR)=0
  400    CONTINUE
C
C
C *** DEBUGGING
C     WRITE(BUFFER,401) IROW,LENROW
C 401 FORMAT(1X,'SYMFCT: row ',I6,' of  L has length=',I6)
C     CALL MYWRT(IOLOG,BUFFER)
C
C
C     Remove the pivot row from the linked list of rows
C     that have the first off-diagonal entry in column  IR.
  450    IF(LENROW.LE.1) GO TO 500
         KBEG=AATPNT(IROW)
         KEND=AATPNT(IROW+1)-1
         IF(KBEG.GT.KEND) GO TO 480
         IR=AATPAT(KBEG)
         IF(IR.EQ.JCOL) GO TO 500
         NEXT=LINKFD(IROW)
         PREVS=LINKBK(IROW)
         IF(NEXT.GT.0) LINKBK(NEXT)=PREVS
         IF(PREVS.LE.0) THEN
            HEADER(IR)=NEXT
         ELSE
            LINKFD(PREVS)=NEXT
         ENDIF
C
C
C     Add the pivot row to the linked list of rows that have
C     the first off-diagonal entry in column  JCOL.
C     Rows of length  1 are not added to the list since they
C     do not influence the sparsity pattern of Cholesky matrix.
  480    IF(LENROW.LE.1) GO TO 500
         NEXT=HEADER(JCOL)
         HEADER(JCOL)=IROW
         LINKFD(IROW)=NEXT
         LINKBK(IROW)=-JCOL
         IF(NEXT.GT.0) LINKBK(NEXT)=IROW
C
C
C
C
C
C     End of main loop.
  500 CONTINUE
      LCLPTS(M+1)=LENOFL+1
C
C
C
C
C
C     Go perform a double transpose sort.
C
C     SUBROUTINE DTSORT(ROWNBS,COLPTS,
C    X ICLNBS,IRWPTS,MAXNZ,MAXM,M,IOLOG)
C
      CALL DTSORT(LRWNBS,LCLPTS,
     X AATPAT(1),TEMP,MAXNZL,MAXM,M,IOLOG)
C
C
C
C
C     Write problem statistics.
      K=AATPNT(M+1)-1
      KX=LENOFL
      A1=LENOFL*200.0
      A2=M*M-M
      IF(M.GT.1) THEN
         A1=A1/A2
      ELSE
         A1=0.0
      ENDIF
      WRITE(BUFFER,501) KX,A1
  501 FORMAT(1X,'SYMFCT: Sparse matrix L has ',I13,
     X  ' subdiagonal elts (density=',F5.1,'%).')
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,502) KX-K
  502 FORMAT(1X,'        Fill-in             ',I13)
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,503) FLOPS
  503 FORMAT(1X,'        Decomposition flops',1PD14.6)
      CALL MYWRT(IOLOG,BUFFER)
C
C     WRITE(BUFFER,504) K,KX,FLOPS
C 504 FORMAT(1X,'qqqb',I10,' &',I10,' &',1PD13.6)
C     CALL MYWRT(99,BUFFER)
C     CALL MYWRT(IOLOG,BUFFER)
C
C
C
C
C
C     Check if it is useful to make a switch to the full code
C     near the end of factorization.
C     IDNSRW indicates the first 'dense' row of Cholesky matrix
C     i.e., a row with the density at least  DENSE.
      IDNSRW=M+1
      DO 600 I=M,1,-1
         LENROW=LCLPTS(I+1)-LCLPTS(I)
         A0=(M-I)*DENSE+0.5
         LENOK=A0
         IF(LENROW.GE.LENOK) THEN
            FLOPS=FLOPS-DBLE(LENROW)*DBLE(LENROW)
            IDNSRW=I
         ELSE
            GO TO 700
         ENDIF
  600 CONTINUE
C
C
C     Check if the dense window is not too small.
  700 IF(M-IDNSRW.LE.10) THEN
         IDNSRW=M+1
         GO TO 1000
      ENDIF
      WRITE(BUFFER,701) M-IDNSRW
  701 FORMAT(1X,'SYMFCT: Dense window found  ',I13)
      CALL MYWRT(IOLOG,BUFFER)
C
C
C     Expand the lower right triangle of the sparse Cholesky
C     factor to a dense matrix.
      IROW=IDNSRW
      LENOFL=LCLPTS(IDNSRW)-1
      LENROW=M-IDNSRW
      LENROW=LENROW*(LENROW+1)/2
      IF(LENOFL+LENROW.GT.MAXNZL) GO TO 9000
      DO 900 IROW=IDNSRW,M
         LCLPTS(IROW)=LENOFL+1
         DO 800 IR=IROW+1,M
            LENOFL=LENOFL+1
            LRWNBS(LENOFL)=IR
  800    CONTINUE
  900 CONTINUE
      LCLPTS(M+1)=LENOFL+1
C
C
C
C
C     Write final problem statistics.
      DFLOPS=M-IDNSRW
      DFLOPS=2.0*DFLOPS+1.0
      DFLOPS=DBLE(LENROW)*DFLOPS/3.
      FLOPS=FLOPS+DFLOPS
      KX=LENOFL
      A1=LENOFL*200.0
      IF(M.GT.1) THEN
         A1=A1/A2
      ELSE
         A1=0.0
      ENDIF
      WRITE(BUFFER,901) KX,A1
  901 FORMAT(1X,'SYMFCT: Final matrix  L has ',I13,
     X  ' subdiagonal elts (density=',F5.1,'%).')
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,902) KX-K
  902 FORMAT(1X,'        Fill-in             ',I13)
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,903) FLOPS,DFLOPS
  903 FORMAT(1X,'        Decomposition flops',1PD14.6,
     X  ' (',1PD12.6,' in dense mode).')
      CALL MYWRT(IOLOG,BUFFER)
C
C
C
C
 1000 CONTINUE
      NZCHL=LENOFL
      RETURN
C
C
C
C
C     Here to write error message.
 9000 WRITE(BUFFER,9001) LENOFL+LENROW
 9001 FORMAT(1X,'SYMFCT ERROR: Matrix L overflow ',I10)
      CALL ERRWRT(IOLOG,BUFFER)
      WRITE(BUFFER,9002) MAXNZL
 9002 FORMAT(1X,'    space was provided for only ',I10,' nonzeros.')
      CALL ERRWRT(IOLOG,BUFFER)
C
      IF(IREG.EQ.-1) THEN
C *** Here for HOPDM: STOP the program.
         CALL MYSTOP(7)
      ENDIF
      IF(IREG.GE.0) THEN
C *** Here for HYBRID: Do not STOP the program.
         LENROW=M-IROW+1
         LENROW=LENROW*(LENROW+1)/2
         NZCHL=LENOFL+LENROW
         RTCD=0
         RETURN
      ENDIF
C
C
C
C *** LAST CARD OF (SYMFCT) ***
      END
