C**************************************************************
C     *  WRTSOL ... WRITE MPS (or NONSTANDARD) SULUTION FILE  *
C**************************************************************
C
      SUBROUTINE WRTSOL(FILSOL,IROBJ,M,M0,MFINAL,N,NSTRCT,MAXM,MAXN,
     X STAVAR,RWSTAT,STAROW,RWNAME,RWCOPY,CLNAME,NAMMPS,MULT,
     X LOBND,UPBND,B,C,PRLVAR,DLVAR,YORGNL,
     X RWORK,IWORK,RMAP,IMAP,IROW,RELT,
     X PRLACT,BNEW,IWRITE,IOLOG)
C
C
C
C *** PARAMETERS
      CHARACTER*13 FILSOL
      INTEGER*4 MAXM,MAXN,IROBJ,M,M0,MFINAL,N,NSTRCT
      INTEGER*2 STAVAR(MAXN),STAROW(MAXM),RWSTAT(MAXM)
      CHARACTER*8 RWNAME(MAXM),RWCOPY(MAXM),CLNAME(MAXN),NAMMPS
      DOUBLE PRECISION MULT,UPBND(MAXN),LOBND(MAXN)
      DOUBLE PRECISION B(MAXM),C(MAXN),PRLVAR(MAXN),DLVAR(MAXM)
      INTEGER*4 IROW(MAXN)
      DOUBLE PRECISION YORGNL(MAXM),RELT(MAXN),PRLACT(MAXM),BNEW(MAXM)
      INTEGER*4 IWRITE,IOLOG
C
C
C
C *** HIDDEN DATA STRUCTURES
      INTEGER*4 IWORK(*),IMAP(*),RMAP(*)
      DOUBLE PRECISION RWORK(*)
C
C
C
C *** LOCAL VARIABLES
      INTEGER*4 I,I1,IKX,J,K,NFREE,OUTMPS,STATUS
      DOUBLE PRECISION VRBLJ,BNDLJ,PROD,PRFSBT
      CHARACTER*100 BUFFER
C
C
C
C     An indicator if the elimination routine has been used.
      COMMON /ELMNTE/  IELIM
      INTEGER*4        IELIM
C
C
C
C *** PARAMETERS DESCRIPTION
C     FILSOL  Solution file name.
C     MAXM    Maximum number of constraints.
C     MAXN    Maximum number of variables.
C     IROBJ   Index of the objective row (original row order).
C     M       Current number of constraints.
C     M0      Original number of constraints.
C     MFINAL  Final number of constraints.
C     N       Number of variables (total, i.e. including slacks, surplus
C             and artificials).
C     NSTRCT  Number of structural variables (excluding slacks, surplus
C             and artificials).
C     RWSTAT  Array of row types:
C             1  row type is = ;
C             2  row type is >= ;
C             3  row type is <= ;
C             4  free row (may be objective for example).
C     STAROW  Array of row status:
C             0  row has been removed (it indicates a free row);
C             1  row has not been removed.
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C            -k  FREE variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicates the position of the original variable.
C     RWNAME  Array of row names.
C     RWCOPY  A copy of row names (keeps trace of the original order).
C     CLNAME  Array of column names.
C     NAMMPS  The name of the  LP problem.
C     MULT    Direction of optimization:
C             +1 means minimization;
C             -1 means maximization.
C     LOBND   Array of lower bounds.
C     UPBND   Array of upper bounds. Note that the upper bound
C             is changed when the variable has a lower bound. 
C     B       Right hand side of the linear program.
C     C       Objective function coefficients.
C     PRLVAR  Primal variables of the LP problem.
C     DLVAR   Dual variables of the LP problem.
C     YORGNL  Dual solution in original row order.
C
C     IROW and  RELT are the arrays for temporary handling of rows
C             and columns of the constraint matrix. They are primarily
C             intended to handle sparse vectors (in packed form)
C             but may also be used for storing dense ones.
C     PRLACT  Primal activity ((Ai)transp*X).
C     BNEW    New right hand side of the (modified) linear program.
C     IWRITE  Solution output parameter:
C             0  no MPS file desired;
C             1  produce MPS-like solution file.
C     IOLOG   Output unit number where log messages are to be written.
C
C
C
C *** OUTPUT FILE DESCRIPTION
C
C     ROWS SECTION DESCRIPTION:
C        column ROW: index of the row
C        column EQUATION: name of the row
C        column LOWER LIMIT: lower limit of the constraint
C        column UPPER LIMIT: upper limit of the constraint
C        column ACTIVITY: value of the constraint
C        column STATUS:
C           LL: constraint is at its lower limit
C           UL: constraint is at its upper limit
C           EQ: constraint is equal to RHS
C        column SLACK: index of the slack variable added
C               to the constraint
C        column SLK ACTIVITY: value of the slack variable
C
C     COLUMNS SECTION DESCRIPTION:
C        column COLUMN: index of the variable
C        column VARIABLE: name of the variable
C        column LOWER BOUND: lower bound of the variable
C        column UPPER BOUND: upper bound of the variable
C        column ACTIVITY: value of the variable
C
C
C
C *** PURPOSE
C     This routine writes an MPS (or nonstandard) solution file.
C
C
C
C *** SUBROUTINES CALLED
C     GETROW,GETCOL,DABS,DBLE
C
C
C
C *** NOTES
C     Since the elimination routine might have earlier been
C     called, some of the original  LP data may have been lost.
C     We will not be able to write complete MPS output in such
C     case, so a (nonstandard) output file will be produced.
C
C
C
C
C *** REFERENCES:
C     Altman A., Gondzio J. (1993). An efficient implementation of 
C        a higher order primal-dual interior point method for large 
C        sparse linear programs, Archives of Control Sciences 2, 
C        No 1-2, pp. 23-40. 
C     Altman A., Gondzio J. (1993). HOPDM - A higher order primal-
C        dual method for large scale linear programmming, European
C        Journal of Operational Research 66 (1993) pp 158-160.
C     Gondzio J., Tachat D. (1994). The design and application of 
C        IPMLO - a FORTRAN library for linear optimization with 
C        interior point methods, RAIRO Recherche Operationnelle 28, 
C        No 1, pp. 37-56.
C     Gondzio J. (1995). HOPDM (version 2.12) - A fast LP solver 
C        based on a primal-dual interior point method, European
C        Journal of Operational Research 85, pp. 221-225.
C     Murtagh B. (1981). Advanced Linear Programming, McGrew-Hill,
C        New York, 1981.
C     Murtagh B., Saunders M. (1983). MINOS 5.0 User's guide,
C        Technical Report SOL 83-20, Department of Operations
C        Research, Stanford University, Stanford, 1983.
C
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  May 29, 1991
C     Last modified: February 7, 1996
C
C
C
C
C
C
C *** BODY OF (WRTSOL) ***
C
C     Zero dual variables corresponding to eliminated rows.
      DO 50 I=M+1,M0
         DLVAR(I)=0.0D0
C        WRITE(BUFFER,51) I,RWSTAT(I),RWNAME(I)
C  51    FORMAT(1X,'eliminated rw=',I6,',  st=',I2,',  name=',A8)
C        CALL MYWRT(IOLOG,BUFFER)
   50 CONTINUE
C
C
C
C     Open the MPS output file.
      OUTMPS=13
      OPEN(OUTMPS,FILE=FILSOL,ACCESS='SEQUENTIAL')
C
C
C
C     Set up (local) feasibility tolerance.
      PRFSBT=1.0D-6
C
C
C
C
C     computing the slack activity and the LP objective
C     *************************************************
C
      PROD=0.0
      DO 100 I=1,M
         PRLACT(I)=0.0D0
         BNEW(I)=B(I)
  100 CONTINUE
      DO 120 I=M+1,MAXM
         PRLACT(I)=0.0D0
         DLVAR(I)=0.0D0
         BNEW(I)=0.0D0
  120 CONTINUE
      DO 160 J=1,NSTRCT
C
C     Restore the original status of all variables.
         STATUS=STAVAR(J)
         IF(STATUS.EQ.14) THEN
C
C     Here for slack variable eliminated in HOELIM routine.
            STAVAR(J)=0
            GO TO 160
         ENDIF
         IF(STATUS.EQ.15) THEN
C
C     Here for structural variable eliminated as a FREE one.
C     Its correct value has already been restored by POSTSL routine.
            STAVAR(J)=6
            STATUS=STAVAR(J) 
         ENDIF
         IF(STATUS.GE.7) THEN
C
C     Here for structural variable eliminated in HOELIM routine.
            STAVAR(J)=STAVAR(J)-7
         ENDIF
C
         VRBLJ=PRLVAR(J)
         BNDLJ=0.0D0
         IF(STAVAR(J).EQ.2.OR.STAVAR(J).EQ.3.OR.STAVAR(J).EQ.6) THEN
            VRBLJ=LOBND(J)+PRLVAR(J)
            BNDLJ=LOBND(J)
         ENDIF
         PROD=PROD+VRBLJ*C(J)
C
C     Columns refering to FIXED variables were not reordered so their
C     contribution to the right hand side cannot be taken into account.
         IF(STATUS.GE.6) GO TO 160
         CALL GETCOL(J,RWORK,IWORK,RMAP,IMAP,
     X    IROW,RELT,K,MAXN,IOLOG)
         DO 140 I1=1,K
            IKX=IROW(I1)
            PRLACT(IKX)=PRLACT(IKX)+VRBLJ*RELT(I1)
            BNEW(IKX)=BNEW(IKX)+BNDLJ*RELT(I1)
  140    CONTINUE
  160 CONTINUE
      DO 180 I=1,M
         IF(DABS(BNEW(I)).LE.1.0D-12) BNEW(I)=0.0
  180 CONTINUE
C
C
C
C
C     writing the output file
C     ***********************
C
C     writing the title of the original problem
C     -----------------------------------------
C
      IF(IWRITE.EQ.1) THEN
         WRITE(BUFFER,103) NAMMPS
  103    FORMAT(1X,'TITLE OF THE PROBLEM: ',A8)
         CALL MYWRT(OUTMPS,BUFFER)
         IF(IELIM.EQ.1) THEN
            WRITE(BUFFER,104)
  104       FORMAT(1X,'Incomplete report (rows/cols elimination).')
            CALL MYWRT(OUTMPS,BUFFER)
         ENDIF
         WRITE(BUFFER,105)
  105    FORMAT(1X)
         CALL MYWRT(OUTMPS,BUFFER)
         CALL MYWRT(OUTMPS,BUFFER)
      ENDIF
C
C     writing the optimal value of the objective function
C     ---------------------------------------------------
C
      PROD=MULT*PROD
      PRLACT(MAXM)=PROD
      WRITE(BUFFER,107) PROD
  107 FORMAT(1X,'OBJECTIVE FUNCTION VALUE = ',D18.10)
      CALL MYWRT(IOLOG,BUFFER)
      CALL MYWRT(99,BUFFER)
      IF(IWRITE.EQ.1) THEN
         CALL MYWRT(OUTMPS,BUFFER)
      ENDIF
      IF(IWRITE.EQ.1) THEN
         WRITE(BUFFER,108)
  108    FORMAT(1X)
         CALL MYWRT(OUTMPS,BUFFER)
      ENDIF
C
C
C
C
C     writing the ROWS section
C     ------------------------
C
      WRITE(BUFFER,109)
  109 FORMAT(2X,'  ROW EQUATION LOWER LIMIT UPPER ',
     X 'LIMIT    ACTIVITY STAT  SLACK  SL ACTIVITY')
      IF(IWRITE.EQ.1) THEN
         CALL MYWRT(OUTMPS,BUFFER)
      ENDIF
C
C
C
C     Main loop over constraints.
      DO 300 I=1,MFINAL
         PROD=BNEW(I)-PRLACT(I)
         CALL GETROW(I,RWORK,IWORK,RMAP,IMAP,
     X    IROW,RELT,K,MAXN,IOLOG)
C
C
         IF(RWSTAT(I).EQ.1) THEN
C
C     Here for EQUALITY constraint.
            IF(IWRITE.EQ.1) THEN
               WRITE(BUFFER,301) I,RWNAME(I),BNEW(I),BNEW(I),
     X          PRLACT(I),PROD
  301          FORMAT(1X,I6,1X,A8,1X,D11.5,1X,D11.5,1X,D11.5,
     X          3X,'EQ',3X,'NONE',2X,D11.5)
               CALL MYWRT(OUTMPS,BUFFER)
            ENDIF
            GO TO 300
         ENDIF
C
C     Here for INEQUALITY constraint.
         IF(K.EQ.0) THEN
            IF(IWRITE.EQ.1) THEN
               WRITE(BUFFER,302) I,RWNAME(I)
  302          FORMAT(1X,I6,1X,A8,2X,'was eliminated.')
               CALL MYWRT(OUTMPS,BUFFER)
            ENDIF
            GO TO 300
         ENDIF
         J=IROW(1)
C
C
         IF(RWSTAT(I).EQ.2) THEN
C
C     Here for GREATER OR EQUAL type constraint (surplus variable).
            PROD=-PROD
            PRLVAR(J)=PROD
            IF(DABS(PROD)/(DABS(BNEW(I))+1.0).LT.PRFSBT) THEN
C
C     Row  I is at its LOWER bound.
               IF(IWRITE.EQ.1) THEN
                  WRITE(BUFFER,303) I,RWNAME(I),BNEW(I),PRLACT(I),
     X             J,PROD
  303             FORMAT(1X,I6,1X,A8,1X,D11.5,8X,'NONE',1X,D11.5,
     X             3X,'LL',1X,I6,2X,D11.5)
                  CALL MYWRT(OUTMPS,BUFFER)
               ENDIF
            ELSE
C
C     Row  I is not at its LOWER bound.
               IF(IWRITE.EQ.1) THEN
                  WRITE(BUFFER,304) I,RWNAME(I),BNEW(I),PRLACT(I),
     X             J,PROD
  304             FORMAT(1X,I6,1X,A8,1X,D11.5,8X,'NONE',1X,D11.5,
     X             6X,I6,2X,D11.5)
                  CALL MYWRT(OUTMPS,BUFFER)
               ENDIF
            ENDIF
C
            GO TO 300
         ENDIF
C
C
         IF(RWSTAT(I).EQ.3) THEN
            PRLVAR(J)=PROD
C
C     Here for LESS OR EQUAL type constraint (slack variable).
            IF(DABS(PROD)/(DABS(BNEW(I))+1.0).LT.PRFSBT) THEN
C
C     Row  I is at its UPPER bound.
               IF(IWRITE.EQ.1) THEN
                  WRITE(BUFFER,305) I,RWNAME(I),BNEW(I),PRLACT(I),
     X             J,PROD
  305             FORMAT(1X,I6,1X,A8,8X,'NONE',1X,D11.5,1X,D11.5,
     X             3X,'UL',1X,I6,2X,D11.5)
                  CALL MYWRT(OUTMPS,BUFFER)
               ENDIF
            ELSE
C
C     Row  I is not at its UPPER bound.
               IF(IWRITE.EQ.1) THEN
                  WRITE(BUFFER,306) I,RWNAME(I),BNEW(I),PRLACT(I),
     X             J,PROD
  306             FORMAT(1X,I6,1X,A8,8X,'NONE',1X,D11.5,1X,D11.5,
     X             6X,I6,2X,D11.5)
                  CALL MYWRT(OUTMPS,BUFFER)
               ENDIF
            ENDIF
C
            GO TO 300
         ENDIF
C
C
C
C     End of main loop over constraints.
  300 CONTINUE
C
C
C
C     Write information on eliminated constraints.
      DO 320 I=MFINAL+1,MAXM-1
         PRLACT(I)=-1.0D+36
  320 CONTINUE
      IF(IWRITE.EQ.1) THEN
         IF(IELIM.EQ.1) THEN
            DO 340 I=MFINAL+1,M
               WRITE(BUFFER,339) I,RWNAME(I)
  339          FORMAT(1X,I6,1X,A8,2X,'was eliminated.')
               CALL MYWRT(OUTMPS,BUFFER)
  340       CONTINUE
         ENDIF
      ENDIF
C
C
C
C
C     writing the dual activity of constraints
C     ----------------------------------------
C
      IF(IWRITE.EQ.1) THEN
         WRITE(BUFFER,341)
  341    FORMAT(1X)
         CALL MYWRT(OUTMPS,BUFFER)
         WRITE(BUFFER,342)
  342    FORMAT(1X,'   ROW EQUATION DUAL ACTIVITY')
         CALL MYWRT(OUTMPS,BUFFER)
      ENDIF
C
C
C
C     Main loop over constraints.
      IF(IWRITE.EQ.1) THEN
         DO 400 I=1,M
C
            WRITE(BUFFER,399) I,RWNAME(I),DLVAR(I)
  399       FORMAT(1X,I6,1X,A8,3X,D11.5)
            CALL MYWRT(OUTMPS,BUFFER)
C
C     End of main loop over constraints.
  400    CONTINUE
      ENDIF
C
C
C
C
C     writing the COLUMNS section
C     ---------------------------
C
      IF(IWRITE.EQ.1) THEN
         WRITE(BUFFER,401)
  401    FORMAT(1X)
         CALL MYWRT(OUTMPS,BUFFER)
         WRITE(BUFFER,402)
  402    FORMAT(1X,'COLUMN VARIABLE LOWER BOUND ',
     X    'UPPER BOUND    ACTIVITY')
         CALL MYWRT(OUTMPS,BUFFER)
      ENDIF
C
C
C
C     Main loop over variables.
      NFREE=0
      DO 500 J=1,NSTRCT
C
C
         IF(STAVAR(J).EQ.0) THEN
C
C     Here for the STANDARD variable.
            IF(IWRITE.EQ.0) GO TO 500
            WRITE(BUFFER,501) J,CLNAME(J),DBLE(0.),PRLVAR(J)
  501       FORMAT(1X,I6,1X,A8,1X,D11.5,8X,'NONE',1X,D11.5)
            CALL MYWRT(OUTMPS,BUFFER)
            GO TO 500
         ENDIF
C
C
         IF(STAVAR(J).EQ.1) THEN
C
C     Here if the variable has only an UPPER bound.
            IF(IWRITE.EQ.0) GO TO 500
            WRITE(BUFFER,502) J,CLNAME(J),DBLE(0.),UPBND(J),PRLVAR(J)
  502       FORMAT(1X,I6,1X,A8,1X,D11.5,1X,D11.5,1X,D11.5)
            CALL MYWRT(OUTMPS,BUFFER)
            GO TO 500
         ENDIF
C
C
         IF(STAVAR(J).EQ.2) THEN
C
C     Here if the variable has only a LOWER bound.
            PRLVAR(J)=LOBND(J)+PRLVAR(J)
            IF(IWRITE.EQ.0) GO TO 500
            WRITE(BUFFER,503) J,CLNAME(J),LOBND(J),PRLVAR(J)
  503       FORMAT(1X,I6,1X,A8,1X,D11.5,8X,'NONE',1X,D11.5)
            CALL MYWRT(OUTMPS,BUFFER)
            GO TO 500
         ENDIF
C
C
         IF(STAVAR(J).EQ.3) THEN
C
C     Here if the variable has both LOWER and UPPER bounds.
C     Observe that it has already been pushed to a zero LOWER bound.
C     Consequently, its UPPER bound was equal to UPBND(J)-LOBND(J).
            UPBND(J)=LOBND(J)+UPBND(J)
            PRLVAR(J)=LOBND(J)+PRLVAR(J)
            IF(IWRITE.EQ.0) GO TO 500
            WRITE(BUFFER,504) J,CLNAME(J),LOBND(J),UPBND(J),PRLVAR(J)
  504       FORMAT(1X,I6,1X,A8,1X,D11.5,1X,D11.5,1X,D11.5)
            CALL MYWRT(OUTMPS,BUFFER)
            GO TO 500
         ENDIF
C
C
         IF(STAVAR(J).EQ.6) THEN
C
C     Here if the variable is FIXED.
            UPBND(J)=LOBND(J)+UPBND(J)
            PRLVAR(J)=LOBND(J)+PRLVAR(J)
            IF(IWRITE.EQ.0) GO TO 500
            WRITE(BUFFER,505) J,CLNAME(J),LOBND(J),UPBND(J),PRLVAR(J)
  505       FORMAT(1X,I6,1X,A8,1X,D11.5,1X,D11.5,1X,D11.5)
            CALL MYWRT(OUTMPS,BUFFER)
            GO TO 500
         ENDIF
C
C
         IF(STAVAR(J).LT.0) THEN
C
C     Here if the variable is FREE.
            K=STAVAR(J)
            IF(J.GT.-K) GO TO 500
            PRLVAR(J)=PRLVAR(J)-PRLVAR(-K)
            NFREE=NFREE+1
            IF(IWRITE.EQ.0) GO TO 500
            WRITE(BUFFER,506) J,CLNAME(J),PRLVAR(J)
  506       FORMAT(1X,I6,1X,A8,8X,'NONE',8X,'NONE',1X,D11.5)
            CALL MYWRT(OUTMPS,BUFFER)
            GO TO 500
         ENDIF
C
C
C
C     End of main loop over variables.
  500 CONTINUE
      NSTRCT=NSTRCT-NFREE
C
C     Troubles in the problem 'marek.mps'. 
C     DO 600 J=638,643
C        WRITE(BUFFER,601) J,CLNAME(J),PRLVAR(J),STAVAR(J)
C 601    FORMAT(1X,I6,1X,A8,8X,'NONE',8X,'NONE',1X,D11.5,' st=',I5)
C        CALL MYWRT(OUTMPS,BUFFER)
C        CALL MYWRT(IOLOG,BUFFER)
C 600 CONTINUE
C
C
C     Print dual variables in the original order (as in the MPS file).
C     Save them in YORGNL work array.
C     DO 760 I1=1,M0
C        YORGNL(I1)=1.0D+30
C        DO 720 I=1,M0
C           IF(RWNAME(I).EQ.RWCOPY(I1)) THEN
C              YORGNL(I1)=DLVAR(I)
C              GO TO 740
C           ENDIF
C 720    CONTINUE
C 740    CONTINUE
C        WRITE(BUFFER,741) I1,RWSTAT(I),RWCOPY(I1),YORGNL(I1)
C 741    FORMAT(1X,'rw=',I6,'  st=',I2,'  name=',A8,'  y=',D12.5)
C        CALL MYWRT(IOLOG,BUFFER)
  760 CONTINUE
C
C     Save the value of the objective row.
      YORGNL(IROBJ)=PRLACT(MAXM)
C     WRITE(BUFFER,761) IROBJ,RWCOPY(IROBJ),YORGNL(IROBJ)
C 761 FORMAT(1X,'obj rw=',I6,'  name=',A8,'  y=',D18.10)
C     CALL MYWRT(IOLOG,BUFFER)
C
C
C
C     Close the MPS-like output file.
      CLOSE(OUTMPS)
      RETURN
C
C
C *** LAST CARD OF (WRTSOL) ***
      END
