C**************************************************************
C     **** XGTCOL ... GET THE  J-th COLUMN OF MATRIX  A ****
C**************************************************************
C
      SUBROUTINE XGTCOL(J,ACOEFF,
     X CLPNTS,RWNMBS,LENCOL,
     X IROW,RELT,COLLEN,IOLOG)
C
C
C
C *** PARAMETERS
      INTEGER*4 J,COLLEN,IOLOG
      INTEGER*4 CLPNTS(*),IROW(*)
      INTEGER*2 RWNMBS(*),LENCOL(*)
      DOUBLE PRECISION ACOEFF(*),RELT(*)
C
C
C
C *** LOCAL VARIABLES
      INTEGER*4 K,KBEG,IKX
C
C
C
C *** PARAMETERS DESCRIPTION
C     ON INPUT:
C     J       Index of the column to be created.
C     ACOEFF  Array of non zero elements for each column.
C     CLPNTS  Pointers to the beginning of columns of matrix A.
C     RWNMBS  Row numbers of nonzeros in columns of matrix A.
C     LENCOL  Lengths of (sparse) columns of matrix A.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     IROW    Row indices of nonzero entries of the column to be created.
C     RELT    Nonzero entries of the column to be created.
C     COLLEN  Number of nonzero entries of the column to be created.
C
C
C
C *** SUBROUTINES CALLED
C     NONE
C
C
C *** NOTES
C
C
C
C *** REFERENCES:
C     Gondzio J., Tachat D. (1994). The design and application of 
C        IPMLO - a FORTRAN library for linear optimization with 
C        interior point methods, RAIRO Recherche Operationnelle 28, 
C        No 1, pp. 37-56.
C
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  June 10, 1992
C     Last modified: May 17, 1995
C
C
C
C
C *** BODY OF (XGTCOL) ***
C
      COLLEN=LENCOL(J)
      KBEG=CLPNTS(J)-1
      DO 100 IKX=1,COLLEN
         K=KBEG+IKX
         IROW(IKX)=RWNMBS(K)
         RELT(IKX)=ACOEFF(K)
  100 CONTINUE
C
      RETURN
C
C *** LAST CARD OF (XGTCOL) ***
      END
